/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.dtypes;

import java.util.Locale;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import org.apache.commons.math3.complex.Complex;
import org.apache.commons.math3.linear.BlockRealMatrix;
import org.apache.commons.math3.linear.RealMatrix;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;

/**
 * Data representation of numeric-matrix
 * 
 * @author Werner Koppelsttter
 * @author Gnther Weidenholzer
 */
public class TNumeric extends TData {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Real part of numeric matrix */
	private BlockRealMatrix re = null; // Double, null if not real!!

	/** Imaginary part of numeric matrix */
	private BlockRealMatrix im = null; // Double, null if not imag!!

	/**
	 * Convert from Double to double
	 * 
	 * @param val
	 *            Double-object to convert
	 * @return Primitive double
	 */
	private static double convert(Double val) {
		return val == null ? 0.0 : val.doubleValue();
	}

	/**
	 * Convert from Double-array to double-array
	 * 
	 * @param arrayObj
	 *            Double-object-array to convert
	 * @return Primitive double-object-array
	 */
	private static double[] convert(Double[] arrayObj) {
		double[] array = null;

		if (arrayObj != null) {
			array = new double[arrayObj.length];
			for (int k = 0; k < arrayObj.length; k++) {
				array[k] = TNumeric.convert(arrayObj[k]);
			}
		}

		return array;
	}

	/**
	 * Convert from Double-array to double-array
	 * 
	 * @param arrayObj
	 *            Double-object-array to convert
	 * @return Primitive double-object-array
	 */
	private static double[][] convert(Double[][] arrayObj) {
		double[][] array = null;

		if (arrayObj != null) {
			array = new double[arrayObj.length][];
			for (int k = 0; k < arrayObj.length; k++) {
				array[k] = TNumeric.convert(arrayObj[k]);
			}
		}

		return array;
	}

	/**
	 * Represents (complex) numeric data matrices for loading
	 */
	protected TNumeric() {
	}

	/**
	 * Initializes a real double matrix
	 * 
	 * @param re
	 *            two-dimensional double array of real part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double[][] re) throws TDataException {
		this(re, null);
	}

	/**
	 * Initializes a real double row vector
	 * 
	 * @param re
	 *            double array of real part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double[] re) throws TDataException {
		this(re, null);
	}

	/**
	 * Initializes a real double scalar
	 * 
	 * @param re
	 *            double real value
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double re) throws TDataException {
		this(re, 0.0);
	}

	/**
	 * Initializes a real double matrix
	 * 
	 * @param re
	 *            two-dimensional double array of real part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double[][] re) throws TDataException {
		this(re, null);
	}

	/**
	 * Initializes a real double row vector
	 * 
	 * @param re
	 *            double array of real part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double[] re) throws TDataException {
		this(re, null);
	}

	/**
	 * Initializes a real double scalar
	 * 
	 * @param re
	 *            double real value
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double re) throws TDataException {
		this(re, null);
	}

	/**
	 * Initializes a complex value
	 * 
	 * @param val
	 *            complex value
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Complex val) throws TDataException {
		this(val.getReal(), val.getImaginary());
	}

	/**
	 * Initializes a complex double matrix
	 * 
	 * @param re
	 *            two-dimensional double array of real part
	 * @param im
	 *            two-dimensional double array of imaginary part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double[][] re, double[][] im) throws TDataException {

		if ((re == null) && (im == null)) {
			throw new TDataException("Empty Numeric is not allowed");
		} else if ((re != null) && (im != null)) {
			if ((re.length != im.length) || (re[0].length != im[0].length)) {
				throw new TDataException("Dimension mismatch");
			}
		}

		// real
		if (re != null) {
			if ((re.length != 0) && (re[0].length != 0)) {
				this.re = new BlockRealMatrix(re);
			} else {
				throw new TDataException("Dimension 0 is not allowed");
			}
		}

		// imaginary
		if (im != null) {
			if ((im.length != 0) && (im[0].length != 0)) {
				this.im = new BlockRealMatrix(im);
			} else {
				throw new TDataException("Dimension 0 is not allowed");
			}
		}

		this.compact();
	}

	/**
	 * Initializes a complex double row vector
	 * 
	 * @param re
	 *            double array of real part
	 * @param im
	 *            double array of imaginary part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double[] re, double[] im) throws TDataException {

		if ((re == null) && (im == null)) {
			throw new TDataException("Empty Numeric is not allowed");
		} else if ((re != null) && (im != null) && (re.length != im.length)) {
			throw new TDataException("Dimension mismatch");
		}

		// real
		if (re != null) {
			if (re.length != 0) {
				this.re = new BlockRealMatrix(new double[][] { re });
			} else {
				throw new TDataException("Dimension 0 is not allowed");
			}
		}

		// imaginary
		if (im != null) {
			if (im.length != 0) {
				this.im = new BlockRealMatrix(new double[][] { im });
			} else {
				throw new TDataException("Dimension 0 not allowed");
			}
		}

		this.compact();
	}

	/**
	 * Initializes a complex double number
	 * 
	 * @param re
	 *            real part of number
	 * @param im
	 *            imaginary part of number
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(double re, double im) throws TDataException {

		this.re = new BlockRealMatrix(new double[][] { { re } });
		this.im = new BlockRealMatrix(new double[][] { { im } });

		this.compact();
	}

	/**
	 * Initializes a complex double matrix
	 * 
	 * @param re
	 *            two-dimensional double array of real part
	 * @param im
	 *            two-dimensional double array of imaginary part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double[][] re, Double[][] im) throws TDataException {
		this(TNumeric.convert(re), TNumeric.convert(im));
	}

	/**
	 * Initializes a complex double row vector
	 * 
	 * @param re
	 *            double array of real part
	 * @param im
	 *            double array of imaginary part
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double[] re, Double[] im) throws TDataException {
		this(TNumeric.convert(re), TNumeric.convert(im));
	}

	/**
	 * Initializes a complex double number
	 * 
	 * @param re
	 *            real part of number
	 * @param im
	 *            imaginary part of number
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(Double re, Double im) throws TDataException {
		this(TNumeric.convert(re), TNumeric.convert(im));
	}

	/**
	 * Initializes a complex matrix
	 * 
	 * @param re
	 *            real part of matrix
	 * @param im
	 *            imaginary part of matrix
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(RealMatrix re, RealMatrix im) throws TDataException {

		if ((re == null) && (im == null)) {
			throw new TDataException("Empty Numeric is not allowed");
		} else if ((re != null) && (im != null)) {
			if ((re.getRowDimension() != im.getRowDimension())
					|| (re.getColumnDimension() != im.getColumnDimension())) {
				throw new TDataException("Dimension mismatch");
			}
		}

		// copy data
		if (re != null) {
			if ((re.getRowDimension() != 0) && (re.getColumnDimension() != 0)) {
				this.re = new BlockRealMatrix(re.copy().getData());
			} else {
				throw new TDataException("Dimension 0 is not allowed");
			}
		}
		if (im != null) {
			if ((im.getRowDimension() != 0) && (im.getColumnDimension() != 0)) {
				this.im = new BlockRealMatrix(im.copy().getData());
			} else {
				throw new TDataException("Dimension 0 is not allowed");
			}
		}
	}

	/**
	 * Initializes a real matrix
	 * 
	 * @param re
	 *            real part of number
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(RealMatrix re) throws TDataException {

		this(re, null);
	}

	/**
	 * Initializes a numeric from string
	 * 
	 * @param str
	 *            string similar to "[1.2,3; 12,-2]" or "[1.2-2i,3+0i;
	 *            0+0i,-2-1i]"
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public TNumeric(String str) throws TDataException {
		this.fromString(str);
	}

	/**
	 * String interpreter
	 * 
	 * @param str
	 *            string representation of numeric
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	private void fromString(String str) throws TDataException {
		int m, n;

		str = str.trim();
		if (str.equals(new String("[]"))) {
			throw new TDataException("Empty Numeric is not allowed");
		}
		if (str.startsWith("[") && str.endsWith("]")) {
			str = str.substring(1, str.length() - 1);
		}

		String[] rows = str.split(";");
		String[][] items = new String[rows.length][];

		// check dimension
		m = rows.length;
		for (int k = 0; k < m; k++) {
			items[k] = rows[k].trim().split("[ ,]+");
			if (items[k].length != items[0].length) {
				throw new TDataException("Numeric dimension mismatch!");
			}
		}
		n = items[0].length;

		this.re = new BlockRealMatrix(m, n);
		this.im = new BlockRealMatrix(m, n);

		for (int row = 0; row < m; row++) {
			for (int col = 0; col < n; col++) {
				this.setStringValEntry(row, col, items[row][col]);
			}
		}

		this.compact();
	}

	/**
	 * Creates Zero Matrix
	 * 
	 * @return Zero Matrix
	 */
	protected BlockRealMatrix zeroMatrix() {
		return new BlockRealMatrix(this.size()[0], this.size()[1]);
	}

	/**
	 * Set matrix to null if unused
	 */
	protected void compact() {
		if ((this.im == null) || (this.im.getNorm() == 0.0)) {
			this.im = null;
		} else if ((this.re == null) || (this.re.getNorm() == 0.0)) {
			this.re = null;
		}
	}

	/**
	 * expands null matrices
	 */
	protected void expand() {
		if (this.re == null) {
			this.re = this.zeroMatrix();
		} else if (this.im == null) {
			this.im = this.zeroMatrix();
		}
	}

	/**
	 * @return true if 1 row and 1 column
	 */
	public boolean isScalar() {
		return (this.size()[0] == 1) && (this.size()[1] == 1);
	}

	/**
	 * Get row dimension
	 * 
	 * @return number of rows
	 */
	public int getNumRows() {
		// FIXME remove
		return this.size()[0];
	}

	/**
	 * Get column dimension
	 * 
	 * @return number of columns
	 */
	public int getNumColumns() {
		// FIXME remove
		return this.size()[1];
	}

	/**
	 * Get cell of numeric
	 * 
	 * @param row
	 *            Index of row
	 * @param col
	 *            Index of column
	 * @return Numeric with row of this
	 * @throws TDataException
	 *             Wrong index
	 */
	public TNumeric get(int row, int col) throws TDataException {
		TNumeric num;

		if ((row < 0) || (this.size()[0] <= row)) {
			throw new TDataException("Index out of bounds!");
		}
		if ((col < 0) || (this.size()[1] <= col)) {
			throw new TDataException("Index out of bounds!");
		}

		num = new TNumeric();
		if (this.re != null) {
			num.re = this.re.getSubMatrix(row, row, col, col);
		}
		if (this.im != null) {
			num.im = this.im.getSubMatrix(row, row, col, col);
		}

		return num;
	}

	/**
	 * Get row of numeric
	 * 
	 * @param row
	 *            Index of row
	 * @return Numeric with row of this
	 * @throws TDataException
	 *             Wrong index
	 */
	public TNumeric getRow(int row) throws TDataException {
		TNumeric num;

		if ((row < 0) || (this.size()[0] <= row)) {
			throw new TDataException("Index out of bounds!");
		}

		num = new TNumeric();
		if (this.re != null) {
			num.re = this.re.getRowMatrix(row);
		}
		if (this.im != null) {
			num.im = this.im.getRowMatrix(row);
		}

		return num;
	}

	/**
	 * Get column of numeric
	 * 
	 * @param col
	 *            Index of column
	 * @return Numeric with column of this
	 * @throws TDataException
	 *             Wrong index
	 */
	public TNumeric getColumn(int col) throws TDataException {
		TNumeric num;

		if ((col < 0) || (this.size()[1] <= col)) {
			throw new TDataException("Index out of bounds!");
		}

		num = new TNumeric();
		if (this.re != null) {
			num.re = this.re.getColumnMatrix(col);
		}
		if (this.im != null) {
			num.im = this.im.getColumnMatrix(col);
		}

		return num;
	}

	/**
	 * Get sub-matrix of numeric
	 * 
	 * @param startRow
	 *            Initial row index
	 * @param endRow
	 *            Final row index (inclusive)
	 * @param startCol
	 *            Initial column index
	 * @param endCol
	 *            Final column index (inclusive)
	 * @return Sub-matrix
	 * @throws TDataException
	 *             Wrong index
	 */
	public TNumeric getSubMatrix(int startRow, int endRow, int startCol, int endCol) throws TDataException {
		TNumeric num;

		if ((startRow < 0) || (this.size()[0] <= endRow)) {
			throw new TDataException("Index out of bounds!");
		} else if ((startCol < 0) || (this.size()[1] <= endCol)) {
			throw new TDataException("Index out of bounds!");
		} else if (startRow > endRow) {
			throw new TDataException("Start-index of row greater than end-index!");
		} else if (startCol > endCol) {
			throw new TDataException("Start-index of column  greater than end-index!");
		}

		num = new TNumeric();
		if (this.re != null) {
			num.re = this.re.getSubMatrix(startRow, endRow, startCol, endCol);
		}
		if (this.im != null) {
			num.im = this.im.getSubMatrix(startRow, endRow, startCol, endCol);
		}

		return num;
	}

	/**
	 * Get real value
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @return Real item at (m,n);
	 */
	public double getReal(int m, int n) {
		if (this.re != null) {
			return this.re.getEntry(m, n);
		} else {
			return 0.0;
		}
	}

	/**
	 * Get imaginary value
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @return Imaginary item at (m,n);
	 */
	public double getImag(int m, int n) {
		if (this.im != null) {
			return this.im.getEntry(m, n);
		} else {
			return 0.0;
		}

	}

	/**
	 * Get value in string representation
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @return String representation of value
	 */
	public String getStringVal(int m, int n) {
		return this.getStringVal(m, n, -1);
	}

	/**
	 * Get value in string representation
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @param decimals
	 *            Number of significant digits
	 * @return String representation of value
	 */
	public String getStringVal(int m, int n, int decimals) {
		double real, imag;
		String str;

		real = this.getReal(m, n);
		str = TNumeric.num2str(real, decimals);

		imag = this.getImag(m, n);
		if (imag != 0.0) {
			str += (imag >= 0) ? "+i*" : "-i*";
			str += TNumeric.num2str(Math.abs(imag), decimals);
		}

		return str.replace("Infinity", "Inf");
	}

	/**
	 * Set value form string
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @param val
	 *            String representaion of value
	 * @throws TDataException
	 *             Unable to create data-object
	 */
	public void setStringVal(int m, int n, String val) throws TDataException {
		this.setStringValEntry(m, n, val);
		this.compact();
	}

	/**
	 * Set value form string
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @param val
	 *            String representaion of value
	 * @throws TDataException
	 *             Wrong index
	 */
	private void setStringValEntry(int m, int n, String val) throws TDataException {
		int index, start;
		Double re, im;

		val = val.trim().replace("Inf", "Infinity").replace("Infinityinity", "Infinity");

		if (val.contains("i*")) {
			index = val.indexOf("i*");
			re = Double.parseDouble(val.substring(0, index - 1));
			im = Double.parseDouble(val.charAt(index - 1) + val.substring(index + 2, val.length()));
		} else if (val.endsWith("i")) {
			if (val.endsWith("*i")) {
				val = val.substring(0, val.length() - 2);
			} else {
				val = val.substring(0, val.length() - 1);
			}

			index = val.lastIndexOf('+');
			if (index == -1) {
				index = val.lastIndexOf('-');
			}
			if ((index > 0) && ((val.charAt(index - 1) == 'e') || (val.charAt(index - 1) == 'E'))) {
				start = index;
				index = val.lastIndexOf('+', start - 1);
				if (index == -1) {
					index = val.lastIndexOf('-', start - 1);
				}
			}
			// if (index == -1) {
			// re = 0.0;
			// im = Double.parseDouble(val);
			// } else {
			re = Double.parseDouble(val.substring(0, index));
			im = Double.parseDouble(val.substring(index));
			// }
		} else {
			re = Double.parseDouble(val);
			im = 0.0;
		}

		this.setRealEntry(m, n, re);
		this.setImagEntry(m, n, im);
	}

	/**
	 * Returns real matrix data
	 * 
	 * @return real matrix
	 */
	public BlockRealMatrix getRealMatrix() {

		if (this.re == null) {
			return this.zeroMatrix();
		} else {
			return this.re.copy();
		}
	}

	/**
	 * Sets real matrix data
	 * 
	 * @param re
	 *            real matrix data
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public void setRealMatrix(BlockRealMatrix re) throws TDataException {
		if (re != null) {
			if ((re.getColumnDimension() != this.size()[1]) || (re.getRowDimension() != this.size()[0])) {
				throw new TDataException("wrong dimension!");
			}
			this.re = re.copy();
		} else {
			throw new TDataException("Illegal arguments for TNumeric");
		}

	}

	/**
	 * Returns real matrix data
	 * 
	 * @return real matrix
	 */
	public BlockRealMatrix getImagMatrix() {
		if (this.im == null) {
			return this.zeroMatrix();
		} else {
			return this.im.copy();
		}
	}

	/**
	 * Sets real matrix data
	 * 
	 * @param im
	 *            imaginary matrix data
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public void setImagMatrix(BlockRealMatrix im) throws TDataException {
		if (im != null) {
			if ((im.getColumnDimension() != this.size()[1]) || (im.getRowDimension() != this.size()[0])) {
				throw new TDataException("wrong dimension!");
			}
			this.im = im.copy();
		} else {
			throw new TDataException("Illegal arguments for TNumeric");
		}
	}

	/**
	 * Get complex value
	 * 
	 * @param m
	 *            number of row
	 * @param n
	 *            number of column
	 * @return complex item at (m,n);
	 */
	public Complex getComplex(int m, int n) {
		return new Complex(this.getReal(m, n), this.getImag(m, n));
	}

	/**
	 * Sets real entry
	 * 
	 * @param m
	 *            row
	 * @param n
	 *            column
	 * @param val
	 *            real value
	 * @throws TDataException
	 *             Wrong index
	 */
	public void setReal(int m, int n, double val) throws TDataException {
		this.setRealEntry(m, n, val);
		this.compact();
	}

	/**
	 * Sets real entry
	 * 
	 * @param m
	 *            row
	 * @param n
	 *            column
	 * @param val
	 *            real value
	 * @throws TDataException
	 *             Wrong index
	 */
	private void setRealEntry(int m, int n, double val) throws TDataException {
		if ((m < this.size()[0]) || (n < this.size()[1])) {
			if (this.re == null) {
				this.re = this.zeroMatrix();
			}
			this.re.setEntry(m, n, val);
		} else {
			throw new TDataException("Index exceeds matrix dimension!");
		}
	}

	/**
	 * Sets imaginary entry
	 * 
	 * @param m
	 *            row
	 * @param n
	 *            column
	 * @param val
	 *            imaginary value
	 * @throws TDataException
	 *             Wrong index
	 */
	public void setImag(int m, int n, double val) throws TDataException {
		this.setImagEntry(m, n, val);
		this.compact();
	}

	/**
	 * Sets imaginary entry
	 * 
	 * @param m
	 *            row
	 * @param n
	 *            column
	 * @param val
	 *            imaginary value
	 * @throws TDataException
	 *             Wrong index
	 */
	private void setImagEntry(int m, int n, double val) throws TDataException {
		if ((m < this.size()[0]) || (n < this.size()[1])) {
			if (this.im == null) {
				this.im = this.zeroMatrix();
			}
			this.im.setEntry(m, n, val);
		} else {
			throw new TDataException("Index exceeds matrix dimension!");
		}
	}

	/**
	 * Sets complex entry
	 * 
	 * @param m
	 *            row
	 * @param n
	 *            column
	 * @param val
	 *            complex value
	 * @throws TDataException
	 *             Wrong index
	 */
	public void setComplex(int m, int n, Complex val) throws TDataException {
		this.setReal(m, n, val.getReal());
		this.setImag(m, n, val.getImaginary());
	}

	/**
	 * Scales the given numeric. Source object remains unchanged!
	 * 
	 * @param scale
	 *            scale factor
	 * @return Scaled numeric
	 */
	public TNumeric scale(double scale) {
		TNumeric result = new TNumeric();

		if (this.re != null) {
			result.re = new BlockRealMatrix(this.re.scalarMultiply(scale).getData());
		}

		if (this.im != null) {
			result.im = new BlockRealMatrix(this.im.scalarMultiply(scale).getData());
		}

		return result;
	}

	/**
	 * Adds a scalar offset to the given numeric. Source object remains
	 * unchanged!
	 * 
	 * @param real
	 *            Real offset
	 * @param imag
	 *            Imaginary offset
	 * @return numeric with offset
	 */
	public TNumeric offset(double real, double imag) {

		TNumeric result = new TNumeric();

		this.expand();

		result.re = this.re.scalarAdd(real);
		result.im = this.im.scalarAdd(imag);

		this.compact();
		result.compact();

		return result;
	}

	/**
	 * Adds a scalar offset to the given numeric. Source object remains
	 * unchanged!
	 * 
	 * @param real
	 *            Real offset
	 * @return numeric with offset
	 */
	public TNumeric offset(double real) {
		return this.offset(real, 0);
	}

	/**
	 * Flips the given numeric from LEFT to RIGHT. Source object remains
	 * unchanged!
	 * 
	 * @return Flipped numeric
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TNumeric fliplr() throws CloneException {

		TNumeric theClone;
		theClone = this.deepClone();
		int[] size = this.size();

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				if (theClone.re != null) {
					double help = this.re.getEntry(row, size[1] - col - 1);
					theClone.re.setEntry(row, col, help);
				}
				if (theClone.im != null) {
					double help = this.im.getEntry(row, size[1] - col - 1);
					theClone.im.setEntry(row, col, help);
				}
			}
		}
		return theClone;
	}

	/**
	 * Flips the given numeric from UP to DOWN. Source object remains unchanged!
	 * 
	 * @return Flipped numeric
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TNumeric flipud() throws CloneException {

		TNumeric theClone;
		theClone = this.deepClone();
		int[] size = this.size();

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {

				if (theClone.re != null) {
					double help = this.re.getEntry(size[0] - row - 1, col);
					theClone.re.setEntry(row, col, help);
				}
				if (theClone.im != null) {
					double help = this.im.getEntry(size[0] - row - 1, col);
					theClone.im.setEntry(row, col, help);
				}
			}
		}
		return theClone;
	}

	/**
	 * Matrix multiplication, source matrix remains unchanged!
	 * 
	 * @param other
	 *            Second matrix
	 * @return result Result matrix
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix multiply(TPeriodicMatrix other) throws TDataException, CloneException {

		return other.r_multiply(this);

	}

	/**
	 * Matrix multiplication (other * this), source matrix remains unchanged!
	 * 
	 * @param other
	 *            Second matrix
	 * @return result Result matrix
	 * @throws TDataException
	 *             Wrong dimension
	 * @throws CloneException
	 *             Unable to clone (serialize)
	 */
	public TPeriodicMatrix r_multiply(TPeriodicMatrix other) throws TDataException, CloneException {

		return other.multiply(this);

	}

	/**
	 * Given numeric will be transposed. Source object remains unchanged!
	 * 
	 * @return Transposed numeric
	 */
	public TNumeric transpose() {

		TNumeric result = new TNumeric();

		if (this.re != null) {
			result.re = this.re.transpose();
		}

		if (this.im != null) {
			result.im = this.im.transpose();
		}

		return result;

	}

	@Override
	public TNumeric deepClone() throws CloneException {
		return (TNumeric) Clone.deepClone(this);
	}

	/**
	 * Returns a full string representation with umlimited precision
	 * 
	 * @return string similar to "[1.2,3; 12,-2]" or "[1.2-2i,3+0i; 0+0i,-2-1i]"
	 */
	public String toFullString() {
		return (this.toFullString(-1)); // unlimited precision
	}

	/**
	 * Number formatter
	 * 
	 * @param val
	 *            Value to format
	 * @param precision
	 *            Number of significant digits
	 * @return String representation
	 */
	public static String num2str(double val, int precision) {
		if (precision > 0) {
			return Double.valueOf(String.format(Locale.ENGLISH, "%." + (precision - 1) + "e", val)).toString()
					.replace('E', 'e');
		} else {
			return Double.toString(val).replace('E', 'e');
		}
	}

	/**
	 * Returns a full string representation with limited precision
	 * 
	 * @param decimals
	 *            Number of significant digits
	 * 
	 * @return string similar to "[1.2,3; 12,-2]" or "[1.2-2i,3+0i; 0+0i,-2-1i]"
	 */
	public String toFullString(int decimals) {

		this.expand();
		StringBuffer str = new StringBuffer();

		int[] size = this.size();

		if (!this.isScalar()) {
			str.append("[");
		}

		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {
				str.append(this.getStringVal(row, col, decimals) + " ");
			}
			str.deleteCharAt(str.length() - 1);
			str.append(";");

		}
		str.deleteCharAt(str.length() - 1);
		if (!this.isScalar()) {
			str.append("]");
		}

		this.compact();
		return str.toString().replace("Infinity", "Inf");
	}

	@Override
	public String toString() {

		String str;
		try {
			str = this.toFullString(6);
		} catch (Exception e) {
			str = "String generation failed!";
		}

		if (str.length() > 70) {
			int[] size = this.size();

			str = this.isReal() ? new String("[real ") : new String("[complex ");

			if (this.isScalar()) {
				str += " scalar]";
			} else if (size[0] == 1) {
				str += "1x" + size[1] + " row vector]";
			} else if (size[1] == 1) {
				str += size[0] + "x1 column vector]";
			} else {
				str += size[0] + "x" + size[1] + " matrix]";
			}
		}
		return str;
	}

	@Override
	public void xml2data(XMLStreamReader xmlReader) throws XMLStreamException, TDataException {

		String str = new String();

		xmlReader.next();
		while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
			str += xmlReader.getText();
			xmlReader.next();
		}
		xmlReader.next();

		this.fromString(str);
	}

	@Override
	public void data2xml(XMLStreamWriter xmlWriter) throws XMLStreamException {

		xmlWriter.writeStartElement("Numeric");
		xmlWriter.writeCharacters(this.toFullString());
		xmlWriter.writeEndElement();
	}

	/**
	 * compares two matrices
	 * 
	 * @param m1
	 *            matrix 1
	 * @param m2
	 *            matrix 2
	 * @return true if equal
	 */
	protected boolean compareMatrix(BlockRealMatrix m1, BlockRealMatrix m2) {
		int rows, cols;
		double d1, d2;

		if (m1 == null) {
			return ((m2 == null) || (m2.getNorm() == 0.0));
		} else if (m2 == null) {
			return (m1.getNorm() == 0.0);
		} else {
			rows = m1.getRowDimension();
			cols = m1.getColumnDimension();
			if ((rows != m2.getRowDimension()) || (cols != m2.getColumnDimension())) {
				return false;
			}
			for (int r = 0; r < rows; r++) {
				for (int c = 0; c < cols; c++) {
					d1 = m1.getEntry(r, c);
					d2 = m2.getEntry(r, c);
					if (d1 != d2) {
						// except NaN
						if (!(Double.isNaN(d1) && Double.isNaN(d2))) {
							return false;
						}
					}
				}
			}
		}
		return true;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		TNumeric other = (TNumeric) obj;

		if (!this.compareMatrix(this.re, other.re)) {
			return false;
		}

		if (!this.compareMatrix(this.im, other.im)) {
			return false;
		}

		return true;
	}

	/**
	 * Check if matrix is real
	 * 
	 * @return matrix is real
	 */
	public boolean isReal() {
		this.compact();
		return (this.im == null);
	}

	/**
	 * Get a array of the size (two dimension)
	 * 
	 * @return Size of matrix
	 */
	public int[] size() {
		int[] size = new int[2];

		if (this.re != null) {
			size[0] = this.re.getRowDimension();
			size[1] = this.re.getColumnDimension();
		} else {
			size[0] = this.im.getRowDimension();
			size[1] = this.im.getColumnDimension();
		}

		return size;
	}

	/**
	 * returns the M-by-N matrix whose elements are taken columnwise from X.
	 * 
	 * @param m
	 *            new row dimension
	 * @param n
	 *            new column dimension
	 * 
	 * @return reshaped numeric
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TNumeric reshape(int m, int n) throws TDataException {

		TNumeric result = new TNumeric();

		int[] size = this.size();

		if ((size[0] * size[1]) != (m * n)) {
			throw new TDataException("matrix dimension error");
		}

		if (this.re != null) {
			result.re = new BlockRealMatrix(m, n);
		}

		if (this.im != null) {
			result.im = new BlockRealMatrix(m, n);
		}

		int idx = 0;
		for (int row = 0; row < size[0]; row++) {
			for (int col = 0; col < size[1]; col++) {

				int row2 = idx / n;
				int col2 = idx % n;

				if (this.re != null) {
					result.re.setEntry(row2, col2, this.re.getEntry(row, col));
				}
				if (this.im != null) {
					result.im.setEntry(row2, col2, this.im.getEntry(row, col));
				}

				idx++;
			}
		}
		return result;
	}

	/**
	 * Add matrix to matrix. If matrices has same size then add each element
	 * will added by the element of other matrix. If one matrix is a scalar then
	 * the scalar is will added to each element of other matrix.
	 * 
	 * @param num
	 *            Matrix to add
	 * @return Added matrix
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TNumeric add(TNumeric num) throws TDataException {
		TNumeric result;

		if (num.isScalar()) {
			result = this.addScalar(num);
		} else if (this.isScalar()) {
			result = num.addScalar(this);
		} else if ((this.size()[0] != num.size()[0]) || (this.size()[1] != num.size()[1])) {
			throw new TDataException("matrix dimension error");
		} else {
			result = new TNumeric();
			if ((this.re != null) && (num.re != null)) {
				result.re = this.re.add(num.re);
			} else if (this.re != null) {
				result.re = this.re.copy();
			} else if (num.re != null) {
				result.re = num.re.copy();
			}

			if ((this.im != null) && (num.im != null)) {
				result.im = this.im.add(num.im);
			} else if (this.im != null) {
				result.im = this.im.copy();
			} else if (num.im != null) {
				result.im = num.im.copy();
			}
		}

		return result;
	}

	/**
	 * Add scalar matrix to matrix. The scalar is will added to each element of
	 * other matrix.
	 * 
	 * @param scalar
	 *            Matrix with on entry
	 * @return Added matrix
	 * @throws TDataException
	 *             Wrong dimension
	 */
	private TNumeric addScalar(TNumeric scalar) throws TDataException {
		TNumeric result = new TNumeric();

		if (!scalar.isScalar()) {
			throw new TDataException("Numeric is not a scalar");
		}

		if (this.re != null) {
			result.re = this.re.scalarAdd(scalar.getReal(0, 0));
		} else {
			result.re = new BlockRealMatrix(this.getNumRows(), this.getNumColumns()).scalarAdd(scalar.getReal(0, 0));
		}

		if (this.im != null) {
			result.im = this.im.scalarAdd(scalar.getImag(0, 0));
		} else {
			result.im = new BlockRealMatrix(this.getNumRows(), this.getNumColumns()).scalarAdd(scalar.getImag(0, 0));
		}

		return result;
	}

	/**
	 * Multiply matrix to matrix. If matrix size matched from colum and row then
	 * multiply by matrix multiply. If one matrix is a scalar then the scalar
	 * will multiplied to each element of other matrix.
	 * 
	 * @param num
	 *            Matrix to multiply
	 * @return Multiplied matrix
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TNumeric multiply(TNumeric num) throws TDataException {
		TNumeric result;
		BlockRealMatrix tmp;

		if (num.isScalar()) {
			result = this.multiplyScalar(num);
		} else if (this.isScalar()) {
			result = num.multiplyScalar(this);
		} else if (this.size()[1] != num.size()[0]) {
			throw new TDataException("matrix dimension error");
		} else {
			result = new TNumeric();
			// (a+b*i)*(c+d*i) = (a*c-b*d) + (b*c+a*d)i
			// real
			if ((this.re != null) && (num.re != null)) {
				// a*c
				result.re = this.re.multiply(num.re);
			}
			if ((this.im != null) && (num.im != null)) {
				// -b*d
				tmp = new BlockRealMatrix(this.im.multiply(num.im).scalarMultiply(-1).getData());
				if (result.re == null) {
					result.re = tmp;
				} else {
					result.re = result.re.add(tmp);
				}
			}
			// imag
			if ((this.im != null) && (num.re != null)) {
				// b*c
				result.im = this.im.multiply(num.re);
			}
			if ((this.re != null) && (num.im != null)) {
				// a*d
				tmp = this.re.multiply(num.im);
				if (result.im == null) {
					result.im = tmp;
				} else {
					result.im = result.im.add(tmp);
				}
			}
		}

		return result;
	}

	/**
	 * Multiply scalar matrix to matrix. The scalar will multiplied to each
	 * element of other matrix.
	 * 
	 * @param scalar
	 *            Matrix with on entry
	 * @return Multiplied matrix
	 * @throws TDataException
	 *             Wrong dimension
	 */
	private TNumeric multiplyScalar(TNumeric scalar) throws TDataException {
		TNumeric result = new TNumeric();

		if (!scalar.isScalar()) {
			throw new TDataException("Numeric is not a scalar");
		}

		if (this.re == null) {
			// (b*i)*(c+d*i) = (-b*d) + (b*c)i
			result.re = new BlockRealMatrix(this.im.scalarMultiply(-scalar.getImag(0, 0)).getData());
			result.im = new BlockRealMatrix(this.im.scalarMultiply(scalar.getReal(0, 0)).getData());
		} else if (this.im == null) {
			// (a)*(c+d*i) = (a*c) + (a*d)i
			result.re = new BlockRealMatrix(this.re.scalarMultiply(scalar.getReal(0, 0)).getData());
			result.im = new BlockRealMatrix(this.re.scalarMultiply(scalar.getImag(0, 0)).getData());
		} else {
			// (a+b*i)*(c+d*i) = (a*c-b*d) + (b*c+a*d)i
			result.re = new BlockRealMatrix(this.re.scalarMultiply(scalar.getReal(0, 0))
					.add(this.im.scalarMultiply(-scalar.getImag(0, 0))).getData());
			result.im = new BlockRealMatrix(this.im.scalarMultiply(scalar.getReal(0, 0))
					.add(this.re.scalarMultiply(scalar.getImag(0, 0))).getData());
		}

		return result;
	}

	/**
	 * Multiply elements of matrix.
	 * 
	 * @param num
	 *            Matrix to multiply elements
	 * @return Multiplied matrix
	 * @throws TDataException
	 *             Wrong dimension
	 */
	public TNumeric multiplyElem(TNumeric num) throws TDataException {
		double[][] re, im;
		double a, b, c, d;
		TNumeric result;

		if (num.isScalar()) {
			result = this.multiplyScalar(num);
		} else if (this.isScalar()) {
			result = num.multiplyScalar(this);
		} else if ((this.size()[0] != num.size()[0]) || (this.size()[1] != num.size()[1])) {
			throw new TDataException("matrix dimension error");
		} else {

			re = new double[this.getNumRows()][this.getNumColumns()];
			im = new double[this.getNumRows()][this.getNumColumns()];
			for (int row = 0; row < this.getNumRows(); row++) {
				for (int col = 0; col < this.getNumColumns(); col++) {
					// (a+b*i)*(c+d*i) = (a*c-b*d) + (b*c+a*d)i
					a = this.re == null ? 0 : this.re.getEntry(row, col);
					b = this.im == null ? 0 : this.im.getEntry(row, col);
					c = num.re == null ? 0 : num.re.getEntry(row, col);
					d = num.im == null ? 0 : num.im.getEntry(row, col);
					re[row][col] = (a * c) - (b * d);
					im[row][col] = (b * c) + (a * d);
				}
			}
			result = new TNumeric(re, im);
		}

		return result;
	}

	// /**
	// * Divide elements of matrix.
	// *
	// * @param num
	// * Matrix to divide elements
	// * @return Divided matrix
	// * @throws TDataException
	// * Wrong dimension
	// */
	// public TNumeric divideElem(TNumeric num) throws TDataException {
	// double[][] re, im;
	// double a, b, c, d, div;
	//
	// if ((this.size()[0] != num.size()[0])
	// || (this.size()[1] != num.size()[1])) {
	// throw new TDataException("matrix dimension error");
	// }
	//
	// re = new double[this.getNumRows()][this.getNumColumns()];
	// im = new double[this.getNumRows()][this.getNumColumns()];
	// for (int row = 0; row < this.getNumRows(); row++) {
	// for (int col = 0; col < this.getNumColumns(); col++) {
	// // (a+b*i)/(c+d*i) = (a*c+b*d)/(c^2+d^2) + (b*c-a*d)/(c^2+d^2)i
	// a = this.re == null ? 0 : this.re.getEntry(row, col);
	// b = this.im == null ? 0 : this.im.getEntry(row, col);
	// c = num.re == null ? 0 : num.re.getEntry(row, col);
	// d = num.im == null ? 0 : num.im.getEntry(row, col);
	// if ((c == 0) && (d == 0)) {
	// throw new TDataException("division by zero");
	// }
	// div = (c * c) + (d * d); // c^2+d^2
	// re[row][col] = ((a * c) + (b * d)) / div;
	// im[row][col] = ((b * c) - (a * d)) / div;
	// }
	// }
	//
	// return new TNumeric(re, im);
	// }

	/**
	 * Invert elements of matrix.
	 * 
	 * @return Inverted matrix
	 * @throws TDataException
	 *             Division by zero
	 */
	public TNumeric invertElem() throws TDataException {
		double[][] re, im;
		double a, b, div;

		re = new double[this.getNumRows()][this.getNumColumns()];
		im = new double[this.getNumRows()][this.getNumColumns()];
		for (int row = 0; row < this.getNumRows(); row++) {
			for (int col = 0; col < this.getNumColumns(); col++) {
				if (this.im == null) {
					re[row][col] = 1.0 / this.re.getEntry(row, col);
				} else if (this.re == null) {
					im[row][col] = -1.0 / this.im.getEntry(row, col);
				} else {
					// 1/(a+b*i) = a/(a^2+b^2) - b/(a^2+b^2)i
					a = this.re == null ? 0 : this.re.getEntry(row, col);
					b = this.im == null ? 0 : this.im.getEntry(row, col);
					if ((a == 0) && (b == 0)) {
						throw new TDataException("division by zero");
					}
					div = (a * a) + (b * b); // a^2+b^2
					re[row][col] = a / div;
					im[row][col] = -b / div;
				}
			}
		}

		return new TNumeric(re, im);
	}
}
