/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import java.awt.image.ImageObserver;
import java.awt.print.PrinterException;
import java.awt.print.PrinterJob;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.MessageFormat;
import java.util.Vector;

import javax.swing.AbstractListModel;
import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTable.PrintMode;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListModel;
import javax.swing.ListSelectionModel;
import javax.swing.event.TableModelListener;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;
import javax.xml.stream.XMLStreamException;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.TPeriodicMatrix;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.dtypes.editor.icons.IconEditor;
import at.lcm.bu21.general.plot.Plot2D;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.Plot2XY;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.general.utils.CloneException;

/**
 * MatrixPanel for displaying TArray and TPeriodixMatrix
 * 
 * @author Christoph Halbmayr
 */
class EditorMatrixPanel implements EditorPanelInterface {

	/** Action Listener */
	private ActionListener al;

	/** Button Vector (for Toolbars) */
	private Vector<JButton> buttons;

	/** column intput field */
	private TextField col;

	/** coliumn count */
	private int columnCount = 0;

	/** column headers */
	private Vector<String> columnHeaders;

	/** TData */
	private TData data;

	/** datatype selection combo box */
	private JComboBox dataTypes;

	/** Key Listener */
	private KeyListener kl;

	/** Matrix Table Model */
	private TableModelMatrixPanel matrixTableModel;

	/** Mouse Listener */
	private MouseListener ml;

	/** toolbar for creating new TData */
	private JToolBar newvar;

	/** JPanel */
	private final JPanel panel;

	/** row intput field */
	private TextField row;

	/** row count */
	private int rowCount = 0;

	/** row Headers */
	private String[] rowHeaders;

	/** Button for TPeriodicMatrix preview */
	private JButton subpPrev;

	/** table */
	private JTable table;

	/** toolbar for navigation/selection... */
	private JToolBar tb;

	/** Frame for TPeriodicMatrix preview */
	private JFrame tpmPrev;

	/**
	 * Creats a Matrix Panel
	 * 
	 * @param data
	 *            data source
	 * @param tml
	 *            Table Model Listener
	 */
	public EditorMatrixPanel(final TData data, final TableModelListener tml) {

		this.buttons = new Vector<JButton>();
		this.panel = new JPanel();
		this.panel.setLayout(new BorderLayout());

		this.data = data;
		this.createHeaders();
		this.createToolbars();

		this.createMatrix();

		if (data.getClass() == TPeriodicMatrix.class) {
			this.subpPrev.setVisible(true);
		}
	}

	/**
	 * Adds Action, Key and Mouse Listener to the Table
	 * 
	 * @param al
	 *            ActionListenerv
	 * @param kl
	 *            KeyListener
	 * @param ml
	 *            MouseListener
	 */
	public void addListener(final ActionListener al, final KeyListener kl, final MouseListener ml) {
		this.table.addKeyListener(kl);
		this.table.addMouseListener(ml);

		for (final JButton i : this.buttons) {
			i.addActionListener(al);
		}

		this.kl = kl;
		this.ml = ml;
		this.al = al;
	}

	/**
	 * Adds a new TData to the Panel
	 */
	public void addNewTData() {

		int row;
		int col;
		try {
			row = Integer.parseInt(this.row.getText());
			col = Integer.parseInt(this.col.getText());
		} catch (final NumberFormatException e1) {
			return;
		}

		final String datatype = this.dataTypes.getSelectedItem().toString();

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
			((TArray) this.data).set(row, col, this.newTData(datatype));
		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			try {
				((TPeriodicMatrix) this.data).set(row - 1, col - 1, new TPeriodic());
			} catch (final Exception e) {
				return;
			}
		}
		this.reload();

	}

	@Override
	public void delete() {
		final int row = this.table.getSelectedRow();
		final int col = this.table.getSelectedColumn();

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {

			if ((row != -1) && (col != -1)) {
				((TArray) this.data).set(row, col, new TNull());
			}

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			try {
				if ((row != -1) && (col != -1)) {
					((TPeriodicMatrix) this.data).set(row, col, new TPeriodic());
				}
			} catch (final CloneException e) {
				e.printStackTrace();
			}
		}
		this.reload();
	}

	/**
	 * deletes the selected column
	 */
	@Override
	public void deleteColumn() {
		final int col = this.table.getSelectedColumn();
		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {

			final TArray temp = (TArray) this.data;
			this.data = new TArray();

			for (int m = 0; m < temp.size()[0]; m++) {
				int index = 0;
				for (int n = 0; n < temp.size()[1]; n++) {
					try {
						if (n == col) {
							index = 1;
							if ((n + index) < temp.size()[1]) {
								((TArray) this.data).set(m, n, temp.get(m, n + index));
							}
						} else if ((n + index) < temp.size()[1]) {
							((TArray) this.data).set(m, n, temp.get(m, n + index));
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {

			final TPeriodicMatrix temp = (TPeriodicMatrix) this.data;
			this.data = new TPeriodicMatrix();

			for (int m = 0; m < temp.size()[0]; m++) {
				int index = 0;
				for (int n = 0; n < temp.size()[1]; n++) {
					try {
						if (n == col) {
							index = 1;
							if ((n + index) < temp.size()[1]) {
								try {
									((TPeriodicMatrix) this.data).set(m, n, temp.get(m, n + index));
								} catch (final CloneException e) {

									e.printStackTrace();
								}
							}
						} else if ((n + index) < temp.size()[1]) {
							try {
								((TPeriodicMatrix) this.data).set(m, n, temp.get(m, n + index));
							} catch (final CloneException e) {

								e.printStackTrace();
							}
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		}
		this.reload();
	}

	/**
	 * deletes the selectet Row
	 */
	@Override
	public void deleteRow() {
		final int row = this.table.getSelectedRow();
		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {

			final TArray temp = (TArray) this.data;
			this.data = new TArray();

			int index = 0;
			for (int m = 0; m < temp.size()[0]; m++) {
				for (int n = 0; n < temp.size()[1]; n++) {
					try {
						if (m == row) {
							index = 1;
							if ((m + index) < temp.size()[0]) {
								((TArray) this.data).set(m, n, temp.get(m + index, n));
							}
						} else if ((m + index) < temp.size()[0]) {
							((TArray) this.data).set(m, n, temp.get(m + index, n));
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {

			final TPeriodicMatrix temp = (TPeriodicMatrix) this.data;
			this.data = new TPeriodicMatrix();

			int index = 0;
			for (int m = 0; m < temp.size()[0]; m++) {
				for (int n = 0; n < temp.size()[1]; n++) {
					try {
						if (m == row) {
							index = 1;
							if ((m + index) < temp.size()[0]) {
								try {
									((TPeriodicMatrix) this.data).set(m, n, temp.get(m + index, n));
								} catch (final CloneException e) {

									e.printStackTrace();
								}
							}
						} else if ((m + index) < temp.size()[0]) {
							try {
								((TPeriodicMatrix) this.data).set(m, n, temp.get(m + index, n));
							} catch (final CloneException e) {

								e.printStackTrace();
							}
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		}
		this.reload();
	}

	/**
	 * @return the panel
	 */
	@Override
	public JPanel getPanel() {
		return this.panel;
	}

	/**
	 * @return selected row
	 */
	public int getSelectedRow() {

		return this.table.getSelectedRow();

	}

	/**
	 * @return selected column
	 */
	public int getSelectedColumn() {

		return this.table.getSelectedColumn();

	}

	@Override
	public TData getSelectedTData()

	{

		TData selection = null;

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
			try {
				if ((this.table.getSelectedRow() != -1) && (this.table.getSelectedColumn() != -1)) {
					selection = ((TArray) this.data).get(this.table.getSelectedRow(), this.table.getSelectedColumn());
				}
			} catch (final TDataException e) {
				e.printStackTrace();
			}
		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			try {
				if ((this.table.getSelectedColumn() != -1) && (this.table.getSelectedRow() != -1)) {
					selection = ((TPeriodicMatrix) this.data).get(this.table.getSelectedRow(),
							this.table.getSelectedColumn());
				}
				if (this.tpmPrev != null) {
					this.tpmPrev.dispose();
				}
			} catch (final TDataException e) {
				e.printStackTrace();
			}
		}
		return selection;

	}

	@Override
	public TData getTData() {
		return this.data;
	}

	@Override
	public void importTData(final File f) {
		TStruct temp = null;

		try {
			temp = (TStruct) TData.read(new FileInputStream(f));
		} catch (final FileNotFoundException e) {
			e.printStackTrace();
		} catch (final TDataException e) {
			e.printStackTrace();
		} catch (final XMLStreamException e) {
			e.printStackTrace();
		} catch (final IOException e) {
			e.printStackTrace();
		}

		this.matrixTableModel.getDataVector().removeAllElements();

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {

			final int rowSize = temp.getKeys().length;
			int columnSize = ((TArray) this.data).size()[1];

			if (columnSize == 0) {
				columnSize = 2;
			}

			this.data = new TArray();

			final String keys[] = temp.getKeys();
			int keyIndex = 0;

			for (int m = 0; m < rowSize; m++) {
				for (int n = 0; n < columnSize; n++) {
					if (keyIndex < keys.length) {
						((TArray) this.data).set(m, n, temp.get(keys[keyIndex]));
						keyIndex++;
					}
				}
			}

			this.matrixTableModel.addTData(this.data);

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			final int rowSize = temp.getKeys().length;
			int columnSize = ((TPeriodicMatrix) this.data).size()[1];

			if (columnSize == 0) {
				columnSize = 2;
			}
			this.data = new TPeriodicMatrix();

			final String keys[] = temp.getKeys();
			int keyIndex = 0;

			for (int m = 0; m < rowSize; m++) {
				for (int n = 0; n < columnSize; n++) {
					if (keyIndex < keys.length) {

						if (temp.get(keys[keyIndex]).getClass().getName()
								.equals("at.lcm.bu21.general.dtypes.TPeriodic")) {
							try {
								((TPeriodicMatrix) this.data).set(m, n, (TPeriodic) temp.get(keys[keyIndex]));
							} catch (final CloneException e) {
								e.printStackTrace();
							}
						}
						keyIndex++;

					}
				}
			}

			this.matrixTableModel.addTData(this.data);
		}
		this.reload();
	}

	/**
	 * inserts a column
	 */
	@Override
	public void insertColumn() {
		final int col = this.table.getSelectedColumn();
		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray") && (col != -1)) {

			TArray temp = null;
			try {
				temp = (TArray) this.data.deepClone();
			} catch (final CloneException e1) {
				e1.printStackTrace();
			}

			for (int m = 0; m < temp.size()[0]; m++) {
				for (int n = 0; n <= temp.size()[1]; n++) {
					try {
						if (n == col) {
							((TArray) this.data).set(m, n, new TNull());
						} else if (n < col) {
							((TArray) this.data).set(m, n, temp.get(m, n));
						}
						if (n > col) {
							((TArray) this.data).set(m, n, temp.get(m, n - 1));
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix") && (col != -1)) {
			TPeriodicMatrix temp = null;
			try {
				temp = (TPeriodicMatrix) this.data.deepClone();
			} catch (final CloneException e1) {
				e1.printStackTrace();
			}

			for (int m = 0; m < temp.size()[0]; m++) {
				for (int n = 0; n <= temp.size()[1]; n++) {
					try {
						if (n == col) {
							((TPeriodicMatrix) this.data).set(m, n, new TPeriodic());
						} else if (n < col) {
							((TPeriodicMatrix) this.data).set(m, n, temp.get(m, n));
						}
						if (n > col) {
							((TPeriodicMatrix) this.data).set(m, n, temp.get(m, n - 1));
						}
					} catch (final Exception e) {
						e.printStackTrace();
					}
				}
			}

		}
		this.reload();
	}

	/**
	 * inserts a Row
	 */
	@Override
	public void insertRow() {
		final int row = this.table.getSelectedRow();
		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray") && (row != -1)) {

			TArray temp = null;
			try {
				temp = (TArray) this.data.deepClone();
			} catch (final CloneException e1) {
				e1.printStackTrace();
			}

			for (int m = 0; m <= temp.size()[0]; m++) {
				for (int n = 0; n < temp.size()[1]; n++) {
					try {

						if (m == row) {
							((TArray) this.data).set(m, n, new TNull());
						} else if (m < row) {
							((TArray) this.data).set(m, n, temp.get(m, n));
						}
						if (m > row) {
							((TArray) this.data).set(m, n, temp.get(m - 1, n));
						}
					} catch (final TDataException e) {

						e.printStackTrace();
					}
				}
			}

		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix") && (row != -1)) {
			TPeriodicMatrix temp = null;
			try {
				temp = (TPeriodicMatrix) this.data.deepClone();
			} catch (final CloneException e1) {
				e1.printStackTrace();
			}

			for (int m = 0; m <= temp.size()[0]; m++) {
				for (int n = 0; n < temp.size()[1]; n++) {
					try {
						if (m == row) {
							((TPeriodicMatrix) this.data).set(m, n, new TPeriodic());
						} else if (m < row) {
							((TPeriodicMatrix) this.data).set(m, n, temp.get(m, n));
						}
						if (m > row) {
							((TPeriodicMatrix) this.data).set(m, n, temp.get(m - 1, n));
						}
					} catch (final Exception e) {

						e.printStackTrace();
					}
				}
			}

		}
		this.reload();

	}

	/**
	 * Prints out the whole Editor View
	 */
	@Override
	public void printme() {
		final String TDataType = this.data.getClass().getName().split("\\.")[5];
		final PrinterJob printJob = PrinterJob.getPrinterJob();
		final MessageFormat headerFormat = new MessageFormat(TDataType + " Page {0}");
		final MessageFormat footerFormat = new MessageFormat("- {0} -");
		printJob.setPrintable(this.table.getPrintable(PrintMode.NORMAL, headerFormat, footerFormat));
		if (printJob.printDialog()) {
			try {
				printJob.print();
			} catch (final PrinterException pe) {
			}
		}
	}

	@Override
	public void reload() {
		this.panel.removeAll();
		this.createToolbars();
		this.createHeaders();
		this.createMatrix();

		if (this.data.getClass() == TPeriodicMatrix.class) {
			this.subpPrev.setVisible(true);
		}

	}

	@Override
	public void saveTData(final File f) {
		final int row = this.table.getSelectedRow();
		final int col = this.table.getSelectedColumn();

		final TStruct saving = new TStruct();

		if ((row == -1) || (col == -1)) {
			saving.put(this.data.getClass().getName(), this.data);
		} else if ((row >= 0) && (col >= 0)) {
			if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
				try {
					saving.put(this.data.getClass().getName(), ((TArray) this.data).get(row, col));
				} catch (final TDataException e) {
					e.printStackTrace();
				}
			} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
				try {
					saving.put(this.data.getClass().getName(), ((TPeriodicMatrix) this.data).get(row, col));
				} catch (final TDataException e) {
					e.printStackTrace();
				}
			}
		}

		try {
			saving.write(new FileOutputStream(f));
		} catch (final FileNotFoundException e) {
			e.printStackTrace();
		} catch (final XMLStreamException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Shows TPeriodic Matrix Preview
	 */
	public void showTPMPreview() {

		final TPeriodicMatrix tpm = (TPeriodicMatrix) this.data;

		final int xlength = tpm.size()[0];
		final int ylength = tpm.size()[1];

		// Style style = new Style();
		// style.setLineVisible(true);
		// style.setMarkerVisible(false);
		// style.setMarkerFilled(true);

		final SubPlot sub = new SubPlot(xlength, ylength);

		final Plot2D[][] plots = new Plot2D[xlength][ylength];

		for (int x = 0; x < xlength; x++) {
			for (int y = 0; y < ylength; y++) {

				final Plot2XY plot = new Plot2Line();
				try {
					plot.addSignal("tpm", tpm.get(x, y), 500);
					// FIXME style
					// plot.setStyle(0, style);
				} catch (final Exception e) {
					e.printStackTrace();
				}
				plots[x][y] = plot;
			}
		}

		sub.setPlots(plots);
		this.tpmPrev = new JFrame("TPeriodicMartrix preview");

		this.tpmPrev.setIconImage(IconEditor.ICON_SUBPLOT.getImage());

		this.tpmPrev.setSize(700, 700);
		this.tpmPrev.setVisible(true);
		this.tpmPrev.setLayout(new BorderLayout());
		this.tpmPrev.add(sub.getPanel());
		this.tpmPrev.repaint();

	}

	@Override
	public void updateChangedSignal(final TData selection) {

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
			for (int m = 0; m < ((TArray) this.data).size()[0]; m++) {
				for (int n = 0; n < ((TArray) this.data).size()[1]; n++) {
					try {
						if ((m == this.table.getSelectedRow()) && (n == this.table.getSelectedColumn())) {
							((TArray) this.data).set(m, n, selection);
						}

					} catch (final Exception e) {
					}
				}
			}
		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			for (int m = 0; m < ((TPeriodicMatrix) this.data).size()[0]; m++) {
				for (int n = 0; n < ((TPeriodicMatrix) this.data).size()[1]; n++) {
					try {
						if ((m == this.table.getSelectedRow()) && (n == this.table.getSelectedColumn())) {
							((TPeriodicMatrix) this.data).set(m, n, (TPeriodic) selection);
						}

					} catch (final Exception e) {
					}
				}
			}
		}
		this.matrixTableModel.getDataVector().clear();
		this.matrixTableModel.addTData(this.data);
	}

	/**
	 * creats the table headers
	 */
	private void createHeaders() {

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
			this.columnCount = ((TArray) this.data).size()[1];
			this.rowCount = ((TArray) this.data).size()[0];
		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			this.columnCount = ((TPeriodicMatrix) this.data).size()[1];
			this.rowCount = ((TPeriodicMatrix) this.data).size()[0];
		}

		this.columnHeaders = new Vector<String>();
		if ((this.columnCount > 0) && (this.rowCount > 0)) {

			this.rowHeaders = new String[this.rowCount];
			for (int n = 0; n < this.columnCount; n++) {
				this.columnHeaders.add("" + n);
			}

			for (int i = 0; i < this.rowCount; i++) {
				this.rowHeaders[i] = "" + i;
			}
		} else {

			this.columnHeaders.add("");
			this.columnHeaders.add("");
		}

	}

	/**
	 * creats the matrix
	 */
	private void createMatrix() {

		this.matrixTableModel = new TableModelMatrixPanel(this.columnHeaders, 0);

		if (this.data != null) {
			this.matrixTableModel.addTData(this.data);
			this.table = new JTable(this.matrixTableModel);

			this.table.setDefaultEditor(Object.class, new DefaultCellEditor(new JTextField()) {

				/** Serialization version number */
				private static final long serialVersionUID = 1L;

				@Override
				public Component getTableCellEditorComponent(final JTable table, final Object value,
						final boolean isSelected, final int row, final int column) {
					Component c;
					c = super.getTableCellEditorComponent(table, value, isSelected, row, column);
					((JTextField) c).selectAll();
					return c;
				}
			});

			if ((this.columnCount > 0) && (this.rowCount > 0)) {
				this.createTable();
			}
			this.table.addKeyListener(this.kl);
			this.table.addMouseListener(this.ml);
		}

	}

	/**
	 * creats the Tabel
	 */
	private void createTable() {
		this.table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		this.table.setColumnSelectionAllowed(true);
		this.table.setRowSelectionAllowed(true);

		this.table.setAutoResizeMode(0);

		this.table.getTableHeader().setReorderingAllowed(false);
		this.table.getTableHeader().setResizingAllowed(false);

		final ListModel lm = new AbstractListModel() {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Object getElementAt(final int index) {
				return EditorMatrixPanel.this.rowHeaders[index];
			}

			@Override
			public int getSize() {
				return EditorMatrixPanel.this.rowHeaders.length;
			}
		};

		final JList rowHeader = new JList(lm);
		rowHeader.setFixedCellWidth(25);

		rowHeader.setFixedCellHeight(this.table.getRowHeight());
		rowHeader.setCellRenderer(new EditorPanelRowRenderer(this.table));

		this.table.getBackground().getRGB();

		rowHeader.setBackground(Color.gray.brighter());

		final JScrollPane scroll = new JScrollPane(this.table);
		scroll.setRowHeaderView(rowHeader);
		scroll.setBackground(Color.gray.brighter());

		this.panel.add(scroll);
		this.panel.repaint();
		this.panel.revalidate();
		this.ColumsResize(this.table, 90);

		this.table.getTableHeader().setReorderingAllowed(false);
		this.ColumsResize(this.table, 90);
	}

	/**
	 * creats the toolbars
	 */
	private void createToolbars() {

		this.tb = new JToolBar();
		this.tb.setPreferredSize(new Dimension(ImageObserver.WIDTH, 25));
		this.buttons = new Vector<JButton>();

		final JButton up = new JButton(IconEditor.ICON_UP);
		up.setActionCommand("up");
		up.setToolTipText("move one level up");
		this.buttons.add(up);

		final ActionListener alis = new ActionListener() {

			@Override
			public void actionPerformed(final ActionEvent e) {
				if (EditorMatrixPanel.this.tpmPrev != null) {
					EditorMatrixPanel.this.tpmPrev.dispose();
				}

			}
		};

		up.addActionListener(alis);

		final JButton impor = new JButton(IconEditor.ICON_IMPORT);
		impor.setActionCommand("Import Data");
		impor.setToolTipText("import data");
		this.buttons.add(impor);

		final JButton save = new JButton(IconEditor.ICON_SAVE);
		save.setActionCommand("Save Data");
		save.setToolTipText("save data");
		this.buttons.add(save);

		final JButton del = new JButton(IconEditor.ICON_REMOVE);
		del.setActionCommand("Delete");
		del.setToolTipText("delete selected signal");
		this.buttons.add(del);

		final JButton print = new JButton(IconEditor.ICON_PRINTER);
		print.setActionCommand("Print");
		print.setToolTipText("print");
		this.buttons.add(print);

		final JButton os = new JButton(IconEditor.ICON_OpenSelection);
		os.setActionCommand("Open Selection");
		os.setToolTipText("open selection");
		this.buttons.add(os);

		final JButton addRow = new JButton("Row", IconEditor.ICON_ADD);
		addRow.setToolTipText("insert a row ");
		addRow.setActionCommand("addRow");
		this.tb.add(addRow);
		this.buttons.add(addRow);

		final JButton delRow = new JButton("Row", IconEditor.ICON_REMOVE);
		delRow.setToolTipText("delete selected row");
		this.tb.add(delRow);
		this.buttons.add(delRow);

		final JButton addCol = new JButton("Column", IconEditor.ICON_ADD);
		addCol.setToolTipText("insert a column");
		addCol.setActionCommand("addCol");
		this.tb.add(addCol);
		this.buttons.add(addCol);

		final JButton delCol = new JButton("Column", IconEditor.ICON_REMOVE);
		delCol.setToolTipText("delete selected column");
		this.tb.add(delCol);
		this.buttons.add(delCol);

		this.subpPrev = new JButton(IconEditor.ICON_SUBPLOT);
		this.subpPrev.setToolTipText("plot subplot preview");
		this.subpPrev.setActionCommand("TPMPreview");
		this.subpPrev.setVisible(false);
		this.tb.add(this.subpPrev);
		this.buttons.add(this.subpPrev);

		this.tb.setVisible(true);
		this.panel.add(this.tb, BorderLayout.NORTH);
		this.panel.repaint();
		this.panel.revalidate();

		this.newvar = new JToolBar();

		final JButton add = new JButton("SET");
		add.addActionListener(alis);

		this.dataTypes = new JComboBox();

		if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TArray")) {
			this.dataTypes.addItem("TNull");
			this.dataTypes.addItem("TArray");
			this.dataTypes.addItem("TBoolean");
			this.dataTypes.addItem("TNumeric");
			this.dataTypes.addItem("TPeriodic");
			this.dataTypes.addItem("TPeriodicMatrix");
			this.dataTypes.addItem("TString");
			this.dataTypes.addItem("TStruct");
		} else if (this.data.getClass().getName().equals("at.lcm.bu21.general.dtypes.TPeriodicMatrix")) {
			this.dataTypes.addItem("TNull");
		}

		this.buttons.add(add);
		this.row = new TextField("row");
		this.col = new TextField("column");

		this.row.addFocusListener(new java.awt.event.FocusAdapter() {
			@Override
			public void focusGained(final java.awt.event.FocusEvent evt) {
				EditorMatrixPanel.this.row.selectAll();
			}
		});

		this.col.addFocusListener(new java.awt.event.FocusAdapter() {
			@Override
			public void focusGained(final java.awt.event.FocusEvent evt) {
				EditorMatrixPanel.this.col.selectAll();
			}
		});

		this.row.setEditable(true);
		this.col.setEditable(true);

		for (final JButton i : this.buttons) {
			this.tb.add(i);
			i.addActionListener(this.al);

		}

		this.newvar.setPreferredSize(new Dimension(ImageObserver.WIDTH, 25));
		this.newvar.add(this.row);
		this.newvar.add(this.col);
		this.newvar.add(this.dataTypes);
		this.newvar.add(add);

		this.panel.add(this.newvar, BorderLayout.SOUTH);

	}

	/**
	 * creats a new TData
	 * 
	 * @param datatype
	 * @return TData
	 */
	private TData newTData(final String datatype) {
		TData data = null;
		try {
			@SuppressWarnings("rawtypes")
			final Class cls = Class.forName("at.lcm.bu21.general.dtypes." + datatype);

			final Object o = cls.newInstance();

			data = (TData) o;

		} catch (final Exception e) {
			if (datatype.equals("TBoolean")) {
				data = new TBoolean(false);
			}
			if (datatype.equals("TString")) {
				data = new TString("");
			}

			if (datatype.equals("TNumeric")) {
				try {
					data = new TNumeric("0,0");
				} catch (final TDataException e1) {
					e1.printStackTrace();
				}
			}

		}
		return data;
	}

	/**
	 * fits column size to the column content
	 * 
	 * @param t
	 *            JTable
	 * @param minimumSize
	 *            minimum column size
	 */
	protected void ColumsResize(final JTable t, final int minimumSize) {
		final JTableHeader header = t.getTableHeader();
		TableCellRenderer defaultHeaderRenderer = null;

		if (header != null) {
			defaultHeaderRenderer = header.getDefaultRenderer();
		}

		final TableColumnModel columns = t.getColumnModel();
		final TableModel data = t.getModel();
		final int margin = columns.getColumnMargin();
		final int rowCount = data.getRowCount();

		for (int i = columns.getColumnCount() - 1; i >= 0; --i) {
			final javax.swing.table.TableColumn column = columns.getColumn(i);
			final int columnIndex = column.getModelIndex();
			int width = -1;

			TableCellRenderer h = column.getHeaderRenderer();

			if (h == null) {
				h = defaultHeaderRenderer;
			}

			if (h != null) {
				final Component c = h.getTableCellRendererComponent(t, column.getHeaderValue(), false, false, -1, i);
				width = c.getPreferredSize().width;

			}

			for (int row = rowCount - 1; row >= 0; --row) {
				final TableCellRenderer r = t.getCellRenderer(row, i);
				final Component c = r.getTableCellRendererComponent(t, data.getValueAt(row, columnIndex), false, false,
						row, i);
				width = Math.max(width, c.getPreferredSize().width);
			}

			if (width >= 0) {
				if ((width + margin) > minimumSize) {
					column.setPreferredWidth(width + margin);
				} else {
					column.setPreferredWidth(minimumSize);
				}
			}
		}
	}
}
