/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1002 $
 * $Date: 2020-04-10 23:31:24 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.general.dtypes.editor2;

import java.util.Arrays;

import javax.swing.table.AbstractTableModel;

import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.general.dtypes.editor2.TDataModel.TStructModel;

/**
 * This is a table model for a JTable to view TStructs. It based on
 * AbstractTableModel.
 * 
 * @author Johannes Kaisinger
 */
public class TStructTableModel extends AbstractTableModel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**  */
	private TStructKeyListener tStructKeyListener;

	/** Column header names. */
	private String[] colNames = { "Name:", "Info:" };

	/**
	 * Local reference of the tStructModel. It allows to manipulate the TStruct.
	 */
	private TStructModel tStructModel;

	/**
	 * Creates an object
	 * 
	 * @param tStructModel
	 *            The TStructModel
	 */
	public TStructTableModel(TStructModel tStructModel) {
		this.tStructModel = tStructModel;
	}

	/**
	 * Set tStructModel
	 * 
	 * @param tStructModel
	 *            The TStructModel
	 */
	public void setData(TStructModel tStructModel) {
		this.tStructModel = tStructModel;
	}

	/**
	 * Set the structKeyListener
	 * 
	 * @param tStructKeyListener
	 *            TStructKeyListener
	 */
	public void setTStructKeyListener(TStructKeyListener tStructKeyListener) {
		this.tStructKeyListener = tStructKeyListener;
	}

	/**
	 * Creates an object
	 * 
	 * @param column
	 *            Index of column.
	 * @return name The name of the column
	 */
	@Override
	public String getColumnName(int column) {
		return this.colNames[column];
	}

	/**
	 * Get row count of tArrayModel.
	 * 
	 * @return rowCount
	 */
	@Override
	public int getRowCount() {
		return this.tStructModel.getKeys().length;
	}

	/**
	 * Get column count of tArrayModel.
	 * 
	 * @return columnCount
	 */
	@Override
	public int getColumnCount() {
		return 2;
	}

	/**
	 * Get column class.
	 * 
	 * @param column
	 *            Index of column.
	 * @return Class
	 */
	@Override
	public Class<String> getColumnClass(int column) {
		return String.class;
	}

	/**
	 * Get if a cell is editable.
	 * 
	 * @param row
	 *            Index of row.
	 * @param column
	 *            Index of column.
	 * @return isEditable
	 */
	@Override
	public boolean isCellEditable(int row, int column) {

		boolean isEditable;

		if (column == 0) {
			isEditable = true;
		} else {
			isEditable = false;
		}

		return isEditable;
	}

	/**
	 * Get a cell value.
	 * 
	 * @param row
	 *            Index of row.
	 * @param column
	 *            Index of column.
	 * @return value Value of cell (String).
	 */
	@Override
	public String getValueAt(int row, int column) {

		TStruct tStruct = this.tStructModel.getTStruct();
		String[] allKeys = tStruct.getKeys();

		Arrays.sort(allKeys, String.CASE_INSENSITIVE_ORDER);

		String name = "";
		String key = allKeys[row];

		if (column == 0) {
			name = key;
		} else if (column == 1) {
			name = this.tStructModel.getValue(key);
		}

		return name;
	}

	/**
	 * Set a cell value.
	 * 
	 * @param value
	 *            New value of cell.
	 * @param row
	 *            Index of row.
	 * @param column
	 *            Index of column.
	 */
	@Override
	public void setValueAt(Object value, int row, int column) {
		if (column == 0) {
			String oldKey = this.getValueAt(row, column);
			String newKey = value.toString();
			this.tStructModel.changeKey(oldKey, newKey);
			this.fireTableDataChanged();

			TStructKeyChangedEvent evt = new TStructKeyChangedEvent(this, newKey);

			if (this.tStructKeyListener != null) {
				this.tStructKeyListener.keyChangedOccurred(evt);
			}
		}
	}
}
