/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.xy.XYDataset;

import at.lcm.bu21.general.utils.MathUtils;

/**
 * Histogram plot. Draw distribution of y-values
 * 
 * @author Werner Koppelsttter
 */
public class Plot2Hist extends Plot2XY {

	/** Number of intervals */
	private int numIntervals = 10;

	/**
	 * Create histogram-plot
	 */
	public Plot2Hist() {
		super();

		this.getDefaultStyle().setSignalStyle(SignalStyle.Step);
		this.getDefaultStyle().setMarkerFilled(true);
		this.getDefaultSelectStyle().setSignalStyle(SignalStyle.Step);
		this.getDefaultSelectStyle().setMarkerFilled(true);
	}

	/**
	 * Translate values to values for histogram
	 * 
	 * @param values
	 *            Signal to display
	 * @param style
	 *            Style of plot
	 * @return Histogram values
	 */
	private double[][] translate(double[][] values, Style style) {
		double[][] newVal;
		double area;

		area = (this.xLimitMax - this.xLimitMin) / this.numIntervals;

		// define array
		if (style.getSignalStyle() == SignalStyle.Step) {
			newVal = new double[3][this.numIntervals + 3];
		} else {
			newVal = new double[3][this.numIntervals + 2];
		}

		// init x-values
		for (int k = -1; k < (newVal[0].length - 1); k++) {
			newVal[0][k + 1] = this.xLimitMin + (k * area);
		}

		// fill y-values (number of occurs)
		for (double val : values[1]) {
			int index;
			if (val == this.xLimitMax) {
				index = this.numIntervals;
			} else if (val < this.xLimitMin) {
				index = 0;
			} else if (val > this.xLimitMax) {
				index = this.numIntervals + 1;
			} else {
				index = ((int) Math.floor((val - this.xLimitMin) / area)) + 1;
			}
			newVal[1][index]++;
		}

		// add last value for step stlye
		if (style.getSignalStyle() == SignalStyle.Step) {
			newVal[1][this.numIntervals + 2] = newVal[1][this.numIntervals + 1];
		}

		return newVal;
	}

	/**
	 * Get dataset of values
	 * 
	 * @param name
	 *            Name of dataset
	 * @param style
	 *            Style of dataset
	 * @param values
	 *            Values of dataset
	 * @return Dataset of values
	 */
	@Override
	protected XYDataset getDataset(String name, Style style, double[][] values) {
		XYDataset ds = null;

		if (values[0].length > 0) {
			ds = super.getDataset(name, style, this.translate(values, style));
		}

		return ds;
	}

	@Override
	void redraw() {
		int sigs;

		// draw signal
		sigs = this.getNumSignals();
		for (int k = 0; k < sigs; k++) {
			Signal sig;
			Style style;
			XYItemRenderer rend;

			sig = this.getSignal(k);

			// values
			style = sig.getStyle();
			this.plot.setRenderer(k + 1, this.getRenderer(style));
			this.plot.setDataset(k + 1,
					this.getDataset(sig.getName(), style, sig.getValues()));

			// selected values
			style = sig.getSelectStyle();
			rend = this.getRenderer(style);
			rend.setSeriesVisibleInLegend(0, false);
			this.plot.setRenderer(k + 1 + sigs, rend);
			this.plot.setDataset(
					k + 1 + sigs,
					this.getDataset(sig.getName() + "_Selected", style,
							sig.getSelectedValues()));
		}

		// remove old datasets
		for (int k = 1 + (2 * sigs); k < this.plot.getDatasetCount(); k++) {
			this.plot.setRenderer(k, null);
			this.plot.setDataset(k, null);
		}

		this.updateBounds();

		this.plot.getDomainAxis().setLabel(this.getXLabel());
		this.plot.getRangeAxis().setLabel(this.getYLabel());
	}

	@Override
	public void zoom() {
		final int min = 0;
		final int max = 1;
		int sigs;

		// x limits
		if (!this.xLimitSet) {
			double[] limitsX = new double[] { 0.0, 1.0 };

			sigs = this.getNumSignals();
			if (this.getNumSignals() > 0) {
				double[] minX, maxX;

				minX = new double[sigs];
				maxX = new double[sigs];

				for (int k = 0; k < sigs; k++) {
					Signal sig = this.getSignal(k);
					minX[k] = sig.getMinY();
					maxX[k] = sig.getMaxY();
				}

				limitsX[min] = MathUtils.min(minX);
				limitsX[max] = MathUtils.max(maxX);
			}

			// limitsX = MathUtils.roundLimits(limitsX);

			this.xLimitMin = limitsX[0];
			this.xLimitMax = limitsX[1];
		}
		this.xBoundMin = this.xLimitMin;
		this.xBoundMax = this.xLimitMax;

		// y limits
		if (!this.yLimitSet) {
			double[] limitsY = new double[] { 0.0, 1.0 };

			sigs = this.getNumSignals();
			if (this.getNumSignals() > 0) {
				double maxY;

				maxY = Double.MIN_VALUE;

				for (int k = 0; k < sigs; k++) {
					Signal sig = this.getSignal(k);
					double[][] values = this.translate(sig.getValues(),
							sig.getStyle());
					for (int n = 1; n <= this.numIntervals; n++) {
						if (values[1][n] > maxY) {
							maxY = values[1][n];
						}
					}
				}

				limitsY[min] = 0.0;
				limitsY[max] = maxY;
			}

			// limitsY = MathUtils.roundLimits(limitsY);

			this.yLimitMin = limitsY[0];
			this.yLimitMax = limitsY[1];

		}
		this.yBoundMin = this.yLimitMin;
		this.yBoundMax = this.yLimitMax;

		this.redraw();
		// this.updateBounds();
	}

	@Override
	void select(int signal, double xmin, double xmax, double ymin, double ymax,
			boolean select) {
		double[] y;
		LinkPlot link;
		Signal sig;

		link = this.getLink();
		sig = this.getSignal(signal);
		if (sig != null) {
			y = sig.getYRef();

			for (int k = 0; k < y.length; k++) {
				if ((xmin <= y[k]) && (y[k] <= xmax)) {
					link.select(signal, k, select);
				}
			}
		}
	}

	/**
	 * Get number of intervals
	 * 
	 * @return Number of intervals
	 */
	public int getNumIntervals() {
		return this.numIntervals;
	}

	/**
	 * Set number of intervals
	 * 
	 * @param numIntervals
	 *            Number of intervals
	 */
	public void setNumIntervals(int numIntervals) {
		this.numIntervals = numIntervals;
		this.redraw();
	}

	@Override
	public void setXLimit(double min, double max) {
		super.setXLimit(min, max);
		this.zoom();
	}

	@Override
	public void setYLimit(double min, double max) {
		super.setYLimit(min, max);
		this.zoom();
	}

}
