/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 997 $
 * $Date: 2020-03-19 13:49:34 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.general.plot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.SwingConstants;

import org.freehep.graphicsio.AbstractVectorGraphicsIO;
import org.freehep.graphicsio.pdf.PDFGraphics2D;
import org.freehep.graphicsio.ps.PSGraphics2D;
import org.freehep.graphicsio.svg.SVGGraphics2D;

import at.lcm.bu21.general.utils.JVLabel;

/**
 * Class for Creating SubPlots zoom
 * 
 * @author Christoph Halbmayr
 * @author Werner Koppelsttter
 */

public class SubPlot extends Plot {

	/** Plot Array - contains all Plots */
	private Plot[][] plots;

	/** Grid panel for all plots */
	private JPanel grid = new JPanel();

	/** Label for title */
	private JLabel titleLabel;

	/** Label for x-axis */
	private JLabel xLabel;

	/** Label for y-axis */
	private JLabel yLabel;

	/**
	 * Creates a new SubPlot
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of columns
	 */
	public SubPlot(int row, int col) {
		super();
		JPanel panel;

		// title
		this.titleLabel = new JLabel();
		this.titleLabel.setHorizontalAlignment(SwingConstants.CENTER);
		this.titleLabel.setAutoscrolls(true);
		this.titleLabel.setFont(new Font("Helvetica", Font.PLAIN, 20));
		this.titleLabel.setBackground(Color.WHITE);
		this.titleLabel.setOpaque(true);

		// x-label
		this.xLabel = new JLabel();
		this.xLabel.setHorizontalAlignment(SwingConstants.CENTER);
		this.xLabel.setAutoscrolls(true);
		this.xLabel.setFont(new Font("Helvetica", Font.PLAIN, 20));
		this.xLabel.setBackground(Color.WHITE);
		this.xLabel.setOpaque(true);

		// y-label
		this.yLabel = new JVLabel();
		// this.yLabel = new JLabelRotate();
		// ((JLabelRotate) this.yLabel).setRotation(-90.0);
		this.yLabel.setHorizontalAlignment(SwingConstants.CENTER);
		this.yLabel.setVerticalAlignment(SwingConstants.CENTER);
		this.yLabel.setAutoscrolls(true);
		this.yLabel.setFont(new Font("Helvetica", Font.PLAIN, 20));
		this.yLabel.setBackground(Color.WHITE);
		this.yLabel.setOpaque(true);

		this.plots = new Plot[row][col];
		panel = this.getPlotPanel();
		panel.add(this.grid, BorderLayout.CENTER);
		this.addPlots();
	}

	/**
	 * Get plot of positon
	 * 
	 * @param row
	 *            Row index
	 * @param col
	 *            Column index
	 * @return Plot Plot
	 */
	public Plot getPlot(int row, int col) {
		return this.plots[row][col];
	}

	/**
	 * Sets one particular Plot
	 * 
	 * @param row
	 *            Row position
	 * @param col
	 *            Column position
	 * @param plot
	 *            plot
	 */
	public void setPlot(int row, int col, Plot plot) {
		this.plots[row][col] = plot;
		plot.setParent(this);
		this.addPlots();
	}

	/**
	 * Sets the plot array which contains all plots
	 * 
	 * @param plots
	 *            Array contains all plots
	 */
	public void setPlots(Plot[][] plots) {
		this.plots = plots;
		this.setSize(plots.length, plots[0].length);
	}

	/**
	 * Return the size of the SubPlot (row,col)
	 * 
	 * @return Size of SubPlot
	 */
	public int[] getSize() {
		return new int[] { this.plots.length, this.plots[0].length };
	}

	/**
	 * Sets the Size of the SubPlot
	 * 
	 * @param row
	 *            Number of rows
	 * @param col
	 *            Number of columns
	 */
	public void setSize(int row, int col) {
		Plot[][] oldPlots;

		oldPlots = this.plots;
		this.plots = new Plot[row][col];

		for (int r = 0; r < row; r++) {
			for (int c = 0; c < col; c++) {
				if ((r < oldPlots.length) && (c < oldPlots[r].length)) {
					this.plots[r][c] = oldPlots[r][c];
					if (this.plots[r][c] != null) {
						this.plots[r][c].setParent(this);
					}
				} else {
					this.plots[r][c] = null;
					if (this.plots[r][c] != null) {
						this.plots[r][c].setParent(null);
					}
				}

			}
		}
		this.addPlots();
	}

	/**
	 * Adds all plots to the local panel
	 */
	private void addPlots() {
		int rows, cols;
		JPanel p;

		rows = this.getSize()[0];
		cols = this.getSize()[1];

		this.grid.removeAll();
		this.grid.setLayout(new GridLayout(rows, cols));

		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				if (this.plots[r][c] == null) {
					p = new JPanel();
					p.setBackground(Color.white);
				} else {
					p = this.plots[r][c].getPanel();
				}
				this.grid.add(p);
			}
		}
		this.redraw();
	}

	@Override
	void select(int signal, int index, boolean select) {
	}

	@Override
	public void reset() {
		for (Plot[] row : this.plots) {
			for (Plot plot : row) {
				if (plot != null) {
					plot.reset();
				}
			}
		}
	}

	@Override
	void repaint() {
		for (Plot[] row : this.plots) {
			for (Plot plot : row) {
				if (plot != null) {
					plot.repaint();
				}
			}
		}
	}

	@Override
	void redraw() {
		// title
		if (this.getTitle() == null) {
			this.getPlotPanel().remove(this.titleLabel);
		} else {
			this.titleLabel.setText(this.getTitle());
			this.getPlotPanel().add(this.titleLabel, BorderLayout.NORTH);
		}

		// x-label
		if (this.getXLabel() == null) {
			this.getPlotPanel().remove(this.xLabel);
		} else {
			this.xLabel.setText(this.getXLabel());
			this.getPlotPanel().add(this.xLabel, BorderLayout.SOUTH);
		}

		// y-label
		if (this.getYLabel() == null) {
			this.getPlotPanel().remove(this.yLabel);
		} else {
			this.yLabel.setText(this.getYLabel());
			this.getPlotPanel().add(this.yLabel, BorderLayout.WEST);
		}

		this.getPlotPanel().validate();

		for (Plot[] row : this.plots) {
			for (Plot plot : row) {
				if (plot != null) {
					plot.redraw();
				}
			}
		}
	}

	@Override
	protected BufferedImage getBufferedImage() {
		int rows, cols;
		BufferedImage im;
		Plot plot;
		JImagePanel[][] images;

		rows = this.plots.length;
		cols = this.plots[0].length;
		images = new JImagePanel[rows][cols];

		// display images
		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				if (this.plots[r][c] != null) {
					plot = this.plots[r][c];
					images[r][c] = new JImagePanel(plot.getBufferedImage());
					images[r][c].setSize(plot.getPlotPanel().getSize());
					plot.getPanel().remove(plot.getPlotPanel());
					plot.getPanel().add(images[r][c], BorderLayout.CENTER);
				}
			}
		}

		// get image
		im = super.getBufferedImage();

		// restore structure
		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				if (this.plots[r][c] != null) {
					plot = this.plots[r][c];
					plot.getPanel().remove(images[r][c]);
					plot.getPanel().add(plot.getPlotPanel(),
							BorderLayout.CENTER);
				}
			}
		}

		return im;
	}

	@Override
	public void save(File f, int width, int height) throws IOException {
		String fileType = "";
		int index;
		AbstractVectorGraphicsIO g2d = null;
		Dimension dim;

		index = f.getName().lastIndexOf('.');
		if (index > 0) {
			fileType = f.getName().substring(index + 1);
		}

		if (fileType.equalsIgnoreCase("eps")
				|| fileType.equalsIgnoreCase("svg")
				|| fileType.equalsIgnoreCase("pdf")) {
			if (this.isVectorGraphicPossible()) {
				// Color c = (Color) this.chart.getBackgroundPaint();
				//
				// this.chart.setBackgroundPaint(Color.WHITE);

				OutputStream os;
				os = new FileOutputStream(f);
				dim = new Dimension(width, height);
				if (fileType.equalsIgnoreCase("eps")) {
					g2d = new PSGraphics2D(os, dim);
				} else if (fileType.equalsIgnoreCase("svg")) {
					g2d = new SVGGraphics2D(os, dim);
				} else if (fileType.equalsIgnoreCase("pdf")) {
					g2d = new PDFGraphics2D(os, dim);
				}
				Rectangle2D rect = new Rectangle2D.Double(0, 0, width, height);

				g2d.startExport();
				int[] size = this.getSize();
				for (int r = 0; r < size[0]; r++) {
					for (int c = 0; c < size[1]; c++) {
						if (this.getPlot(r, c) != null) {
							rect = new Rectangle2D.Double(
									(width / size[1]) * c, (height / size[0])
											* r, width / size[1], height
											/ size[0]);
							((Plot2D) this.getPlot(r, c)).chart.draw(g2d, rect);
						}
					}
				}
				// this.draw(g2d, rect);
				g2d.endExport();
				os.close();
				// this.chart.setBackgroundPaint(c);
			} else {
				super.save(f, width, height);
			}
		} else {
			super.save(f, width, height);
		}
	}

	@Override
	protected void setParent(SubPlot parent) {
		int rows, cols;

		super.setParent(parent);

		rows = this.getSize()[0];
		cols = this.getSize()[1];

		for (int r = 0; r < rows; r++) {
			for (int c = 0; c < cols; c++) {
				if (this.plots[r][c] != null) {
					this.plots[r][c].setParent(this);
				}
			}
		}
	}

	@Override
	protected void addPopupMenu(JMenu popup) {
		JMenuItem tmp;

		super.addPopupMenu(popup);

		for (int k = 0; k < popup.getItemCount(); k++) {
			tmp = popup.getItem(k);
			if (tmp != null) {
				if (tmp.getText().equals("Clear Selection")
						|| tmp.getText().equals("Suppress")) {
					tmp.setVisible(false);
				}
			}
		}
	}

	@Override
	protected boolean isVectorGraphicPossible() {
		boolean possible = true;
		int[] size;

		size = this.getSize();
		for (int r = 0; r < size[0]; r++) {
			for (int c = 0; c < size[1]; c++) {
				possible &= (this.getPlot(r, c) == null)
						|| (this.getPlot(r, c).isVectorGraphicPossible());
			}
		}
		return possible;
	}

}
