/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 1031 $
 * $Date: 2020-09-01 13:19:19 +0200 (Tue, 01 Sep 2020) $
 */
package at.lcm.bu21.general.utils;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.LinkedList;

import javax.swing.JOptionPane;

/**
 * TexWriter for writing tex-reports
 * 
 * @author Werner Koppelsttter
 */
public class TexWriter {

	/** System new-line characters */
	static final String NEW_LINE = System.getProperty("line.separator");

	/** Content of the file */
	protected StringBuffer content = new StringBuffer();

	/** Directory where the tex-file should write */
	protected File file = null;

	/**
	 * Main function of class
	 * 
	 * @param args Name of files to extract
	 */
	public static void main(String[] args) {

		// evaluate parameter
		if (args.length == 0) {
			System.out.println("USAGE:");
			System.out.println("\tGeneral.jar <file>");
			System.out.println();
		} else {
			for (String str : args) {
				TexWriter.extractBody(str);
			}
		}

	}

	/**
	 * Extract body of tex-file
	 * 
	 * @param file File to convert
	 */
	static protected void extractBody(String file) {
		TexWriter tex;
		int index;
		File source, target;
		String errMsg = null;

		index = file.toLowerCase().indexOf(".source.");
		if (file.toLowerCase().endsWith(".tex") && (index != -1)) {
			source = new File(file);
			target = new File(file.substring(0, index) + file.substring(index + 7, file.length()));

			if (source.exists()) {
				tex = new TexWriter(target);
				try {
					tex.readBody(source);
					tex.write();
				} catch (IOException e) {
					errMsg = "Error occur on file: '" + file + "'\n" + e.getMessage();
				}
			} else {
				errMsg = "File: '" + file + "' doesn't exist";
			}
		} else {
			errMsg = "File: '" + file + "' doesn't match the pattern 'xxx.source.xx.tex'";
		}

		if (errMsg != null) {
			JOptionPane.showMessageDialog(null, errMsg, "Extract tex body", JOptionPane.WARNING_MESSAGE);
		}

	}

	/**
	 * Create TexWriter
	 */
	protected TexWriter() {
	}

	/**
	 * Create TexWriter for given file
	 * 
	 * @param file directory for writer
	 */
	public TexWriter(String file) {
		this(new File(file));
	}

	/**
	 * Create TexWriter for given file
	 * 
	 * @param file directory for writer
	 */
	public TexWriter(File file) {
		this.setFile(file);
	}

	/**
	 * Get file of report
	 * 
	 * @return File
	 */
	public File getFile() {
		return this.file;
	}

	/**
	 * Set file of report
	 * 
	 * @param file File to write
	 */
	protected void setFile(File file) {
		this.file = file;
	}

	/**
	 * Get directory-string of the report for paths or files
	 * 
	 * @return directory-string
	 */
	public String getDirectory() {
		return this.file.getParent();
	}

	/**
	 * Get directory of the report for paths or files
	 * 
	 * @return Directory
	 */
	public File getDirectoryFile() {
		return this.file.getParentFile();
	}

	/**
	 * Add string to content
	 * 
	 * @param str String to add
	 */
	public void add(String str) {
		this.content.append(str);
	}

	/**
	 * Replaces each '\MOREPLACE' substring with parameter that matches the
	 * target string with replacement string
	 * 
	 * @param target String to be replaced
	 * @param replacement Replacement string
	 */
	public void replace(String target, String replacement) {
		this.replaceString(target, this.replaceSpecialSigns(replacement));
	}

	/**
	 * Replaces each '\MOREPLACE' substring with parameter that matches the
	 * target string with replacement string
	 * 
	 * @param target String to be replaced
	 * @param replacement Replacement string
	 */
	protected void replaceString(String target, String replacement) {
		int index;

		index = this.content.length();
		while ((index = this.content.lastIndexOf("\\SYMREPLACE{" + target + "}", index)) != -1) {
			this.content.replace(index, index + target.length() + 13, replacement);
		}

		index = this.content.length();
		while ((index = this.content.lastIndexOf("\\MOREPLACE{" + target + "}", index)) != -1) {
			this.content.replace(index, index + target.length() + 12, replacement);
		}
	}

	/**
	 * Add string to content
	 * 
	 * @param str String to add
	 */
	public void addString(String str) {
		this.content.append(this.replaceSpecialSigns(str));
	}

	/**
	 * Replace special character
	 * 
	 * @param str String to add
	 * @return New String
	 */
	public String replaceSpecialSigns(String str) {
		if (str == null) {
			str = "[]";
		}

		str = str.replace("\"", "''");
		str = str.replace("\\", "\\textbackslash\\-");
		str = str.replace("#", "\\#");
		str = str.replace("%", "\\%");
		str = str.replace("^", "\\^");
		str = str.replace("/", "/\\-");
		str = str.replace(".", ".\\-");
		str = str.replace("_", "\\_\\-");
		str = str.replace("", "\\euro{}");
		str = str.replace("&", "\\&");

		return str;
	}

	/**
	 * Add string to content
	 * 
	 * @param str String to add
	 */
	public void addName(String str) {
		str = str.replace(" ", "_");
		str = str.replace("\\", "/");
		this.content.append(str);
	}

	/**
	 * Add file to content
	 * 
	 * @param file File to add
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(String file) {
		return this.addGraphic(file, null);
	}

	/**
	 * Add file to content
	 * 
	 * @param file File to add
	 * @param width Width of figure (unit cm)
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(String file, double width) {
		return this.addGraphic(file, "width=" + width + "cm");
	}

	/**
	 * Add file to content
	 * 
	 * @param file File to add
	 * @param property Property of graphic
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(String file, String property) {
		return this.addGraphic(new File(file), property);
	}

	/**
	 * Add file to content
	 * 
	 * @param file File to add
	 * @param property Property of graphic
	 * @return Status of insert graphic (false if file not found)
	 */
	public boolean addGraphic(File file, String property) {
		boolean status = true;
		String str;
		File relFile;

		if (file != null) {
			relFile = FileUtils.getRelativePath(this.getDirectoryFile(),
					new File(this.getDirectoryFile(), file.getPath()));

			if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {

				str = this.includeGraphic(relFile, property);
				if (str != null) {
					this.add(str);
				} else {
					this.add("\\href{");
					this.add(relFile.getPath().replace(File.separatorChar, '/'));
					this.add("}{");
					this.addString(file.getName());
					this.add("}\\newline");
				}

			} else {
				status = false;
			}
		} else {
			status = false;
		}

		if (!status) {
			this.add("\\missingfigure{");
			this.addString(file.getPath().replace("\\", "/"));
			this.add("} ");
		}

		return status;
	}

	/**
	 * Get string with graphic
	 * 
	 * @param file File to add
	 * @return String for file
	 */
	protected String GraphicPath(File file) {
		String[] parts = this.convertGraphics(file);
		return parts[0].replace(File.separatorChar, '/') + parts[1];
	}

	/**
	 * Get string with include graphic
	 * 
	 * @param file Graphic to include
	 * @param property Property for include
	 * @return String for including file
	 */
	protected String includeGraphic(File file, String property) {
		StringBuffer buffer;
		String strfile, ext = "";
		String[] tmp;

		buffer = new StringBuffer();

		tmp = this.convertGraphics(file);
		strfile = tmp[0];
		ext = tmp[1];

		// add files
		strfile = strfile.replace(File.separatorChar, '/');
		if ((ext.compareToIgnoreCase(".pdf") == 0) || (ext.compareToIgnoreCase(".png") == 0)
				|| (ext.compareToIgnoreCase(".jpg") == 0)) {

			buffer.append("\\includegraphics");
			if (property != null) {
				buffer.append("[" + property + "]");
			}
			buffer.append("{");
			buffer.append(strfile);
			// if ((ext.compareToIgnoreCase(".jpg") == 0)
			// || (ext.compareToIgnoreCase(".jpeg") == 0)) {
			// buffer.append(ext);
			// }
			buffer.append("}");
		} else if (ext.compareToIgnoreCase(".avi") == 0) {
			buffer.append("\\includemovie[");
			if (property == null) {
				buffer.append("poster,text=(Movie),mouse,repeat");
			} else {
				buffer.append(property);
			}
			buffer.append("]{.8\\linewidth}{.5\\linewidth}{");
			buffer.append("\\InputPath/" + strfile + ext);
			buffer.append("}");
		}

		if (buffer.length() > 0) {
			return buffer.toString();
		} else {
			return null;
		}

	}

	/**
	 * Convert graphic and return path and extension
	 * 
	 * @param file Path of file
	 * @return List with path and extension
	 */
	private String[] convertGraphics(File file) {
		String strfile, name, ext = "";
		int index;

		name = file.getName();
		index = name.lastIndexOf('.');
		if (index >= 0) {
			ext = name.substring(index);
		}
		strfile = file.getPath();
		strfile = strfile.substring(0, strfile.length() - ext.length());

		// emf-files
		if (ext.compareToIgnoreCase(".emf") == 0) {
			if (ImageConvert.emf2pdf(new File(this.getDirectoryFile(), strfile + ext)) != null) {
				ext = ".pdf";
			}
			// eps-files
		} else if ((ext.compareToIgnoreCase(".eps") == 0) || (ext.compareToIgnoreCase(".ps") == 0)) {
			if (ImageConvert.eps2pdf(new File(this.getDirectoryFile(), strfile + ext)) != null) {
				ext = ".pdf";
			}
		}

		return new String[] { strfile, ext };
	}

	/**
	 * Add tex-file to content
	 * 
	 * @param file File to add
	 * @return Status of insert tex (false if file not found)
	 */
	public boolean addTex(String file) {
		return this.addTex(new File(file));
	}

	/**
	 * Add tex-file to content
	 * 
	 * @param file File to add
	 * @return Status of insert tex (false if file not found)
	 */
	public boolean addTex(File file) {
		boolean status = true;
		File relFile;
		String str, path;

		if (file != null) {
			relFile = FileUtils.getRelativePath(this.getDirectoryFile(),
					new File(this.getDirectoryFile(), file.getPath()));

			if (new File(this.getDirectoryFile(), relFile.getPath()).exists()) {
				str = relFile.getPath();
				if (str.toLowerCase().endsWith(".tex")) {
					str = str.substring(0, str.length() - 4);
					str = str.replace(File.separatorChar, '/');
					path = relFile.getParent();
					if (path != null) {
						path = path.replace(File.separatorChar, '/');
						this.add("\\graphicspath{{" + path + "/}} ");
						this.add("\\renewcommand{\\InputPath}{" + path + "/} ");
					}
					this.add("\\input{" + str + "} ");
					if (path != null) {
						this.add("\\graphicspath{{.}} ");
						this.add("\\renewcommand{\\InputPath}{.} ");
					}
				} else {
					status = false;
				}
			} else {
				status = false;
			}
		} else {
			status = false;
		}

		if (!status) {
			this.add("\\missingfigure{");
			this.addString(file.getPath().replace("\\", "/"));
			this.add("} ");
		}

		return status;
	}

	/**
	 * Add tex-table to content
	 * 
	 * @param file File to add
	 * @param caption Caption of table
	 * @param label Label of table
	 * @return Status of insert table (false if file not found)
	 */
	public boolean addTable(String file, String caption, String label) {
		return this.addTable(new File(file), caption, label);
	}

	/**
	 * Add tex-table to content
	 * 
	 * @param file File to add
	 * @param caption Caption of table
	 * @param label Label of table
	 * @return Status of insert table (false if file not found)
	 */
	public boolean addTable(File file, String caption, String label) {
		boolean status = true;
		File relFile, absFile;
		String str = null, path;
		TexWriter tex;
		int index;

		if (file != null) {
			relFile = FileUtils.getRelativePath(this.getDirectoryFile(),
					new File(this.getDirectoryFile(), file.getPath()));
			absFile = new File(this.getDirectoryFile(), relFile.getPath());
			if (absFile.exists()) {
				if (file.getName().toLowerCase().endsWith(".tex")) {
					tex = new TexWriter("");
					try {
						tex.readContent(absFile);
						str = tex.toString();
					} catch (IOException e) {
						status = false;
					}

					if (str != null) {
						// modify table
						tex = new TexWriter("");
						while ((index = str.indexOf("\\end{table}")) != -1) {
							tex.add(str.substring(0, index));
							if (caption != null) {
								tex.add("\\caption{" + caption + "}");
							}
							if (label != null) {
								tex.add("\\label{" + label + "}");
							}
							tex.add("\\");
							str = str.substring(index + 1);
						}
						tex.add(str);
						str = tex.toString();

						// modify longtable
						tex = new TexWriter("");
						while ((index = str.indexOf("\\end{longtable}")) != -1) {
							tex.add(str.substring(0, index));
							if (caption != null) {
								tex.add("\\caption{" + caption + "}");
							}
							if (label != null) {
								tex.add("\\label{" + label + "}");
							}
							tex.add("\\");
							str = str.substring(index + 1);
						}
						tex.add(str);
						str = tex.toString();

						// add to content
						path = relFile.getParent();
						if (path != null) {
							path = path.replace(File.separatorChar, '/');
							this.add("\\graphicspath{{" + path + "/}} ");
							this.add("\\renewcommand{\\InputPath}{" + path + "/} ");
						}
						this.add(str);
						if (path != null) {
							this.add("\\graphicspath{{.}} ");
							this.add("\\renewcommand{\\InputPath}{.} ");
						}
					}
				} else {
					status = false;
				}
			} else {
				status = false;
			}
		} else {
			status = false;
		}

		if (!status) {
			this.add("\\missingfigure{");
			this.addString(file.getPath().replace("\\", "/"));
			this.add("} ");
		}

		return status;
	}

	/**
	 * Length of content in writer
	 * 
	 * @return Length
	 */
	public int length() {
		return this.content.length();
	}

	/**
	 * Add content of TexWriter to content
	 * 
	 * @param report TexWriter to add
	 */
	public void addWriter(TexWriter report) {
		this.content.append(report.content);
	}

	/**
	 * Write content to tex-file
	 * 
	 * @throws IOException File exception
	 */
	public void write() throws IOException {
		FileWriter fw;

		this.file.getParentFile().mkdirs();
		fw = new FileWriter(this.file);

		// write report
		fw.write(this.toString());
		fw.close();
	}

	public int compile(boolean open) throws IOException {
		return this.compile(open, true);
	}

	/**
	 * Compile file
	 * 
	 * @param open Open after compilation
	 * 
	 * @throws IOException IO exception
	 */
	public int compile(boolean open, boolean console) throws IOException {
		ProcessBuilder pb;
		Process p;
		LinkedList<String> cmd;
		int ret = -1;
		String[] nameParts;
		boolean useTexify, useLatexmk;

		cmd = new LinkedList<String>();
		if (console) {
			if (System.getProperty("os.name").toLowerCase().startsWith("windows")) {
				cmd.add("cmd");
				cmd.add("/c");
				cmd.add("start");
				cmd.add("/wait");
				cmd.add("Generate Pdf"); // name of window
				cmd.add("cmd");
				cmd.add("/c");
			}
		}

		// check available
		try {
			useTexify = new ProcessBuilder("texify", "-version").start().waitFor() == 0;
		} catch (Exception e) {
			useTexify = false;
		}
		try {
			useLatexmk = new ProcessBuilder("latexmk", "-version").start().waitFor() == 0;
		} catch (Exception e) {
			useLatexmk = false;
		}

		if (useTexify) {
			cmd.add("texify");
			// cmd.add("-s");
			cmd.add("-p");
			cmd.add("-b");
			cmd.add("-c");
			if (open) {
				cmd.add("--run-viewer");
			}
		} else if (useLatexmk) {
			cmd.add("latexmk");
			cmd.add("-pdf");
			cmd.add("-silent");
			if (open) {
				cmd.add("-pv");
			}
		}
		cmd.add(this.file.getName());

		pb = new ProcessBuilder(cmd.toArray(new String[0]));
		pb.directory(this.getDirectoryFile());
		p = pb.start();

		// output-stream
		new Thread() {
			@Override
			public void run() {
				String line;
				BufferedReader in;

				in = new BufferedReader(new InputStreamReader(p.getInputStream()));
				try {
					while ((line = in.readLine()) != null) {
						System.out.println(line);
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					try {
						in.close();
					} catch (IOException e) {
					}
				}
			}
		}.start();

		// error-stream
		new Thread() {
			@Override
			public void run() {
				String line;
				BufferedReader in;

				in = new BufferedReader(new InputStreamReader(p.getErrorStream()));
				try {
					while ((line = in.readLine()) != null) {
						System.err.println(line);
					}
				} catch (IOException e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				} finally {
					try {
						in.close();
					} catch (IOException e) {
					}
				}

			}
		}.start();

		try {
			ret = p.waitFor();
		} catch (InterruptedException e) {
		}

		if (ret == 0) {
			nameParts = FileUtils.splitName(this.file.getName());
			if (!new File(this.getDirectoryFile(), nameParts[0] + ".pdf").exists()) {
				ret = -1;
			}
		}

		return ret;
	}

	/**
	 * Read body of file. Part between begin{document} and end{document}
	 * 
	 * @param file Tex-file
	 * @throws IOException File exception
	 */
	public void readBody(String file) throws IOException {
		this.readBody(new File(file));
	}

	/**
	 * Read body of file. Part between begin{document} and end{document}
	 * 
	 * @param file Tex-file
	 * @throws IOException File exception
	 */
	public void readBody(File file) throws IOException {
		BufferedReader fr;
		String line;
		StringBuffer buffer;
		int start, end;

		if (file.exists()) {
			buffer = new StringBuffer();
			fr = new BufferedReader(new FileReader(file));

			// write report
			while ((line = fr.readLine()) != null) {
				buffer.append(line);
				buffer.append(TexWriter.NEW_LINE);
			}
			fr.close();

			start = buffer.indexOf("\\begin{document}");
			end = buffer.lastIndexOf("\\end{document}");

			if ((start == -1) || (end == -1)) {
				this.content.append(buffer);
			} else {
				this.content.append(buffer.subSequence(start + 16, end - 1));
			}
		}
	}

	/**
	 * Read content of file
	 * 
	 * @param file Tex-file
	 * @throws IOException File exception
	 */
	public void readContent(String file) throws IOException {
		this.readContent(new File(file));
	}

	/**
	 * Read content of file
	 * 
	 * @param file Tex-file
	 * @throws IOException File exception
	 */
	public void readContent(File file) throws IOException {
		BufferedReader fr;
		String line;

		if (file.exists()) {
			fr = new BufferedReader(new FileReader(file));

			// write report
			while ((line = fr.readLine()) != null) {
				this.content.append(line);
				this.content.append(TexWriter.NEW_LINE);
			}
			fr.close();
		}
	}

	/**
	 * Trim file (remove all outside of document)
	 * 
	 * @param file File to trim
	 */
	public static void trim(String file) {
		TexWriter.trim(new File(file));
	}

	/**
	 * Trim file (remove all outside of document)
	 * 
	 * @param file File to trim
	 */
	public static void trim(File file) {
		TexWriter tex;

		tex = new TexWriter(file);
		try {
			tex.readBody(file);
			tex.write();
		} catch (IOException e) {
		}
	}

	/**
	 * Get string representation of content
	 * 
	 * @return Content
	 */
	@Override
	public String toString() {
		return this.content.toString();
	}
}
