#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 955 $
$LastChangedDate:: 2019-06-12 15:50:17 +0200#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

import numpy as np
import scipy.spatial
import matplotlib.pyplot as plt
from abc import abstractmethod


class Mesh(scipy.spatial.Delaunay):
    def __init__(self, points, max_dist=0.5, refine=True, *args, **kwargs):
        super().__init__(points, incremental=True, *args, **kwargs)
        self.initial_points = points
        self.initial_simplices = self.simplices.copy()
        self.max_dist = max_dist
        if refine:
            self.refine()
        self.close()

    @abstractmethod
    def refine(self):
        raise NotImplementedError


class Mesh2D(Mesh):

    def refine(self):
        new_points = []

        for k in range(self.nsimplex):

            p = self.points[self.simplices[k, :]]

            l1 = np.linalg.norm(p[0] - p[1])
            l2 = np.linalg.norm(p[1] - p[2])
            l3 = np.linalg.norm(p[2] - p[0])

            if l1 > self.max_dist:
                new_points.append((p[0] + p[1]) / 2)

            if l2 > self.max_dist:
                new_points.append((p[1] + p[2]) / 2)

            if l3 > self.max_dist:
                new_points.append((p[2] + p[0]) / 2)

        if new_points:
            self.add_points(new_points)
            self.refine()

    def refineCentroid(self, p, new_points):
        c = np.sum(p, axis=0) / 3

        l1 = np.linalg.norm(p[0] - c)
        l2 = np.linalg.norm(p[1] - c)
        l3 = np.linalg.norm(p[2] - c)

        if np.min([l1, l2, l3]) > self.max_dist:
            new_points.append(c)

    def plotMPL(self):
        x = self.initial_points[:, 0]
        y = self.initial_points[:, 1]

        xi = self.points[:, 0]
        yi = self.points[:, 1]

        plt.triplot(x, y, self.initial_simplices)
        plt.triplot(xi, yi, self.simplices)

        plt.scatter(xi, yi, color='r')
        plt.scatter(x, y, color='b')


class Mesh3D(Mesh):

    def refine(self):
        new_points = []

        for k in range(self.nsimplex):

            p = self.points[self.simplices[k, :]]

            l1 = np.linalg.norm(p[0] - p[1])
            l2 = np.linalg.norm(p[1] - p[2])
            l3 = np.linalg.norm(p[2] - p[0])
            l4 = np.linalg.norm(p[3] - p[0])
            l5 = np.linalg.norm(p[3] - p[1])
            l6 = np.linalg.norm(p[3] - p[2])

            if l1 > self.max_dist:
                new_points.append((p[0] + p[1]) / 2)

            if l2 > self.max_dist:
                new_points.append((p[1] + p[2]) / 2)

            if l3 > self.max_dist:
                new_points.append((p[2] + p[0]) / 2)

            if l4 > self.max_dist:
                new_points.append((p[3] + p[0]) / 2)

            if l5 > self.max_dist:
                new_points.append((p[3] + p[1]) / 2)

            if l6 > self.max_dist:
                new_points.append((p[3] + p[2]) / 2)

        if new_points:
            self.add_points(new_points)
            self.refine()

    def refineCentroid(self, p, new_points):
        c = np.sum(p, axis=0) / 4

        l1 = np.linalg.norm(p[0] - c)
        l2 = np.linalg.norm(p[1] - c)
        l3 = np.linalg.norm(p[2] - c)
        l4 = np.linalg.norm(p[3] - c)
        l5 = np.linalg.norm(p[4] - c)
        l6 = np.linalg.norm(p[5] - c)

        if np.min([l1, l2, l3, l4, l5, l6]) > self.max_dist:
            new_points.append(c)

    def plotMPL(self):

        from mpl_toolkits import mplot3d

        fig = plt.figure()
        ax = fig.add_subplot(111, projection=mplot3d.Axes3D.name)

        for simplex in self.simplices:
            pts = self.points[simplex, :]
            ax.plot(pts[[0, 1], 0], pts[[0, 1], 1], pts[[0, 1], 2], color='g', lw='0.1')
            ax.plot(pts[[0, 2], 0], pts[[0, 2], 1], pts[[0, 2], 2], color='g', lw='0.1')
            ax.plot(pts[[0, 3], 0], pts[[0, 3], 1], pts[[0, 3], 2], color='g', lw='0.1')
            ax.plot(pts[[1, 2], 0], pts[[1, 2], 1], pts[[1, 2], 2], color='g', lw='0.1')
            ax.plot(pts[[1, 3], 0], pts[[1, 3], 1], pts[[1, 3], 2], color='g', lw='0.1')
            ax.plot(pts[[2, 3], 0], pts[[2, 3], 1], pts[[2, 3], 2], color='g', lw='0.1')

        ax.scatter(self.points[:, 0], self.points[:, 1], self.points[:, 2], color='r')
        ax.scatter(self.initial_points[:, 0], self.initial_points[:, 1], self.initial_points[:, 2], color='b')


if __name__ == '__main__':


    n_in = 3
    n_lp = 10

    np.random.seed(1234)
    xc = np.random.rand(n_in, n_lp)

    points = xc.transpose()
    if n_in == 2:
        mesh = Mesh2D(points, refine=True, max_dist=0.5)
    elif n_in == 3:
        mesh = Mesh3D(points, refine=True, max_dist=0.5)
    else:
        raise Exception

    mesh.plotMPL()

    plt.show()
