"""
Created on 05.10.2016 14:51

@author: kkaspar
"""

# Import Python Classes
import numpy as np
from lcm.utils.pytperiodic import PyTPeriodicMatrix
from lcm.tdata_interface.pyjavadata_interface import PyJavaDataInterface
from lcm.tdata_interface.pyjlist_interface import PyJListInterface
from lcm.tdata_interface.pytnull_interface import PyTNullInterface
from lcm.tdata_interface.pytboolean_interface import PyTBooleanInterface
from lcm.tdata_interface.pytnumeric_interface import PyTNumericInterface
from lcm.tdata_interface.pytstring_interface import PyTStringInterface
from lcm.tdata_interface.pytperiodic_interface import PyTPeriodicInterface


class PyJDataInterface(PyJavaDataInterface):
    """Interface for python (via jep) <-> java

    Java                    | Python
    --------------------------------
    TData/jep.PyJObject     | ...
    Map/jep.PyJMap          | dict
    jep.PyJArray            | ...
    jep.PyJList             | ...
    ...                     | ...


    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Create a Python object equivalent to given Java object
        """

        if str(type(jobj)) == "<class 'jep.PyJObject'>":
            '''Assuming an jep.PyJObject can only be a TData'''
            pyobj = PyTDataInterface.convert_J2Py(jobj)
        elif str(type(jobj)) == "<class 'jep.PyJMap'>":
            pyobj = PyJMapInterface.convert_J2Py(jobj)
        elif str(type(jobj)) == "<class 'jep.PyJArray'>":
            pyobj = PyJArrayInterface.convert_J2Py(jobj)
        elif str(type(jobj)) == "<class 'jep.PyJList'>":
            pyobj = PyJListInterface.convert_J2Py(jobj)
        elif (jobj is None) or isinstance(jobj, bytes) or isinstance(jobj, bool) or isinstance(jobj, str) or isinstance(
                jobj, int) or isinstance(jobj, float):
            '''In this case jep has already automatically converted the java variable into its python counterpart'''
            pyobj = jobj
        else:
            msg = "Matching conversion from Java Class of class: {} to a Python Object not found".format(type(jobj))
            raise TypeError(msg)
        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Create a Java object equivalent to the given python object
        """

        try:
            jobj = PyTDataInterface.convert_Py2J(pyobj)
        except Exception as e:
            msg = "Matching conversion from Java Class of class: {} " \
                  "to a Python Object not found\n{}".format(type(pyobj), e)
            raise TypeError(msg).with_traceback(e.__traceback__)
        return jobj


class PyJMapInterface(PyJavaDataInterface):
    """Java Map conversion

    Java            | Python
    -------------------------
    Map/jep.PyJMap  | dict

    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "dict"
        equivalent to given Java object of type "Map"
        """

        keys = []
        values = []
        for key in jobj.keySet():
            keys.append(key)
            values.append(PyJDataInterface.convert_J2Py(jobj.get(key)))
        pyobj = dict(zip(keys, values))

        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "Map"
        equivalent to given python object of type "dict"
        """
        # noinspection PyUnresolvedReferences
        from java.util import Map, Hashtable

        jobj = Hashtable()
        for key, value in pyobj.items():
            jobj.put(key, PyJDataInterface.convert_Py2J(value))
        return jobj


class PyJArrayInterface(PyJavaDataInterface):
    """JArray conversion

    Java            | Python
    -------------------------
    jep.PyJArray    | list

    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "list"
        equivalent to given Java object of type "jep.PyJArray"
        """
        pyobj = []
        for entry in jobj:
            try:
                # use recursive call go get into deeper structure levels
                pyobj.append(PyJDataInterface.convert_J2Py(entry))
            except:
                pyobj.append(entry)

        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "jep.PyJArray"
        equivalent to given python object of type "list"
        """

        raise NotImplementedError("Conversion of a python object", type(pyobj),
                                  "to a  Java object of type 'jep.PyJArray' has not been implemented.")
        # return jobj


class PyTDataInterface(PyJavaDataInterface):
    """
    Only datatypes, with an existing equivalent of the Java TData Class can be assigned.

    Java            | Python
    -------------------------
    TNull           | NoneType
    TBool           | bool
    TNumeric        | numpy.ndarray, int, float
    TString         | string
    TStruct         | dict
    TArray          | numpy.ndarray with numpy.dtype(object)
    TPeriodic       | PyTPeriodicMatrix
    TPeriodicMatrix | PyTPeriodicMatrix
    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object
        equivalent to given Java object of type "TData"
        """
        if jobj.getClassName() == 'TNull':
            pyobj = PyTNullInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TBoolean':
            pyobj = PyTBooleanInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TString':
            pyobj = PyTStringInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TNumeric':
            pyobj = PyTNumericInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TStruct':
            pyobj = PyTStructInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TArray':
            pyobj = PyTArrayInterface.convert_J2Py(jobj)
        elif jobj.getClassName() == 'TPeriodic' or jobj.getClassName() == 'TPeriodicMatrix':
            pyobj = PyTPeriodicInterface.convert_J2Py(jobj)
        else:
            msg = "Matching PyObject (from Java TData) class for conversion of class: {} " \
                  "for conversion to a Python class not found".format(type(jobj))
            raise TypeError(msg)
        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "TData"
        equivalent to given python object
        """
        if pyobj is None:
            """
            TNull
            """
            jobj = PyTNullInterface.convert_Py2J(pyobj)
        elif isinstance(pyobj, bool):
            """
            TBool
            """
            jobj = PyTBooleanInterface.convert_Py2J(pyobj)
        elif isinstance(pyobj, str):
            """
            TString
            """
            jobj = PyTStringInterface.convert_Py2J(pyobj)
        elif isinstance(pyobj, int) or isinstance(pyobj, float) or isinstance(pyobj, complex) \
                or isinstance(pyobj, np.ndarray) and not pyobj.dtype == object:
            """
            TNumeric
            """
            jobj = PyTNumericInterface.convert_Py2J(pyobj)
        elif isinstance(pyobj, dict):
            """
            TStruct
            """
            jobj = PyTStructInterface.convert_Py2J(pyobj)
        elif (isinstance(pyobj, np.ndarray) and pyobj.dtype == object) or isinstance(pyobj, list) or isinstance(pyobj,
                                                                                                                tuple):
            """
            TArray
            """
            jobj = PyTArrayInterface.convert_Py2J(pyobj)
        elif isinstance(pyobj, PyTPeriodicMatrix):
            """
            TPeriodic
            """
            jobj = PyTPeriodicInterface.convert_Py2J(pyobj)
        else:
            '''try native conversion'''
            jobj = pyobj
        return jobj


class PyTStructInterface(PyJavaDataInterface):
    """TStruct conversion

    Java            | Python
    -------------------------
    TStruct         | dict

    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "dict"
        equivalent to given Java object of type "TStruct"
        """

        keys = []
        values = []
        for key in jobj.getKeys():
            """
            TStruct values are of class TData,
            hence the values need to be converted to the equivalent python types
            """
            keys.append(key)
            value = jobj.get(key)
            values.append(PyTDataInterface.convert_J2Py(value))
        pyobj = dict(zip(keys, values))

        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "TStruct"
        equivalent to given python object of type "dict"
        """
        # noinspection PyUnresolvedReferences
        from at.lcm.bu21.general.dtypes import TStruct

        jobj = TStruct()
        for key, value in pyobj.items():
            """
            TStruct has to contain only values of TData
            """
            try:
                jobj.put(key, PyTDataInterface.convert_Py2J(value))
            except Exception as ex:
                raise Exception("['{}'] {}".format(key, ex))
        return jobj


class PyTArrayInterface(PyJavaDataInterface):
    """TArray conversion

    Java            | Python
    -------------------------
    TArray          | numpy.ndarray with numpy.dtype(object)

    """

    @staticmethod
    def convert_J2Py(jobj):
        """
        Return a Python object of type "numpy.ndarray" with "dtype=np.dtype(object)"
        equivalent to given Java object of type "TArray"
        """

        shape = jobj.size()
        pyobj = np.zeros(shape, dtype=np.dtype(object))
        for row in range(shape[0]):
            for col in range(shape[1]):
                pyobj[row, col] = PyTDataInterface.convert_J2Py(jobj.get(row, col))

        return pyobj

    @staticmethod
    def convert_Py2J(pyobj):
        """
        Return a Java object of type "TArray"
        equivalent to given python object of type "numpy.ndarray" with "dtype=np.dtype(object)"
        """
        # noinspection PyUnresolvedReferences
        from at.lcm.bu21.general.dtypes import TArray

        if isinstance(pyobj, list) or isinstance(pyobj, tuple):
            pyobj_array = np.asarray(pyobj, dtype=np.dtype(object))
        else:
            pyobj_array = pyobj

        jobj = TArray()
        if pyobj_array.ndim == 1:
            for index in range(pyobj_array.size):
                value = PyTDataInterface.convert_Py2J(pyobj[index])
                jobj.set(0, index, value)

        elif pyobj_array.ndim == 2:
            for row in range(pyobj_array.shape[0]):
                for col in range(pyobj_array.shape[1]):
                    value = PyTDataInterface.convert_Py2J(pyobj_array[row, col])
                    jobj.set(row, col, value)
        else:
            raise ValueError("Number of dimensions for TArray constructor arguments needs to be 1 or 2, "
                             "instead pyobj.ndim is:", pyobj_array.ndim)
        return jobj


if __name__ == "__main__":
    # pyobj_in = np.array([1, 2, 3, 4, 'str'], dtype=np.dtype(object))
    pyobj_in = [1, 2, 3, 4, 'str', [1, 2]]

    jobj_out = PyJDataInterface.convert_Py2J(pyobj_in)
    pyobj_out = PyJDataInterface.convert_J2Py(jobj_out)
    print('java:' + str(jobj_out))
    print('python:' + str(pyobj_out))
    print('python type:' + str(type(pyobj_out)))
