import time
import numpy as np
import matplotlib.pyplot as plt
import scipy.interpolate
import lcm.math.rbf as rbf
from lcm.math.rbf import kernels
from lcm.math.rbf import transformations


def run_rbfRandom(n_in=2, n_out=3, n_lp=4):
    """
    Example for periodic RBF Interpolation
    using random input and output values with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc = np.random.rand(n_in, n_lp)

    '''output matrix'''
    zc = np.random.rand(n_out, n_lp)

    '''create rbf'''
    t = transformations.normalization(xc)
    rbf1 = rbf.RBF(xc, zc, kernels.KernelTPS(), t=t, gamma=1e-3, max_dist=0.5)

    '''test'''
    x = np.zeros((n_in, 1))
    x[0, 0] = 0.5
    x[1, 0] = 0.5

    '''evaluate rbf'''
    z = rbf1(x)

    print('xc', xc)
    print('zc', zc)
    print('x', x)
    print('z', z)


def run_rbfConst(n_in=2, n_out=3, n_lp=4):
    """
    Example for periodic RBF Interpolation
    using constant input and output values with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    '''input matrix'''
    xc = np.ones((n_in, n_lp))
    xc[0, 0] = 3
    xc[-1, -1] = -3

    '''output matrix'''
    zc = np.ones((n_out, n_lp))
    zc[0, 0] = 5
    zc[-1, -1] = -5

    '''create rbf'''
    t = transformations.normalization(xc)
    rbf1 = rbf.RBF(xc, zc, kernels.KernelTPS(), t=t, gamma=1e-3, max_dist=1.0)

    '''test'''
    x = np.zeros((n_in, 1))
    x[0, 0] = 0.5
    x[1, 0] = 0.5

    '''evaluate rbf'''
    z = rbf1(x)

    print('xc', xc)
    print('zc', zc)
    print('x', x)
    print('z', z)


def run_plotMaxDist(n_in=2, n_out=1, n_lp=10):
    """
    Example for periodic RBF Interpolation
    using random input and output values with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """

    import plotly.offline as ply

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc = np.random.rand(n_in, n_lp) * 10
    # xc[:, 0] = np.random.rand(n_in) * 10

    '''output matrix'''
    zc = np.random.rand(n_out, n_lp)

    '''create rbf'''
    t = transformations.normalization(xc)
    rbf1 = rbf.RBF(xc, zc, kernels.KernelTPS(), t=t, gamma=1e-3, max_dist=0.5)

    if n_in == 2:
        data = rbf1.plotMinRadius2DPLY()
    elif n_in == 3:
        data = rbf1.plotMinRadius3DPLY(z_values=np.linspace(np.min(xc[2, :]), np.max(xc[2, :]), 3))
    else:
        raise Exception

    ply.plot(data)


def run_plotMaxDistTriangulatePLY(n_in=2, n_out=1, n_lp=10):
    """
    Example for periodic RBF Interpolation
    using random input and output values with the defined size

    Parameters
    ----------
    n_in : int
        number of inputs
    n_out : int
        number of outputs
    n_lp : int
        number of loadpoints

    """
    import lcm.math.spatial.meshing as tri
    import plotly.offline as ply
    import plotly.graph_objs as go
    from scipy.interpolate import griddata

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc = np.random.rand(n_in, n_lp) * 10
    # xc[:, 0] = np.random.rand(n_in) * 10

    '''output matrix'''
    zc = np.random.rand(n_out, n_lp)

    '''create rbf'''
    t = transformations.normalization(xc)
    rbf1 = rbf.RBF(xc, zc, kernels.KernelTPS(), t=t, gamma=1e-3, max_dist=0.5)

    points = rbf1.xc_t.transpose()
    #
    # points = (np.linalg.inv(rbf1.t) @ rbf1.xc_t).transpose()

    if n_in == 2:
        mesh = tri.Mesh2D(points, refine=True, max_dist=0.5)
    elif n_in == 3:
        mesh = tri.Mesh3D(points, refine=True, max_dist=0.5)
    else:
        raise Exception

    xc_t_tri = mesh.points.transpose()
    xc_tri = np.linalg.inv(rbf1.t) @ xc_t_tri

    xaxis = 0
    yaxis = 1
    x = xc_tri[xaxis, :]
    y = xc_tri[yaxis, :]

    r_array = np.zeros((1, xc_tri.shape[1]))
    for index in range(xc_tri.shape[1]):
        r_array[0, index] = rbf1.evalRadius(xc_tri[:, index, np.newaxis])

    xi = np.linspace(np.min(x), np.max(x), x.size)
    yi = np.linspace(np.min(y), np.max(y), y.size)
    X, Y = np.meshgrid(xi, yi)

    zi = griddata(xc_tri.transpose(), r_array.flatten(), (X, Y), method='linear')
    # grid_z1 = griddata(points, values, (grid_x, grid_y), method='linear')
    # grid_z2 = griddata(points, values, (grid_x, grid_y), method='cubic')

    contour = go.Contour(z=zi, x=xi, y=yi, colorscale='Jet',
                         showscale=True,
                         zmin=0, zmax=rbf1.max_dist,
                         connectgaps=True,

                         contours=dict(
                             coloring='heatmap',
                             showlabels=True,
                             labelfont=dict(
                                 family='Raleway',
                                 size=12,
                                 color='white',
                             )
                         )
                         )

    centers = go.Scatter(x=x, y=y, mode='markers')

    ply.plot([contour, centers])


def run_example_1():
    """
    In this example we generate synthetic scattered data with added noise
    and then fit it with a smoothed RBF interpolant. The interpolant in
    this example is equivalent to a thin plate spline.
    """

    np.random.seed(1)

    # observation points
    x_obs = np.random.random((100, 2))
    # values at the observation points
    u_obs = np.sin(2 * np.pi * x_obs[:, 0]) * np.cos(2 * np.pi * x_obs[:, 1])
    u_obs += np.random.normal(0.0, 0.1, 100)
    # create a thin-plate spline interpolant, where the data is assumed to
    # be noisy
    # I = RBFInterpolant(x_obs, u_obs, sigma=0.1, phi='phs2', order=1)
    # I_scipy = scipy.interpolate.Rbf(x_obs[:, 0], x_obs[:, 1], u_obs)
    I_lcm = rbf.RBF(x_obs.transpose(), u_obs[np.newaxis, :])
    # create the interpolation points, and evaluate the interpolant
    x1, x2 = np.linspace(0, 1, 200), np.linspace(0, 1, 200)
    x_itp = np.reshape(np.meshgrid(x1, x2), (2, 200 * 200)).T
    # u_itp = I(x_itp[])
    # u_itp = I_scipy(x_itp[:, 0], x_itp[:, 1])
    u_itp = I_lcm(x_itp.transpose()).squeeze()
    # plot the results
    plt.tripcolor(x_itp[:, 0], x_itp[:, 1], u_itp, vmin=-1.1, vmax=1.1, cmap='viridis')
    plt.scatter(x_obs[:, 0], x_obs[:, 1], s=100, c=u_obs, vmin=-1.1, vmax=1.1,
                cmap='viridis', edgecolor='k')
    plt.xlim((0.05, 0.95))
    plt.ylim((0.05, 0.95))
    plt.colorbar()
    plt.tight_layout()
    # plt.savefig('../figures/interpolate.a.png')
    plt.show()


def run_example_2():
    """
    RBF Interpolation on a unit sphere. This is done by converting theta
    (azimuthal angle) and phi (polar angle) into cartesian coordinates and
    then interpolating over R^3.
    """
    np.random.seed(1)

    def spherical_to_cartesian(theta, phi):
        x = np.sin(phi) * np.cos(theta)
        y = np.sin(phi) * np.sin(theta)
        z = np.cos(phi)
        return np.array([x, y, z]).T

    def true_function(theta, phi):
        # create some arbitary function that we want to recontruct with
        # interpolation
        cart = spherical_to_cartesian(theta, phi)
        out = (np.cos(cart[:, 2] - 1.0) *
               np.cos(cart[:, 0] - 1.0) *
               np.cos(cart[:, 1] - 1.0))
        return out

    # number of observation points
    nobs = 50
    # make the observation points in spherical coordinates
    theta_obs = np.random.uniform(0.0, 2 * np.pi, nobs)
    phi_obs = np.random.uniform(0.0, np.pi, nobs)
    # get the cartesian coordinates for the observation points
    cart_obs = spherical_to_cartesian(theta_obs, phi_obs)
    # get the latent function at the observation points
    val_obs = true_function(theta_obs, phi_obs)
    # make a grid of interpolation points in spherical coordinates
    theta_itp, phi_itp = np.meshgrid(np.linspace(0.0, 2 * np.pi, 100),
                                     np.linspace(0.0, np.pi, 100))
    theta_itp = theta_itp.flatten()
    phi_itp = phi_itp.flatten()
    # get the catesian coordinates for the interpolation points
    cart_itp = spherical_to_cartesian(theta_itp, phi_itp)
    # create an RBF interpolant from the cartesian observation points. I
    # am just use the default `RBFInterpolant` parameters here, nothing
    # special.
    # I = RBFInterpolant(cart_obs, val_obs)
    I_scipy = scipy.interpolate.rbf.Rbf(cart_obs[:, 0], cart_obs[:, 1], cart_obs[:, 2], val_obs)
    # evaluate the interpolant on the interpolation points
    # val_itp = I(cart_itp)
    val_itp = I_scipy(cart_itp[:, 0], cart_itp[:, 1], cart_itp[:, 2])

    ## PLOTTING

    # plot the true function in spherical coordinates
    val_true = true_function(theta_itp, phi_itp)
    plt.figure(1)
    plt.title('True function')
    p = plt.tripcolor(theta_itp, phi_itp, val_true,
                      cmap='viridis')
    plt.colorbar(p)
    plt.xlabel('theta (azimuthal angle)')
    plt.ylabel('phi (polar angle)')
    plt.xlim(0, 2 * np.pi)
    plt.ylim(0, np.pi)
    plt.grid(ls=':', color='k')
    plt.tight_layout()
    # plt.savefig('../figures/interpolate.b.1.png')

    # plot the interpolant in spherical coordinates
    plt.figure(2)
    plt.title('RBF interpolant (points are observations)')
    # plot the interpolated function
    p = plt.tripcolor(theta_itp, phi_itp, val_itp,
                      cmap='viridis')
    # plot the observations
    plt.scatter(theta_obs, phi_obs, c=val_obs,
                s=50, edgecolor='k', cmap='viridis',
                vmin=p.get_clim()[0], vmax=p.get_clim()[1])
    plt.colorbar(p)
    plt.xlabel('theta (azimuthal angle)')
    plt.ylabel('phi (polar angle)')
    plt.xlim(0, 2 * np.pi)
    plt.ylim(0, np.pi)
    plt.grid()
    plt.grid(ls=':', color='k')
    plt.tight_layout()
    # plt.savefig('../figures/interpolate.b.2.png')

    # compute and print the mean L2 error
    mean_error = np.mean(np.abs(val_true - val_itp))
    print('mean interpolation error: %s' % mean_error)
    plt.show()


def run_loop(n_times):
    n_in = 2
    n_out = 3
    n_lp = 4

    '''create reproducible random numbers'''
    np.random.seed(1234)

    '''input matrix'''
    xc = np.random.rand(n_in, n_lp)

    '''output matrix'''
    zc = np.random.rand(n_out, n_lp)

    '''create rbf'''
    t = transformations.normalization(xc)
    # rbf1 = rbf.RBF(xc, zc, kernels.KernelTPS(), t=t, gamma=1e-3, max_dist=0.5)
    rbf1 = rbf.RBF(xc, zc, kernels.KernelCS(), t=t, gamma=1e-3, max_dist=0.5)

    '''test'''
    x = np.zeros((n_in, 1))
    x[0, 0] = 0.5
    x[1, 0] = 0.5

    '''evaluate rbf'''
    dt_sum = 0.0
    for _ in range(n_times):
        t1 = time.time()
        z = rbf1(x)
        dt_sum += time.time() - t1

    print('\nrbf: Total={:.2f} s  Avg={:.2f} s'.format(dt_sum, dt_sum / n_times))


if __name__ == '__main__':
    # run_rbfRandom()
    # run_rbfConst()
    # run_plotMaxDist()
    # run_plotMaxDistTriangulatePLY()
    # run_example_1()
    # run_example_2()
    run_loop(n_times=100000)  # for profiling
