/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10156 $
 * $Date: 2021-04-10 17:32:56 +0200 (Sat, 10 Apr 2021) $
 */
package at.lcm.symspace.optimizer;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import javax.swing.border.EmptyBorder;
import javax.swing.filechooser.FileNameExtensionFilter;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.magopt.structure.jep.JepProxy;
import at.lcm.bu21.magopt.utils.MagOptConfig;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Optimization;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.SavePoint.SavePointType;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.event.OptimizerListener;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.plot.PlotJson;
import at.lcm.bu21.optim.core.plot.PlotJson.ResultType;
import at.lcm.bu21.optim.core.plot.WebPlotInterface;
import at.lcm.bu21.optim.core.project.ComputeProject;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import at.lcm.bu21.optim.gui.setting.ChooseProblem;
import at.lcm.bu21.optim.gui.setting.ChooseResults;
import at.lcm.bu21.optim.gui.setting.ChooseVariables;
import at.lcm.bu21.optim.gui.setting.GeneralSettings;
import at.lcm.bu21.optim.gui.setting.SolverSettings;
import at.lcm.symspace.Launcher;
import at.lcm.symspace.center.CenterController;
import at.lcm.symspace.center.CenterGui;
import at.lcm.symspace.gui.About;
import at.lcm.symspace.gui.Controller;
import at.lcm.symspace.gui.utils.ProgressBar;
import at.lcm.symspace.gui.utils.Storage;
import at.lcm.symspace.optimizer.icons.opt.OptIcons;
import at.lcm.symspace.optimizer.utils.DialogStopCondition;
import at.lcm.symspace.optimizer.utils.OptimizerMacroDialog;
import at.lcm.symspace.optimizer.views.OptimizerStatusBar;
import at.lcm.symspace.optimizer.views.OptimizerViewManager;
import at.lcm.symspace.optimizer.views.settings.SettingsDialog;
import jobmanager.JobDesc;
import magopt.gui.console.PythonConsole;
import magopt.gui.helpers.SendToDialog;
import magopt.utils.Execute;
import magopt.utils.MopFile;

public class OptimizerController extends Controller {

	private Problem problem = null;

	private Optimization optim;

	private boolean saveInProgress = false;

	private boolean plotVariables = false;

	private PlotJson.ResultType plotResults = PlotJson.ResultType.OBJECTIVES;

	public OptimizerController(OptimizerGui gui) {
		super(gui);
	}

	@Override
	protected void showProgressBar(final String init) {
		this.showProgressBar(init, 0);
	}

	@Override
	protected void showProgressBar(final String init, final int max) {
		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				@Override
				public void run() {
					OptimizerController.this.getViewManager().setEnabled(false);
				}
			});
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		((OptimizerStatusBar) this.getViewManager().getStatusBar()).showProgressBar(init, max);
	}

	@Override
	protected void updateProgressBar(final int val, final String status) {
		((OptimizerStatusBar) this.getViewManager().getStatusBar()).updateProgressBar(val, status);
	}

	@Override
	protected void hideProgressBar() {
		((OptimizerStatusBar) this.getViewManager().getStatusBar()).hideProgressBar();

		try {
			SwingUtilities.invokeAndWait(new Runnable() {
				@Override
				public void run() {
					OptimizerController.this.getViewManager().setEnabled(true);
				}
			});
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void openOptimization(Problem prob) {
		OptimizerController.this.problem = prob;

		Storage s = new Storage();
		s.addRecentOptimization(s.new OptimizationDescription(this.problem.getArchive().getMop().getAbsolutePath(),
				this.problem.getArchive().getName(), this.problem.getName()));

		try {
			this.optim = this.problem.getOptimization();
		} catch (Exception e) {
			System.err.println("Unable to create Controller");
			e.printStackTrace();
		}

		// update script directory
		OptimizerViewManager views = this.getViewManager();
		views.pythonConsole.setScriptDirectory(this.problem.getArchive().getMop().getParentFile());

		if (this.optim != null) {
			this.optim.addOptimizerListener(new OptimizerListener() {
				@Override
				public void optimizerPerformed(final OptimizerEvent event) {
					final OptimizerViewManager view = OptimizerController.this.getViewManager();

					switch (event.getType()) {
					case OptimizerEvent.LOG:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								if (view != null) {
									view.refresh(view.EVENT);
								}
							}
						});
						break;
					case OptimizerEvent.PROBLEM:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// refresh overview panel
								if (OptimizerController.this.optim != null) {
									OptimizerController.this.optim
											.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.STATUS));
								}
								// refresh objective plots
								if ((view != null) && (view.plots != null)) {
									view.plots.refresh();
								}
							}
						});
						break;
					case OptimizerEvent.STATUS:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// Update project status
								if (view != null) {
									view.refresh(view.POOL);
									// view.refresh(view.MAPPING);
									// view.refresh(view.SENSITIVITY);
									view.getToolBar().refresh();
									view.getStatusBar().refresh();
								}
							}
						});
						break;
					case OptimizerEvent.POOL:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// Update project status
								if (view != null) {
									view.refresh(view.POOL);
								}
							}
						});
						break;
					case OptimizerEvent.PLOT:
						if ((view != null) && (view.plots != null)) {
							view.plots.performEvent(event);
						}
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// refresh plots
								if (OptimizerController.this.optim != null) {
									OptimizerController.this.optim
											.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
								}
							}
						});
						break;
					case OptimizerEvent.MAPPING:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// refresh mapping panel
								if (view != null) {
									view.refresh(view.MAPPING);
								}
							}
						});
						break;
					case OptimizerEvent.SENSITIVITY:
						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								// refresh sensitivity panel
								if (view != null) {
									view.refresh(view.SENSITIVITY);
								}
							}
						});
						break;
					}
				}
			});
		}

	}

	public static void openOptimizationExtern(Problem prob) {
		Execute exec;
		File java;

		java = new File(new File(System.getProperty("java.home"), "bin"), "java");

		exec = new Execute(java.getPath(), "-cp",
				FileUtils.getFile(MagOptConfig.getHome(), "System", "Java", "magopt.jar").getAbsolutePath(),
				Launcher.class.getName(), "SyMSpaceOptimizerGui", prob.getArchive().getMop().getAbsolutePath(),
				prob.getArchive().getName(), prob.getName());
		exec.setRedirectStreams(true);
		exec.start();
	}

	public Problem getProblem() {
		return this.problem;
	}

	private File getWorkDir() {
		List<String> recentFiles;
		File workDir = null;

		if (this.problem != null) {
			// try {
			workDir = this.problem.getArchive().getMop().getParentFile();
			// } catch (RemoteException e) {
			// System.err.println(e.getMessage());
			// e.printStackTrace();
			// }
		}
		if (workDir == null) {
			recentFiles = new Storage().getRecentProjects();
			if ((recentFiles != null) && !recentFiles.isEmpty()) {
				workDir = new File(recentFiles.get(0)).getParentFile();
			}
		}

		return workDir;
	}

	public void mnu_file_wizard() {
		this.mnu_file_wizard(null, null, null);
	}

	public void mnu_file_wizard(final File project, final String archive, final String problem) {

		new Thread() {
			@Override
			public void run() {
				final SettingsDialog dia;
				MopFile mop;
				Archive arch;
				Problem prob = null;

				OptimizerController.this.getGuiComponent().setEnabled(false);

				OptimizerController.this.showProgressBar("Wizard");
				if ((project != null) && project.exists()) {
					if ((archive != null) && (problem != null)) {
						if (project.getName().toLowerCase().endsWith(".mop")) {
							mop = new MopFile(project);
							try {
								arch = Archive.load(mop, archive);
								prob = arch.readProblem(problem);
							} catch (Exception e) {
							}
						}
					}
				}
//				OptimizerController.this.hideProgressBar();

				if (prob != null) {
					dia = new SettingsDialog(OptimizerController.this.getGuiComponent(), prob);
				} else if ((project != null) && project.exists()) {
					dia = new SettingsDialog(OptimizerController.this.getGuiComponent(), project);
				} else {
					dia = new SettingsDialog(OptimizerController.this.getGuiComponent(),
							OptimizerController.this.getWorkDir());
				}

				dia.addWindowListener(new WindowAdapter() {
					@Override
					public void windowClosed(WindowEvent event) {
						new Thread() {
							@Override
							public void run() {
								OptimizerController.this.getGuiComponent().setEnabled(true);
								OptimizerController.this.hideProgressBar();
								if (dia.getProblem() != null) {
									OptimizerController.this.mnu_file_open(dia.getProblem());
								}
							}
						}.start();
					}

					@Override
					public void windowIconified(WindowEvent e) {
						if (OptimizerController.this.getGuiComponent() instanceof JFrame) {
							((JFrame) OptimizerController.this.getGuiComponent()).setState(Frame.ICONIFIED);
						}
					}

					@Override
					public void windowDeiconified(WindowEvent e) {
						if (OptimizerController.this.getGuiComponent() instanceof JFrame) {
							((JFrame) OptimizerController.this.getGuiComponent()).setState(Frame.NORMAL);
						}
					}
				});
			}
		}.start();
	}

	public void mnu_file_open(final File project, final String archive, final String problem) {

		new Thread() {
			@Override
			public void run() {
				MopFile mop;
				Archive arch;
				Problem prob;

				OptimizerController.this.showProgressBar("Open optimization");

				try {
					mop = new MopFile(project);
					arch = Archive.load(mop, archive);
					prob = arch.readProblem(problem);

					if (OptimizerController.this.problem != null) {
						// open in new vm
						OptimizerController.openOptimizationExtern(prob);
					} else {
						// open in new vm
						OptimizerController.this.openOptimization(prob);
					}
				} catch (Exception e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}

				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();
			}
		}.start();

	}

	public void mnu_file_open(final Problem problem) {

		new Thread() {
			@Override
			public void run() {
				OptimizerController.this.showProgressBar("Open problem");

				try {
					problem.getArchive().save();

					if (OptimizerController.this.problem != null) {
						// open in new vm
						OptimizerController.openOptimizationExtern(problem);
					} else {
						// open in new vm
						OptimizerController.this.openOptimization(problem);
					}
				} catch (Exception e) {
					System.err.println(e.getMessage());
					e.printStackTrace();
				}

				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();
			}
		}.start();

	}

	public void mnu_file_save() {
		new Thread() {
			@Override
			public void run() {
				OptimizerController.this.showProgressBar("Create savepoint");
				try {
					OptimizerController.this.problem.createSavePoint();
					OptimizerController.this.problem.getArchive().save();
				} catch (Exception e) {
					// FIXME event
					System.err.println("Unable to save");
					e.printStackTrace();
				}
				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_file_close() {
		new Thread() {
			@Override
			public void run() {
				OptimizerGui gui;

				OptimizerController.this.showProgressBar("Close");

				if (OptimizerController.this.optim != null) {
					OptimizerController.this.optim.shutdown();
				}

				OptimizerController.this.optim = null;
				OptimizerController.this.problem = null;

				// update script directory
				OptimizerViewManager views = OptimizerController.this.getViewManager();
				views.pythonConsole.setScriptDirectory(null);

				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();

				gui = (OptimizerGui) OptimizerController.this.getGuiComponent();
				if (gui != null) {
					gui.close();
				}

			}
		}.start();
	}

	public void mnu_file_exit() {
		if (!OptimizerController.this.saveInProgress) {
			new Thread() {
				@Override
				public void run() {
					OptimizerGui gui;

					OptimizerController.this.showProgressBar("Exit");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.saveInProgress = true;
						OptimizerController.this.optim.shutdown();
						OptimizerController.this.saveInProgress = false;
					}

					OptimizerController.this.hideProgressBar();

					gui = (OptimizerGui) OptimizerController.this.getGuiComponent();
					if (gui != null) {
						gui.exit();
					}
				}
			}.start();
		}
	}

	public void mnu_display_refresh() {
		new Thread() {
			@Override
			public void run() {
				ComputeProject calcProject;

				OptimizerController.this.showProgressBar("Refresh project");

				if (OptimizerController.this.optim != null) {
					try {
						calcProject = OptimizerController.this.optim.getProjectControl();
						calcProject.getProblem().createSavePoint(SavePointType.USER, true, calcProject.getNumberValid(),
								calcProject.getNumberInvalid(), calcProject.getNumberError());
						OptimizerController.this.optim
								.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PROBLEM));
						OptimizerController.this.optim
								.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.MAPPING));
					} catch (Exception e) {
						System.err.println(e.getMessage());
						e.printStackTrace();
					}
				}
				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_display_jobmanager() {
		new Thread() {
			@Override
			public void run() {
				OptimizerController.this.showProgressBar("Open JobManager");

				JobDesc.setVisible(true);

				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_display_add_plot() {
		String name;

		if (this.problem != null) {
			name = JOptionPane.showInputDialog(this.getGuiComponent(), "Name for new plot", "Add plot",
					JOptionPane.QUESTION_MESSAGE);
			OptimizerController.this.mnu_display_add_plot(name);
		}
	}

	public void mnu_display_add_plot(final String name) {
		if ((name != null) && !name.isEmpty()) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Add plot");
					OptimizerController.this.problem.addUserPlot(name);

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_display_remove_plot() {
		String name;
		String[] plots;

		if (this.problem != null) {
			plots = this.problem.listUserPlots();
			name = (String) JOptionPane.showInputDialog(this.getGuiComponent(), "Select Plot to remove", "Remove Plot",
					JOptionPane.INFORMATION_MESSAGE, null, plots, plots[plots.length - 1]);
			this.mnu_display_remove_plot(name);
		}
	}

	public void mnu_display_remove_plot(final String name) {
		if ((name != null)) {
			new Thread() {
				@Override
				public void run() {
					ProgressBar pb;

					OptimizerController.this.showProgressBar("Remove plot");
					OptimizerController.this.problem.rmUserPlot(name);

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_project_run() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Run");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.runProject();
					}

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();

					OptimizerController.this.updateStopContition();
				}
			}.start();
		}
	}

	public void mnu_project_finish() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Finish");

					OptimizerController.this.getOptimization().finishProject();

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_project_stop() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Stop");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.stopProject();
					}

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_project_init_buffer() {

		if (this.optim != null) {
			new Thread() {
				@Override
				public void run() {
					final JSpinner spinnerInit;
					JButton btnApply;
					JPanel spinnerPanel, buttonPanel;
					final JDialog dia;

					dia = new JDialog();
					dia.setTitle("Initialize buffer");
					dia.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
					dia.getContentPane().setLayout(new BorderLayout());
					dia.setModal(true);
					dia.setIconImages(OptIcons.getIcons());

					// spinner
					spinnerInit = new JSpinner();
					if (OptimizerController.this.optim != null) {
						spinnerInit.setModel(new SpinnerNumberModel(
								OptimizerController.this.optim.getProblem().getAutosaveInterval(), 1, null, 10));
					}
					spinnerPanel = new JPanel(new BorderLayout());
					spinnerPanel.setBorder(new EmptyBorder(5, 5, 5, 5));
					spinnerPanel.add(spinnerInit);
					dia.getContentPane().add(spinnerPanel, BorderLayout.CENTER);

					// button
					btnApply = new JButton("Apply");
					btnApply.addActionListener(new ActionListener() {
						@Override
						public void actionPerformed(ActionEvent e) {
							int num;

							dia.setVisible(false);
							num = (Integer) spinnerInit.getValue();
							dia.dispose();
							OptimizerController.this.mnu_project_init_buffer(num);

						}
					});
					buttonPanel = new JPanel(new BorderLayout());
					buttonPanel.setBorder(new EmptyBorder(0, 5, 5, 5));
					buttonPanel.add(btnApply);
					dia.getContentPane().add(buttonPanel, BorderLayout.SOUTH);

					// this.setSize(100, 100);
					dia.pack();
					dia.setLocationByPlatform(true);
					dia.setVisible(true);
				}
			}.start();
		}
	}

	public void mnu_project_init_buffer(final int number) {
		if (number > 0) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Init buffer");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.initProjectBuffer(number);
					}

					if (OptimizerController.this.getViewManager() != null) {
						OptimizerController.this.getViewManager().getStatusBar().refresh();
					}
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}

	}

	public void mnu_project_clear_buffer() {
		new Thread() {
			@Override
			public void run() {
				OptimizerController.this.showProgressBar("Clear buffer");

				if (OptimizerController.this.optim != null) {
					OptimizerController.this.optim.clearProjectBuffer();
				}

				if (OptimizerController.this.getViewManager() != null) {
					OptimizerController.this.getViewManager().getStatusBar().refresh();
				}
				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_project_create_excel() {
		JFileChooser chooser;

		chooser = new JFileChooser();
		if (this.optim != null) {
			chooser.setCurrentDirectory(this.optim.getProblem().getArchive().getMop().getParentFile());
		}
		chooser.setFileFilter(new FileNameExtensionFilter("Excel File (*.xlsx / *.xls)", "xls", "xlsx"));

		if (chooser.showSaveDialog(this.getGuiComponent()) == JFileChooser.APPROVE_OPTION) {
			this.mnu_project_create_excel(chooser.getSelectedFile());
		}
	}

	public void mnu_project_create_excel(final File excel) {
		new Thread() {
			@Override
			public void run() {
				File file;

				OptimizerController.this.showProgressBar("Clear buffer");
				try {
					file = OptimizerController.this.optim.writeProjectToExcel(excel);

					if (file.exists()) {
						try {
							if (Desktop.isDesktopSupported()) {
								Desktop.getDesktop().open(file);
							}
						} catch (Exception ex) {
						}
					}
				} catch (Exception ex) {
					Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
				}

				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_project_select() {
		if (this.getViewManager() != null) {
			new Thread() {
				@Override
				public void run() {
					Individual[] list;
					int[] selected, suppressed;
					MopFile mop;
					boolean doc;
					JFileChooser chooser;
					PanelPlots polts;
					File path = null;
					polts = OptimizerController.this.getViewManager().plots;
					if (polts != null) {
						list = polts.getListEnabledIndividuals();
						selected = polts.getSelectedIndices();
						suppressed = polts.getSuppressedIndices();

						if (selected.length > 0) {
							if (OptimizerController.this.problem != null) {
								path = OptimizerController.this.problem.getArchive().getMop().getParentFile();
							}

							chooser = new JFileChooser("Create SyMSpace-Project");
							chooser.setCurrentDirectory(path);
							chooser.setFileFilter(new FileNameExtensionFilter("SyMSpace project file (*.mop)", "mop"));
							if (chooser.showSaveDialog(
									OptimizerController.this.getGuiComponent()) == JFileChooser.APPROVE_OPTION) {
								mop = new MopFile(chooser.getSelectedFile());
								if (!mop.exists()
										|| (JOptionPane.showConfirmDialog(OptimizerController.this.getGuiComponent(),
												"File already exists. Do yout want to override the file?",
												"Confirm override",
												JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)) {
									doc = JOptionPane.showConfirmDialog(OptimizerController.this.getGuiComponent(),
											"Create doc files?", "Documentation",
											JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION;

									OptimizerController.this.mnu_project_select(mop, list, selected, suppressed, doc,
											true);
								}
							}
						} else {
							JOptionPane.showMessageDialog(OptimizerController.this.getGuiComponent(),
									"No global pareto is selected!", "Select optimum", JOptionPane.OK_OPTION);
						}
					}
				}
			}.start();
		}
	}

	public void mnu_project_select(final MopFile mop, final Individual[] list, final int[] selected,
			final int[] suppressed, final boolean doc, final boolean open) {
		if (selected.length > 0) {
			new Thread() {
				@Override
				public void run() {
					Project project;

					try {
						project = OptimizerController.this.getOptimization().selectPareto(mop, list, selected,
								suppressed, doc);
						project.save();

						if ((project != null) && open) {
							CenterController.openProjectExtern(project.getMopFile());
						}
					} catch (Exception ex) {
						Logger.getLogger(OptimizerGui.class.getName()).log(Level.SEVERE, null, ex);
					}
				}
			}.start();
		}

	}

	public void mnu_mapping_create_all() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Create all");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.createMapping();
					}

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_mapping_optimize() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Run mapping");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.runMapping();
					}

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_mapping_sensitivity() {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					final int[] selected;
					Individual[] list;

					selected = OptimizerController.this.getViewManager().plots.getSelectedIndices();
					list = OptimizerController.this.getViewManager().plots.getListEnabledIndividuals();

					if (selected.length < 1) {
						JOptionPane.showMessageDialog(OptimizerController.this.getGuiComponent(),
								"No global Pareto-optimal design is selected!",
								"Select design for Sensitivity Analysis", JOptionPane.OK_OPTION);
					} else if (selected.length > 1) {
						JOptionPane.showMessageDialog(OptimizerController.this.getGuiComponent(),
								"Multiple Pareto-optimal designs selected! Please choose just one.",
								"Select design for Sensitivity Analysis", JOptionPane.OK_OPTION);
					} else {
						// JOptionPane.showMessageDialog(this.getGuiComponent(),
						// "Design nr." + selected[0] + " selected for SA.",
						// "Select design for Sensitivity Analysis",
						// JOptionPane.OK_OPTION);

						OptimizerController.this.mnu_mapping_sensitivity(list[selected[0]]);
					}
				}
			}.start();
		}
	}

	public void mnu_mapping_sensitivity(final Individual chosenInd) {
		if (this.getOptimization() != null) {
			new Thread() {
				@Override
				public void run() {
					OptimizerController.this.showProgressBar("Run mapping");

					if (OptimizerController.this.optim != null) {
						OptimizerController.this.optim.initializeSensitivityAnalysis(chosenInd);
					}

					OptimizerController.this.refreshViews();
					OptimizerController.this.hideProgressBar();
				}
			}.start();
		}
	}

	public void mnu_settings_problem() {
		new Thread() {
			@Override
			public void run() {
				JFrame frame;
				ChooseProblem panel;
				//
				// OptimizerController.this.showProgressBar("Settings",
				// "Open problem");

				frame = new JFrame("Problem");
				// OptimizerGui.this,
				panel = new ChooseProblem(OptimizerController.this.problem);
				panel.setSettingEnabled(false);
				frame.getContentPane().add(panel);
				frame.setSize(670, 500);
				frame.setLocationByPlatform(true);
				frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				frame.setIconImages(OptIcons.getIcons());

				// OptimizerController.this.hideProgressBar();

				frame.setVisible(true);
			}
		}.start();
	}

	public void mnu_settings_variables() {
		new Thread() {
			@Override
			public void run() {
				JFrame frame;
				ChooseVariables panel;
				//
				// OptimizerController.this.showProgressBar("Settings",
				// "Open variables");

				frame = new JFrame("Variables");
				// OptimizerGui.this,
				panel = new ChooseVariables(OptimizerController.this.problem);
				panel.setSettingEnabled(false);
				frame.getContentPane().add(panel);
				frame.setSize(670, 500);
				frame.setLocationByPlatform(true);
				frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				frame.setIconImages(OptIcons.getIcons());

				// OptimizerController.this.hideProgressBar();

				frame.setVisible(true);
			}
		}.start();
	}

	public void mnu_settings_results() {
		new Thread() {
			@Override
			public void run() {
				JFrame frame;
				ChooseResults panel;
				//
				// OptimizerController.this.showProgressBar("Settings",
				// "Open results");

				frame = new JFrame("Results");
				// OptimizerGui.this,
				panel = new ChooseResults(OptimizerController.this.problem);
				panel.setSettingEnabled(false);
				frame.getContentPane().add(panel);
				frame.setSize(670, 500);
				frame.setLocationByPlatform(true);
				frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				frame.setIconImages(OptIcons.getIcons());

				// OptimizerController.this.hideProgressBar();

				frame.setVisible(true);
			}
		}.start();
	}

	public void mnu_settings_general() {
		new Thread() {
			@Override
			public void run() {
				JFrame frame;
				GeneralSettings panel;
				//
				// OptimizerController.this.showProgressBar("Settings",
				// "Open general settings");

				frame = new JFrame("General settings");
				// OptimizerGui.this,
				panel = new GeneralSettings(OptimizerController.this.problem);
				panel.setSettingEnabled(false);
				frame.getContentPane().add(panel);
				frame.setSize(670, 500);
				frame.setLocationByPlatform(true);
				frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
				frame.setIconImages(OptIcons.getIcons());

				// OptimizerController.this.hideProgressBar();

				frame.setVisible(true);
			}
		}.start();
	}

	public void mnu_settings_solver() {
		new Thread() {
			@Override
			public void run() {
				JFrame frame;
				SolverSettings panel;
				ComputeProject cProject;
				//
				// OptimizerController.this.showProgressBar("Settings",
				// "Open sovler settings");

				if (OptimizerController.this.optim != null) {
					cProject = OptimizerController.this.optim.getProjectControl();
					frame = new JFrame("Solver settings");
					// OptimizerGui.this,
					panel = new SolverSettings(cProject.getProblem().getSolver());
					panel.setSettingEnabled(false);
					frame.getContentPane().add(panel);
					frame.setSize(670, 500);
					frame.setLocationByPlatform(true);
					frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
					frame.setIconImages(OptIcons.getIcons());

					// OptimizerController.this.hideProgressBar(pb);

					frame.setVisible(true);
				}
			}
		}.start();
	}

	public void mnu_debug_assign_problem_python() {
		new Thread() {
			@Override
			public void run() {
				SendToDialog dia;
				JepProxy jep;

				try {
					dia = new SendToDialog(null);
					dia.show("Assign problem handle to Python console", "Problem");
					if (dia.getVariableName() != null) {

						OptimizerController.this.showProgressBar("Assign problem to python");

						jep = OptimizerController.this.getViewManager().pythonConsole.getJep();
						if (jep == null) {
							OptimizerController.this.getOptimization().getEventLog().warning("",
									"No python-console avaiable", new Exception().getStackTrace());
						} else {
							if (OptimizerController.this.optim != null) {
								jep.eval("from symspace.wrapper.symoptwrapper import SymProblem");
								jep.set(dia.getVariableName(), OptimizerController.this.optim.getProblem());
								jep.eval(dia.getVariableName() + " = SymProblem(" + dia.getVariableName() + ")");
								jep.eval("del SymProblem");
							}
						}
					}
				} catch (Exception ex) {
					OptimizerController.this.getOptimization().getEventLog().warning("", ex.getMessage(),
							ex.getStackTrace());
				} finally {
					OptimizerController.this.hideProgressBar();
				}
			}
		}.start();
	}

	public void mnu_debug_assign_project_python() {
		new Thread() {
			@Override
			public void run() {
				SendToDialog dia;
				JepProxy jep;

				try {
					dia = new SendToDialog(null);
					dia.show("Assign project handle to Python console", "OptProject");
					if (dia.getVariableName() != null) {

						OptimizerController.this.showProgressBar("Assign project to python");

						jep = OptimizerController.this.getViewManager().pythonConsole.getJep();
						if (jep == null) {
							OptimizerController.this.getOptimization().getEventLog().warning("",
									"No python-console avaiable", new Exception().getStackTrace());
						} else {
							if (OptimizerController.this.optim != null) {
								jep.eval("from symspace.wrapper.symwrapper import SymParameter");
								jep.set(dia.getVariableName(),
										OptimizerController.this.optim.getProjectControl().getProject());
								jep.eval(dia.getVariableName() + " = SymParameter._convertObject("
										+ dia.getVariableName() + ")");
								jep.eval("del SymParameter");
							}
						}
					}
				} catch (Exception ex) {
					OptimizerController.this.getOptimization().getEventLog().warning("", ex.getMessage(),
							ex.getStackTrace());
				} finally {
					OptimizerController.this.hideProgressBar();
				}
			}
		}.start();
	}

	public void mnu_debug_assign_controller_python() {
		new Thread() {
			@Override
			public void run() {
				SendToDialog dia;
				JepProxy jep;

				try {
					dia = new SendToDialog(null);
					dia.show("Assign controller handle to Python console", "Controller");
					if (dia.getVariableName() != null) {

						OptimizerController.this.showProgressBar("Assign controller to python");

						jep = OptimizerController.this.getViewManager().pythonConsole.getJep();
						if (jep == null) {
							OptimizerController.this.getOptimization().getEventLog().warning("",
									"No python-console avaiable", new Exception().getStackTrace());
						} else {
							jep.set(dia.getVariableName(), OptimizerController.this);
						}
					}
				} catch (Exception ex) {
					OptimizerController.this.getOptimization().getEventLog().warning("", ex.getMessage(),
							ex.getStackTrace());
				} finally {
					OptimizerController.this.hideProgressBar();
				}
			}
		}.start();
	}

	public void mnu_debug_assign_open_project() {
		new Thread() {
			@Override
			public void run() {
				Project proj;

				try {
					OptimizerController.this.showProgressBar("Open optimizer project");

					if (OptimizerController.this.optim != null) {
						proj = OptimizerController.this.optim.getProjectControl().getProject();
						if (proj != null) {
							CenterGui gui = new CenterGui();
							gui.getController().openProject(proj);
							gui.setVisible(true);
						}
					}
				} catch (Exception ex) {
					OptimizerController.this.getOptimization().getEventLog().warning("", ex.getMessage(),
							ex.getStackTrace());
				} finally {
					OptimizerController.this.hideProgressBar();
				}
			}
		}.start();
	}

	public void mnu_debug_macro() {
		OptimizerViewManager view;
		Component comp = null;

		view = this.getViewManager();
		if (view != null) {
			comp = this.getGuiComponent();
		}

		this.mnu_debug_macro(new OptimizerMacroDialog(comp).getMacro());
	}

	public void mnu_debug_macro(final File script) {
		new Thread() {
			@Override
			public void run() {
				OptimizerViewManager view;
				PythonConsole py;
				JepProxy jep;
				Project proj = null;

				OptimizerController.this.showProgressBar("Run macro");

				view = OptimizerController.this.getViewManager();
				if (view != null) {
					py = view.pythonConsole;
					if (py != null) {
						if (script != null) {
							try {
								jep = py.getJep();
								jep.set("__controller__", OptimizerController.this);
								if (OptimizerController.this.optim != null) {
									jep.eval("from symspace.wrapper.symoptwrapper import SymProblem");
									jep.set("__problem__", OptimizerController.this.optim.getProblem());
									jep.eval("__problem__" + " = SymProblem(" + "__problem__" + ")");
									jep.eval("del SymProblem");

									proj = OptimizerController.this.optim.getProjectControl().getProject();
									if (proj != null) {
										jep.eval("from symspace.wrapper.symwrapper import SymParameter");
										jep.set("__project__", proj);
										jep.eval("__project__ = SymParameter._convertObject(__project__)");
										jep.eval("del SymParameter");
									}
								}
								py.getConsole().writeInput(
										"Run macro: " + script.getName() + " (" + script.getParent() + ")\n");
								jep.eval("exec(open(r'" + script.getAbsolutePath() + "').read())");
								jep.eval("del __controller__");
								if (OptimizerController.this.optim != null) {
									jep.eval("del __problem__");
									if (proj != null) {
										jep.eval("del __project__");
									}
								}
							} catch (Exception ex) {
								py.getConsole().getErrorPrintStream().println("Error by run macro (" + script + ")");
								py.getConsole().getErrorPrintStream().println(ex.getMessage());
								ex.printStackTrace(py.getConsole().getErrorPrintStream());
							}
						}
					} else {
						System.err.println("Python console not available");
					}
					view.refresh();
				}
				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_plot_browse() {
		new Thread() {
			@Override
			public void run() {
				WebPlotInterface.setProblem(OptimizerController.this.problem, OptimizerController.this.plotVariables,
						OptimizerController.this.plotResults);
				WebPlotInterface.browse();
			}
		}.start();
	}

	public boolean isPlotVariables() {
		return this.plotVariables;
	}

	public void mnu_plot_variables() {
		this.plotVariables = !this.plotVariables;
	}

	public boolean isPlotObjectives() {
		return this.plotResults == ResultType.OBJECTIVES;
	}

	public void mnu_plot_objectives() {
		this.plotResults = ResultType.OBJECTIVES;
	}

	public boolean isPlotConstraints() {
		return this.plotResults == ResultType.CONSTAINTS;
	}

	public void mnu_plot_constraints() {
		this.plotResults = ResultType.CONSTAINTS;
	}

	public boolean isPlotResults() {
		return this.plotResults == ResultType.RESULTS;
	}

	public void mnu_plot_results() {
		this.plotResults = ResultType.RESULTS;
	}

	public void mnu_plot_json() {
		// boolean open = true;
		String name;
		int index;

		JFileChooser chooser = new JFileChooser();
		FileNameExtensionFilter fJson = new FileNameExtensionFilter("JavaScript Object Notation (*.json)", "json");
		chooser.addChoosableFileFilter(fJson);
		chooser.setFileFilter(fJson);
		chooser.setAcceptAllFileFilterUsed(false);
		chooser.setDialogTitle("Choose File");
		chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
		chooser.setMultiSelectionEnabled(false);
		chooser.setFileHidingEnabled(true);

		// generate new name
		name = this.problem.getArchive().getMop().getName();
		index = name.lastIndexOf('.');
		if (index != -1) {
			name = name.substring(0, index);
		}
		name += "_" + this.problem.getArchive().getName() + "_" + this.problem.getName();
		if (!this.plotVariables) {
			name += "-minimal";
		}
		name += ".json";

		chooser.setSelectedFile(new File(this.problem.getArchive().getMop().getParentFile(), name));
		if (chooser.showSaveDialog(this.getGuiComponent()) == JFileChooser.APPROVE_OPTION) {
			this.mnu_plot_json(chooser.getSelectedFile());
		}
	}

	public void mnu_plot_json(File json) {
		new Thread() {
			@Override
			public void run() {
				OptimizerController.this.showProgressBar("Export JSON");
				try {
					PlotJson.exportJson(OptimizerController.this.problem, OptimizerController.this.plotVariables,
							OptimizerController.this.plotResults, json);
				} catch (IOException e) {
					// FIXME event
					System.err.println("Unable to export json");
					e.printStackTrace();
				}
				OptimizerController.this.refreshViews();
				OptimizerController.this.hideProgressBar();
			}
		}.start();
	}

	public void mnu_help_about() {
		new Thread() {
			@Override
			public void run() {
				// ProgressBar pb;
				//
				// pb = OptimizerController.this.showProgressBar("Help",
				// "Open about");

				new About(OptimizerController.this.getGuiComponent(), OptIcons.getIcons()).setVisible(true);

				// OptimizerController.this.hideProgressBar(pb);
			}
		}.start();
	}

	public Optimization getOptimization() {
		return this.optim;
	}

	// public EventLog getEventLog() {
	// return this.logger;
	// }
	//
	// public ComputeProject getComputeProject() {
	// return this.calcProject;
	// }

	public boolean isRunEnabled() {
		ComputeProject proj;
		boolean enable = false;

		if (this.optim != null) {
			proj = this.optim.getProjectControl();
			enable = proj.isStop() || !proj.isRun() || !proj.isAdd();
		}

		return enable;
	}

	public boolean isFinishEnabled() {
		ComputeProject proj;
		boolean enable = false;

		if (this.optim != null) {
			proj = this.optim.getProjectControl();
			enable = proj.isStop() || !proj.isRun() || proj.isAdd();
		}

		return enable;
	}

	public boolean isStopEnabled() {
		ComputeProject proj;
		boolean enable = false;

		if (this.optim != null) {
			proj = this.optim.getProjectControl();
			enable = proj.isRun();
		}

		return enable;
	}

	public boolean isCreateMappingEnabled() {
		MappingManager map;

		boolean enable = false;

		if (this.optim != null) {
			map = this.optim.getMappingControl();

			enable = !map.areAnyMappingHoldersBusy() && !map.isOptimizationRunning() && map.hasSufficientSamples()
					&& map.isSensitivityAnalysisIdle();
		}

		return enable;
	}

	public boolean isRunMappingEnabled() {
		MappingManager map;

		boolean enable = false;

		if (this.optim != null) {
			map = this.optim.getMappingControl();

			enable = map.areAllMappingHoldersReady() && !map.isOptimizationRunning() && map.isOptimizationIdle()
					&& map.isSensitivityAnalysisIdle();
		}

		return enable;
	}

	public boolean isSensitivityInitializationEnabled() {
		MappingManager map;

		boolean enable = false;

		if (this.optim != null) {
			map = this.optim.getMappingControl();

			enable = map.areAllMappingHoldersExceptErrorsClsReady() && map.isSensitivityAnalysisIdle();
		}

		return enable;
	}

	public void updateStopContition() {
		if ((this.optim != null) && (this.optim.getProjectControl() != null)) {
			this.optim.getProjectControl().clearStopCondition();
			new DialogStopCondition(this.optim.getProjectControl(), this.getGuiComponent());
			this.refreshViews();
		}
	}

	@Override
	public OptimizerViewManager getViewManager() {
		return (OptimizerViewManager) super.getViewManager();
	}

}
