/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.docu;

import java.awt.Color;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import org.apache.poi.hssf.usermodel.HSSFWorkbook;
import org.apache.poi.hssf.util.HSSFColor;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CellStyle;
import org.apache.poi.ss.usermodel.Font;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.xssf.usermodel.XSSFCellStyle;
import org.apache.poi.xssf.usermodel.XSSFColor;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.ParamDirectory;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.ParamFile;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;

/**
 * Create Excel report of MagOpt project
 * 
 * @author Werner Koppelsttter
 */
public class ReportExcel {

	/** Blue for caption */
	static final Color BLUE = new Color(0, 159, 227);
	/** Gray for container */
	static final Color GRAY = new Color(217, 217, 217);
	/** Yellow for uncertain */
	static final Color YELLOW = new Color(255, 255, 103);

	/** Style for caption */
	private CellStyle captionStyle;

	/** Style for container */
	private CellStyle containerStyle;

	/** Style for uncertain */
	private CellStyle uncertainStyle;

	/** Style for italic uncertain */
	private CellStyle uncertainItalicStyle;

	/** Style with a font of size 14 */
	private CellStyle font14Style;

	/** Style with a bold fond of size 18 */
	private CellStyle font18BStyle;

	/** Style with italic style */
	private CellStyle fontItalicStyle;

	/**
	 * Report project in a new excel workbook
	 * 
	 * @param proj
	 *            Project for report
	 * @param excel
	 *            Wanted name of report file
	 * @return Name of created report (with right extension)
	 * @throws IOException
	 */
	public File reportExcel(Project proj, File excel) throws IOException {
		Workbook wb;
		Sheet sheet;
		Font font;
		FileOutputStream out;
		boolean xssf;

		// open file
		if (!excel.getName().toLowerCase().endsWith(".xls")) {
			xssf = true;
			if (!excel.getName().toLowerCase().endsWith(".xlsx")) {
				excel = new File(excel.getPath() + ".xlsx");
			}
			wb = new XSSFWorkbook();
		} else {
			xssf = false;
			wb = new HSSFWorkbook();
		}

		// caption style
		this.captionStyle = wb.createCellStyle();
		this.captionStyle.setFillPattern(CellStyle.SOLID_FOREGROUND);
		font = wb.createFont();
		font.setBoldweight(Font.BOLDWEIGHT_BOLD);
		this.captionStyle.setFont(font);
		if (xssf) {
			((XSSFCellStyle) this.captionStyle).setFillForegroundColor(new XSSFColor(ReportExcel.BLUE));
		} else {
			((HSSFWorkbook) wb).getCustomPalette().setColorAtIndex(HSSFColor.GREEN.index,
					(byte) ReportExcel.BLUE.getRed(), (byte) ReportExcel.BLUE.getGreen(),
					(byte) ReportExcel.BLUE.getBlue());
			this.captionStyle.setFillForegroundColor(HSSFColor.GREEN.index);
		}

		// container style
		this.containerStyle = wb.createCellStyle();
		this.containerStyle.setFillPattern(CellStyle.SOLID_FOREGROUND);
		font = wb.createFont();
		font.setBoldweight(Font.BOLDWEIGHT_BOLD);
		// font.setBold(true);
		this.containerStyle.setFont(font);
		if (xssf) {
			((XSSFCellStyle) this.containerStyle).setFillForegroundColor(new XSSFColor(ReportExcel.GRAY));
		} else {
			((HSSFWorkbook) wb).getCustomPalette().setColorAtIndex(HSSFColor.GREY_25_PERCENT.index,
					(byte) ReportExcel.GRAY.getRed(), (byte) ReportExcel.GRAY.getGreen(),
					(byte) ReportExcel.GRAY.getBlue());
			this.containerStyle.setFillForegroundColor(HSSFColor.GREY_25_PERCENT.index);
		}

		// uncertain style
		this.uncertainStyle = wb.createCellStyle();
		this.uncertainStyle.setFillPattern(CellStyle.SOLID_FOREGROUND);
		if (xssf) {
			((XSSFCellStyle) this.uncertainStyle).setFillForegroundColor(new XSSFColor(ReportExcel.YELLOW));
		} else {
			((HSSFWorkbook) wb).getCustomPalette().setColorAtIndex(HSSFColor.YELLOW.index,
					(byte) ReportExcel.YELLOW.getRed(), (byte) ReportExcel.YELLOW.getGreen(),
					(byte) ReportExcel.YELLOW.getBlue());
			this.uncertainStyle.setFillForegroundColor(HSSFColor.YELLOW.index);
		}

		// uncertain italic style
		this.uncertainItalicStyle = wb.createCellStyle();
		this.uncertainItalicStyle.setFillPattern(CellStyle.SOLID_FOREGROUND);
		font = wb.createFont();
		font.setItalic(true);
		this.uncertainItalicStyle.setFont(font);
		if (xssf) {
			((XSSFCellStyle) this.uncertainItalicStyle).setFillForegroundColor(new XSSFColor(ReportExcel.YELLOW));
		} else {
			((HSSFWorkbook) wb).getCustomPalette().setColorAtIndex(HSSFColor.YELLOW.index,
					(byte) ReportExcel.YELLOW.getRed(), (byte) ReportExcel.YELLOW.getGreen(),
					(byte) ReportExcel.YELLOW.getBlue());
			this.uncertainItalicStyle.setFillForegroundColor(HSSFColor.YELLOW.index);
		}

		// style with font of 14
		this.font14Style = wb.createCellStyle();
		font = wb.createFont();
		font.setFontHeightInPoints((short) 14);
		this.font14Style.setFont(font);

		// style with font of 18 and bold
		this.font18BStyle = wb.createCellStyle();
		font = wb.createFont();
		font.setBoldweight(Font.BOLDWEIGHT_BOLD);
		font.setFontHeightInPoints((short) 18);
		this.font18BStyle.setFont(font);

		// style with font of italic
		this.fontItalicStyle = wb.createCellStyle();
		font = wb.createFont();
		font.setItalic(true);
		this.fontItalicStyle.setFont(font);

		sheet = wb.createSheet();

		this.reportProject(proj, sheet);

		out = new FileOutputStream(excel);
		wb.write(out);
		out.close();

		return excel;

	}

	/**
	 * Report project with header and footer
	 * 
	 * @param proj
	 *            Project for report
	 * @param sheet
	 *            Spreadsheet for report
	 */
	private void reportProject(Project proj, Sheet sheet) {
		Cell cell;
		String str;
		int row;
		int indizes;

		indizes = proj.getLength();

		this.getCell(sheet, 0, 0).setCellValue("SyMSpace V1.0 - Report");
		this.setRowStyle(sheet, 0, this.captionStyle, indizes);

		cell = this.getCell(sheet, 3, 0);
		cell.setCellValue(proj.getName() + " - " + proj.getDescription());
		cell.setCellStyle(this.font18BStyle);

		for (int index = 0; index < indizes; index++) {
			cell = this.getCell(sheet, 5, index + 3);
			cell.setCellValue("Index #" + (index + 1));
			cell.setCellStyle(this.font14Style);

			str = proj.getDSComment(index);
			if ((str != null) && !str.isEmpty()) {
				cell = this.getCell(sheet, 6, index + 3);
				cell.setCellValue(str);
				cell.setCellStyle(this.font14Style);
			}
		}

		row = 8;
		for (Parameter p : proj.getParam()) {
			row = this.reportParam(sheet, row, p, "", indizes);
		}

		this.getCell(sheet, row + 2, 0)
				.setCellValue("(c) 2008 Linz Center of Mechatronics GmbH # Altenberger Str. 69 # A-4040 Linz");
		this.setRowStyle(sheet, row + 2, this.captionStyle, indizes);

	}

	/**
	 * Report all parameter
	 * 
	 * @param sheet
	 *            Spreadsheet for report
	 * @param row
	 *            Current row to insert
	 * @param param
	 *            Parameter to report
	 * @param indent
	 *            Indent suitable to the structure deepness
	 * @param indices
	 *            Number of data indices
	 * @return New number of current row
	 */
	private int reportParam(Sheet sheet, int row, Parameter param, String indent, int indices) {
		String str;
		TData data;
		Cell cell;

		if (param instanceof ParamContainer) {
			// Name and style
			this.getCell(sheet, row, 0).setCellValue(indent + param.getName());
			this.setRowStyle(sheet, row, this.containerStyle, indices);
			row++;

			// sub parameter
			for (Parameter p : ((ParamContainer) param).getParam()) {
				row = this.reportParam(sheet, row, p, indent + "     ", indices);
			}

		} else if (param.getPrint()) {
			if (param instanceof ParamField) {
				// line style
				if (!param.getUsed()) {
					this.setRowStyle(sheet, row, this.fontItalicStyle, indices);
				}

				// Name
				this.getCell(sheet, row, 0).setCellValue(indent + param.getName());

				// Description
				str = param.getDescription();
				if ((str != null) && !str.isEmpty()) {
					this.getCell(sheet, row, 1).setCellValue(str);
				}

				// Tex
				this.getCell(sheet, row, 2).setCellValue(indent + param.getName());

				// Value
				for (int index = 0; index < indices; index++) {
					data = param.getData(index);
					cell = this.getCell(sheet, row, index + 3);
					if ((data instanceof TNumeric) && ((TNumeric) data).isScalar()) {
						cell.setCellValue(((TNumeric) data).getReal(0, 0));
					} else {
						cell.setCellValue(data.toString());
					}
					if (param.getUncertain(index)) {
						if (param.getUsed()) {
							cell.setCellStyle(this.uncertainStyle);
						} else {
							cell.setCellStyle(this.uncertainItalicStyle);
						}
					}
				}

				// Unit
				str = ((ParamField) param).getUnit();
				if ((str != null) && !str.isEmpty()) {
					this.getCell(sheet, row, indices + 3).setCellValue(str);
				}

				row++;

			} else if (param instanceof ParamFile) {
				// Name
				this.getCell(sheet, row, 0).setCellValue(indent + param.getName());

				// Description
				str = param.getDescription();
				if ((str != null) && !str.isEmpty()) {
					this.getCell(sheet, row, 1).setCellValue(str);
				}

				// file
				for (int index = 0; index < indices; index++) {
					str = ((ParamFile) param).getRelDirPath(index);
					if ((str != null) && !str.isEmpty()) {
						str = new File(str, ((ParamFile) param).getFileName(index)).getPath();
					} else {
						str = ((ParamFile) param).getFileName(index);
					}
					cell = this.getCell(sheet, row, index + 3);
					if ((str != null) && !str.isEmpty()) {
						cell.setCellValue(str);
					}
					if (param.getUncertain(index)) {
						if (param.getUsed()) {
							cell.setCellStyle(this.uncertainStyle);
						} else {
							cell.setCellStyle(this.uncertainItalicStyle);
						}
					}
				}

				row++;

			} else if (param instanceof ParamDirectory) {
				// Name
				this.getCell(sheet, row, 0).setCellValue(indent + param.getName());

				// Description
				str = param.getDescription();
				if ((str != null) && !str.isEmpty()) {
					this.getCell(sheet, row, 1).setCellValue(str);
				}

				// Directory
				for (int index = 0; index < indices; index++) {
					str = ((ParamDirectory) param).getRelDirPath(index);
					cell = this.getCell(sheet, row, index + 3);
					if ((str != null) && !str.isEmpty()) {
						cell.setCellValue(str);
					}
					if (param.getUncertain(index)) {
						if (param.getUsed()) {
							cell.setCellStyle(this.uncertainStyle);
						} else {
							cell.setCellStyle(this.uncertainItalicStyle);
						}
					}
				}

				row++;
			}
		}

		return row;
	}

	/**
	 * Set the style of a row (5 cells)
	 * 
	 * @param sheet
	 *            Spreadsheet
	 * @param rowIndex
	 *            Index of row
	 * @param style
	 *            Style of cell
	 * @param indices
	 *            Number of data indices
	 */
	private void setRowStyle(Sheet sheet, int rowIndex, CellStyle style, int indices) {
		for (int colIndex = 0; colIndex < (indices + 4); colIndex++) {
			this.getCell(sheet, rowIndex, colIndex).setCellStyle(style);
		}
	}

	/**
	 * Get the cell of a spreadsheet
	 * 
	 * @param sheet
	 *            Spreadsheet
	 * @param rowIndex
	 *            Row index of cell
	 * @param columnIndex
	 *            Column index of cell
	 * @return Selected cell
	 */
	private Cell getCell(Sheet sheet, int rowIndex, int columnIndex) {
		Row row;
		Cell cell;

		// get or create row
		row = sheet.getRow(rowIndex);
		if (row == null) {
			row = sheet.createRow(rowIndex);
		}

		// get or create cell
		cell = row.getCell(columnIndex);
		if (cell == null) {
			cell = row.createCell(columnIndex);
		}

		return cell;
	}

}
