/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10035 $
 * $Date: 2020-10-22 13:14:36 +0200 (Thu, 22 Oct 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.security.NoSuchAlgorithmException;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamFile;
import at.lcm.bu21.magopt.structure.rmi.RParamFile;

/**
 * Parameter representation of File
 * 
 * @author Werner Koppelsttter
 */
public class ParamFile extends ParamDirectory implements IParamFile {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** File from constant parameter */
	private String file = "";

	/** Checksum from constant parameter */
	private String checkSum = "";

	private long checkSumModified = 0l;

	@Override
	public List<String> listTabNames() {
		List<String> list = super.listTabNames();
		list.remove(0);
		list.add(0, "File Info");
		return list;
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamFile() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamFile(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamFile(String name, String description) throws Exception {
		this(name, description, "");
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        File
	 * @throws Exception
	 */
	public ParamFile(String name, String description, String file) throws Exception {
		super(name, description);
		this.setAbsoluteFilePath(file);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamFile other = (ParamFile) obj;
		if (this.file == null) {
			if (other.file != null) {
				return false;
			}
		} else if (!this.file.equals(other.file)) {
			return false;
		}
		if (this.checkSum == null) {
			if (other.checkSum != null) {
				return false;
			}
		} else if (!this.checkSum.equals(other.checkSum)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamFile deepClone() throws CloneException {
		return (ParamFile) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamFile) {
			ParamFile other = (ParamFile) clone;

			other.file = this.file;
			other.checkSum = this.checkSum;
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamFile) {
			ParamFile other = (ParamFile) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.FILE);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.FILE, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.FILE);
			}

			data = dsThis.getValue(this.getPId(), ParamSetKey.CHECKSUM);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.CHECKSUM, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.CHECKSUM);
			}
		}
	}

	/**
	 * Get checksum of current file
	 * 
	 * @return checksum
	 */
	public String getCheckSum() {
		Dataset ds;
		TData value;
		String checksum = "";

		if (this.getConstant()) {
			checksum = this.checkSum;
		} else {
			ds = this.getDataset();
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.CHECKSUM);
				if (value != null) {
					checksum = ((TString) value).getString();
				}
			}
		}

		return checksum;
	}

	/**
	 * Set checksum of current file
	 * 
	 * @param date     Last Modified
	 * @param checkSum checksum
	 */
	protected void setCheckSum(long date, String checkSum) {
		Dataset ds;

		// last modified date
		if (this.getConstant()) {
			this.checkSumModified = date;
		} else {
			ds = this.getDataset();
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.CHECKSUM_MODIFIED, new TString(Long.toString(date)));
			}
		}

		// check-sum
		if (!this.getCheckSum().equals(checkSum)) {

			if (this.getConstant()) {
				this.checkSum = checkSum;
				for (int i = 0; i < this.getLength(); i++) {
					this.setModified(i, true);
				}
			} else {
				ds = this.getDataset();
				if (ds != null) {
					ds.setValue(this.getPId(), ParamSetKey.CHECKSUM, new TString(checkSum));
				}
				this.setModified(true);
			}
		}
	}

	protected long getCheckSumModified() {
		Dataset ds;
		TData value;
		long lastModified = 0l;

		if (this.getConstant()) {
			lastModified = this.checkSumModified;
		} else {
			ds = this.getDataset();
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.CHECKSUM_MODIFIED);
				if (value != null) {
					lastModified = Long.parseLong(((TString) value).getString());
				}
			}
		}

		return lastModified;
	}

	@Override
	public String getFileName() {
		return this.getFileName(this.getIndex());
	}

	/**
	 * Get file-name of dataset
	 * 
	 * @param index Index of dataset
	 * @return File
	 */
	public String getFileName(int index) {
		Dataset ds;
		TData value;
		String file = "";

		if (this.getConstant()) {
			file = this.file;
		} else {
			ds = this.getDataset(index);
			if (ds != null) {
				value = ds.getValue(this.getPId(), ParamSetKey.FILE);
				if (value != null) {
					file = ((TString) value).getString();
				}
			}
		}

		return file;
	}

	@Override
	public void setFileName(String name) {
		Dataset ds;
		String ext;
		int i;

		if (name == null) {
			name = "";
		} else {
			while (name.startsWith(File.separator)) {
				name = name.substring(1);
			}
		}

		// check for proE files
		i = name.lastIndexOf(".");
		if (i != -1) {
			ext = name.substring(i);
			if (ext.matches("[.]\\d+")) {
				name = name.substring(0, name.lastIndexOf("."));
			}
		}

		// set file
		if (!this.getFileName().equals(name)) {
			if (this.getConstant()) {
				this.file = name;
				for (int k = 0; k < this.getLength(); k++) {
					this.setModified(k, true);
				}
			} else {
				ds = this.getDataset();
				if (ds != null) {
					ds.setValue(this.getPId(), ParamSetKey.FILE, new TString(name));
				}
				this.setModified(true);
			}
		}

		this.refreshFile();
	}

	/**
	 * Get absolute-file string
	 * 
	 * @return Absolute-file
	 */
	public String getAbsoluteFilePath() {
		return this.getAbsoluteFilePath(this.getIndex());
	}

	/**
	 * Get absolute-file-path of dataset
	 * 
	 * @param index Index of dataset
	 * @return Absolute-file
	 */
	public String getAbsoluteFilePath(int index) {
		return this.getAbsoluteFile(index).getPath();
	}

	/**
	 * Set absolute-file-path
	 * 
	 * @param path Absolute-path of file
	 */
	public void setAbsoluteFilePath(String path) {
		if (path == null) {
			path = "";
		}
		this.setAbsoluteFile(new File(path));
	}

	@Override
	public File getAbsoluteFile() {
		return this.getAbsoluteFile(this.getIndex());
	}

	/**
	 * Get absolute-file
	 * 
	 * @param index Index of dataset
	 * @return Absolute-file
	 */
	public File getAbsoluteFile(int index) {
		String path, file;
		File f;

		path = this.getAbsoluteDirectoryPath(index);
		file = this.getFileName(index);
		if (path.isEmpty()) {
			f = new File(file);
		} else {
			f = new File(path, file);
		}

		return f;
	}

	@Override
	public void setAbsoluteFile(File file) {
		if (file == null) {
			file = new File("");
			// } else {
			// while (file.getPath().startsWith(File.separator)) {
			// file = new File(file.getPath().substring(1));
			// }
		}

		this.setAbsoluteDirectory(file.getParentFile());
		this.setFileName(file.getName());
	}

	@Override
	public TData getData(int index) {
		return new TString(this.getAbsoluteFilePath(index));
	}

	@Override
	public void setData(TData data) throws Exception {
		if (data instanceof TString) {
			this.setAbsoluteFilePath(((TString) data).getString());
		} else {
			throw new Exception("Only Strings are allowed to set to " + this.getClass().getSimpleName());
		}
	}

	/**
	 * Refresh File and check checksum
	 */
	protected void refreshFile() {
		String strFile, strTmp;
		File file;
		File[] files;
		int index = -1;
		int version = 0;

		file = this.getAbsoluteFile();
		strFile = file.getPath();
		if ((strFile.contains(":") || strFile.startsWith("\\\\")) && file.getParentFile().exists()) {
			// FIXME root directory

			// find latest ProE file
			files = file.getParentFile().listFiles();
			for (int i = 0; i < files.length; i++) {
				strTmp = files[i].getAbsolutePath();
				if (strTmp.startsWith(strFile)) {
					strTmp = strTmp.substring(strFile.length());
					if (strTmp.isEmpty()) {
						if (version == 0) {
							index = i;
						}
					} else if (strTmp.startsWith(".")) {
						strTmp = strTmp.substring(1);
						try {
							int ver = Integer.parseInt(strTmp);
							if (ver > version) {
								index = i;
								version = ver;
							}
						} catch (NumberFormatException ex) {
						}
					}
				}
			}
			if (index != -1) {
				file = files[index];
			}

			if (file.exists() && file.isFile()) {
				try {
					if ((this.getCheckSumModified() == 0) || (this.getCheckSumModified() != file.lastModified())) {
						this.setCheckSum(file.lastModified(), FileUtils.checkSum(file));
					}
				} catch (NoSuchAlgorithmException e) {
					this.error(e.getMessage());
					this.setCheckSum(0l, "");
				} catch (IOException e) {
					this.error(e.getMessage());
					this.setCheckSum(0l, "");
				}
			} else {
				this.setCheckSum(0l, "");
			}
		} else {
			this.setCheckSum(0l, "");
		}
	}

	@Override
	public void setConstant(boolean constant) {
		Dataset ds;
		TData value;
		String strFile, strCheckSum;

		if (this.getConstant() != constant) {
			if (constant) {
				this.file = this.getFileName();
				this.checkSum = this.getCheckSum();
				for (int i = 0; i < this.getLength(); i++) {
					strFile = "";
					strCheckSum = "";
					ds = this.getDataset(i);
					if (ds != null) {
						value = ds.getValue(this.getPId(), ParamSetKey.FILE);
						ds.rmValue(this.getPId(), ParamSetKey.FILE);
						if (value != null) {
							strFile = ((TString) value).getString();
						}
						value = ds.getValue(this.getPId(), ParamSetKey.CHECKSUM);
						ds.rmValue(this.getPId(), ParamSetKey.CHECKSUM);
						if (value != null) {
							strCheckSum = ((TString) value).getString();
						}

					}
					if ((!this.file.equals(strFile)) || (!this.checkSum.equals(strCheckSum))) {
						this.setModified(i, true);
					}
				}
			} else {
				for (int i = 0; i < this.getLength(); i++) {
					ds = this.getDataset(i);
					if (ds != null) {
						ds.setValue(this.getPId(), ParamSetKey.FILE, new TString(this.file));
						ds.setValue(this.getPId(), ParamSetKey.CHECKSUM, new TString(this.checkSum));
					}
				}
				this.file = "";
				this.checkSum = "";
			}
			super.setConstant(constant);
		}
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamFile\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tFile: " + this.getAbsoluteFile());

		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamDirectory")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("File")) {
				xmlReader.next();
				this.file = "";
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.file += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("CheckSum")) {
				xmlReader.next();
				this.checkSum = "";
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.checkSum += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamFile");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// File
		xmlWriter.writeStartElement("File");
		xmlWriter.writeCharacters(this.file);
		xmlWriter.writeEndElement();

		// CheckSum
		xmlWriter.writeStartElement("CheckSum");
		xmlWriter.writeCharacters(this.checkSum);
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	void reportInfo(DocTexWriter content) {
		String path, projectPath;
		Parameter param;

		super.reportInfo(content);

		content.add("\tFile & ");
		content.addString(this.getFileName());
		content.add(" \\\\\n");

		path = this.getAbsoluteFile().getPath();
		param = this.getRoot();
		if (param instanceof Project) {
			projectPath = ((Project) param).getPath();
			if (path.indexOf(projectPath) == 0) {
				if (path.length() == projectPath.length()) {
					path = "." + File.separator;
				} else if (path.charAt(projectPath.length()) == File.separatorChar) {
					path = "." + path.substring(projectPath.length());
				}
			}
		}

		content.add("\tWhole file & \\mopath{");
		content.addString(path.replace(File.separatorChar, '/'));
		content.add("} \\\\\n");

		content.add("\tChecksum & ");
		content.add(this.getCheckSum());
		content.add(" \\\\\n");
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.FILE);
		list.add(ParamSetKey.CHECKSUM);

		return list;
	}

	@Override
	public IParamFile getRMIImpl() throws RemoteException {
		return new RParamFile(this);
	}

	@Override
	public IParamFile getClientImpl() {
		return new CParamFile(this);
	}
}
