/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9749 $
 * $Date: 2020-04-10 16:00:50 +0200 (Fri, 10 Apr 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.rmi.RemoteException;
import java.util.List;
import java.util.Map;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamLinkedField;
import at.lcm.bu21.magopt.structure.rmi.RParamLinkedField;

/**
 * A field that update its value from other field
 * 
 * @author Werner Koppelsttter
 */
public class ParamLinkedField extends ParamField implements IParamLinkedField, Function {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Dependency of Parameter */
	private Depend depend = new Depend(this);

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamLinkedField() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name
	 *            Name of Parameter
	 * @throws Exception
	 */
	public ParamLinkedField(String name) throws Exception {
		this(name, name);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name
	 *            Name of Parameter
	 * @param description
	 *            Description of Parameter
	 * @throws Exception
	 */
	public ParamLinkedField(String name, String description) throws Exception {
		super(name, description);
		this.addSetBy(this);
		this.getDepend().setImportVars(new PortDesc[] { new PortDesc("link", "Source of value", "", false) });
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamLinkedField other = (ParamLinkedField) obj;
		if (this.depend == null) {
			if (other.depend != null) {
				return false;
			}
		} else if (!this.depend.equals(other.depend)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamLinkedField deepClone() throws CloneException {
		return (ParamLinkedField) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException, InstantiationException, IllegalAccessException {

		super.cloneParam(clone);

		if (clone instanceof ParamLinkedField) {
			ParamLinkedField other = (ParamLinkedField) clone;

			this.depend.cloneDepend(other.depend);
		}
	}

	@Override
	void cloneParam(Parameter clone, Dataset dsThis, Dataset dsClone, boolean docFiles) throws CloneException {
		TData data;

		super.cloneParam(clone, dsThis, dsClone, docFiles);

		if (clone instanceof ParamLinkedField) {
			ParamLinkedField other = (ParamLinkedField) clone;

			data = dsThis.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (data != null) {
				dsClone.setValue(other.getPId(), ParamSetKey.UPDATEREQUIRED, data.deepClone());
			} else {
				dsClone.rmValue(other.getPId(), ParamSetKey.UPDATEREQUIRED);
			}
		}
	}

	@Override
	public Depend getDepend() {
		return this.depend;
	}

	@Override
	public void refreshDependency() {
	}

	@Override
	public boolean getReadOnly() {
		return this.getDepend().getImports()[0].getHandle() != null;
	}

	@Override
	public void setName(String name) throws Exception {
		super.setName(name);
	}

	@Override
	public boolean getUpdateRequired() {
		return this.getUpdateRequired(this.getIndex());
	}

	@Override
	public boolean getUpdateRequired(int index) {
		Dataset ds;
		TData value;
		boolean bool = true;

		ds = this.getDataset(index);
		if (ds != null) {
			value = ds.getValue(this.getPId(), ParamSetKey.UPDATEREQUIRED);
			if (value != null) {
				bool = ((TBoolean) value).getBoolean();
			}
		}

		return bool;
	}

	@Override
	public void setUpdateRequired(boolean update) {
		this.setUpdateRequired(this.getIndex(), update);
		// FIXME setUncertain (refresh)
		this.refreshFlags();
	}

	@Override
	public void setUpdateRequired(int index, boolean update) {
		Dataset ds;
		if (update != this.getUpdateRequired(index)) {
			ds = this.getDataset(index);
			if (ds != null) {
				ds.setValue(this.getPId(), ParamSetKey.UPDATEREQUIRED, new TBoolean(update));
			}
			this.resetDocFiles(index);
		}
	}

	@Override
	public void setConstant(boolean constant) {
		if (constant) {
			this.warning("Set linked-parameter to constant is not allowed!");
		}
	}

	@Override
	public boolean isEnable() {
		boolean enable = true;

		if (this.getParent() != null) {
			enable = this.getParent().isEnable();
		}
		return enable;
	}

	@Override
	protected boolean refreshIncomplete() {
		return super.refreshIncomplete() || this.getUpdateRequired();
	}

	@Override
	public void refreshDepend(boolean dirty, boolean load) {
		this.depend.refreshDepend(dirty, load);
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();

		str.append("ParamLinkedField\n");
		str.append("\n\tName        : ");
		str.append(this.getName());
		str.append("\n\tDescription : ");
		str.append(this.getDescription());
		str.append("\n\tData        : ");
		str.append(this.getData());
		str.append("\n\tUnit        : ");
		str.append(this.getUnit());
		str.append("\n\tTex String  : ");
		str.append(this.getTexString());
		str.append("\n\tType        : ");
		str.append(this.getType());
		str.append("\n\tLinked        : ");
		if (this.depend.getImports()[0].getHandle() != null) {
			str.append(this.depend.getImports()[0].getHandle().getIdentifier());
		}

		return str.toString();
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamField")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("Depend")) {
				this.depend.xml2obj(xmlReader);
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();
	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamLinkedField");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		// super class
		super.obj2xml(xmlWriter);

		// Depend
		this.depend.obj2xml(xmlWriter);

		xmlWriter.writeEndElement();
	}

	@Override
	public void reset() {
		this.setUpdateRequired(true);
	}

	@Override
	public void reportDepend(DocTexWriter content) {
		this.getDepend().reportImport(content);
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tLinked & ");
		if (this.depend.getImports()[0].getHandle() != null) {
			content.addString(this.depend.getImports()[0].getHandle().getIdentifier());
		}
		content.add(" \\\\\n");
	}

	@Override
	List<ParamSetKey> listParamSetKey() {
		List<ParamSetKey> list;

		list = super.listParamSetKey();
		list.add(ParamSetKey.UPDATEREQUIRED);

		return list;
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		int status = 0;
		Parameter root;
		DependItem[] depends;
		TData data = null;
		String unit = null;

		// get data from linked parameter
		depends = this.depend.getImports();
		for (DependItem depend : depends) {
			if ((depend.getVarName().compareTo("link") == 0) && (depend.getHandle() != null)) {
				if (depend.getHandle() == this) {
					this.warning("Linked-field references to itself! (" + this.getName() + ")");
				} else {
					data = depend.getHandle().getData();
					if (depend.getHandle() instanceof ParamField) {
						unit = ((ParamField) depend.getHandle()).getUnit();
					}
					this.setData(data, unit);
				}
			}
		}

		root = this.getRoot();
		if (root instanceof Project) {
			((Project) root).check();
		}
		if (status == 0) {
			this.reset();
			this.setUpdateRequired(false);
		}
		return status;
	}

	@Override
	public IParamLinkedField getRMIImpl() throws RemoteException {
		return new RParamLinkedField(this);
	}

	@Override
	public IParamLinkedField getClientImpl() {
		return new CParamLinkedField(this);
	}

}
