/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.BorderLayout;
import java.awt.Graphics;
import java.awt.Image;
import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Map;

import javax.imageio.ImageIO;
import javax.swing.JPanel;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamMatlabFunction;
import at.lcm.bu21.magopt.structure.rmi.RParamMatlabFunction;
import at.lcm.bu21.magopt.utils.MatlabControl;
import matlabcontrol.MatlabInvocationException;
import matlabcontrol.MatlabProxy;

/**
 * Parameter for user-defined matlab-function
 * 
 * @author Werner Koppelsttter
 */
public class ParamMatlabFunction extends ParamFunction implements IParamMatlabFunction {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Separate class-path and class-name
	 * 
	 * @param file Class-file
	 * @return Array with class-path and class-name
	 */
	public static String[] separatePathName(String file) {
		String classPath = "";
		String className = "";
		int index;

		if (file != null) {
			classPath = file;
			if (classPath.toLowerCase().endsWith(".m")) {
				classPath = file.substring(0, classPath.length() - 2);
			}

			// separating className
			index = classPath.lastIndexOf(File.separator);
			if (index == -1) {
				className = classPath;
				classPath = "";
			} else {
				className = classPath.substring(index + 1);
				classPath = classPath.substring(0, index);
			}

			// remove class-folder
			index = classPath.lastIndexOf(File.separator);
			if ((classPath.length() != 0) && (classPath.charAt(index + 1) == '@')) {
				if (index == -1) {
					classPath = "";
				} else {
					classPath = classPath.substring(0, index);
				}
			}

			// separating packages
			index = classPath.lastIndexOf(File.separator);
			while ((classPath.length() != 0) && (classPath.charAt(index + 1) == '+')) {
				if (index == -1) {
					className = classPath.substring(1) + "." + className;
					classPath = "";
				} else {
					className = classPath.substring(index + 2) + "." + className;
					classPath = classPath.substring(0, index);
				}
				index = classPath.lastIndexOf(File.separator);
			}
		}

		return new String[] { classPath, className };

	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamMatlabFunction() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamMatlabFunction(String name) throws Exception {
		this(name, name);

		// FIXME Default description???

		// FIXME refreshDependency()
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamMatlabFunction(String name, String description) throws Exception {
		this(name, description, null, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        File of function
	 * @throws Exception
	 */
	public ParamMatlabFunction(String name, String description, String file) throws Exception {
		this(name, description, file, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param classPath   Class-path of function
	 * @param className   Class-name of function
	 * @throws Exception
	 */
	public ParamMatlabFunction(String name, String description, String classPath, String className) throws Exception {
		super(name, description);
		if ((className == null) || className.isEmpty()) {
			String[] parts = ParamMatlabFunction.separatePathName(classPath);
			classPath = parts[0];
			className = parts[1];
		}
		this.setClassName(className);
		this.setClassPath(classPath);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamMatlabFunction other = (ParamMatlabFunction) obj;
		if (this.classLocation != other.classLocation) {
			return false;
		}
		if (this.classPath == null) {
			if (other.classPath != null) {
				return false;
			}
		} else if (!this.classPath.equals(other.classPath)) {
			return false;
		}
		if (this.className == null) {
			if (other.className != null) {
				return false;
			}
		} else if (!this.className.equals(other.className)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamMatlabFunction deepClone() throws CloneException {
		return (ParamMatlabFunction) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {

		super.cloneParam(clone);

		if (clone instanceof ParamMatlabFunction) {
			ParamMatlabFunction other = (ParamMatlabFunction) clone;

			other.className = this.className;
			other.classPath = this.classPath;
			other.classLocation = this.classLocation;
		}
	}

	@Override
	public String getFunctionID() {
		MatlabProxy proxy;
		Object[] ret = null;
		String id = "";

		proxy = MatlabControl.getProxy(this);
		try {
			ret = proxy.returningFeval("EvalFncMethode", 3,
					new Object[] { this.getWrapper(proxy), 1, "getFunctionID" });

			if (ret != null) {
				if (!(ret[0] instanceof double[])) {
					this.warningWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
				} else {
					id = (String) ret[2];
				}
			}
		} catch (MatlabInvocationException e) {
			this.warning("Error by read function-id", e.getStackTrace());
		} catch (RemoteException e) {
			this.warning("Exception by remote call", e.getStackTrace());
		}

		return id;
	}

	@Override
	public void refreshDependency() {
		MatlabProxy proxy;
		Object[] ret = null;

		proxy = MatlabControl.getProxy(this);
		try {
			ret = proxy.returningFeval("EvalFncMethode", 2,
					new Object[] { this.getWrapper(proxy), 0, "refreshDependency" });

			if ((ret != null) && !(ret[0] instanceof double[])) {
				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			}
		} catch (MatlabInvocationException e) {
			this.error("Error by refresh matlab-function", e.getStackTrace());
		} catch (RemoteException e) {
			this.error("Exception by remote call", e.getStackTrace());
		}
	}

	@Override
	public String getClassName() {
		return this.className;
	}

	@Override
	public void setClassName(String className) {
		if (className == null) {
			className = "";
		}
		this.className = className;
	}

	@Override
	public String getLocation() {
		return this.getLocation(this.classLocation);
	}

	/**
	 * Get location-string of class
	 * 
	 * @return Location
	 */
	public String getLocationString() {
		return this.getLocationString(this.classLocation);
	}

	@Override
	public String getLibraryPackage() {
		String path = null;
		if (this.classLocation == Location.LIBRARY) {
			path = new File(this.classPath).getParentFile().getParent();
		}
		return path;
	}

	@Override
	void refreshClass() {
//		MatlabProxy proxy;
//		Object[] ret = null;
//		String path = null;
//
//		proxy = MatlabControl.getProxy(this);
//		try {
//			ret = proxy.returningFeval("EvalMethode", 3,
//					new Object[] { this.getWrapper(proxy), 1, "which", this.getClassName() });
//
//			if ((ret != null) && !(ret[0] instanceof double[])) {
//				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
//			} else {
//				path = (String) ret[2];
//			}
//		} catch (MatlabInvocationException e) {
//			this.error("Error by show of matlab-function", e.getStackTrace());
//		} catch (RemoteException e) {
//			this.error("Exception by remote call", e.getStackTrace());
//		}
//
//		if ((path != null) && (path.length() != 0)) {
//			this.setClassPath(ParamMatlabFunction.separatePathName(path)[0]);
//		}
	}

	/**
	 * Refresh class-path
	 */
	void refreshPath() {
		String path = this.getClassPath();
		if (path.contains(":")) {
			this.setClassPath(this.getClassPath());
		}
	}

	@Override
	public void edit() {
		MatlabProxy proxy;
		Object[] ret = null;

		proxy = MatlabControl.getProxy(this);
		try {
			ret = proxy.returningFeval("EvalMethode", 2,
					new Object[] { this.getWrapper(proxy), 0, "edit", this.getClassName() });

			if ((ret != null) && !(ret[0] instanceof double[])) {
				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			} else {
			}
		} catch (MatlabInvocationException e) {
			this.error("Error by show of matlab-function", e.getStackTrace());
		} catch (RemoteException e) {
			this.error("Exception by remote call", e.getStackTrace());
		}
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamMatlabFunction\n");
		str.append("\n\tName          : ");
		str.append(this.getName());
		str.append("\n\tDescription   : ");
		str.append(this.getDescription());
		str.append("\n\tClassNamen    : ");
		str.append(this.getClassName());
		str.append("\n\tClassPath     : ");
		str.append(this.classPath);
		str.append("\n\tClassLocation : ");
		str.append(this.getLocationString());
		str.append("\n\tParams:       ");

		for (Parameter p : this.getParam()) {
			str.append("\n\t\tName: " + p.getName());
		}

		return str.toString();
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		int status = 0;
		MatlabProxy proxy;
		Object[] ret = null;
		LinkedList<Object> paramList;

		proxy = MatlabControl.getProxy(this);
		paramList = new LinkedList<Object>();
		paramList.add(this.getWrapper(proxy));
		paramList.add(1);
		paramList.add("run");
		for (String key : args.keySet()) {
			paramList.add(key);
			paramList.add(args.get(key));
		}

		ret = proxy.returningFeval("EvalFncMethode", 3, paramList.toArray(new Object[0]));

		if ((ret != null) && !(ret[0] instanceof double[])) {
			this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			status = 1;
		} else {
			status = (int) ((double[]) ret[2])[0];
		}

		this.update(status);
		return status;
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;
		// Integer length = 1;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamFunction")) {
				super.xml2obj(xmlReader);
				// length = this.getLength();
			} else if (element.equals("ClassName")) {
				xmlReader.next();
				this.className = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.className += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassPath")) {
				xmlReader.next();
				this.classPath = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.classPath += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassLocation")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.classLocation = Location.valueOf(tmp);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamMatlabFunction");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		this.refreshPath();

		// super class
		super.obj2xml(xmlWriter);

		// Class-name
		xmlWriter.writeStartElement("ClassName");
		if (this.className != null) {
			xmlWriter.writeCharacters(this.className);
		}
		xmlWriter.writeEndElement();

		// Class-path
		xmlWriter.writeStartElement("ClassPath");
		if (this.classPath != null) {
			xmlWriter.writeCharacters(this.classPath);
		}
		xmlWriter.writeEndElement();

		// ClassLocation
		xmlWriter.writeStartElement("ClassLocation");
		xmlWriter.writeCharacters(this.classLocation.name());
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public void reset() {
		MatlabProxy proxy;
		Object[] ret = null;

		if (!this.getUpdateRequired()) {
			this.setUpdateRequired(true);
		}

		proxy = MatlabControl.getProxy(this);
		try {
			ret = proxy.returningFeval("EvalFncMethode", 2, new Object[] { this.getWrapper(proxy), 0, "reset" });

			if ((ret != null) && !(ret[0] instanceof double[])) {
				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			} else {
				this.setResetFlag(false);
			}

		} catch (MatlabInvocationException e) {
			this.error("Error by reset matlab-function", e.getStackTrace());
			e.printStackTrace();
		} catch (RemoteException e) {
			this.error("Exception by remote call", e.getStackTrace());
		}

	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tMatlab function & ");
		content.addString((this.getLocationString() + "#" + this.getClassPath() + "->" + this.getClassName())
				.replace(File.separatorChar, '/'));
		content.add(" \\\\\n");
	}

	@Override
	protected void generateDocFiles(File directory) throws Exception {
		MatlabProxy proxy;
		Object[] ret = null;

		super.generateDocFiles(directory);
		proxy = MatlabControl.getProxy(this);
		try {
			ret = proxy.returningFeval("EvalFncMethode", 2,
					new Object[] { this.getWrapper(proxy), 0, "generateDocFilesJava", directory.getAbsolutePath() });

			if ((ret != null) && !(ret[0] instanceof double[])) {
				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			}
		} catch (MatlabInvocationException e) {
			this.error("Error by regenerate DocFiles", e.getStackTrace());
		} catch (RemoteException e) {
			this.error("Exception by remote call", e.getStackTrace());
		}

		// Generate Template
		try {
			ret = proxy.returningFeval("EvalFncMethode", 2,
					new Object[] { this.getWrapper(proxy), 0, "generateTemplate", directory.getAbsolutePath() });

			if ((ret != null) && !(ret[0] instanceof double[])) {
				this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
			}

		} catch (MatlabInvocationException e) {
			this.error("Error by regenerate Template", e.getStackTrace());
		} catch (RemoteException e) {
			this.error("Exception by remote call", e.getStackTrace());
		}
	}

	@Override
	public IParamMatlabFunction getRMIImpl() throws RemoteException {
		return new RParamMatlabFunction(this);
	}

	@Override
	public IParamMatlabFunction getClientImpl() {
		return new CParamMatlabFunction(this);
	}

	@Override
	public boolean hasPreview() {
		MatlabProxy proxy;
		Object[] ret = null;
		boolean exist = false;

		if (this.getClassName() != "") {
			proxy = MatlabControl.getProxy(this);
			try {
				ret = proxy.returningFeval("EvalMethode", 3,
						new Object[] { this.getWrapper(proxy), 1, "ismethod", this.getClassName(), "preview" });

				if ((ret != null) && !(ret[0] instanceof double[])) {
					this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
				} else {
					exist = ((boolean[]) ret[2])[0];
				}
			} catch (MatlabInvocationException e) {
				this.error("Error by show of matlab-function", e.getStackTrace());
			} catch (RemoteException e) {
				this.error("Exception by remote call", e.getStackTrace());
			}
		}

		return exist;
	}

	@Override
	public void preview(JPanel panel) {
		MatlabProxy proxy;
		Object[] ret = null;
		File tmpFile;
		String path;
		final Image image;
		final JPanel preview;
		int width, height;

		if (this.hasPreview()) {
			proxy = MatlabControl.getProxy(this);
			if (panel == null) {
				tmpFile = null;
				path = null;
				width = 0;
				height = 0;
			} else {
				tmpFile = new File(System.getProperty("java.io.tmpdir"), "__preview__.jpg");
				path = tmpFile.getPath();
				width = panel.getSize().width;
				height = panel.getSize().height;
			}
			try {
				ret = proxy.returningFeval("MatlabFunctionPreview", 2,
						new Object[] { this.getWrapper(proxy), width, height, path });

				if ((ret != null) && !(ret[0] instanceof double[])) {
					this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
				} else {
					if ((tmpFile != null) && tmpFile.exists()) {
						image = ImageIO.read(tmpFile);
						preview = new JPanel() {
							/** Serialization version number */
							private static final long serialVersionUID = 1L;

							@Override
							public void paint(Graphics g) {
								super.paint(g);
								g.drawImage(image, 0, 0, this);
							}
						};

						panel.setLayout(new BorderLayout());
						panel.add(preview);
						tmpFile.delete();
						preview.revalidate();
					}
				}
			} catch (MatlabInvocationException e) {
				this.error("Error by preview of matlab-function", e.getStackTrace());
			} catch (RemoteException e) {
				this.error("Exception by remote call", e.getStackTrace());
			} catch (IOException e) {
				this.error("Error by load preview of matlab-function", e.getStackTrace());
			}
		}
	}

	@Override
	public boolean hasShow() {
		MatlabProxy proxy;
		Object[] ret = null;
		boolean exist = false;

		if (this.getClassName() != "") {
			proxy = MatlabControl.getProxy(this);
			try {
				ret = proxy.returningFeval("EvalMethode", 3,
						new Object[] { this.getWrapper(proxy), 1, "ismethod", this.getClassName(), "show" });

				if ((ret != null) && !(ret[0] instanceof double[])) {
					this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
				} else {
					exist = ((boolean[]) ret[2])[0];
				}
			} catch (MatlabInvocationException e) {
				this.error("Error by show of matlab-function", e.getStackTrace());
			} catch (RemoteException e) {
				this.error("Exception by remote call", e.getStackTrace());
			}
		}

		return exist;
	}

	@Override
	public void show() {
		MatlabProxy proxy;
		Object[] ret = null;

		if (this.hasShow()) {
			proxy = MatlabControl.getProxy(this);
			try {
				ret = proxy.returningFeval("EvalFncMethode", 2, new Object[] { this.getWrapper(proxy), 0, "show" });

				if ((ret != null) && !(ret[0] instanceof double[])) {
					this.errorWithStack((String) ret[0], (StackTraceElement[]) ret[1]);
				}
			} catch (MatlabInvocationException e) {
				this.error("Error by show of matlab-function", e.getStackTrace());
			} catch (RemoteException e) {
				this.error("Exception by remote call", e.getStackTrace());
			}
		}
	}

	@Override
	public void infoWithStack(String message, Object stack) {
		this.info(message, this.getStackTrace(stack));
	}

	@Override
	public void warningWithStack(String message, Object stack) {
		this.warning(message, this.getStackTrace(stack));
	}

	@Override
	public void errorWithStack(String message, Object stack) {
		this.error(message, this.getStackTrace(stack));
	}

	/**
	 * Get a stack-trace of caller
	 *
	 * @param stack List of StackTraceElements from caller
	 *
	 * @return stack-trace
	 */
	private StackTraceElement[] getStackTrace(Object stack) {
		LinkedList<StackTraceElement> list;
		int len;

		list = new LinkedList<StackTraceElement>();
		if (!(stack instanceof double[])) {
			list.addAll(Arrays.asList((StackTraceElement[]) stack));
		}
		len = list.size();
		list.addAll(Arrays.asList(new Exception().getStackTrace()));

		// remove this method from stack
		list.remove(len);
		list.remove(len);

		return list.toArray(new StackTraceElement[0]);
	}

	/**
	 * Get wrapper of this object
	 * 
	 * @param proxy Matlab proxy
	 * @return Wrapper
	 * @throws RemoteException
	 */
	private IParamMatlabFunction getWrapper(MatlabProxy proxy) throws RemoteException {
		if (proxy.isRunningInsideMatlab()) {
			return this;
		} else {
			return this.getRMIImpl().getClientImpl();
		}
	}

}
