/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Desktop;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;

import javax.imageio.ImageIO;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTabbedPane;
import javax.swing.JTextPane;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamPythonFunction;
import at.lcm.bu21.magopt.structure.jep.JepProxy;
import magopt.utils.Base64;

/**
 * Parameter for user-defined python-function
 * 
 * @author Werner Koppelsttter
 */
public class ParamPythonFunction extends ParamFunction implements IParamPythonFunction {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Name of class */
	protected String moduleName;

	/**
	 * Panel to display images
	 *
	 * @author Werner Koppelsttter
	 */
	private class ImagePanel extends JPanel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		/** Image to display */
		BufferedImage image = null;

		/** List of callbacks */
		List<Object> callback = null;

		/** Value of zoom */
		private double zoom = 1.0;

		/** Value to pan in x direction */
		private double transX = 0;

		/** Value to pan in y direction */
		private double transY = 0;

		/** Width for plot area */
		private int plotWidth;

		/** Height for plot area */
		private int plotHeight;

		/** Last point of panning */
		private java.awt.Point lastPoint = null;

		/** Popup-menu of plot */
		private JPopupMenu popup;

		/**
		 * Create image-panel
		 */
		public ImagePanel() {
			JMenuItem zoomReset;

			// zoom plot
			this.addMouseWheelListener(new MouseWheelListener() {
				@Override
				public void mouseWheelMoved(MouseWheelEvent event) {
					double s, dist;

					if (event.getScrollType() == MouseWheelEvent.WHEEL_UNIT_SCROLL) {
						s = ImagePanel.this.zoom;
						ImagePanel.this.zoom *= (1 + (0.1 * event.getWheelRotation()));

						// limit scale
						if (ImagePanel.this.zoom < 0.5) {
							ImagePanel.this.zoom = 0.5;
						}
						if (ImagePanel.this.zoom > 10.0) {
							ImagePanel.this.zoom = 10.0;
						}

						// scale trans to center
						ImagePanel.this.transX *= ImagePanel.this.zoom / s;
						ImagePanel.this.transY *= ImagePanel.this.zoom / s;

						// trans to mousepointer
						dist = event.getPoint().getX() - (ImagePanel.this.plotWidth / 2.0);
						ImagePanel.this.transX += dist * (1.0 - (ImagePanel.this.zoom / s));

						dist = event.getPoint().getY() - (ImagePanel.this.plotHeight / 2.0);
						ImagePanel.this.transY += dist * (1.0 - (ImagePanel.this.zoom / s));

						ImagePanel.this.repaint();
					}
				}
			});

			// pan plot
			this.addMouseMotionListener(new MouseMotionAdapter() {
				@Override
				public void mouseDragged(MouseEvent me) {
					ImagePanel.this.transX += me.getPoint().x - ImagePanel.this.lastPoint.x;
					ImagePanel.this.transY += me.getPoint().y - ImagePanel.this.lastPoint.y;
					ImagePanel.this.lastPoint = me.getPoint();
					ImagePanel.this.repaint();
				}
			});

			// start pan
			this.addMouseListener(new MouseAdapter() {
				@Override
				public void mousePressed(MouseEvent event) {
					ImagePanel.this.lastPoint = event.getPoint();
				}

				@Override
				public void mouseReleased(MouseEvent me) {
					if (me.isPopupTrigger()) {
						ImagePanel.this.popup.show(me.getComponent(), me.getX(), me.getY());
					}
				}

				@SuppressWarnings("unchecked")
				@Override
				public void mouseClicked(MouseEvent me) {
					if (me.getClickCount() > 1) {
						if ((ImagePanel.this.callback != null) && !ImagePanel.this.callback.isEmpty()
								&& (ImagePanel.this.callback.get(0) != null)) {
							List<Object> ret;
							try {
								LinkedList<Object> call = new LinkedList<Object>();
								call.add(ParamPythonFunction.this);
								call.addAll(ImagePanel.this.callback);

								ret = (List<Object>) ParamPythonFunction.this.getJep().invoke("eval_fnc_method",
										call.toArray(new Object[0]));
								if (ret.get(0) != null) {
									ParamPythonFunction.this.errorWithStack((String) ret.get(0),
											(List<StackTraceElement>) ret.get(1));
								}
							} catch (Exception e) {
								ParamPythonFunction.this.error(e.getMessage(), e.getStackTrace());
							}
						}
					}
				}

			});

			this.setOpaque(true);
			this.setBackground(Color.WHITE);

			this.popup = new JPopupMenu();

			// Reset zoom to inital zoom
			zoomReset = new JMenuItem("Reset Zoom");
			this.popup.add(zoomReset);
			zoomReset.addActionListener(new ActionListener() {
				@Override
				public void actionPerformed(ActionEvent e) {
					ImagePanel.this.resetZoom();
				}

			});
		}

		/**
		 * Set panel to display
		 * 
		 * @param image    Image to display
		 * @param callback List of callbacks
		 */
		public void setImage(BufferedImage image, List<Object> callback) {
			this.image = image;
			this.callback = callback;
		}

		/**
		 * Reset zoom to original
		 */
		private void resetZoom() {
			this.transX = 0.0;
			this.transY = 0.0;
			this.zoom = 1.0;
			this.repaint();
		}

		@Override
		protected void paintComponent(Graphics g) {
			Graphics2D g2 = (Graphics2D) g;
			double scale, scaleX, scaleY, max;
			int width, height;

			super.paintComponent(g);
			if (this.image != null) {

				this.plotWidth = this.getWidth();
				this.plotHeight = this.getHeight();

				scaleX = ((double) this.getWidth()) / this.image.getWidth();
				scaleY = ((double) this.getHeight()) / this.image.getHeight();
				scale = scaleX < scaleY ? scaleX : scaleY;
				width = (int) (this.zoom * scale * this.image.getWidth());
				height = (int) (this.zoom * scale * this.image.getHeight());

				// limit pan
				max = Math.max(this.plotWidth, width);
				this.transX = this.transX < max ? this.transX : max;
				this.transX = this.transX > -max ? this.transX : -max;
				max = Math.max(this.plotHeight, height);
				this.transY = this.transY < max ? this.transY : max;
				this.transY = this.transY > -max ? this.transY : -max;

				g2.drawImage(this.image.getScaledInstance(width, height, Image.SCALE_DEFAULT),
						((this.getWidth() - width) / 2) + (int) this.transX,
						((this.getHeight() - height) / 2) + (int) this.transY, null);
			}
		}
	}

	/**
	 * Get jep-connection
	 * 
	 * @return Jep-proxy
	 * 
	 * @throws InterruptedException
	 * @throws ExecutionException
	 */
	private JepProxy getJep() throws InterruptedException, ExecutionException {
		Parameter param;
		JepProxy jep;

		param = this.getRoot();
		if (param instanceof Project) {
			jep = ((Project) param).getJep();
		} else {
			jep = new JepProxy();
		}

		return jep;
	}

	/**
	 * Separate path and module
	 * 
	 * @param filePath Module-file
	 * @return Array with class-path and class-name
	 */
	public static String[] separatePathName(String filePath) {
		StringBuffer module = new StringBuffer();
		String classPath = "", moduleName;
		File file;

		if (filePath != null) {
			file = new File(filePath);

			// get class
			moduleName = file.getName();
			file = file.getParentFile();
			if (moduleName.toLowerCase().endsWith(".py")) {
				moduleName = moduleName.substring(0, moduleName.length() - 3);
			}
			module.append(moduleName);

			// separating package and path
			while (new File(file, "__init__.py").exists()) {
				module.insert(0, ".");
				module.insert(0, file.getName());
				file = file.getParentFile();
			}

			if (file != null) {
				classPath = file.getPath();
			} else {
				classPath = "";
			}
		}

		return new String[] { classPath, module.toString() };
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamPythonFunction() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamPythonFunction(String name) throws Exception {
		this(name, name);

		// FIXME Default description???

		// FIXME refreshDependency()
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamPythonFunction(String name, String description) throws Exception {
		this(name, description, null, null, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        File of function
	 * @param className   Class-name of function
	 * @throws Exception
	 */
	public ParamPythonFunction(String name, String description, String file, String className) throws Exception {
		this(name, description, file, null, className);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param classPath   Class-path of function
	 * @param moduleName  Name of module of function
	 * @param className   Class-name of function
	 * @throws Exception
	 */
	public ParamPythonFunction(String name, String description, String classPath, String moduleName, String className)
			throws Exception {
		super(name, description);
		if ((moduleName == null) || moduleName.isEmpty()) {
			String[] parts = ParamPythonFunction.separatePathName(classPath);
			this.setModuleName(parts[1]);
			this.setClassPath(parts[0]);
		} else {
			this.setModuleName(moduleName);
			this.setClassPath(classPath);
		}
		this.setClassName(className);
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamPythonFunction other = (ParamPythonFunction) obj;
		if (this.classLocation != other.classLocation) {
			return false;
		}
		if (this.classPath == null) {
			if (other.classPath != null) {
				return false;
			}
		} else if (!this.classPath.equals(other.classPath)) {
			return false;
		}
		if (this.moduleName == null) {
			if (other.moduleName != null) {
				return false;
			}
		} else if (!this.moduleName.equals(other.moduleName)) {
			return false;
		}
		if (this.className == null) {
			if (other.className != null) {
				return false;
			}
		} else if (!this.className.equals(other.className)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamPythonFunction deepClone() throws CloneException {
		return (ParamPythonFunction) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {

		super.cloneParam(clone);

		if (clone instanceof ParamPythonFunction) {
			ParamPythonFunction other = (ParamPythonFunction) clone;

			other.className = this.className;
			other.moduleName = this.moduleName;
			other.classPath = this.classPath;
			other.classLocation = this.classLocation;
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public String getFunctionID() {
		String id = null;

		try {
			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "getFunctionID");
			if (ret.get(0) != null) {
				this.warningWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
			} else {
				id = (String) ret.get(2);
			}
		} catch (Exception e) {
			this.warning(e.getMessage(), e.getStackTrace());
		}

		return id;
	}

	@SuppressWarnings("unchecked")
	@Override
	public void reset() {

		if (!this.getUpdateRequired()) {
			this.setUpdateRequired(true);
		}

		try {
			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "reset");

			if (ret.get(0) != null) {
				this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
			} else {
				this.setResetFlag(false);
			}
		} catch (Exception e) {
			this.error(e.getMessage(), e.getStackTrace());
		}

	}

	@SuppressWarnings("unchecked")
	@Override
	public int update(Map<String, Object> args) throws Exception, RemoteException {
		int status = 1;

		List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "run", args);
		if (ret.get(0) != null) {
			this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
		} else {
			status = ((Long) ret.get(2)).intValue();
		}

		this.update(status);

		return status;

	}

	@SuppressWarnings("unchecked")
	@Override
	public void refreshDependency() {
		try {
			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "refreshDependency");
			if (ret.get(0) != null) {
				this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
			}
		} catch (Exception e) {
			this.error(e.getMessage(), e.getStackTrace());
		}
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamFunction")) {
				super.xml2obj(xmlReader);
			} else if (element.equals("ClassName")) {
				xmlReader.next();
				this.className = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.className += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ModuleName")) {
				xmlReader.next();
				this.moduleName = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.moduleName += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassPath")) {
				xmlReader.next();
				this.classPath = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.classPath += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassLocation")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.classLocation = Location.valueOf(tmp);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamPythonFunction");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		this.refreshPath();

		// super class
		super.obj2xml(xmlWriter);

		// Class-name
		xmlWriter.writeStartElement("ClassName");
		if (this.className != null) {
			xmlWriter.writeCharacters(this.className);
		}
		xmlWriter.writeEndElement();

		// Module-name
		xmlWriter.writeStartElement("ModuleName");
		if (this.moduleName != null) {
			xmlWriter.writeCharacters(this.moduleName);
		}
		xmlWriter.writeEndElement();

		// Class-path
		xmlWriter.writeStartElement("ClassPath");
		if (this.classPath != null) {
			xmlWriter.writeCharacters(this.classPath);
		}
		xmlWriter.writeEndElement();

		// ClassLocation
		xmlWriter.writeStartElement("ClassLocation");
		xmlWriter.writeCharacters(this.classLocation.name());
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public void edit() {
		File file;

		file = new File(this.getClassPath(), this.getModuleName().replace('.', File.separatorChar) + ".py");
		if (file.exists()) {
			if (Desktop.isDesktopSupported()) {
				try {
					Desktop.getDesktop().open(file);
				} catch (IOException e) {
					// XXX winopen ?
					this.warning("Unable to open (Exception while open)");
				}
			} else {
				this.warning("Unable to open (Unsupported)");
			}
		} else {
			this.warning("Unable to open (file doesn't exist)");
		}
	}

	@Override
	void reportInfo(DocTexWriter content) {

		super.reportInfo(content);

		content.add("\tPython function & ");
		content.addString((this.getLocationString() + "#" + this.getClassPath() + "#" + this.getModuleName() + "->"
				+ this.getClassName()).replace(File.separatorChar, '/'));
		content.add(" \\\\\n");
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void generateDocFiles(File directory) throws Exception {
		super.generateDocFiles(directory);
		List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "generateDocFiles",
				directory.getAbsolutePath());
		if (ret.get(0) != null) {
			this.warningWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
		}
	}

	@Override
	public IParamPythonFunction getClientImpl() {
		return new CParamPythonFunction(this);
	}

	@Override
	public String getClassName() {
		return this.className;
	}

	@Override
	public void setClassName(String className) {
		if (className == null) {
			className = "";
		}
		this.className = className;
	}

	/**
	 * Get name of module
	 * 
	 * @return Module name
	 */
	public String getModuleName() {
		return this.moduleName;
	}

	/**
	 * Set name of module
	 * 
	 * @param moduleName Module name
	 */
	public void setModuleName(String moduleName) {
		if (moduleName == null) {
			moduleName = "";
		}
		this.moduleName = moduleName;
	}

	@Override
	public String getLibraryPackage() {
		String path = null;

		if (this.classLocation == Location.LIBRARY) {
			path = new File(this.classPath).getParentFile().getParentFile().getParent();
		}
		return path;
	}

	@Override
	void refreshClass() {
//		String path;
//
//		try {
//			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "getModule");
//			if (ret.get(0) != null) {
//				this.warningWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
//			} else {
//				path = (String) ret.get(2);
//				if ((path != null) && (path.length() != 0)) {
//					this.setClassPath(ParamPythonFunction.separatePathName(path)[0]);
//				}
//			}
//		} catch (Exception e) {
//			this.error(e.getMessage(), e.getStackTrace());
//		}
	}

	/**
	 * Refresh class-path
	 */
	void refreshPath() {
		String path = this.getClassPath();
		if (path.contains(":")) {
			this.setClassPath(this.getClassPath());
		}
	}

	@Override
	public String getLocation() {
		return this.getLocation(this.classLocation);
	}

	/**
	 * Get location-string of class
	 * 
	 * @return Location
	 */
	public String getLocationString() {
		return this.getLocationString(this.classLocation);
	}

	@SuppressWarnings("unchecked")
	@Override
	public void show() {
		if (this.hasShow()) {
			try {
				List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "show");
				if (ret.get(0) != null) {
					this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
				}
			} catch (Exception e) {
				this.error(e.getMessage(), e.getStackTrace());
			}
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public void preview(JPanel panel) {
		JTabbedPane tabs;
		String name, mime, content;

		if (this.hasPreview()) {
			try {
				List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "preview", 700, 600);
				// FIXME panel size
				// List<Object> ret = (List<Object>) proxy.invoke(
				// "eval_fnc_method", this, "preview", panel.getWidth(),
				// panel.getHeight());
				if (ret != null) {
					if (ret.get(0) != null) {
						this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
					} else {
						List<List<Object>> views = (List<List<Object>>) ret.get(2);

						if (views != null) {
							// if (views.size() > 1) {
							tabs = new JTabbedPane();
							panel.setLayout(new BorderLayout());
							panel.add(tabs);

							for (List<Object> view : views) {
								name = (String) view.get(0);
								mime = (String) view.get(1);

								if ("text/plain".equals(mime)) {
									content = (String) view.get(2);
									JLabel label = new JLabel(content);
									tabs.add(name, label);
								} else if ("text/html".equals(mime)) {
									content = (String) view.get(2);
									double ver = Double.parseDouble(System.getProperty("java.specification.version"));
									// check url exist
									URL url = null;
									File file = new File(content);
									if (file.exists()) {
										url = file.toURI().toURL();
									} else {
										try {
											url = new URL(content);
											url.toURI();
										} catch (Exception e) {
											url = null;
										}
									}

									if (ver >= 1.7) {
										Class<?> cl = Class.forName("magopt.gui.web.WebPanel");
										JComponent browser = (JComponent) cl.getConstructor().newInstance();
										if (url != null) {
											Method m = cl.getMethod("load", String.class);
											m.invoke(browser, url.toString());
										} else {
											Method m = cl.getMethod("loadContent", String.class);
											m.invoke(browser, content);
										}
										Thread.sleep(100);
										tabs.add(name, browser);
									} else {
										JTextPane browser = new JTextPane();
										browser.setEditable(false);
										if (url != null) {
											browser.setPage(url);
										} else {
											browser.setText(content);
										}
										tabs.add(name, browser);
									}
								} else if ("image/png".equals(mime)) {
									content = (String) view.get(2);
									ImagePanel pan = new ImagePanel();
									tabs.add(name, pan);
									try {
										byte[] imageByte = Base64.decode(content);
										ByteArrayInputStream bis = new ByteArrayInputStream(imageByte);
										BufferedImage image = ImageIO.read(bis);
										bis.close();
										pan.setImage(image, view.subList(3, view.size()));
									} catch (Exception e) {
										this.warning(e.getMessage(), e.getStackTrace());
									}
									// FIXME HTML, Text, SVG, ...
								} else if ("application/java".equals(mime)) {
									JPanel pan = (JPanel) view.get(2);
									tabs.add(name, pan);
								}
							}
							//
							// } else {
							// panel.add(comp, constraints);
							// }
						}
					}
				}

			} catch (Exception e) {
				this.warning(e.getMessage(), e.getStackTrace());
			}
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean hasPreview() {
		boolean exist = false;

		try {
			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "hasPreview");
			if (ret.get(0) != null) {
				this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
			} else {
				exist = (Boolean) ret.get(2);
			}
		} catch (Exception e) {
			this.error(e.getMessage(), e.getStackTrace());
		}

		return exist;
	}

	@SuppressWarnings("unchecked")
	@Override
	public boolean hasShow() {
		boolean exist = false;

		try {
			List<Object> ret = (List<Object>) this.getJep().invoke("eval_fnc_method", this, "hasShow");
			if (ret.get(0) != null) {
				this.errorWithStack((String) ret.get(0), (List<StackTraceElement>) ret.get(1));
			} else {
				exist = (Boolean) ret.get(2);
			}
		} catch (Exception e) {
			this.error(e.getMessage(), e.getStackTrace());
		}

		return exist;
	}

	@Override
	public void infoWithStack(String message, List<StackTraceElement> stack) {
		this.info(message, this.getStackTrace(stack));

	}

	@Override
	public void warningWithStack(String message, List<StackTraceElement> stack) {
		this.warning(message, this.getStackTrace(stack));
	}

	@Override
	public void errorWithStack(String message, List<StackTraceElement> stack) {
		this.error(message, this.getStackTrace(stack));
	}

	/**
	 * Get a stack-trace of caller
	 * 
	 * @param stack List of StackTraceElements from caller
	 * 
	 * @return stack-trace
	 */
	private StackTraceElement[] getStackTrace(List<StackTraceElement> stack) {
		LinkedList<StackTraceElement> list;
		int len;

		list = new LinkedList<StackTraceElement>();
		list.addAll(stack);
		len = list.size();
		list.addAll(Arrays.asList(new Exception().getStackTrace()));

		// remove this method from stack
		list.remove(len);
		list.remove(len);

		return list.toArray(new StackTraceElement[0]);
	}

}
