/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9928 $
 * $Date: 2020-08-03 12:06:12 +0200 (Mon, 03 Aug 2020) $
 */
package at.lcm.bu21.magopt.structure;

import java.awt.Desktop;
import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Map;

import javax.swing.JEditorPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.utils.Clone;
import at.lcm.bu21.general.utils.CloneException;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.client.CParamPythonFunction_old;
import at.lcm.bu21.magopt.structure.python.PythonConnection;
import at.lcm.bu21.magopt.structure.rmi.RParamPythonFunction_old;

/**
 * Parameter for user-defined python-function
 * 
 * @author Werner Koppelsttter
 */
public class ParamPythonFunction_old extends ParamFunction implements IParamPythonFunction_old {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/**
	 * Separate class-path and class-name
	 * 
	 * @param file Class-file
	 * @return Array with class-path and class-name
	 */
	public static String[] separatePathName(String file) {
		String classPath = "";
		String className = "";
		int index;

		if (file != null) {
			classPath = file;
			if (classPath.toLowerCase().endsWith(".py")) {
				classPath = file.substring(0, classPath.length() - 3);
			}

			// separating className
			index = classPath.lastIndexOf(File.separator);
			if (index == -1) {
				className = classPath;
				classPath = "";
			} else {
				className = classPath.substring(index + 1);
				classPath = classPath.substring(0, index);
			}
		}
		return new String[] { classPath, className };
	}

	/**
	 * Constructor for loading a xml-file
	 * 
	 * @throws Exception
	 */
	ParamPythonFunction_old() throws Exception {
		this("XML_Load", "XML load constructor");
	}

	/**
	 * Constructor with name
	 * 
	 * @param name Name of Parameter
	 * @throws Exception
	 */
	public ParamPythonFunction_old(String name) throws Exception {
		this(name, name);

		// FIXME Default description???

		// FIXME refreshDependency()
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @throws Exception
	 */
	public ParamPythonFunction_old(String name, String description) throws Exception {
		this(name, description, null, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param file        File of python function
	 * @throws Exception
	 */
	public ParamPythonFunction_old(String name, String description, String file) throws Exception {
		this(name, description, file, null);
	}

	/**
	 * Constructor with name and description
	 * 
	 * @param name        Name of Parameter
	 * @param description Description of Parameter
	 * @param classPath   Class-path of function
	 * @param className   Class-name of function
	 * @throws Exception
	 */
	public ParamPythonFunction_old(String name, String description, String classPath, String className)
			throws Exception {
		super(name, description);
		if (className == null) {
			String[] parts = ParamPythonFunction_old.separatePathName(classPath);
			this.setClassName(parts[1]);
			this.setClassPath(parts[0]);
		} else {
			this.setClassName(className);
			this.setClassPath(classPath);
		}
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (!super.equals(obj)) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		ParamPythonFunction_old other = (ParamPythonFunction_old) obj;
		if (this.classLocation != other.classLocation) {
			return false;
		}
		if (this.classPath == null) {
			if (other.classPath != null) {
				return false;
			}
		} else if (!this.classPath.equals(other.classPath)) {
			return false;
		}
		if (this.className == null) {
			if (other.className != null) {
				return false;
			}
		} else if (!this.className.equals(other.className)) {
			return false;
		}
		return true;
	}

	@Override
	public ParamPythonFunction_old deepClone() throws CloneException {
		return (ParamPythonFunction_old) Clone.deepClone(this);
	}

	@Override
	void cloneParam(Parameter clone) throws CloneException {

		super.cloneParam(clone);

		if (clone instanceof ParamPythonFunction_old) {
			ParamPythonFunction_old other = (ParamPythonFunction_old) clone;

			other.className = this.className;
			other.classPath = this.classPath;
			other.classLocation = this.classLocation;
		}
	}

	@Override
	public String getFunctionID() {
		String id = "";
		File pyFile;

		pyFile = this.getPythonFile();
		try {
			id = PythonConnection.getFunctionID(pyFile);
		} catch (Exception ex) {
			this.warning(ex.getMessage(), ex.getStackTrace());
		}

		return id;
	}

	@Override
	public void refreshDependency() {
		LinkedList<PortDesc> impVars, expVars;
		File pyFile;

		impVars = new LinkedList<PortDesc>();
		expVars = new LinkedList<PortDesc>();

		pyFile = this.getPythonFile();
		try {
			for (String var : PythonConnection.imports(pyFile)) {
				impVars.add(new PortDesc(var, var, "", false));
			}

			for (String var : PythonConnection.exports(pyFile)) {
				expVars.add(new PortDesc(var, var, "", false));
			}
		} catch (Exception ex) {
			this.warning(ex.getMessage(), ex.getStackTrace());
		}

		this.getDepend().setImportVars(impVars.toArray(new PortDesc[0]));
		this.getDepend().setExportVars(expVars.toArray(new PortDesc[0]));
	}

	@Override
	public String getClassName() {
		return this.className;
	}

	@Override
	public void setClassName(String className) {
		if (className == null) {
			className = "";
		}
		this.className = className;
	}

	@Override
	public String getClassPath() {
		return this.getLocation() + File.separator + this.classPath;
	}

	@Override
	public void setClassPath(String classPath) {
		if (classPath == null) {
			classPath = "";
		}
		this.classPath = this.setLocation(classPath);
	}

	@Override
	public String getLocation() {
		return this.getLocation(this.classLocation);
	}

	/**
	 * Get location-string of class
	 * 
	 * @return Location
	 */
	public String getLocationString() {
		return this.getLocationString(this.classLocation);
	}

	@Override
	public String getLibraryPackage() {
		String path = null;
		if (this.classLocation == Location.LIBRARY) {
			path = new File(this.classPath).getParentFile().getParent();
		}
		return path;
	}

	@Override
	void refreshClass() {
	}

	/**
	 * Refresh class-path
	 */
	void refreshPath() {
		String path = this.getClassPath();
		if (path.contains(":")) {
			this.setClassPath(this.getClassPath());
		}
	}

	/**
	 * Get python module
	 * 
	 * @return Py-file
	 */
	private File getPythonFile() {
		return new File(this.getClassPath(), this.getClassName() + ".py");
	}

	@Override
	public void edit() {
		File file;

		file = this.getPythonFile();
		if (file.exists()) {
			if (Desktop.isDesktopSupported()) {
				try {
					Desktop.getDesktop().open(file);
				} catch (IOException e) {
					// XXX winopen ?
					this.warning("Unable to open (Exception while open)");
				}
			} else {
				this.warning("Unable to open (Unsupported)");
			}
		} else {
			this.warning("Unable to open (file doesn't exist)");
		}
	}

	@Override
	public String toString() {
		StringBuffer str = new StringBuffer();
		str.append("ParamPythonFunction\n");
		str.append("\n\tName          : ");
		str.append(this.getName());
		str.append("\n\tDescription   : ");
		str.append(this.getDescription());
		str.append("\n\tClassNamen    : ");
		str.append(this.getClassName());
		str.append("\n\tClassPath     : ");
		str.append(this.classPath);
		str.append("\n\tClassLocation : ");
		str.append(this.getLocationString());
		str.append("\n\tParams:       ");

		for (Parameter p : this.getParam()) {
			str.append("\n\t\tName: " + p.getName());
		}

		return str.toString();
	}

	/**
	 * Get imports-values of function
	 * 
	 * @return Map with import-values
	 */
	protected Map<String, TData> importValues() {
		Map<String, TData> map;

		map = new Hashtable<String, TData>();
		for (DependItem depend : this.getDepend().getImports()) {
			if (depend.getHandle() != null) {
				map.put(depend.getVarName(), depend.getHandle().getData());
			}
		}

		return map;
	}

	/**
	 * Set export-values of function
	 * 
	 * @param map Map with export-values
	 * @throws Exception
	 */
	protected void exportValues(Map<String, TData> map) throws Exception {
		for (DependItem depend : this.getDepend().getExports()) {
			if (depend.getHandle() != null) {
				depend.getHandle().setData(map.get(depend.getVarName()));
			}
		}

	}

	/**
	 * Get export of exported data
	 * 
	 * @param map Map for export-values
	 */
	protected void revertExportValues(Map<String, TData> map) {
		for (DependItem depend : this.getDepend().getExports()) {
			if (depend.getHandle() != null) {
				map.put(depend.getVarName(), depend.getHandle().getData());
			}
		}
	}

	@Override
	public int update(Map<String, Object> args) throws Exception {
		// FIXME update
		int status = 0;
		Map<String, TData> map, res;
		File pyFile, helpFolder;

		pyFile = this.getPythonFile();

		map = this.importValues();

		// update function
		res = PythonConnection.update(pyFile, map);

		// prepare folder for preview html
		helpFolder = new File(this.getParamDirectory(), "html");
		if (helpFolder.exists()) {
			FileUtils.delete(helpFolder);
		}
		if (PythonConnection.hasHtml(pyFile)) {
			helpFolder.mkdirs();
			map.putAll(res);
			PythonConnection.html(pyFile, helpFolder, map);
		}

		this.exportValues(res);
		this.update(status);
		return status;
	}

	@Override
	void xml2obj(XMLStreamReader xmlReader) throws Exception {
		String element, tmp;
		// Integer length = 1;

		xmlReader.next();
		while (xmlReader.getEventType() == XMLStreamConstants.START_ELEMENT) {
			element = xmlReader.getLocalName();
			if (element.equals("ParamFunction")) {
				super.xml2obj(xmlReader);
				// length = this.getLength();
			} else if (element.equals("ClassName")) {
				xmlReader.next();
				this.className = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.className += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassPath")) {
				xmlReader.next();
				this.classPath = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					this.classPath += xmlReader.getText();
					xmlReader.next();
				}
				xmlReader.next();
			} else if (element.equals("ClassLocation")) {
				xmlReader.next();
				tmp = new String();
				while (xmlReader.getEventType() != XMLStreamConstants.END_ELEMENT) {
					tmp += xmlReader.getText();
					xmlReader.next();
				}
				this.classLocation = Location.valueOf(tmp);
				xmlReader.next();
			} else {
				throw new Exception("Unrecognized property: " + xmlReader.getLocalName());
			}
		}
		xmlReader.next();

	}

	@Override
	void obj2xml(XMLStreamWriter xmlWriter) throws Exception {

		xmlWriter.writeStartElement("ParamPythonFunction_old");
		xmlWriter.writeAttribute("object", this.getClass().getSimpleName());
		xmlWriter.writeAttribute("name", this.getName());

		this.refreshPath();

		// super class
		super.obj2xml(xmlWriter);

		// Class-name
		xmlWriter.writeStartElement("ClassName");
		if (this.className != null) {
			xmlWriter.writeCharacters(this.className);
		}
		xmlWriter.writeEndElement();

		// Class-path
		xmlWriter.writeStartElement("ClassPath");
		if (this.classPath != null) {
			xmlWriter.writeCharacters(this.classPath);
		}
		xmlWriter.writeEndElement();

		// ClassLocation
		xmlWriter.writeStartElement("ClassLocation");
		xmlWriter.writeCharacters(this.classLocation.name());
		xmlWriter.writeEndElement();

		xmlWriter.writeEndElement();
	}

	@Override
	public void reset() {
		if (!this.getUpdateRequired()) {
			this.setUpdateRequired(true);
		}
	}

	@Override
	void reportInfo(DocTexWriter content) {
		super.reportInfo(content);

		content.add("\tPython function & ");
		content.addString((this.getLocationString() + "#" + this.getClassPath() + "->" + this.getClassName())
				.replace(File.separatorChar, '/'));
		content.add(" \\\\\n");
	}

	// XXX python html?
	// @Override
	// public void generateDocFiles(File directory) throws Exception {
	// MatlabProxy proxy;
	//
	// proxy = MatlabControl.getProxy(this);
	// if (proxy.isRunningInsideMatlab()) {
	// proxy.returningFeval("GenerateDocFiles", 1, new Object[] { this,
	// directory.getAbsolutePath() });
	// } else {
	// try {
	// proxy.returningFeval("GenerateDocFiles", 1,
	// new Object[] { this.getRMIImpl().getClientImpl(),
	// directory.getAbsolutePath() });
	// } catch (RemoteException ex) {
	// ex.printStackTrace();
	// }
	// }
	// }

	@Override
	public IParamPythonFunction_old getRMIImpl() throws RemoteException {
		return new RParamPythonFunction_old(this);
	}

	@Override
	public IParamPythonFunction_old getClientImpl() {
		return new CParamPythonFunction_old(this);
	}

	@Override
	public boolean hasPreview() {
		boolean preview = false;

		try {
			preview = PythonConnection.hasHtml(this.getPythonFile());
		} catch (Exception e) {
			e.printStackTrace();
		}

		return preview;
	}

	@Override
	public void preview(JPanel panel) {

		if (this.hasPreview()) {
			// write index.html for preview
			File htmlFile = FileUtils.getFile(this.getParamDirectory().getPath(), "html", "index.html");
			if (htmlFile.exists()) {
				try {

					panel.removeAll();

					JEditorPane ep = new JEditorPane();
					ep.setContentType("text/html");
					ep.setEditable(false);
					ep.setPage(htmlFile.toURI().toURL());
					JScrollPane jsp = new JScrollPane(ep);

					panel.add(jsp);
				} catch (Exception ex) {
					ex.printStackTrace();
				}
			}
		}
	}

	@Override
	public boolean hasShow() {
		return false;
	}

	@Override
	public void show() {
		// xxx python
	}

}
