/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_30.
 * 
 * - Change optimizer structure
 * 
 * @author Werner Koppelsttter
 */
public class Converter_30 implements MopAccess {

	/** Mop-file of project */
	MopFile mop;

	/** Node of load archive */
	XMLNode node = null;

	/**
	 * Convert the given node
	 * 
	 * @param mop
	 *            Mop-file with optimization
	 * @throws Exception
	 */
	static void convertOptimizer(MopFile mop) throws Exception {
		Converter_30 convert;
		String[] filelist;

		convert = new Converter_30(mop);
		filelist = mop.getFileNames();
		for (String name : filelist) {
			if (name.startsWith("GeneticOptimization/")
					&& name.endsWith(".xml")) {
				convert.convertOpt(name);
			}
		}
	}

	/**
	 * Create converter object
	 * 
	 * @param mop
	 *            Mop-file to convert
	 */
	private Converter_30(MopFile mop) {
		this.mop = mop;
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	private void convertOpt(String name) throws Exception {
		this.node = null;
		this.mop.readStream(this, name);
		this.node = this.convertArchive(this.node);
		this.mop.writeStream(this, name);
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param oldNodes
	 *            Node of old structure
	 * @return Node with new structure
	 */
	private XMLNode convertArchive(XMLNode oldNodes) {
		XMLNode newNodes;

		newNodes = new XMLNode("Struct");
		this.createArchive(newNodes);

		// baseIndex
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("curIndex") == 0) {
				newNodes.getNode(0)
						.getNode(0)
						.setText(
								Double.toString(Integer.parseInt(n.getNode(0)
										.getText())));
			}
		}

		// variables
		this.refreshVariables(oldNodes,
				newNodes.getNode(4).getNode(0).getNode(0));
		// results
		this.refreshResults(oldNodes, newNodes.getNode(3).getNode(0).getNode(0));

		// memory
		this.refreshMemory(oldNodes, newNodes.getNode(1).getNode(0).getNode(0),
				newNodes.getNode(3).getNode(0).getNode(0).getNodeSize());

		// problem
		this.refreshProblem(oldNodes, newNodes.getNode(2).getNode(0).getNode(0)
				.getNode(0), newNodes.getNode(1).getNode(0).getNode(0));

		this.finishArchive(newNodes);

		return newNodes;
	}

	/**
	 * Create node structure of archive
	 * 
	 * @param newNode
	 *            Structure of new archive
	 */
	private void createArchive(XMLNode newNode) {
		XMLNode item, parent;

		// baseIndex
		item = new XMLNode("field");
		item.setAttribute("name", "baseIndex");
		newNode.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("1.0");
		parent.addNode(item);

		// memory
		item = new XMLNode("field");
		item.setAttribute("name", "memory");
		newNode.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

		// problems
		item = new XMLNode("field");
		item.setAttribute("name", "problems");
		newNode.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("Struct");
		parent.addNode(item);

		// results
		item = new XMLNode("field");
		item.setAttribute("name", "results");
		newNode.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

		// variables
		item = new XMLNode("field");
		item.setAttribute("name", "variables");
		newNode.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

	}

	/**
	 * Refresh names of variables
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param vars
	 *            List of varables
	 */
	private void refreshVariables(XMLNode oldNodes, XMLNode vars) {
		XMLNode list;

		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("varDesc") == 0) {
				list = n.getNode(0).getNode(0);
				for (XMLNode v : list.getNodes()) {
					for (XMLNode item : v.getNodes()) {
						if (item.getAttribute("name").compareTo("name") == 0) {
							vars.addNode(item.getNode(0));
						}
					}
				}
			}
		}
	}

	/**
	 * Refresh names of results
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param res
	 *            List of results
	 */
	private void refreshResults(XMLNode oldNodes, XMLNode res) {
		XMLNode list;
		String name;
		boolean exist;

		// objectives
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("objDesc") == 0) {
				list = n.getNode(0).getNode(0);
				for (XMLNode r : list.getNodes()) {
					for (XMLNode item : r.getNodes()) {
						if (item.getAttribute("name").compareTo("name") == 0) {
							res.addNode(item.getNode(0));
						}
					}
				}
			}
		}

		// constraints
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("constrDesc") == 0) {
				list = n.getNode(0);
				if (list.getNodeSize() > 0) {
					list = list.getNode(0);
					for (XMLNode r : list.getNodes()) {
						for (XMLNode item : r.getNodes()) {
							if (item.getAttribute("name").compareTo("name") == 0) {
								name = item.getNode(0).getText();
								exist = false;
								for (XMLNode str : res.getNodes()) {
									exist |= (str.getText().compareTo(name) == 0);
								}
								if (!exist) {
									res.addNode(item.getNode(0));
								}
							}
						}
					}
				}
			}
		}

	}

	/**
	 * Refresh memory. Add all individuals
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param mem
	 *            Memory (List of individuals)
	 * @param numRes
	 *            Number of results
	 */
	private void refreshMemory(XMLNode oldNodes, XMLNode mem, int numRes) {
		String var, res;

		// paretos
		for (XMLNode n : oldNodes.getNodes()) {
			// pareto history
			if (n.getAttribute("name").compareTo("paretoHistory") == 0) {
				for (XMLNode p : n.getNode(0).getNode(0).getNodes()) {
					var = null;
					res = null;
					for (XMLNode item : p.getNodes()) {
						if (item.getAttribute("name").compareTo(
								"ParetoObjectives") == 0) {
							res = item.getNode(0).getText();
						} else if (item.getAttribute("name").compareTo(
								"ParetoVariables") == 0) {
							var = item.getNode(0).getText();
						}
					}
					this.addtoMemory(mem, numRes, var, res);
				}
				// resume data
			} else if (n.getAttribute("name").compareTo("resumeData") == 0) {
				for (XMLNode n1 : n.getNode(0).getNodes()) {
					// save data
					if (n1.getAttribute("name").compareTo("saveData") == 0) {
						// archive and pareto
						for (XMLNode n2 : n1.getNode(0).getNodes()) {
							var = null;
							res = null;
							for (XMLNode item : n2.getNode(0).getNodes()) {
								if (item.getAttribute("name").compareTo(
										"objectives") == 0) {
									res = item.getNode(0).getText();
								} else if (item.getAttribute("name").compareTo(
										"variables") == 0) {
									var = item.getNode(0).getText();
								}
							}
							this.addtoMemory(mem, numRes, var, res);
						}
					}
				}
			}
		}
	}

	/**
	 * Get index of memory
	 * 
	 * @param mem
	 *            Memory (List of individuals)
	 * @param varLine
	 *            Line of individuals-variables
	 * @return List of indexes in memory
	 */
	private String getIndexofMemory(XMLNode mem, String varLine) {
		String[] var;
		LinkedList<Integer> list;
		String str;

		list = new LinkedList<Integer>();

		// prepare
		if (varLine.startsWith("[") && varLine.endsWith("]")) {
			varLine = varLine.substring(1, varLine.length() - 1);
		}
		var = varLine.split(";");

		// search individual
		for (int k = 0; k < var.length; k++) {
			var[k] = "[" + var[k] + "]";
			for (int m = 0; m < mem.getNodeSize(); m++) {
				if (mem.getNode(m).getNode(1).getNode(0).getText()
						.compareTo(var[k]) == 0) {
					if (!list.contains(m)) {
						list.add(m);
					}
				}
			}
		}

		str = "[";
		for (int ind : list) {
			str += Double.toString(ind) + " ";
		}
		if (str.length() > 1) {
			str = str.substring(0, str.length() - 1);
		}
		str += "]";

		return str;

	}

	/**
	 * Add individual to memory
	 * 
	 * @param mem
	 *            Memory (List of individuals)
	 * @param numRes
	 *            Number of results
	 * @param varStr
	 *            Variable-values
	 * @param resStr
	 *            Result-values
	 */
	private void addtoMemory(XMLNode mem, int numRes, String varStr,
			String resStr) {
		XMLNode ind, field, val;
		String[] var, res;
		boolean exist;

		// prepare var and res
		if (varStr.startsWith("[") && varStr.endsWith("]")) {
			varStr = varStr.substring(1, varStr.length() - 1);
		}
		var = varStr.split(";");

		if (resStr.startsWith("[") && resStr.endsWith("]")) {
			resStr = resStr.substring(1, resStr.length() - 1);
		}
		res = resStr.split(";");

		// search and add new individuals
		for (int k = 0; k < var.length; k++) {
			exist = false;

			var[k] = "[" + var[k] + "]";
			for (XMLNode m : mem.getNodes()) {
				exist |= (m.getNode(1).getNode(0).getText().compareTo(var[k]) == 0);
			}

			if (!exist) {
				while (res[k].split(" ").length < numRes) {
					res[k] += " NaN";
				}
				res[k] = "[" + res[k] + "]";
				ind = new XMLNode("Struct");
				mem.addNode(ind);

				// results
				field = new XMLNode("field");
				field.setAttribute("name", "res");
				ind.addNode(field);
				val = new XMLNode("Numeric");
				val.setText(res[k]);
				field.addNode(val);

				// variables
				field = new XMLNode("field");
				field.setAttribute("name", "var");
				ind.addNode(field);
				val = new XMLNode("Numeric");
				val.setText(var[k]);
				field.addNode(val);
			}
		}

	}

	/**
	 * Refresh problem
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param problem
	 *            Problem-node
	 * @param mem
	 *            Memory (List of individuals)
	 */
	private void refreshProblem(XMLNode oldNodes, XMLNode problem, XMLNode mem) {
		this.createProblem(problem);

		// autosaveInterval
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("autosaveInterval") == 0) {
				problem.getNode(0).getNode(0).setText(n.getNode(0).getText());
			}
		}

		// functionTimeout
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("timeoutFunction") == 0) {
				problem.getNode(2).getNode(0).setText(n.getNode(0).getText());
			}
		}

		// poolsize
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("popsize") == 0) {
				problem.getNode(5)
						.getNode(0)
						.setText(
								Double.toString(Double.parseDouble(n.getNode(0)
										.getText())));
			}
		}

		// varDesc
		this.refreshVarDesc(oldNodes, problem.getNode(9).getNode(0).getNode(0));

		// resDesc
		this.refreshResDesc(oldNodes, problem.getNode(6).getNode(0).getNode(0));

		// paretos (local and full)
		this.refreshParetos(oldNodes, mem, problem.getNode(1).getNode(0)
				.getNode(0), problem.getNode(3).getNode(0).getNode(0));

		// solver
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("solver") == 0) {
				if (n.getNode(0).getText().compareTo("Spea2Real") == 0) {
					problem.getNode(7).getNode(0).setText("GEN_SPEA2");
				} else {
					problem.getNode(7).getNode(0).setText("GEN_NSGA2");
				}
			}
		}

		// solverSettings
		this.refreshSolverSettings(oldNodes, problem.getNode(8).getNode(0));

		this.finishProblem(problem);
	}

	/**
	 * Create node structure of problem
	 * 
	 * @param prob
	 *            Structure of new problem
	 */
	private void createProblem(XMLNode prob) {
		XMLNode item, parent;

		// autosaveInterval
		item = new XMLNode("field");
		item.setAttribute("name", "autosaveInterval");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("100.0");
		parent.addNode(item);

		// fullParetos
		item = new XMLNode("field");
		item.setAttribute("name", "fullParetos");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

		// functionTimeout
		item = new XMLNode("field");
		item.setAttribute("name", "functionTimeout");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("30.0");
		parent.addNode(item);

		// localParetos
		item = new XMLNode("field");
		item.setAttribute("name", "localParetos");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

		// name
		item = new XMLNode("field");
		item.setAttribute("name", "name");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("String");
		item.setText("problem");
		parent.addNode(item);

		// poolsize
		item = new XMLNode("field");
		item.setAttribute("name", "poolsize");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("100.0");
		parent.addNode(item);

		// resDesc
		item = new XMLNode("field");
		item.setAttribute("name", "resDesc");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);

		// solver
		item = new XMLNode("field");
		item.setAttribute("name", "solver");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("String");
		parent.addNode(item);

		// solverSettings
		item = new XMLNode("field");
		item.setAttribute("name", "solverSettings");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Struct");
		parent.addNode(item);

		// varDesc
		item = new XMLNode("field");
		item.setAttribute("name", "varDesc");
		prob.addNode(item);
		parent = item;
		item = new XMLNode("Array");
		parent.addNode(item);
		parent = item;
		item = new XMLNode("row");
		parent.addNode(item);
	}

	/**
	 * Refresh variable-description
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param varsDesc
	 *            List of variable-descriptions
	 */
	private void refreshVarDesc(XMLNode oldNodes, XMLNode varsDesc) {
		XMLNode list;

		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("varDesc") == 0) {
				list = n.getNode(0).getNode(0);
				for (XMLNode v : list.getNodes()) {
					varsDesc.addNode(v);
				}
			}
		}
	}

	/**
	 * Refresh results-description
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param resDesc
	 *            List of result-descriptions
	 */
	private void refreshResDesc(XMLNode oldNodes, XMLNode resDesc) {
		XMLNode list, res;
		String name, lb, ub;

		// objectives
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("objDesc") == 0) {
				list = n.getNode(0).getNode(0);
				for (XMLNode r : list.getNodes()) {
					name = r.getNode(0).getNode(0).getText();
					resDesc.addNode(this.createResDesc(name, true));
				}
			}
		}

		// constraints
		for (XMLNode n : oldNodes.getNodes()) {
			if (n.getAttribute("name").compareTo("constrDesc") == 0) {
				list = n.getNode(0);
				if (list.getNodeSize() > 0) {
					list = list.getNode(0);
					for (XMLNode r : list.getNodes()) {
						name = "";
						lb = "";
						ub = "";
						for (XMLNode item : r.getNodes()) {
							if (item.getAttribute("name").compareTo("name") == 0) {
								name = item.getNode(0).getText();
							} else if (item.getAttribute("name")
									.compareTo("lb") == 0) {
								lb = item.getNode(0).getText();
							} else if (item.getAttribute("name")
									.compareTo("ub") == 0) {
								ub = item.getNode(0).getText();
							}
						}
						res = null;
						for (XMLNode item : resDesc.getNodes()) {
							if (item.getNode(1).getNode(0).getText()
									.compareTo(name) == 0) {
								res = item;
							}
						}
						if (res == null) {
							res = this.createResDesc(name, false);
							resDesc.addNode(res);
						}

						res.getNode(0).getNode(0).setText(lb);
						res.getNode(3).getNode(0).setText(ub);
					}
				}
			}
		}

	}

	/**
	 * Create node structure of result description
	 * 
	 * @param name
	 *            Name of result
	 * @param obj
	 *            Is objective
	 * @return Structure of new result description
	 */
	private XMLNode createResDesc(String name, boolean obj) {
		XMLNode res, field, val;

		res = new XMLNode("Struct");

		// lb
		field = new XMLNode("field");
		field.setAttribute("name", "lb");
		res.addNode(field);
		val = new XMLNode("Numeric");
		val.setText("NaN");
		field.addNode(val);

		// name
		field = new XMLNode("field");
		field.setAttribute("name", "name");
		res.addNode(field);
		val = new XMLNode("String");
		val.setText(name);
		field.addNode(val);

		// objective
		field = new XMLNode("field");
		field.setAttribute("name", "obj");
		res.addNode(field);
		val = new XMLNode("String");
		if (obj) {
			val.setText("MIN");
		} else {
			val.setText("NONE");
		}
		field.addNode(val);

		// ub
		field = new XMLNode("field");
		field.setAttribute("name", "ub");
		res.addNode(field);
		val = new XMLNode("Numeric");
		val.setText("NaN");
		field.addNode(val);

		return res;
	}

	/**
	 * Refresh individual memory with pareto-history
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param mem
	 *            Memory (List of individuals)
	 * @param fullPareto
	 *            List of full paretos
	 * @param localPareto
	 *            List of local paretos
	 */
	private void refreshParetos(XMLNode oldNodes, XMLNode mem,
			XMLNode fullPareto, XMLNode localPareto) {
		XMLNode pareto;
		String str;

		for (XMLNode n : oldNodes.getNodes()) {
			if ((n.getAttribute("name").compareTo("paretoHistory") == 0)
					&& (n.getNode(0).getNodeSize() > 0)) {
				for (XMLNode p : n.getNode(0).getNode(0).getNodes()) {
					for (XMLNode item : p.getNodes()) {
						if (item.getAttribute("name").compareTo(
								"ParetoVariables") == 0) {
							str = this.getIndexofMemory(mem, item.getNode(0)
									.getText());
							if (str.length() == 2) {
								pareto = new XMLNode("Null");
							} else {
								pareto = new XMLNode("Numeric");
								pareto.setText(str);
							}
							fullPareto.addNode(pareto);
							localPareto.addNode(pareto);
						}
					}
				}
			}
		}
	}

	/**
	 * Refresh settings of solver
	 * 
	 * @param oldNodes
	 *            Old archive
	 * @param settings
	 *            Node with settings
	 */
	private void refreshSolverSettings(XMLNode oldNodes, XMLNode settings) {
		this.createSolverSettings(settings);

		for (XMLNode n : oldNodes.getNodes()) {
			// populationSize
			if (n.getAttribute("name").compareTo("popsize") == 0) {
				settings.getNode(6)
						.getNode(0)
						.setText(
								Double.toString(Double.parseDouble(n.getNode(0)
										.getText())));
			} else if (n.getAttribute("name").compareTo("timeoutGeneration") == 0) {
				// generationTimeout
				settings.getNode(3)
						.getNode(0)
						.setText(
								Double.toString(Double.parseDouble(n.getNode(0)
										.getText())));
			} else if (n.getAttribute("name").compareTo("optsettings") == 0) {
				for (XMLNode s : n.getNode(0).getNodes()) {
					if (s.getAttribute("name").compareTo("ArchiveSize") == 0) {
						// archiveSize
						settings.getNode(0).getNode(0)
								.setText(s.getNode(0).getText());
					} else if (s.getAttribute("name").compareTo(
							"CrossoverDistributionIndex") == 0) {
						// crossover distribution index
						settings.getNode(1).getNode(0)
								.setText(s.getNode(0).getText());
					} else if (s.getAttribute("name").compareTo(
							"CrossoverProbability") == 0) {
						// crossover probability
						settings.getNode(2).getNode(0)
								.setText(s.getNode(0).getText());
					} else if (s.getAttribute("name").compareTo(
							"MutationDistributionIndex") == 0) {
						// mutation distribution index
						settings.getNode(4).getNode(0)
								.setText(s.getNode(0).getText());
					} else if (s.getAttribute("name").compareTo(
							"MutationProbability") == 0) {
						// mutation probability
						settings.getNode(5).getNode(0)
								.setText(s.getNode(0).getText());
					}
				}
			}
		}
	}

	/**
	 * Create node structure of solver-settings
	 * 
	 * @param settings
	 *            Node with solver-settings
	 */
	private void createSolverSettings(XMLNode settings) {
		XMLNode item, parent;

		// archiveSize
		item = new XMLNode("field");
		item.setAttribute("name", "archiveSize");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("200.0");
		parent.addNode(item);

		// crossover distribution index
		item = new XMLNode("field");
		item.setAttribute("name", "crossoverDistributionIndex");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("20.0");
		parent.addNode(item);

		// crossover probability
		item = new XMLNode("field");
		item.setAttribute("name", "crossoverProbability");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("0.9");
		parent.addNode(item);

		// generationTimeout
		item = new XMLNode("field");
		item.setAttribute("name", "generationTimeout");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("120.0");
		parent.addNode(item);

		// mutation distribution index
		item = new XMLNode("field");
		item.setAttribute("name", "mutationDistributionIndex");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("20.0");
		parent.addNode(item);

		// mutation probability
		item = new XMLNode("field");
		item.setAttribute("name", "mutationProbability");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("0.2");
		parent.addNode(item);

		// populationSize
		item = new XMLNode("field");
		item.setAttribute("name", "populationSize");
		settings.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("100.0");
		parent.addNode(item);

	}

	/**
	 * Finish problem. (Refresh size)
	 * 
	 * @param prob
	 *            Node with problem definition
	 */
	private void finishProblem(XMLNode prob) {
		XMLNode item;

		// fullParetos
		item = prob.getNode(1).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// localParetos
		item = prob.getNode(3).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// resDesc
		item = prob.getNode(6).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// varDesc
		item = prob.getNode(9).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));
	}

	/**
	 * Finish archive. (Refresh size)
	 * 
	 * @param archive
	 *            Node with archive definition
	 */
	private void finishArchive(XMLNode archive) {
		XMLNode item;

		// memory
		item = archive.getNode(1).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// problems
		item = archive.getNode(2).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// results
		item = archive.getNode(3).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

		// variables
		item = archive.getNode(4).getNode(0);
		item.setAttribute("nrows", "1");
		item.setAttribute("ncolumns",
				Integer.toString(item.getNode(0).getNodeSize()));

	}

	@Override
	public void readStream(InputStream is) throws Exception {
		XMLStreamReader xmlReader = XMLInputFactory.newInstance()
				.createXMLStreamReader(is, "utf-8");
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		// read objects
		this.node = new XMLNode(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new XMLStreamException("Undefined error by reading file!");
		}
		xmlReader.close();
	}

	@Override
	public void writeStream(OutputStream os) throws Exception {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance()
				.createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");

		// write object to File
		this.node.writeXML(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}

}
