/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.converter;

import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

import magopt.utils.MopAccess;
import magopt.utils.MopFile;
import at.lcm.bu21.general.utils.XMLNode;

/**
 * Convert to V_0_31.
 * 
 * - Change optimizer savepoints <br>
 * - Add periodic savepoints
 * 
 * @author Werner Koppelsttter
 */
public class Converter_31 implements MopAccess {

	/** Mop-file of project */
	MopFile mop;

	/** Node of load archive */
	XMLNode node = null;

	/**
	 * Convert the given node
	 * 
	 * @param mop
	 *            Mop-file with optimization
	 * @throws Exception
	 */
	static void convertOptimizer(MopFile mop) throws Exception {
		Converter_31 convert;
		String[] filelist;

		convert = new Converter_31(mop);
		filelist = mop.getFileNames();
		for (String name : filelist) {
			if (name.startsWith("GeneticOptimization/")
					&& name.endsWith(".xml")) {
				convert.convertOpt(name);
			}
		}
	}

	/**
	 * Create converter object
	 * 
	 * @param mop
	 *            Mop-file to convert
	 */
	private Converter_31(MopFile mop) {
		this.mop = mop;
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param name
	 *            Name of archive
	 * @throws Exception
	 */
	private void convertOpt(String name) throws Exception {
		this.node = null;
		this.mop.readStream(this, name);
		this.convert(this.node);
		this.mop.writeStream(this, name);
	}

	/**
	 * Convert optimizer archive-file
	 * 
	 * @param node
	 *            Archive to convert
	 */
	private void convert(XMLNode node) {
		XMLNode problems;

		for (XMLNode n : node.getNodes()) {
			if (n.getAttribute("name").compareTo("problems") == 0) {
				problems = n.getNode(0).getNode(0);
				for (XMLNode prob : problems.getNodes()) {
					this.convertProblem(prob);
				}
			}
		}

	}

	/**
	 * Convert problem
	 * 
	 * @param node
	 *            Problem node
	 */
	private void convertProblem(XMLNode node) {
		XMLNode full = null;
		XMLNode local = null;
		XMLNode n, item, parent;

		// find paretos
		for (int k = node.getNodeSize() - 1; k >= 0; k--) {
			n = node.getNode(k);
			if (n.getAttribute("name").compareTo("fullParetos") == 0) {
				full = n;
				node.rmNode(k);
			} else if (n.getAttribute("name").compareTo("localParetos") == 0) {
				local = n;
				node.rmNode(k);
			}
		}

		// create savepoint
		if ((full != null) && (local != null)) {
			item = new XMLNode("field");
			item.setAttribute("name", "savepoints");
			node.addNode(item);
			parent = item;
			item = new XMLNode("Array");
			item.setAttribute("nrows", full.getNode(0).getAttribute("nrows"));
			item.setAttribute("ncolumns",
					full.getNode(0).getAttribute("ncolumns"));
			parent.addNode(item);
			parent = item;
			item = new XMLNode("row");
			parent.addNode(item);

			full = full.getNode(0).getNode(0);
			local = local.getNode(0).getNode(0);

			for (int k = 0; k < full.getNodeSize(); k++) {
				item.addNode(this.createSavePoint(full.getNode(k),
						local.getNode(k)));
			}
		}

		// add autosave time enabled
		item = new XMLNode("field");
		item.setAttribute("name", "autosaveTimeEnabled");
		node.addNode(item);
		parent = item;
		item = new XMLNode("Boolean");
		item.setText("false");
		parent.addNode(item);

		// add autosave time
		item = new XMLNode("field");
		item.setAttribute("name", "autosaveTime");
		node.addNode(item);
		parent = item;
		item = new XMLNode("Numeric");
		item.setText("60.0");
		parent.addNode(item);

	}

	/**
	 * Create savepoint
	 * 
	 * @param full
	 *            Full pareto-front
	 * @param local
	 *            Local pareto-front
	 * @return Node of savepoint
	 */
	private XMLNode createSavePoint(XMLNode full, XMLNode local) {
		XMLNode sp, item;

		sp = new XMLNode("Struct");

		// date
		item = new XMLNode("field");
		item.setAttribute("name", "date");
		item.addNode(new XMLNode("Null"));
		sp.addNode(item);

		// type
		item = new XMLNode("field");
		item.setAttribute("name", "type");
		item.addNode(new XMLNode("String"));
		item.getNode(0).setText("INTERVAL");
		sp.addNode(item);

		// local pareto
		item = new XMLNode("field");
		item.setAttribute("name", "full");
		item.addNode(local);
		sp.addNode(item);

		// local pareto
		item = new XMLNode("field");
		item.setAttribute("name", "local");
		item.addNode(local);
		sp.addNode(item);

		return sp;
	}

	@Override
	public void readStream(InputStream is) throws Exception {
		XMLStreamReader xmlReader = XMLInputFactory.newInstance()
				.createXMLStreamReader(is, "utf-8");
		while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
			xmlReader.next();
		}

		// read objects
		this.node = new XMLNode(xmlReader);

		if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
			throw new XMLStreamException("Undefined error by reading file!");
		}
		xmlReader.close();
	}

	@Override
	public void writeStream(OutputStream os) throws Exception {
		XMLStreamWriter xmlWriter = XMLOutputFactory.newInstance()
				.createXMLStreamWriter(os, "utf-8");
		xmlWriter.writeStartDocument("utf-8", "1.0");

		// write object to File
		this.node.writeXML(xmlWriter);

		xmlWriter.writeEndDocument();
		xmlWriter.close();
	}

}
