/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.LinkedList;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNull;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TPeriodic;
import at.lcm.bu21.general.dtypes.TPeriodicMatrix;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Evaluate method
 *
 * @author Werner Koppelsttter
 */
public class EvalMethod {

	/**
	 * Evaluate method
	 * 
	 * @param obj
	 *            Object to eval method
	 * @param name
	 *            Name of Method
	 * @param params
	 *            List of parameters
	 * @return Evaluated value
	 * 
	 * @throws Exception
	 */
	public static TData evalMethod(TData obj, String name, TData[] params)
			throws Exception {
		LinkedList<Method> methods;
		Object[] oParams;
		Object result;

		// check if method exists
		methods = new LinkedList<Method>();
		methods.addAll(Arrays.asList(obj.getClass().getMethods()));
		// check name
		for (int k = methods.size() - 1; k >= 0; k--) {
			if (!name.equals(methods.get(k).getName())) {
				methods.remove(k);
			}
		}
		if (methods.size() == 0) {
			throw new Exception("Method is undefined for '"
					+ obj.getClass().getSimpleName() + "'");
		}

		// check parameter
		oParams = EvalMethod.checkParameterList(methods, params);

		// eval
		result = methods.get(0).invoke(obj, oParams);
		return EvalMethod.convertReturn(result);
	}

	/**
	 * Filter methods by parameter-list and convert parameter-types
	 * 
	 * @param methods
	 *            Method to execute
	 * @param params
	 *            List of parameters
	 * @return List of converted parameters
	 * 
	 * @throws Exception
	 */
	private static Object[] checkParameterList(LinkedList<Method> methods,
			TData[] params) throws Exception {
		Object[] oParams = null, tmp;

		for (int k = methods.size() - 1; k >= 0; k--) {
			Class<?>[] cl = methods.get(k).getParameterTypes();
			if (params.length == cl.length) {
				tmp = EvalMethod.getParameterList(cl, params);
				if (tmp == null) {
					methods.remove(k);
				} else {
					oParams = tmp;
				}
			} else {
				methods.remove(k);
			}
		}

		if (methods.size() == 0) {
			throw new Exception("No method which fit to parameter-list");
		} else if (methods.size() > 1) {
			throw new Exception("More than one method fit to parameter-list");
		}

		return oParams;
	}

	/**
	 * Get converted list of parameter. If list is null then parameter doesn't
	 * match to method
	 * 
	 * @param cl
	 *            List of parameter-types
	 * @param params
	 *            List of parameters
	 * @return Converted list of parameters
	 */
	private static Object[] getParameterList(Class<?>[] cl, TData[] params) {
		Object[] oParams;

		oParams = new Object[params.length];

		for (int k = 0; k < params.length; k++) {
			if (cl[k].isAssignableFrom(Object.class)) {
				oParams[k] = params[k];
			} else if (cl[k].isAssignableFrom(TData.class)) {
				oParams[k] = params[k];
			} else if (params[k] instanceof TNull) {
				if (Object.class.isAssignableFrom(cl[k])) {
					oParams[k] = null;
				}
			} else if (params[k] instanceof TString) {
				if (TString.class.isAssignableFrom(cl[k])) {
					oParams[k] = params[k];
				} else if (String.class.isAssignableFrom(cl[k])) {
					oParams[k] = ((TString) params[k]).getString();
				}
			} else if (params[k] instanceof TBoolean) {
				if (TBoolean.class.isAssignableFrom(cl[k])) {
					oParams[k] = params[k];
				} else if (boolean.class.isAssignableFrom(cl[k])) {
					oParams[k] = ((TBoolean) params[k]).getBoolean();
				} else if (Boolean.class.isAssignableFrom(cl[k])) {
					oParams[k] = ((TBoolean) params[k]).getBoolean();
				}
			} else if (params[k] instanceof TNumeric) {
				oParams[k] = EvalMethod.getNumericParameter(
						(TNumeric) params[k], cl[k]);
			} else if (params[k] instanceof TArray) {
				oParams[k] = EvalMethod.getArrayParameter((TArray) params[k],
						cl[k]);

			} else if (params[k] instanceof TStruct) {
				if (TStruct.class.isAssignableFrom(cl[k])) {
					oParams[k] = params[k];
				}
			} else if (params[k] instanceof TPeriodic) {
				if (TPeriodic.class.isAssignableFrom(cl[k])) {
					oParams[k] = params[k];
				}
			} else if (params[k] instanceof TPeriodicMatrix) {
				if (TPeriodicMatrix.class.isAssignableFrom(cl[k])) {
					oParams[k] = params[k];
				}
			}

			if (oParams[k] == null) {
				return null;
			}
		}

		return oParams;
	}

	/**
	 * Get converted numeric parameter
	 * 
	 * @param num
	 *            Parameter
	 * @param cl
	 *            Type of parameter
	 * @return Converted parameter
	 */
	private static Object getNumericParameter(TNumeric num, Class<?> cl) {
		Object obj = null;

		if (TNumeric.class.isAssignableFrom(cl)) {
			obj = num;
		} else if (byte.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			double val = num.getReal(0, 0);
			if ((Byte.MIN_VALUE <= val) && (val <= Byte.MAX_VALUE)) {
				obj = (byte) val;
			}
		} else if (byte[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			byte[] list;
			double val;
			boolean valid = true;
			if (num.size()[0] == 1) {
				list = new byte[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					val = num.getReal(0, col);
					valid &= (Byte.MIN_VALUE <= val) && (val <= Byte.MAX_VALUE);
					list[col] = (byte) val;
				}
			} else {
				list = new byte[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, 0);
					valid &= (Byte.MIN_VALUE <= val) && (val <= Byte.MAX_VALUE);
					list[row] = (byte) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (byte[][].class.isAssignableFrom(cl) && num.isReal()) {
			byte[][] list = new byte[num.size()[0]][num.size()[1]];
			double val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					val = list[row][col];
					valid &= (Byte.MIN_VALUE <= val) && (val <= Byte.MAX_VALUE);
					list[row][col] = (byte) num.getReal(row, col);
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (short.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			double val = num.getReal(0, 0);
			if ((Short.MIN_VALUE <= val) && (val <= Short.MAX_VALUE)) {
				obj = (short) val;
			}
		} else if (short[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			short[] list;
			double val;
			boolean valid = true;
			if (num.size()[0] == 1) {
				list = new short[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					val = num.getReal(0, col);
					valid &= (Short.MIN_VALUE <= val)
							&& (val <= Short.MAX_VALUE);
					list[col] = (short) val;
				}
			} else {
				list = new short[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, 0);
					valid &= (Short.MIN_VALUE <= val)
							&& (val <= Short.MAX_VALUE);
					list[row] = (short) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (short[][].class.isAssignableFrom(cl) && num.isReal()) {
			short[][] list = new short[num.size()[0]][num.size()[1]];
			double val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					val = list[row][col];
					valid &= (Short.MIN_VALUE <= val)
							&& (val <= Short.MAX_VALUE);
					list[row][col] = (short) num.getReal(row, col);
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (int.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			double val = num.getReal(0, 0);
			if ((Integer.MIN_VALUE <= val) && (val <= Integer.MAX_VALUE)) {
				obj = (int) val;
			}
		} else if (int[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			int[] list;
			double val;
			boolean valid = true;
			if (num.size()[0] == 1) {
				list = new int[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					val = num.getReal(0, col);
					valid &= (Integer.MIN_VALUE <= val)
							&& (val <= Integer.MAX_VALUE);
					list[col] = (int) val;
				}
			} else {
				list = new int[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, 0);
					valid &= (Integer.MIN_VALUE <= val)
							&& (val <= Integer.MAX_VALUE);
					list[row] = (int) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (int[][].class.isAssignableFrom(cl) && num.isReal()) {
			int[][] list = new int[num.size()[0]][num.size()[1]];
			double val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					val = list[row][col];
					valid &= (Integer.MIN_VALUE <= val)
							&& (val <= Integer.MAX_VALUE);
					list[row][col] = (int) num.getReal(row, col);
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (long.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			double val = num.getReal(0, 0);
			if ((Long.MIN_VALUE <= val) && (val <= Long.MAX_VALUE)) {
				obj = (long) val;
			}
		} else if (long[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			long[] list;
			double val;
			boolean valid = true;
			if (num.size()[0] == 1) {
				list = new long[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					val = num.getReal(0, col);
					valid &= (Long.MIN_VALUE <= val) && (val <= Long.MAX_VALUE);
					list[col] = (long) val;
				}
			} else {
				list = new long[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, 0);
					valid &= (Long.MIN_VALUE <= val) && (val <= Long.MAX_VALUE);
					list[row] = (long) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (long[][].class.isAssignableFrom(cl) && num.isReal()) {
			long[][] list = new long[num.size()[0]][num.size()[1]];
			double val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					val = list[row][col];
					valid &= (Long.MIN_VALUE <= val) && (val <= Long.MAX_VALUE);
					list[row][col] = (long) num.getReal(row, col);
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (float.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			double val = num.getReal(0, 0);
			if ((Float.MIN_VALUE <= val) && (val <= Float.MAX_VALUE)) {
				obj = (float) val;
			}
		} else if (float[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			float[] list;
			double val;
			boolean valid = true;
			if (num.size()[0] == 1) {
				list = new float[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					val = num.getReal(0, col);
					valid &= (Float.MIN_VALUE <= val)
							&& (val <= Float.MAX_VALUE);
					list[col] = (float) val;
				}
			} else {
				list = new float[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, 0);
					valid &= (Float.MIN_VALUE <= val)
							&& (val <= Float.MAX_VALUE);
					list[row] = (float) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (float[][].class.isAssignableFrom(cl) && num.isReal()) {
			float[][] list = new float[num.size()[0]][num.size()[1]];
			double val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					val = num.getReal(row, col);
					valid &= (Float.MIN_VALUE <= val)
							&& (val <= Float.MAX_VALUE);
					list[row][col] = (float) val;
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (double.class.isAssignableFrom(cl) && num.isScalar()
				&& num.isReal()) {
			obj = num.getReal(0, 0);
		} else if (double[].class.isAssignableFrom(cl) && num.isReal()
				&& ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			double[] list;
			if (num.size()[0] == 1) {
				list = new double[num.size()[1]];
				for (int col = 0; col < list.length; col++) {
					list[col] = num.getReal(0, col);
				}
			} else {
				list = new double[num.size()[0]];
				for (int row = 0; row < list.length; row++) {
					list[row] = num.getReal(row, 0);
				}
			}
			obj = list;
		} else if (double[][].class.isAssignableFrom(cl) && num.isReal()) {
			double[][] list = new double[num.size()[0]][num.size()[1]];
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					list[row][col] = num.getReal(row, col);
				}
			}
			obj = list;
			// } else if (Number.class.isAssignableFrom(cl) && num.isScalar()
			// && num.isReal()) {
			// // FIXME
			// } else if (Number[].class.isAssignableFrom(cl) && num.isReal()
			// && ((num.size()[0] == 1) || (num.size()[1] == 1))) {
			// // FIXME
			// } else if (Number[][].class.isAssignableFrom(cl) && num.isReal())
			// {
			// // FIXME
		}

		return obj;
	}

	/**
	 * Get converted array parameter
	 * 
	 * @param arr
	 *            Parameter
	 * @param cl
	 *            Type of parameter
	 * @return Converted parameter
	 */
	private static Object getArrayParameter(TArray arr, Class<?> cl) {
		Object obj = null;

		if (TArray.class.isAssignableFrom(cl)) {
			obj = arr;
		} else if (String[].class.isAssignableFrom(cl)
				&& ((arr.size()[0] == 1) || (arr.size()[1] == 1))) {
			String[] list;
			TData val;
			boolean valid = true;
			if (arr.size()[0] == 1) {
				list = new String[arr.size()[1]];
				for (int col = 0; col < list.length; col++) {
					try {
						val = arr.get(0, col);
						if (val instanceof TString) {
							list[col] = ((TString) val).getString();
						} else {
							valid = false;
						}
					} catch (TDataException e) {
						valid = false;
					}
				}
			} else {
				list = new String[arr.size()[0]];
				for (int row = 0; row < list.length; row++) {
					try {
						val = arr.get(row, 0);
						if (val instanceof TString) {
							list[row] = ((TString) val).getString();
						} else {
							valid = false;
						}
					} catch (TDataException e) {
						valid = false;
					}
				}
			}
			if (valid) {
				obj = list;
			}
		} else if (String[][].class.isAssignableFrom(cl)) {
			String[][] list = new String[arr.size()[0]][arr.size()[1]];
			TData val;
			boolean valid = true;
			for (int col = 0; col < list.length; col++) {
				for (int row = 0; row < list.length; row++) {
					try {
						val = arr.get(row, col);
						if (val instanceof TString) {
							list[row][col] = ((TString) val).getString();
						} else {
							valid = false;
						}
					} catch (TDataException e) {
						valid = false;
					}
				}
			}
			if (valid) {
				obj = list;
			}
			// } else if (TData[].class.isAssignableFrom(cl)) {
			// // FIXME
			// } else if (TData[][].class.isAssignableFrom(cl)) {
			// // FIXME
			// } else if (Object[].class.isAssignableFrom(cl)) {
			// // FIXME
			// } else if (Object[][].class.isAssignableFrom(cl)) {
			// // FIXME
		}

		return obj;
	}

	/**
	 * Convert return value (result)
	 * 
	 * @param obj
	 *            Object of convert
	 * @return Converted value
	 * @throws Exception
	 */
	private static TData convertReturn(Object obj) throws Exception {
		if (obj instanceof TData) {
			return (TData) obj;
		} else if (obj instanceof Boolean) {
			return new TBoolean((Boolean) obj);
		} else if (obj instanceof Number) {
			return new TNumeric(((Number) obj).doubleValue());
		} else if (obj instanceof Number[]) {
			Number[] list = (Number[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k].doubleValue();
			}
			return new TNumeric(d);
		} else if (obj instanceof Number[][]) {
			Number[][] list = (Number[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col].doubleValue();
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof byte[]) {
			byte[] list = (byte[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof byte[][]) {
			byte[][] list = (byte[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof short[]) {
			short[] list = (short[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof short[][]) {
			short[][] list = (short[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof int[]) {
			int[] list = (int[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof int[][]) {
			int[][] list = (int[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof long[]) {
			long[] list = (long[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof long[][]) {
			long[][] list = (long[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof float[]) {
			float[] list = (float[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof float[][]) {
			float[][] list = (float[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof double[]) {
			double[] list = (double[]) obj;
			double[] d = new double[list.length];
			for (int k = 0; k < list.length; k++) {
				d[k] = list[k];
			}
			return new TNumeric(d);
		} else if (obj instanceof double[][]) {
			double[][] list = (double[][]) obj;
			double[][] d = new double[list.length][list[0].length];
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					d[row][col] = list[row][col];
				}
			}
			return new TNumeric(d);
		} else if (obj instanceof Character) {
			return new TString(((Character) obj).toString());
		} else if (obj instanceof Character[]) {
			Character[] list = (Character[]) obj;
			StringBuffer buff = new StringBuffer();
			for (int k = 0; k < list.length; k++) {
				buff.append(list[k]);
			}
			return new TString(buff.toString());
		} else if (obj instanceof char[]) {
			char[] list = (char[]) obj;
			StringBuffer buff = new StringBuffer();
			for (int k = 0; k < list.length; k++) {
				buff.append(list[k]);
			}
			return new TString(buff.toString());
		} else if (obj instanceof String) {
			return new TString(((String) obj).toString());
		} else if (obj instanceof String[]) {
			String[] list = (String[]) obj;
			TArray arr = new TArray();
			for (int k = 0; k < list.length; k++) {
				arr.set(0, k, new TString(list[k]));
			}
			return arr;
		} else if (obj instanceof String[][]) {
			String[][] list = (String[][]) obj;
			TArray arr = new TArray();
			for (int row = 0; row < list.length; row++) {
				for (int col = 0; col < list[0].length; col++) {
					arr.set(row, col, new TString(list[row][col]));
				}
			}
			return arr;
		} else {
			throw new Exception("Return value is not allowed");
		}
	}
}
