/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.magopt.structure.formula;

import java.util.Hashtable;
import java.util.Map;

/**
 * Kind of formula-tokens
 *
 * @author Werner Koppelsttter
 */
public enum TokenKind {
	// error token
	/** Undefined */
	NONE("Undefined"),
	// token-classes
	/** Identifier */
	IDENT("Identifier"),
	/** Number */
	NUMBER("Number"),
	/** String */
	STRING("String"),
	// operators end special-sign
	/** Plus (+) */
	PLUS("Plus", "+"),
	/** Minus (-) */
	MINUS("Minus", "-"),
	/** Times (*) */
	TIMES("Times", "*"),
	/** Dot-Times (.*) */
	DOTTIMES("DotTimes", ".*"),
	/** Slash (/) */
	SLASH("Slash", "/"),
	/** Dot-Slash (./) */
	DOTSLASH("DotSlash", "./"),
	/** Persent (%) */
	PERCENT("Percent", "%"),
	/** Dot-Persent (%) */
	DOTPERCENT("DotPercent", ".%"),
	/** Caret (^) */
	CARET("Caret", "^"),
	/** Dot-Caret (.^) */
	DOTCARET("DotCaret", ".^"),
	/** Accent (') */
	ACCENT("Accent", "'"),
	/** Equal (==) */
	EQL("Equal", "=="),
	/** Not equal (!=) */
	NEQ("Not equal", "!="),
	/** Less than (&lt;) */
	LSS("Less", "<"),
	/** Less or equal than (&lt;=) */
	LEQ("Less or equal", "<="),
	/** Greater (&gt;) */
	GTR("Greater", ">"),
	/** Greater or equal than (&gt;=) */
	GEQ("Geater or equal", ">="),
	/** And (&amp;&amp;) */
	AND("And", "&&"),
	/** Or (||) */
	OR("Or", "||"),
	/** Call (!) */
	CALL("Call", "!"),
	/** Semicolon (;) */
	SEMICOLON("Semicolon", ";"),
	/** Comma (,) */
	COMMA("Comma", ","),
	/** Period (.) */
	PERIOD("Period", "."),
	/** Colon (:) */
	COLON("Colon", ":"),
	/** Left parenthesis (() */
	LPAR("Left parenthesis", "("),
	/** Right parenthesis ()) */
	RPAR("Right parenthesis", ")"),
	/** Left bracket ([) */
	LBRACK("Left bracket", "["),
	/** Right bracket (]) */
	RBRACK("Right bracket", "]"),
	/** Left brace ({) */
	LBRACE("Left brace", "{"),
	/** Right brace (}) */
	RBRACE("Right brace", "}"),
	// constants
	/** Number pi */
	PI("Pi", "pi"),
	/** Euler's number */
	E("E", "e"),
	/** Imaginary number */
	I("I", "i"),
	/** True */
	TRUE("True", "true"),
	/** False */
	FALSE("False", "false"),
	/** Null */
	NULL("Null", "null"),
	// keywords
	/** If */
	IF("If", "if"),
	/** Else */
	ELSE("Else", "else"),
	/** End */
	END("End", "end"),
	// Additional kinds
	/** End of file */
	EOF("End of formula"),
	/** Var */
	VAR("Variable"),
	/** Method */
	FUNCTION("Function"),
	/** Method */
	METHOD("Method"),
	/** Numeric definition */
	NUMERIC("Numeric"),
	/** Array definition */
	ARRAY("Array"),
	/** Row of matrix */
	ROW("Row"); // row

	/** Description */
	private final String desc;

	/** Symbol */
	private final String sign;

	/** Map with keywords */
	private static Map<String, TokenKind> listKeywords;

	/**
	 * Create TokenKind with description
	 * 
	 * @param desc
	 *            Description
	 */
	private TokenKind(String desc) {
		this.desc = desc;
		this.sign = null;
	}

	/**
	 * Create TokenKind with description
	 * 
	 * @param desc
	 *            Description
	 * @param sign
	 *            Symbol
	 */
	private TokenKind(String desc, String sign) {
		this.desc = desc;
		this.sign = sign;
	}

	/**
	 * Get a map of keyword
	 * 
	 * @return List with keywords and TokenKind
	 */
	public static Map<String, TokenKind> getKeywords() {
		if (TokenKind.listKeywords == null) {
			TokenKind.listKeywords = new Hashtable<String, TokenKind>();
			TokenKind.listKeywords.put(PI.sign, PI);
			TokenKind.listKeywords.put(E.sign, E);
			TokenKind.listKeywords.put(I.sign, I);
			TokenKind.listKeywords.put(TRUE.sign, TRUE);
			TokenKind.listKeywords.put(FALSE.sign, FALSE);
			TokenKind.listKeywords.put(NULL.sign, NULL);
			TokenKind.listKeywords.put(IF.sign, IF);
			TokenKind.listKeywords.put(ELSE.sign, ELSE);
			TokenKind.listKeywords.put(END.sign, END);
		}

		return TokenKind.listKeywords;
	}

	/**
	 * Get description of token
	 * 
	 * @return Description
	 */
	public String getDescription() {
		return this.desc;
	}
}
