/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9713 $
 * $Date: 2020-03-19 09:28:58 +0100 (Thu, 19 Mar 2020) $
 */

package at.lcm.bu21.magopt.structure.ipython;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.UUID;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import org.zeromq.ZMQ;
import org.zeromq.ZMQ.Socket;

/**
 * Connection to IPython kernel
 * 
 * @author gweidenholzer
 */
public class KernelConnection {

	static final String HMAC_ALG = "HmacSHA256";

	private Mac mac;
	String username;
	String id;

	// ZMQ Sockets
	public Socket shell, iopub, stdin, control;

	/**
	 * Establish connection to IPython kernel
	 * 
	 * @param cf
	 *            Ipython kernel connection file
	 * @param ctx
	 *            ZMQ context. Use only one ZMQ context per application, even if
	 *            there are more kernel connections!
	 * @throws Exception
	 */
	public KernelConnection(ConnectionFile cf, ZMQ.Context ctx)
			throws Exception {

		// get unique session id
		this.id = UUID.randomUUID().toString();

		// get username
		this.username = System.getProperty("user.name");

		// connect shell socket
		this.shell = ctx.socket(ZMQ.DEALER);
		this.shell.connect(String.format("%s://%s:%d", cf.transport, cf.ip,
				cf.shell_port));

		// connect iopub socket
		this.iopub = ctx.socket(ZMQ.SUB);
		this.iopub.connect(String.format("%s://%s:%d", cf.transport, cf.ip,
				cf.iopub_port));

		// connect stdin socket
		this.stdin = ctx.socket(ZMQ.DEALER);
		this.stdin.connect(String.format("%s://%s:%d", cf.transport, cf.ip,
				cf.stdin_port));

		// connect control socket
		this.control = ctx.socket(ZMQ.DEALER);
		this.control.connect(String.format("%s://%s:%d", cf.transport, cf.ip,
				cf.control_port));

		// create Message Authentication Code engine
		if (cf.keyString.isEmpty()) {
			this.mac = null;
		} else {
			SecretKeySpec key = new SecretKeySpec(
					(cf.keyString).getBytes("UTF-8"),
					KernelConnection.HMAC_ALG);
			this.mac = Mac.getInstance(KernelConnection.HMAC_ALG);
			this.mac.init(key);
		}

	}

	/**
	 * close connections
	 */
	public void close() {
		this.shell.close();
		this.iopub.close();
		this.stdin.close();
		this.control.close();
	}

	/**
	 * Send message
	 * 
	 * @param socket
	 *            Socket to send message with
	 * @param msg
	 *            Message object
	 * @throws Exception
	 */
	public void send(ZMQ.Socket socket, Message msg) throws Exception {

		byte[][] toSend = msg.serialize(this);

		for (int k = 0; k < (toSend.length - 1); k++) {
			socket.sendMore(toSend[k]);
		}

		socket.send(toSend[toSend.length - 1]);

	}

	/**
	 * 
	 * @param socket
	 *            Socket to receive message from
	 * @return Message object
	 * @throws Exception
	 */
	public Message recv(ZMQ.Socket socket) throws Exception {

		ArrayList<byte[]> data = new ArrayList<byte[]>();
		data.add(socket.recv());

		while (socket.hasReceiveMore()) {
			data.add(socket.recv());
		}

		byte[][] serialized = data.toArray(new byte[0][]);

		return new Message(this, serialized);

	}

	/**
	 * Check signature
	 * 
	 * @return
	 * @throws Exception
	 */
	boolean checkSignature(byte[] signature, byte[][] signed) throws Exception {

		boolean ok = false;

		if (this.mac != null) {

			byte[] check = this.sign(signed);

			if (Arrays.equals(signature, check)) {
				ok = true;
			}

		} else {
			ok = true;
		}
		return ok;
	}

	/**
	 * Sign a message
	 * 
	 * @param toSign
	 *            the message payload (header, parent header, content, metadata)
	 * @return signature hash hex digest
	 * @throws Exception
	 */
	byte[] sign(byte[][] toSign) throws Exception {

		if (this.mac == null) {
			return new byte[] {};

		} else {

			for (int k = 0; k < toSign.length; k++) {
				this.mac.update(toSign[k]);
			}

			byte[] bytes = this.mac.doFinal();

			String hex = KernelConnection.hexDigest(bytes);
			return hex.getBytes("utf-8");
		}
	}

	/**
	 * converts bytearray to hex digest
	 * 
	 * @param bytes
	 * @return
	 */
	private static String hexDigest(byte[] bytes) {
		String digest;
		StringBuffer hash = new StringBuffer();
		for (int i = 0; i < bytes.length; i++) {
			String hex = Integer.toHexString(0xFF & bytes[i]);
			if (hex.length() == 1) {
				hash.append('0');
			}
			hash.append(hex);
		}
		digest = hash.toString();
		return digest;
	}

}
