/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9789 $
 * $Date: 2020-04-21 17:41:13 +0200 (Tue, 21 Apr 2020) $
 */
package at.lcm.symspace;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Map;

import at.lcm.bu21.magopt.utils.MagOptConfig;

/**
 * Launcher for SyMSpace. Launch Gui and console with dynamic settings
 *
 * @author Werner Koppelsttter
 */
public class Launcher {

	/**
	 * Start from console
	 * 
	 * @param args
	 *            Program arguments
	 */
	public static void main(String[] args) {
		Launcher.launch(args);
	}

	/**
	 * Start SyMSpace
	 * 
	 * @param args
	 *            Program arguments
	 */
	public static void launch(String... args) {
		String config;
		boolean debug = false, console = false;
		ProcessBuilder pb = null;
		LinkedList<String> arguments;
		String arg0;
		Map<String, String> env;

		try {
			config = MagOptConfig.getConfig("LauncherDebug");
			if ((config != null) && !config.isEmpty()) {
				debug = Boolean.parseBoolean(config);
			}

			// start parameter
			if (args != null) {
				arguments = new LinkedList<String>(Arrays.asList(args));
				arg0 = arguments.get(0);
				if ("SyMSpaceCenterGui".equals(arg0)) {
					arguments.set(0, "at.lcm.symspace.center.CenterGui");
					pb = Launcher.runJavaCommand(arguments, false);
				} else if ("SyMSpaceOptimizerGui".equals(arg0)) {
					arguments.set(0, "at.lcm.symspace.optimizer.OptimizerGui");
					pb = Launcher.runJavaCommand(arguments, false);
				} else if ("SyMSpaceConsole".equals(arg0)) {
					arguments.set(0, "at.lcm.symspace.Console");
					if (arguments.size() == 1) {
						pb = Launcher.runJavaCommand(arguments, true);
					} else {
						pb = Launcher.runJavaCommand(arguments, false);
						console = true;
					}
					// } else if ("SyMSpaceJep".equals(arg0)) {
					// arguments.set(0, "jep.Run");
					// if (arguments.size() == 1) {
					// arguments.add(1, new File(MagOptConfig.getJepPath(),
					// "console.py").getAbsolutePath());
					// pb = Launcher.runJavaCommand(arguments, true);
					// } else {
					// pb = Launcher.runJavaCommand(arguments, false);
					// console = true;
					// }
				} else if ("SyMSpaceRegenerate".equals(arg0)) {
					arguments.set(0, "at.lcm.symspace.regenerate.RegenerateTask");
					pb = Launcher.runJavaCommand(arguments, false);
				} else {
					System.out.println("Undefined command: " + arg0);
					System.out.println("Valid commands: SyMSpaceCenterGui, SyMSpaceOptimizerGui, SyMSpaceConsole");
				}
			}

			if (pb != null) {
				// launch
				final Process p = pb.start();

				// debug
				if (debug) {
					System.out.println("\n\nLaunch:");
					for (String c : pb.command()) {
						System.out.println(c);
					}

					System.out.println("\n\nEnvironment:");
					env = pb.environment();
					for (String key : env.keySet()) {
						System.out.println(key + "=" + env.get(key));
					}
				}

//				if (debug || console) {
				new Thread() {
					@Override
					public void run() {
						int b;
						InputStream in = p.getInputStream();
						try {
							while ((b = in.read()) != -1) {
								System.out.write(b);
							}
						} catch (IOException e) {
							System.err.println(e.getMessage());
							e.printStackTrace();
						}
					}
				}.start();

				new Thread() {
					@Override
					public void run() {
						int b;
						InputStream in = p.getErrorStream();
						try {
							while ((b = in.read()) != -1) {
								System.err.write(b);
							}
						} catch (IOException e) {
							System.err.println(e.getMessage());
							e.printStackTrace();
						}
					}
				}.start();
//				}
//				if (console) {
				System.exit(p.waitFor());
//				}
			}
		} catch (Throwable ex) {
			System.err.println(ex.getMessage());
			ex.printStackTrace();
			try {
				Thread.sleep(2 * 60 * 1000);
			} catch (InterruptedException e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
			}
		}

	}

	/**
	 * Run java class
	 * 
	 * @param args
	 *            Arguments for java
	 * @param console
	 *            open console
	 * @return Configured ProcessBuilder
	 */
	private static ProcessBuilder runJavaCommand(LinkedList<String> args, boolean console) {
		ProcessBuilder pb;
		LinkedList<String> cmd;
		Map<String, String> env;
		String config;
		File java = null, path;
		String javaName, javaExt;
		String[] vmArgs;

		cmd = new LinkedList<String>();

		// java version
		if (System.getProperty("os.name").equalsIgnoreCase("Linux")) {
			javaName = "java";
			javaExt = "";

			if (console) {
				// cmd.add("gnome-terminal");
				// cmd.add("-x");
				// cmd.add("sh");
				// cmd.add("-c");
				cmd.add("x-terminal-emulator");
				cmd.add("-e");
			}

		} else {
			if (console) {
				javaName = "java";
			} else {
				javaName = "javaw";
			}

			if (console) {
				cmd.add("cmd");
				cmd.add("/c");
				cmd.add("start");
				cmd.add("Launch SyMSpace"); // name of window
				cmd.add("cmd");
				cmd.add("/c");
			}
		}
		// java path
		config = MagOptConfig.getConfig("JavaPath");
		if ((config != null) && !config.isEmpty()) {
			path = new File(config);
			if (path.exists()) {
				java = new File(path, "bin");
				if (!java.exists()) {
					java = path;
				}
			}
		}
		if ((java == null) || !java.exists()) {
			java = new File(System.getProperty("java.home"), "bin");
		}
		cmd.add(new File(java, javaName).getAbsolutePath());

		// Java virtual machine arguments
		config = MagOptConfig.getConfig("JavaArgs");
		if (config != null) {
			vmArgs = config.split(" ");
			for (String str : vmArgs) {
				if (!str.isEmpty()) {
					cmd.add(str);
				}
			}
		}

		// classpath
		cmd.add("-cp");
		cmd.add(Launcher.getPathString(MagOptConfig.getJavaPath()));

		// classpath
		cmd.add("-Djava.library.path=" + Launcher.getPathString(MagOptConfig.getLibPath()));

		// add arguments
		cmd.addAll(args);

		// create ProcessBuilder
		pb = new ProcessBuilder(cmd);

		// evironment
		env = pb.environment();
		Launcher.addEnvironment(env);

		return pb;
	}

	/**
	 * Add Environment-itmes (home-path, system-path)
	 * 
	 * @param env
	 *            Map with environment
	 */
	public static void addEnvironment(Map<String, String> env) {
		StringBuffer path;
		String name;

		// Home-path
//		env.put("SyMSpace_Home", MagOptConfig.getHome());
		env.put("MagOpt_Home", MagOptConfig.getHome() + File.separator);

		// Path
		path = new StringBuffer();
		for (File f : MagOptConfig.getLibPath()) {
			path.append(f.getAbsolutePath());
			path.append(File.pathSeparator);
		}
		name = "Path";
		if (!env.containsKey(name)) {
			name = "PATH";
		}
		path.append(env.get(name));
		env.put(name, path.toString());

	}

	/**
	 * Get string from file-list
	 * 
	 * @param files
	 *            List of Files
	 * @return String with pathes
	 */
	public static String getPathString(File[] files) {
		StringBuffer cp;
		String str = "";

		if ((files != null) && (files.length > 0)) {
			cp = new StringBuffer();
			for (File f : files) {
				cp.append(File.pathSeparator);
				cp.append(f.getAbsolutePath());
			}
			str = cp.substring(1);
		}

		return str;
	}
}
