/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10155 $
 * $Date: 2021-04-08 16:24:22 +0200 (Thu, 08 Apr 2021) $
 */
package at.lcm.symspace.utils;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;

import at.lcm.bu21.magopt.structure.Depend;
import at.lcm.bu21.magopt.structure.DependItem;
import at.lcm.bu21.magopt.structure.Dependency;
import at.lcm.bu21.magopt.structure.Function;
import at.lcm.bu21.magopt.structure.ParamContainer;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;

/**
 * Compute best position of a component
 *
 * @author Werner Koppelsttter
 */
public class SmartComponentLocation {

	/**
	 * Score of a container
	 *
	 * @author Werner Koppelsttter
	 */
	public static class Score {

		/** Container */
		private ParamContainer cont;

		/** Resolved of unresolved */
		private int score;

		/** Deep of container in project */
		private int deep = 0;

		/**
		 * Create Score for container
		 * 
		 * @param cont  Container
		 * @param score Score-value
		 */
		public Score(ParamContainer cont, int score) {
			String id;
			int index = -1;

			this.cont = cont;
			this.score = score;

			id = cont.getIdentifier();
			while ((index = id.indexOf('.', index + 1)) != -1) {
				this.deep++;
			}
		}

		/**
		 * Get score-value
		 * 
		 * @return score-value
		 */
		public int getScore() {
			return this.score;
		}

		/**
		 * Get deep of container
		 * 
		 * @return Deep-value
		 */
		public int getDeep() {
			return this.deep;
		}

		/**
		 * Get container
		 * 
		 * @return Container
		 */
		public ParamContainer getContainer() {
			return this.cont;
		}

		@Override
		public String toString() {
			return this.score + " / " + this.deep + " : " + this.cont.getIdentifier();
		}
	}

	/**
	 * List names of unresolved imports/exports
	 * 
	 * @param comp Component
	 * @return Names of unresolved dependencies
	 */
	private static List<String> listUnresolved(Parameter comp) {
		Depend dep;
		ArrayList<String> unresolved = new ArrayList<>();

		for (Parameter p : comp.getParameters(Dependency.class)) {
			dep = ((Function) p).getDepend();
			for (DependItem d : dep.getImports()) {
				if ((d.getHandle() == null) && !d.getParam().isEmpty()) {
					unresolved.add(d.getParam());
//					System.out.println(d.getParam());
				}
			}
			for (DependItem d : dep.getExports()) {
				if ((d.getHandle() == null) && !d.getParam().isEmpty()) {
					unresolved.add(d.getParam());
//					System.out.println(d.getParam());
				}
			}
		}

		return unresolved;
	}

	/**
	 * Get container-score for unresolved dependencies
	 * 
	 * @param cont       Container to check
	 * @param unresolved List unresolved names
	 * @return Score of container
	 */
	private static Score getScore(ParamContainer cont, List<String> unresolved) {
		int score = 0;

		// count resolve unresolved
		for (String name : unresolved) {
			if (!cont.getObjectPaths(name).isEmpty()) {
				score++;
			}
		}

		return new Score(cont, score);
	}

	/**
	 * Create score-list for unresolved dependency
	 * 
	 * @param proj       Target project
	 * @param unresolved Names of unresolved dependencies
	 * @return Sorted list of score
	 */
	private static List<Score> createScoreList(Project proj, List<String> unresolved) {
		ArrayList<Score> scoreList = new ArrayList<Score>();

		// create score-list
		for (Parameter p : proj.getParameters(ParamContainer.class)) {
			scoreList.add(SmartComponentLocation.getScore((ParamContainer) p, unresolved));
		}

		// sort score-list
		scoreList.sort(new Comparator<Score>() {

			@Override
			public int compare(Score o1, Score o2) {
				int comp;

				comp = o2.getScore() - o1.getScore();
				if (comp == 0) {
					comp = o1.getDeep() - o2.getDeep();
				}

				return comp;
			}
		});

		return scoreList;
	}

	/**
	 * Get score-list for component
	 * 
	 * @param proj Target project
	 * @param comp Component Project to check
	 * @return List of scores
	 */
	public static List<Score> getScoreList(Project proj, Project comp) {
		return SmartComponentLocation.createScoreList(proj, SmartComponentLocation.listUnresolved(comp));
	}

	/**
	 * Get best container to insert
	 * 
	 * @param proj Target project
	 * @param comp Component Project to check
	 * @return List of scores
	 */
	public static ParamContainer getSmartContainer(Project proj, Project comp) {
		return SmartComponentLocation.getScoreList(proj, comp).get(0).getContainer();
	}
}
