/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.dataset;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.rmi.RemoteException;
import java.util.ArrayList;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

import at.lcm.bu21.magopt.structure.IProject;

/**
 * Displays a list of all Data Sets in Project.<br>
 * The Status of each Data Set is indicated by one of the colors Red (Error),
 * Orange (Incomplete) or Green (OK).<br>
 * Each Data Set gets an ToolTip String from the Data Set's Index Number and
 * Index Comment.<br>
 * Each Data Set can be selected, which results in changing of the current
 * Index.<br>
 * An IndexListener can be registered to get the Selection Process.
 * 
 * @author Aigner Markus
 * 
 */
public class DataSetView extends JPanel implements ActionListener {
	private static final long serialVersionUID = 1L;

	private IProject mProject;
	private JPanel mDataSet;
	private ArrayList<DataSetListener> mListener = new ArrayList<DataSetListener>();

	public void addDataSetListener(DataSetListener l) {
		this.mListener.add(l);
	}

	public void removeDataSetListener(DataSetListener l) {
		this.mListener.remove(l);
	}

	private void fireDataSetChanged(int index) {
		for (DataSetListener l : this.mListener) {
			l.fireDataSetChanged(index);
		}
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		this.fireDataSetChanged(
				Integer.parseInt(((JButton) e.getSource()).getActionCommand()));
	}

	public DataSetView() {
		GroupLayout layout = new GroupLayout(this);
		layout.setAutoCreateGaps(true);
		this.setLayout(layout);
		this.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		this.mDataSet = new JPanel();
		FlowLayout indicesLayout = new DataSetLayout(FlowLayout.LEFT, 1, 1);
		this.mDataSet.setLayout(indicesLayout);

		JScrollPane sp = new JScrollPane(this.mDataSet);

		/* legend */
		JLegendLabel lError = new JLegendLabel(IndexType.ERROR.getLabel());
		lError.setOpaque(true);
		lError.setBackground(IndexType.ERROR.getColor());
		JLegendLabel lIncomplete = new JLegendLabel(
				IndexType.INCOMPLETE.getLabel());
		lIncomplete.setOpaque(true);
		lIncomplete.setBackground(IndexType.INCOMPLETE.getColor());
		JLegendLabel lOk = new JLegendLabel(IndexType.OK.getLabel());
		lOk.setOpaque(true);
		lOk.setBackground(IndexType.OK.getColor());

		layout.setHorizontalGroup(layout.createSequentialGroup()
				.addGroup(layout.createParallelGroup().addComponent(sp)
						.addGroup(layout.createSequentialGroup()
								.addComponent(lOk).addComponent(lIncomplete)
								.addComponent(lError))));
		layout.setVerticalGroup(layout.createSequentialGroup()
				.addGroup(layout
						.createParallelGroup(GroupLayout.Alignment.BASELINE)
						.addComponent(sp))
				.addGroup(layout
						.createParallelGroup(GroupLayout.Alignment.BASELINE)
						.addComponent(lOk).addComponent(lIncomplete)
						.addComponent(lError)));
		layout.linkSize(lOk, lIncomplete, lError);
	}

	public void setProject(IProject project) {
		this.mProject = project;
		this.refresh();
	}

	public void refresh() {
		this.mDataSet.removeAll();
		try {
			int prevIndex = this.mProject.getIndex();

			JButton tmp;
			for (int i = 0; i < this.mProject.getLength(); i++) {
				this.mProject.setIndex(i);
				tmp = new JButton();
				tmp.setPreferredSize(new Dimension(15, 15));

				/* only display ds comment if it's not empty */
				if (this.mProject.getDSComment().isEmpty() == true) {
					tmp.setToolTipText("Index #" + String.valueOf(i));
				} else {
					tmp.setToolTipText("Index #" + String.valueOf(i) + " ("
							+ this.mProject.getDSComment() + ")");
				}
				tmp.setActionCommand(String.valueOf(i));
				tmp.addActionListener(this);
				if (i == prevIndex) {
					tmp.setBorderPainted(true);
				} else {
					tmp.setBorderPainted(false);
				}

				/* color the button background */
				if (this.mProject.getError() == true) {
					tmp.setBackground(IndexType.ERROR.getColor());
				} else if (this.mProject.getIncomplete() == true) {
					tmp.setBackground(IndexType.INCOMPLETE.getColor());
				} else {
					tmp.setBackground(IndexType.OK.getColor());
				}
				this.mDataSet.add(tmp);
			}
			this.mProject.setIndex(prevIndex);
		} catch (RemoteException e) {
			e.printStackTrace();
		}

		this.revalidate();
		this.repaint();
	}
}

/**
 * Default JLabel Class with following differences:<br>
 * - Opaque == true to see Background Color<br>
 * - Empty Border with Width = 1px on every side<br>
 * - Horizontal Alignment == Center
 * 
 * @author Aigner Markus
 * 
 */
class JLegendLabel extends JLabel {
	private static final long serialVersionUID = 1L;

	protected JLegendLabel() {
		super();
		this.setOpaque(true);
		this.setBorder(BorderFactory.createEmptyBorder(1, 1, 1, 1));
		this.setHorizontalAlignment(SwingConstants.CENTER);
	}

	protected JLegendLabel(String str) {
		this();
		this.setText(str);
	}
}

/**
 * Data Set states:<br>
 * - Error (red background)<br>
 * - Incomplete (orange background)<br>
 * - OK (green background)
 * 
 * @author Aigner Markus
 * 
 */
enum IndexType {
	ERROR("Error", Color.RED), INCOMPLETE("Incomplete", Color.ORANGE), OK("OK",
			Color.GREEN);

	private String mLabel;
	private Color mColor;

	private IndexType(String label, Color color) {
		this.mLabel = label;
		this.mColor = color;
	}

	protected String getLabel() {
		return (this.mLabel);
	}

	protected Color getColor() {
		return (this.mColor);
	}
}
