/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 9714 $
 * $Date: 2020-03-19 09:37:13 +0100 (Thu, 19 Mar 2020) $
 */
package magopt.gui.view;

import java.awt.Component;
import java.util.Hashtable;
import java.util.LinkedList;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;

import at.lcm.bu21.magopt.structure.IParamContainer;
import at.lcm.bu21.magopt.structure.IParamFunction;
import at.lcm.bu21.magopt.structure.IParameter;
import at.lcm.bu21.magopt.structure.IProject;
import magopt.gui.view.details.DetailView;
import magopt.gui.view.docfiles.DocFilesView;
import magopt.gui.view.port.ImportExportView;
import magopt.gui.view.report.ReportParameterView;
import magopt.gui.view.summary.SummaryView;

/**
 * Manages the MagOpt UI View Component (North-East).
 * 
 * @author Aigner Markus
 * 
 */
public class View extends JPanel {
	private static final long serialVersionUID = 1L;

	private static final String NAME_VFLAGS = "Flags";
	private static final String NAME_VDETAIL = "Detail";
	private static final String NAME_VSUMMARY = "Summary";
	private static final String NAME_VPORT = "Imports/Exports";
	private static final String NAME_VDOCFILES = "DocFiles";
	private static final String NAME_VHELP = "Help";
	private static final String NAME_VREPORT = "Report";
	private static final String NAME_VPREVIEW = "Preview";

	protected IParameter mParam;

	private JTabbedPane mJTabPane;
	private LinkedList<ParamListener> mListener = new LinkedList<ParamListener>();

	public void addParamListener(ParamListener newListener) {
		this.mListener.add(newListener);
	}

	public void removeParamListener(ParamListener listener) {
		this.mListener.remove(listener);
	}

	private void fireParamChanged() {
		for (ParamListener l : this.mListener) {
			l.parameterChanged();
		}
	}

	/**
	 * Contains every View connected with its Name.
	 */
	private Hashtable<String, ParamBaseNoScroll> mViewTable;

	public View() {
		this.mJTabPane = new JTabbedPane(SwingConstants.TOP);
		this.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		this.setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		this.add(this.mJTabPane);

		this.mViewTable = new Hashtable<String, ParamBaseNoScroll>();

		/* Parameter details */
		DetailView vDetail = new DetailView();
		vDetail.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VDETAIL, vDetail);

		/* Function preview */
		PreviewView vPreview = new PreviewView();
		vPreview.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VPREVIEW, vPreview);

		FlagsView vFlags = new FlagsView();
		this.mViewTable.put(View.NAME_VFLAGS, vFlags);

		/* Summary */
		SummaryView vSummary = new SummaryView();
		vSummary.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VSUMMARY, vSummary);

		/* Import/Export */
		ImportExportView vPort = new ImportExportView();
		vPort.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VPORT, vPort);

		/* DocFiles */
		DocFilesView vDocFiles = new DocFilesView();
		vDocFiles.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VDOCFILES, vDocFiles);

		/* Help */
		HelpView vHelp = new HelpView();
		vHelp.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VHELP, vHelp);

		/* Report */
		ReportParameterView vReport = new ReportParameterView();
		vReport.addParamListener(new ParamListener() {
			@Override
			public void parameterChanged() {
				View.this.fireParamChanged();
			}
		});
		this.mViewTable.put(View.NAME_VREPORT, vReport);
	}

	public void set(IParameter param) {
		this.mParam = param;

		Component prevComp = this.mJTabPane.getSelectedComponent();

		this.mJTabPane.removeAll();

		((FlagsView) this.mViewTable.get(View.NAME_VFLAGS)).setParameter(param);
		this.mJTabPane.addTab(View.NAME_VFLAGS, this.mViewTable.get(View.NAME_VFLAGS));
		((DetailView) this.mViewTable.get(View.NAME_VDETAIL)).setParameter(param);
		this.mJTabPane.addTab(View.NAME_VDETAIL, this.mViewTable.get(View.NAME_VDETAIL));

		if (param instanceof IParamContainer) {
			((SummaryView) this.mViewTable.get(View.NAME_VSUMMARY)).setContainer(param);
			this.mJTabPane.addTab(View.NAME_VSUMMARY, this.mViewTable.get(View.NAME_VSUMMARY));
			if (param instanceof IProject) {
				((ReportParameterView) this.mViewTable.get(View.NAME_VREPORT)).setProject(param);
				this.mJTabPane.addTab(View.NAME_VREPORT, this.mViewTable.get(View.NAME_VREPORT));
			} else if (param instanceof IParamFunction) {
				((ImportExportView) this.mViewTable.get(View.NAME_VPORT)).setFunction(param);
				this.mJTabPane.addTab(View.NAME_VPORT, this.mViewTable.get(View.NAME_VPORT));
				((HelpView) this.mViewTable.get(View.NAME_VHELP)).setFunction(param);
				this.mJTabPane.addTab(View.NAME_VHELP, this.mViewTable.get(View.NAME_VHELP));
				((PreviewView) this.mViewTable.get(View.NAME_VPREVIEW)).setFunction(param);
				this.mJTabPane.addTab(View.NAME_VPREVIEW, this.mViewTable.get(View.NAME_VPREVIEW));
			}
		}

		((DocFilesView) this.mViewTable.get(View.NAME_VDOCFILES)).setParameter(param);
		this.mJTabPane.addTab(View.NAME_VDOCFILES, this.mViewTable.get(View.NAME_VDOCFILES));

		/* try to select prev selected tab */
		if (this.mJTabPane.indexOfComponent(prevComp) != -1) {
			this.mJTabPane.setSelectedComponent(prevComp);
		}

	}

	/**
	 * Gets Parameter.
	 * 
	 * @return Parameter
	 */
	public IParameter getParameter() {
		return this.mParam;
	}

}
