/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.Serializable;
import java.util.Arrays;

import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TDataException;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TStruct;

/**
 * Individual with values of variables and results
 * 
 * @author Werner Koppelsttter
 */
public class Individual implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Archive of the individual */
	private Archive archive;

	/** Values of variables */
	private double[] variables;

	/** Values of results */
	private double[] results;

	/** Is created by mapping */
	private boolean fromMapping = false;

	/** Is individual currently recomputed */
	private boolean recompute = false;

	/**
	 * Resume individual
	 * 
	 * @param archive
	 *            Archive of savepoint
	 * @param resumeData
	 *            ResumeData
	 */
	Individual(Archive archive, TStruct resumeData) {
		this.archive = archive;
		this.setResumeData(resumeData);
	}

	/**
	 * Create Individual of the archive
	 * 
	 * @param archive
	 *            Archive
	 */
	public Individual(Archive archive) {
		this.archive = archive;

		this.variables = new double[this.archive.getVariables().length];
		for (int k = 0; k < this.variables.length; k++) {
			this.variables[k] = Double.NaN;
		}

		this.results = new double[this.archive.getResults().length];
		for (int k = 0; k < this.results.length; k++) {
			this.results[k] = Double.NaN;
		}
	}

	/**
	 * Get archive of individual
	 * 
	 * @return Archive
	 */
	public Archive getArchive() {
		return this.archive;
	}

	/**
	 * Get value of variable
	 * 
	 * @param name
	 *            Name of variable
	 * @return Value of variable
	 */
	public double getVariable(String name) {
		int index = -1;
		String[] vars;

		vars = this.archive.getVariables();
		for (int k = 0; k < vars.length; k++) {
			if (vars[k].substring(vars[k].indexOf('.')).compareTo(name.substring(name.indexOf('.'))) == 0) {
				index = k;
			}
		}

		return this.getVariable(index);
	}

	/**
	 * Get value of variable
	 * 
	 * @param index
	 *            Index of variable
	 * @return Value of variable
	 */
	public double getVariable(int index) {
		double value = Double.NaN;

		if ((0 <= index) && (index < this.variables.length)) {
			value = this.variables[index];
		}

		return value;
	}

	/**
	 * Set value of variable
	 * 
	 * @param name
	 *            Name of variable
	 * @param value
	 *            Value of variable
	 * @throws Exception
	 */
	public void setVariable(String name, double value) throws Exception {
		int index = -1;
		String[] vars;

		vars = this.archive.getVariables();
		for (int k = 0; k < vars.length; k++) {
			if (vars[k].compareTo(name) == 0) {
				index = k;
			}
		}

		this.setVariable(index, value);
	}

	/**
	 * Set value of variable
	 * 
	 * @param index
	 *            Index of variable
	 * @param value
	 *            Value of variable
	 * @throws Exception
	 */
	public void setVariable(int index, double value) throws Exception {
		int k = 0;
		boolean hasResult = false;

		while (!hasResult && (k < this.results.length)) {
			hasResult |= !Double.isNaN(this.results[k++]);
		}

		if (!hasResult) {
			if ((0 <= index) && (index < this.variables.length)) {
				this.variables[index] = value;
			}
		} else {
			throw new Exception("Unable to set variable because there is a reslut");
		}
	}

	/**
	 * Get value of result
	 * 
	 * @param name
	 *            Name of result
	 * @return Value of result
	 */
	public double getResult(String name) {
		int index = -1;
		String[] res;

		res = this.archive.getResults();
		for (int k = 0; k < res.length; k++) {
			if (res[k].substring(res[k].indexOf('.')).compareTo(name.substring(name.indexOf('.'))) == 0) {
				index = k;
			}
		}

		return this.getResult(index);
	}

	/**
	 * Get value of result
	 * 
	 * @param index
	 *            Index of result
	 * @return Value of result
	 */
	public double getResult(int index) {
		double value = Double.NaN;

		if ((0 <= index) && (index < this.results.length)) {
			value = this.results[index];
		}

		return value;
	}

	/**
	 * Set value of result
	 * 
	 * @param name
	 *            Name of result
	 * @param value
	 *            Value of result
	 */
	public void setResult(String name, double value) {
		int index = -1;
		String[] res;

		res = this.archive.getResults();
		for (int k = 0; k < res.length; k++) {
			if (res[k].compareTo(name) == 0) {
				index = k;
			}
		}

		this.setResult(index, value);
	}

	/**
	 * Set value of result
	 * 
	 * @param index
	 *            Index of result
	 * @param value
	 *            Value of result
	 */
	public void setResult(int index, double value) {
		if ((0 <= index) && (index < this.results.length)) {
			this.results[index] = value;
		}
	}

	/**
	 * Is created by mapping
	 * 
	 * @return From mapping
	 */
	public boolean isFromMapping() {
		return this.fromMapping;
	}

	/**
	 * Set created by mapping
	 * 
	 * @param fromMapping
	 *            From mapping
	 */
	public void setFromMapping(boolean fromMapping) {
		this.fromMapping = fromMapping;
	}

	/**
	 * Check if individual is valid variable range for problem
	 * 
	 * @param problem
	 *            Problem to check
	 * @return Is Valid
	 */
	public boolean isValidVar(Problem problem) {
		int k;
		boolean valid = false;

		if (problem.getArchive() == this.archive) {
			valid = true;

			// check variables
			k = 0;
			while (valid && (k < this.variables.length)) {
				valid = problem.getVariable(k).isValidValue(this.variables[k]);
				k++;
			}
		}

		return valid;
	}

	/**
	 * Check if individual is valid for problem
	 * 
	 * @param problem
	 *            Problem to check
	 * @return Is Valid
	 */
	public boolean isValid(Problem problem) {
		int k;
		boolean valid = false;

		if (problem.getArchive() == this.archive) {
			valid = true;

			// check constraints
			k = 0;
			while (valid && (k < this.results.length)) {
				valid = problem.getResult(k).isValidValue(this.results[k]);
				k++;
			}

			// check variables
			if (!this.isRecompute()) {
				k = 0;
				while (valid && (k < this.variables.length)) {
					valid = problem.getVariable(k).isValidValue(this.variables[k]);
					k++;
				}
			}
		}

		return valid;
	}

	/**
	 * Check if individual has a error
	 * 
	 * @return Is an error individual
	 */
	public boolean isError() {
		int k;
		boolean error = false;

		// check results
		k = 0;
		while (!error && (k < this.results.length)) {
			error = Double.isNaN(this.results[k++]);
		}

		// check variables
		k = 0;
		while (!error && (k < this.variables.length)) {
			error = Double.isNaN(this.variables[k++]);
		}

		return error;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (this.getClass() != obj.getClass()) {
			return false;
		}
		Individual other = (Individual) obj;
		if (this.archive != other.archive) {
			return false;
		}
		if (!Arrays.equals(this.variables, other.variables)) {
			return false;
		}

		return true;
	}

	/**
	 * Compare two individuals
	 * 
	 * @param other
	 *            Individual to check
	 * @param problem
	 *            Problem for comparing
	 * @return Comparing individuals. -1 if this dominates the other, 1 if other
	 *         dominates this or 0 if both are non-dominated
	 */
	public int dominate(Individual other, Problem problem) {
		boolean dominateThis, dominateOther;
		double valThis, valOther;
		int comp = 0;
		ResultDesc[] resDesc;

		dominateThis = false;
		dominateOther = false;
		resDesc = problem.getResults();

		if ((this.getArchive() == other.getArchive())) {
			for (int k = 0; k < resDesc.length; k++) {
				if (resDesc[k].isObjective()) {
					valThis = this.getResult(k);
					valOther = other.getResult(k);

					if (Double.isNaN(valThis) || Double.isNaN(valOther)) {
						dominateThis = true;
						dominateOther = true;
					} else if (valThis < valOther) {
						if (resDesc[k].getObjective() == ObjectiveType.MIN) {
							dominateThis = true;
						} else if (resDesc[k].getObjective() == ObjectiveType.MAX) {
							dominateOther = true;
						}
					} else if (valThis > valOther) {
						if (resDesc[k].getObjective() == ObjectiveType.MIN) {
							dominateOther = true;
						} else if (resDesc[k].getObjective() == ObjectiveType.MAX) {
							dominateThis = true;
						}
					}
					// exit loop is nothing can change
					if (dominateThis && dominateOther) {
						return 0;
					}
				}
			}
		}

		// get
		if (dominateThis != dominateOther) {
			if (dominateThis) {
				comp = -1;
			} else {
				comp = 1;
			}
		}
		return comp;
	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append("Variables:\n");
		for (int k = 0; k < this.variables.length; k++) {
			str.append("\t" + this.getArchive().getVariables()[k] + ":\t" + this.variables[k] + "\n");
		}
		str.append("Results:\n");
		for (int k = 0; k < this.results.length; k++) {
			str.append("\t" + this.getArchive().getResults()[k] + ":\t" + this.results[k] + "\n");
		}

		return str.toString();

	}

	/**
	 * Get resume data
	 * 
	 * @return Resume-data
	 * @throws TDataException
	 */
	TStruct getResumeData() throws TDataException {
		TStruct resumeData;

		resumeData = new TStruct();
		resumeData.put("var", new TNumeric(this.variables));
		resumeData.put("res", new TNumeric(this.results));
		resumeData.put("mapping", new TBoolean(this.fromMapping));

		return resumeData;

	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume-data
	 */
	private void setResumeData(TStruct resumeData) {
		TData tmp;

		tmp = resumeData.get("var");
		if (tmp != null) {
			this.variables = ((TNumeric) tmp).getRealMatrix().getRow(0);
		}

		tmp = resumeData.get("res");
		if (tmp != null) {
			this.results = ((TNumeric) tmp).getRealMatrix().getRow(0);
		}

		tmp = resumeData.get("mapping");
		if (tmp != null) {
			this.fromMapping = ((TBoolean) tmp).getBoolean();
		}
	}

	/**
	 * Get values of objectives
	 * 
	 * @param problem
	 *            Problem definition
	 * @return List of objective values
	 */
	public double[] getObjectiveValues(Problem problem) {
		double[] values;
		int index = 0;

		values = new double[problem.getNumObjectives()];
		for (int k = 0; k < problem.getNumResults(); k++) {
			if (problem.getResult(k).isObjective()) {
				values[index++] = this.results[k];
			}
		}

		return values;
	}

	/**
	 * Get values of results
	 * 
	 * @return List of result values
	 */
	public double[] getResultValues() {
		return Arrays.copyOf(this.results, this.results.length);
	}

	/**
	 * Get values of variables
	 * 
	 * @return List of variable values
	 */
	public double[] getVariableValues() {
		return Arrays.copyOf(this.variables, this.variables.length);
	}

	/**
	 * Is currently recomputed
	 * 
	 * @return Is recomputed
	 */
	public boolean isRecompute() {
		return this.recompute;
	}

	/**
	 * Set recompute
	 * 
	 * @param recompute
	 *            Recompute
	 */
	public void setRecompute(boolean recompute) {
		this.recompute = recompute;
	}
}
