/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core;

import java.io.File;
import java.io.FileOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import at.lcm.bu21.general.dtypes.TArray;
import at.lcm.bu21.general.dtypes.TBoolean;
import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.dtypes.TString;
import at.lcm.bu21.general.dtypes.TStruct;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.algorithm.Solver;
import at.lcm.bu21.optim.algorithm.SolverList;
import at.lcm.bu21.optim.core.SavePoint.SavePointType;
import at.lcm.bu21.optim.core.event.OptimizerEvent;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants;
import at.lcm.bu21.optim.core.project.PlotDescription;

/**
 * Problem description
 * 
 * @author Werner Koppelsttter
 */
public class Problem implements Serializable {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Name of problem */
	private String name;

	/** Flag if temporary savepoint exist */
	private boolean existTempSavepoint = false;

	/** List of savepoints */
	private LinkedList<SavePoint> savepoints = new LinkedList<SavePoint>();

	/** List of user-plots definition */
	private LinkedList<PlotDescription> userplots = new LinkedList<PlotDescription>();

	/** Archive of problem */
	private Archive archive;

	/** Description of variables */
	private VariableDesc[] variables;

	/** Description of results */
	private ResultDesc[] results;

	/** Serialized mapping objects */
	private String[] serializedMappings;

	/** Serialized errors classifier */
	// XXX serialize for applet
	private transient String serializedErrorsClassifier = "";

	/** Serialized sensitivity analysis objects */
	// XXX serialize for applet
	private transient List<String> serializedSensitivityAnalyses;

	/** Pool size */
	private int poolSize = 50;

	/** Function timeout */
	private int functionTimeout = 20;

	/** Autosave-interval */
	private int autosaveInterval = 100;

	/** Solver to the problem */
	// XXX serialize for applet
	private transient Solver solver = null;

	/** Init empty solver */
	private boolean initSolver = true;

	/** Reattempt individuals with error */
	private boolean reattemptError = true;

	/** Autosave-time enabled */
	private boolean autosaveTimeEnabled = false;

	/** Autosave-time */
	private int autosaveTime = 60;

	/** Optimization */
	// XXX serialize for applet
	transient Optimization optim = null;

	/**
	 * Create problem to archive
	 * 
	 * @param name
	 *            Name of problem
	 * @param archive
	 *            Based archive
	 */
	Problem(Archive archive, String name) {
		this.name = name;
		this.archive = archive;
		this.variables = new VariableDesc[this.archive.getVariables().length];
		for (int k = 0; k < this.variables.length; k++) {
			this.variables[k] = new VariableDesc(archive.getVariables()[k]);
		}
		this.results = new ResultDesc[this.archive.getResults().length];
		this.serializedMappings = new String[this.archive.getResults().length];
		for (int k = 0; k < this.results.length; k++) {
			this.results[k] = new ResultDesc(archive.getResults()[k]);
			this.serializedMappings[k] = "";
		}
	}

	/**
	 * Resume problem
	 * 
	 * @param archive
	 *            Archive of problem
	 * @param resumeData
	 *            Resume data
	 * @throws Exception
	 */
	Problem(Archive archive, TStruct resumeData) throws Exception {
		this.archive = archive;
		this.setResumeData(resumeData);
	}

	/**
	 * Add result to problem and archive
	 * 
	 * @param name
	 *            Name of new result
	 * @return Result-description
	 */
	public ResultDesc addResult(String name) {
		boolean found = false;
		String[] resNames;
		Parameter[] params;
		int index, k;
		ResultDesc result = null;

		if (!this.getArchive().isLocked()) {
			resNames = this.getArchive().getResults();
			for (String res : resNames) {
				found |= (res.compareTo(name) == 0);
			}

			if (!found) {
				resNames = Arrays.copyOf(resNames, resNames.length + 1);
				this.results = Arrays.copyOf(this.results, this.results.length + 1);
				this.serializedMappings = Arrays.copyOf(this.serializedMappings, this.serializedMappings.length + 1);

				// find index to insert
				params = this.archive.getProject().getParameters(ParamField.class);
				index = 0;
				k = 0;
				while ((k < params.length) && (params[k].getIdentifier().compareTo(name) != 0)) {
					if ((index < (resNames.length - 1))
							&& (params[k].getIdentifier().compareTo(resNames[index]) == 0)) {
						index++;
					}
					k++;
				}

				// insert new result
				k = resNames.length - 2;
				while (k >= index) {
					resNames[k + 1] = resNames[k];
					this.results[k + 1] = this.results[k];
					this.serializedMappings[k + 1] = this.serializedMappings[k];
					k--;
				}
				resNames[index] = name;
				result = new ResultDesc(name);
				this.results[index] = result;
				this.serializedMappings[index] = "";

				this.getArchive().setResults(resNames);
			}
		}

		return result;
	}

	/**
	 * Add variable to problem and archive
	 * 
	 * @param name
	 *            Name of new variable
	 * @return Variable-description
	 */
	public VariableDesc addVariable(String name) {
		boolean found = false;
		String[] varNames;
		Parameter[] params;
		int index, k;
		VariableDesc variable = null;

		if (!this.getArchive().isLocked()) {
			varNames = this.getArchive().getVariables();
			for (String var : varNames) {
				found |= (var.compareTo(name) == 0);
			}

			if (!found) {
				varNames = Arrays.copyOf(varNames, varNames.length + 1);
				this.variables = Arrays.copyOf(this.variables, this.variables.length + 1);

				// find index to insert
				params = this.archive.getProject().getParameters(ParamField.class);
				index = 0;
				k = 0;
				while ((k < params.length) && (params[k].getIdentifier().compareTo(name) != 0)) {
					if ((index < (varNames.length - 1))
							&& (params[k].getIdentifier().compareTo(varNames[index]) == 0)) {
						index++;
					}
					k++;
				}

				// insert new variable
				k = varNames.length - 2;
				while (k >= index) {
					varNames[k + 1] = varNames[k];
					this.variables[k + 1] = this.variables[k];
					k--;
				}
				varNames[index] = name;
				variable = new VariableDesc(name);
				this.variables[index] = variable;

				this.getArchive().setVariables(varNames);
			}
		}

		return variable;
	}

	/**
	 * Clear solutions
	 * 
	 * @throws Exception
	 */
	public void clearSolution() throws Exception {
		this.savepoints.clear();
		if (this.solver != null) {
			this.solver.setArchive(new Individual[0]);
		}
	}

	/**
	 * Clear mapping objects
	 */
	public void clearMapping() {
		for (int k = 0; k < this.serializedMappings.length; k++) {
			this.serializedMappings[k] = "";
		}
	}

	/**
	 * Clear error classiffier object
	 */
	public void clearSerializedErrorsClassiffier() {
		this.serializedErrorsClassifier = "";
	}

	/**
	 * Clear sensitivity analysis objects
	 */
	public void clearSensitivityAnalyses() {
		if (this.serializedSensitivityAnalyses != null) {
			for (int k = 0; k < this.serializedSensitivityAnalyses.size(); k++) {
				this.serializedSensitivityAnalyses.set(k, "");
			}
		}
	}

	/**
	 * Create init distribution of individual with latin hypercube-sampling
	 * 
	 * @param num
	 *            Number of individuals
	 * 
	 * @return new individuals
	 */
	public ArrayList<Individual> createInitDistribution(int num) {
		ArrayList<Individual> list;
		ArrayList<LinkedList<Double>> dist;
		LinkedList<Double> values;
		Individual ind;
		VariableDesc varDesc;
		double val;

		// create distribution of variables
		dist = new ArrayList<LinkedList<Double>>(this.getNumVariables());
		for (int k = 0; k < this.getNumVariables(); k++) {
			varDesc = this.getVariable(k);
			values = new LinkedList<Double>();
			for (int n = 0; n < num; n++) {
				if (varDesc.isVaried() && (num > 1)) {
					val = varDesc.round(varDesc.getLB() + (((varDesc.getUB() - varDesc.getLB()) * n) / (num - 1)));
				} else {
					val = varDesc.getLB();
				}
				values.add(val);
			}
			dist.add(values);
		}

		// combine values from distribution
		list = new ArrayList<Individual>(num);
		while (list.size() < num) {
			try {
				ind = new Individual(this.getArchive());
				for (int k = 0; k < dist.size(); k++) {
					values = dist.get(k);
					ind.setVariable(k, values.remove((int) (Math.random() * values.size())));
				}
				list.add(ind);
			} catch (Exception ex) {
				System.out.println("Error by creation individium");
			}
		}

		return list;
	}

	/**
	 * Create init distribution of individuals for the binary problem
	 * 
	 * @param num
	 *            Number of individuals
	 * 
	 * @return new individuals
	 */
	public ArrayList<Individual> createBinaryInitDistribution(int num) {
		ArrayList<Individual> list;
		Individual ind;
		// combine values from distribution
		list = new ArrayList<Individual>(num);
		int i = -1;
		Random rndGen = new Random(System.currentTimeMillis());
		while (list.size() < num) {
			try {
				ind = new Individual(this.getArchive());
				i++;
				int type = i % 3;
				for (int k = 0; k < this.getNumVariables(); k++) {
					switch (type) {
					case 0:
						ind.setVariable(k, 0.0);
						break;
					case 1:
						ind.setVariable(k, 1.0);
						break;
					default:
						ind.setVariable(k, rndGen.nextBoolean() ? 1.0 : 0.0);
						break;
					}
				}
				list.add(ind);
			} catch (Exception ex) {
				System.out.println("Error by creation individium");
			}
		}

		return list;
	}

	/**
	 * Create random individual
	 * 
	 * @return new individuals
	 */
	public Individual createRandom() {
		Individual ind = null;
		VariableDesc var;
		double val;

		try {
			ind = new Individual(this.getArchive());
			for (int k = 0; k < this.getNumVariables(); k++) {
				var = this.variables[k];
				if (var.isVaried()) {
					val = var.round(var.getLB() + ((var.getUB() - var.getLB()) * Math.random()));
				} else {
					val = var.getLB();
				}
				ind.setVariable(k, val);
			}
		} catch (Exception ex) {
			System.out.println("Error by creation individium");
		}

		return ind;
	}

	/**
	 * Create save-point (calculate paretofront)
	 * 
	 * @throws Exception
	 */
	public void createSavePoint() throws Exception {
		this.createSavePoint(SavePointType.USER, false, 0, 0, 0, true);
	}

	/**
	 * Create save-point (calculate paretofront)
	 * 
	 * @param type
	 *            Type of savepoint
	 * @param createTempSavepoint
	 *            Create a temporary savepoint
	 * @param nValid
	 *            Number of valid individuals
	 * @param nInvalid
	 *            Number of invalid individuals
	 * @param nError
	 *            Number of individuals with errors
	 * @throws Exception
	 */
	public void createSavePoint(SavePointType type, boolean createTempSavepoint, int nValid, int nInvalid, int nError)
			throws Exception {
		this.createSavePoint(type, createTempSavepoint, nValid, nInvalid, nError, true);
	}

	/**
	 * Create save-point (calculate paretofront)
	 * 
	 * @param type
	 *            Type of savepoint
	 * @param createTempSavepoint
	 *            Create a temporary savepoint
	 * @param nValid
	 *            Number of valid individuals
	 * @param nInvalid
	 *            Number of invalid individuals
	 * @param nError
	 *            Number of individuals with errors
	 * @param calcPareto
	 *            Calculate pareto or user all individuals
	 * @throws Exception
	 */
	public void createSavePoint(SavePointType type, boolean createTempSavepoint, int nValid, int nInvalid, int nError,
			boolean calcPareto) throws Exception {
		this.addSavePoint(new SavePoint(type, this, nValid, nInvalid, nError, calcPareto), createTempSavepoint);
	}

	/**
	 * Add save-point
	 * 
	 * @throws Exception
	 */
	public void addSavePoint(SavePoint sp, boolean tempSavepoint) throws Exception {

		if (this.existTempSavepoint) {
			this.savepoints.set(this.savepoints.size() - 1, sp);
		} else {
			this.savepoints.add(sp);
		}
		this.existTempSavepoint = tempSavepoint;
	}

	/**
	 * Duplicate problem
	 * 
	 * @param name
	 *            Name of new problem
	 * @return New problem
	 * @throws Exception
	 */
	public Problem duplicate(String name) throws Exception {
		Problem prob = null;
		TStruct resumeData;

		if (!Arrays.asList(this.getArchive().listProblems()).contains(name)) {
			resumeData = this.getResumeData().deepClone();
			resumeData.put("name", new TString(name));
			prob = this.getArchive().createProblem(name);
			prob.setResumeData(resumeData);

			// clear solutions
			prob.clearSolution();

			this.getArchive().writeProblem(prob);
		}

		return prob;
	}

	/**
	 * Check if temporary savepoint exists
	 * 
	 * @return Exist temporary savepoint
	 */
	public boolean existTempSavePoint() {
		return this.existTempSavepoint;
	}

	/**
	 * Get archive of problem
	 * 
	 * @return Archive
	 */
	public Archive getArchive() {
		return this.archive;
	}

	/**
	 * Get autosave-interval
	 * 
	 * @return Autosave-interval
	 */
	public int getAutosaveInterval() {
		return this.autosaveInterval;
	}

	/**
	 * Get autosave-time
	 * 
	 * @return Autosave-time
	 */
	public int getAutosaveTime() {
		return this.autosaveTime;
	}

	/**
	 * Get autosave-time is enabled
	 * 
	 * @return Autosave-time enabled
	 */
	public boolean getAutosaveTimeEnabled() {
		return this.autosaveTimeEnabled;
	}

	/**
	 * Get defined constraints
	 * 
	 * @return Constraints
	 */
	public ResultDesc[] getConstraints() {
		return this.getConstraintsList().toArray(new ResultDesc[0]);
	}

	/**
	 * Get defined constraints
	 * 
	 * @return Constraints
	 */
	private LinkedList<ResultDesc> getConstraintsList() {
		LinkedList<ResultDesc> constaints;

		constaints = new LinkedList<ResultDesc>();
		for (ResultDesc res : this.results) {
			if (res.isConstraint()) {
				constaints.add(res);
			}
		}

		return constaints;
	}

	/**
	 * Get function-timeout
	 * 
	 * @return Function-timeout
	 */
	public int getFunctionTimeout() {
		return this.functionTimeout;
	}

	/**
	 * Get initialize empty solver
	 * 
	 * @return Init solver
	 */
	public boolean getInitSolver() {
		return this.initSolver;
	}

	/**
	 * Get name of problem
	 * 
	 * @return Name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Get number of defined constraints
	 * 
	 * @return Number of constraints
	 */
	public int getNumConstraints() {
		return this.getConstraintsList().size();
	}

	/**
	 * Get number of defined objectives
	 * 
	 * @return Number of objectives
	 */
	public int getNumObjectives() {
		return this.getObjectivesList().size();
	}

	/**
	 * Get number of defined results
	 * 
	 * @return Number of results
	 */
	public int getNumResults() {
		return this.results.length;
	}

	/**
	 * Get number of existing savepoints
	 * 
	 * @return Number of savepoints
	 */
	public int getNumSavePoints() {
		return this.savepoints.size();
	}

	/**
	 * Get number of real existing savepoints (without temporary savepoint)
	 * 
	 * @return Number of savepoints
	 */
	public int getNumRealSavePoints() {
		int sp;

		sp = this.savepoints.size();
		if (this.existTempSavePoint()) {
			sp--;
		}

		return sp;
	}

	/**
	 * Get number of defined variables
	 * 
	 * @return Number of variables
	 */
	public int getNumVariables() {
		return this.variables.length;
	}

	/**
	 * Get number of defined varied variables
	 * 
	 * @return Number of Variables
	 */
	public int getNumVariedVariables() {
		return this.getVariedVariablesList().size();
	}

	/**
	 * Get defined objectives
	 * 
	 * @return Objectives
	 */
	public ResultDesc[] getObjectives() {
		return this.getObjectivesList().toArray(new ResultDesc[0]);
	}

	/**
	 * Get defined objectives
	 * 
	 * @return Objectives
	 */
	private LinkedList<ResultDesc> getObjectivesList() {
		LinkedList<ResultDesc> objectives;

		objectives = new LinkedList<ResultDesc>();
		for (ResultDesc res : this.results) {
			if (res.isObjective()) {
				objectives.add(res);
			}
		}

		return objectives;
	}

	/**
	 * Get pool-size
	 * 
	 * @return Pool-size
	 */
	public int getPoolSize() {
		return this.poolSize;
	}

	/**
	 * Get reattempt individuals with error
	 * 
	 * @return Reattempt on error
	 */
	public boolean getReattemptError() {
		return this.reattemptError;
	}

	/**
	 * Get result description
	 * 
	 * @param index
	 *            Index of result
	 * @return Result-description
	 */
	ResultDesc getResult(int index) {
		ResultDesc result = null;

		if ((0 <= index) && (index < this.results.length)) {
			result = this.results[index];
		}

		return result;
	}

	/**
	 * Get result description
	 * 
	 * @param name
	 *            Name of result
	 * @return Result-description
	 */
	public ResultDesc getResult(String name) {
		ResultDesc result = null;

		for (int k = 0; k < this.results.length; k++) {
			if (this.results[k].getName().compareTo(name) == 0) {
				result = this.results[k];
			}
		}

		return result;
	}

	/**
	 * Get defined results
	 * 
	 * @return Results
	 */
	public ResultDesc[] getResults() {
		return this.results;
	}

	/**
	 * Get resume data
	 * 
	 * @return Resume data
	 * @throws Exception
	 */
	TStruct getResumeData() throws Exception {
		TStruct resumeData;
		TArray array;
		int end;

		resumeData = this.getSettings();

		// name
		resumeData.put("name", new TString(this.name));

		// mapping objects
		if (this.results.length != 0) {
			array = new TArray(1, this.results.length);
			resumeData.put("mappings", array);
			for (int k = 0; k < this.results.length; k++) {
				array.set(0, k, new TString(this.serializedMappings[k]));
			}
		}

		// serialized errors classifier
		resumeData.put("errorsClassifier", new TString(this.serializedErrorsClassifier));

		// serialized sensitivity analysis objects
		int notEmptyObjs = 0;
		if (this.serializedSensitivityAnalyses != null) {
			for (String saObj : this.serializedSensitivityAnalyses) {
				if (!saObj.equals("")) {
					notEmptyObjs++;
				}
			}
			array = new TArray(1, notEmptyObjs);
			resumeData.put("sensitivityAnalyses", array);
			int kSA = 0;
			for (String saObj : this.serializedSensitivityAnalyses) {
				if (!saObj.equals("")) {
					array.set(0, kSA, new TString(saObj));
					kSA++;
				}
			}
		}

		// solver
		if (this.solver != null) {
			resumeData.put("solverState", this.solver.getState());
		}

		// full savepoints
		if (!this.savepoints.isEmpty()) {
			end = this.savepoints.size();
			if (this.existTempSavepoint) {
				end--;
			}
			array = new TArray(1, end);
			resumeData.put("savepoints", array);
			for (int k = 0; k < end; k++) {
				array.set(0, k, this.savepoints.get(k).getResumeData());
			}
		}

		return resumeData;
	}

	/**
	 * Get savepoint
	 * 
	 * @param index
	 *            Index of savepoint
	 * @return Savepoint
	 */
	public SavePoint getSavePoint(int index) {
		return this.savepoints.get(index);
	}

	/**
	 * Get the serialized mappings
	 * 
	 * @return Serialized mappings
	 */
	public String[] getSerializedMappings() {
		return this.serializedMappings;
	}

	/**
	 * Get the serialized mapping classifier
	 * 
	 * @return Serialized mapping classifier
	 */
	public String getSerializedErrorsClassifier() {
		return this.serializedErrorsClassifier;
	}

	/**
	 * Get the serialized sensitivity analyses
	 * 
	 * @return Serialized sensitivity analyses
	 */
	public List<String> getSerializedSensitivityAnalyses() {
		return this.serializedSensitivityAnalyses;
	}

	/**
	 * Get solver
	 * 
	 * @return Solver
	 */
	public Solver getSolver() {
		if (this.solver == null) {
			this.solver = SolverList.getValues()[0].getSolver(this);
		}
		return this.solver;
	}

	/**
	 * Get variable description
	 * 
	 * @param index
	 *            Index of variable
	 * @return Variable-description
	 */
	VariableDesc getVariable(int index) {
		VariableDesc variable = null;

		if ((0 <= index) && (index < this.variables.length)) {
			variable = this.variables[index];
		}

		return variable;
	}

	/**
	 * Get variable description
	 * 
	 * @param name
	 *            Name of variable
	 * @return Variable-description
	 */
	public VariableDesc getVariable(String name) {
		VariableDesc variable = null;

		for (int k = 0; k < this.variables.length; k++) {
			if (this.variables[k].getName().compareTo(name) == 0) {
				variable = this.variables[k];
			}
		}

		return variable;
	}

	/**
	 * Get defined variables
	 * 
	 * @return Variables
	 */
	public VariableDesc[] getVariables() {
		return this.variables;
	}

	/**
	 * Get defined varied variables
	 * 
	 * @return Variables
	 */
	public VariableDesc[] getVariedVariables() {
		return this.getVariedVariablesList().toArray(new VariableDesc[0]);
	}

	/**
	 * Get defined varied variables
	 * 
	 * @return Variables
	 */
	private LinkedList<VariableDesc> getVariedVariablesList() {
		LinkedList<VariableDesc> vars;

		vars = new LinkedList<VariableDesc>();
		for (VariableDesc var : this.variables) {
			if (var.isVaried()) {
				vars.add(var);
			}
		}

		return vars;
	}

	/**
	 * Check if problem is locked (no changes in settings)
	 * 
	 * @return Is problem locked
	 */
	public boolean isLocked() {
		return !this.savepoints.isEmpty();
	}

	/**
	 * Rename Problem
	 * 
	 * @param name
	 *            New name
	 */
	public void rename(String name) {
		if (!Arrays.asList(this.getArchive().listProblems()).contains(name)) {
			this.getArchive().renameProblem(this.name, name);
			this.name = name;
		}
	}

	/**
	 * Remove result from problem and archive
	 * 
	 * @param name
	 *            Name of result
	 */
	public void rmResult(String name) {
		boolean found = false;
		String[] resNames;

		if (!this.getArchive().isLocked()) {
			resNames = this.getArchive().getResults();
			for (int k = 0; k < resNames.length; k++) {
				if (found) {
					resNames[k - 1] = resNames[k];
					this.results[k - 1] = this.results[k];
					this.serializedMappings[k - 1] = this.serializedMappings[k];
				} else {
					found = (resNames[k].compareTo(name) == 0);
				}
			}

			if (found) {
				resNames = Arrays.copyOf(resNames, resNames.length - 1);
				this.results = Arrays.copyOf(this.results, this.results.length - 1);
				this.serializedMappings = Arrays.copyOf(this.serializedMappings, this.serializedMappings.length - 1);

				this.getArchive().setResults(resNames);
			}

		}
	}

	/**
	 * Remove variable from problem and archive
	 * 
	 * @param name
	 *            Name of variable
	 */
	public void rmVariable(String name) {
		boolean found = false;
		String[] varNames;

		if (!this.getArchive().isLocked()) {
			varNames = this.getArchive().getVariables();
			for (int k = 0; k < varNames.length; k++) {
				if (found) {
					varNames[k - 1] = varNames[k];
					this.variables[k - 1] = this.variables[k];
				} else {
					found = (varNames[k].compareTo(name) == 0);
				}
			}

			if (found) {
				varNames = Arrays.copyOf(varNames, varNames.length - 1);
				this.variables = Arrays.copyOf(this.variables, this.variables.length - 1);

				this.getArchive().setVariables(varNames);
			}

		}
	}

	/**
	 * Set autosave-interval
	 * 
	 * @param autosave
	 *            Autosave-interval
	 */
	public void setAutosaveInterval(int autosave) {
		this.autosaveInterval = autosave;
	}

	/**
	 * Set autosave-time
	 * 
	 * @param autosave
	 *            Autosave-time
	 */
	public void setAutosaveTime(int autosave) {
		this.autosaveTime = autosave;
	}

	/**
	 * Get autosave-time enabled
	 * 
	 * @param enabled
	 *            Autsave-time enabled
	 */
	public void setAutosaveTimeEnabled(boolean enabled) {
		this.autosaveTimeEnabled = enabled;
	}

	/**
	 * Set function-timeout
	 * 
	 * @param timeout
	 *            Function-timeout
	 */
	public void setFunctionTimeout(int timeout) {
		this.functionTimeout = timeout;
	}

	/**
	 * Set initialize empty solver
	 * 
	 * @param initSolver
	 *            Init solver
	 */
	public void setInitSolver(boolean initSolver) {
		this.initSolver = initSolver;
	}

	/**
	 * Set pool-size
	 * 
	 * @param poolSize
	 *            Pool-size
	 */
	public void setPoolSize(int poolSize) {
		this.poolSize = poolSize;
	}

	/**
	 * Set reattempt individuals with error
	 * 
	 * @param reattempt
	 *            Reattempt on error
	 */
	public void setReattemptError(boolean reattempt) {
		this.reattemptError = reattempt;
	}

	/**
	 * Set resume data
	 * 
	 * @param resumeData
	 *            Resume Data
	 * @throws Exception
	 */
	private void setResumeData(TStruct resumeData) throws Exception {
		TArray array;
		TString str;

		// name
		this.name = ((TString) resumeData.get("name")).getString();

		// variable description
		array = (TArray) resumeData.get("varDesc");
		if (array == null) {
			this.variables = new VariableDesc[0];
		} else {
			this.variables = new VariableDesc[array.size()[1]];
		}
		for (int k = 0; k < this.variables.length; k++) {
			this.variables[k] = new VariableDesc((TStruct) array.get(0, k));
		}

		// result description
		array = (TArray) resumeData.get("resDesc");
		if (array == null) {
			this.results = new ResultDesc[0];
		} else {
			this.results = new ResultDesc[array.size()[1]];
		}
		for (int k = 0; k < this.results.length; k++) {
			this.results[k] = new ResultDesc((TStruct) array.get(0, k));
		}

		this.setSettings(resumeData);

		// mapping objects
		array = (TArray) resumeData.get("mappings");
		this.serializedMappings = new String[this.results.length];
		for (int k = 0; k < this.results.length; k++) {
			this.serializedMappings[k] = "";
		}
		if (array != null) {
			for (int k = 0; k < this.results.length; k++) {
				this.serializedMappings[k] = ((TString) array.get(0, k)).getString();
			}
		}

		// serialized errors classifier
		str = ((TString) resumeData.get("errorsClassifier"));
		if (str != null) {
			this.serializedErrorsClassifier = str.getString();
		}

		// sensitivity analysis objects
		array = (TArray) resumeData.get("sensitivityAnalyses");
		if (array != null) {
			this.serializedSensitivityAnalyses = new ArrayList<String>();
			for (int k = 0; k < array.size()[1]; k++) {
				this.serializedSensitivityAnalyses.add(((TString) array.get(0, k)).getString());
			}
		}

		// optimizer settings
		str = (TString) resumeData.get("solver");
		if (str != null) {
			try {
				this.solver.setState((TStruct) resumeData.get("solverState"));
			} catch (Exception ex) {
			}
		}

		// savePoints;
		this.savepoints.clear();
		array = (TArray) resumeData.get("savepoints");
		if (array != null) {
			for (int k = 0; k < array.size()[1]; k++) {
				this.savepoints.add(new SavePoint(this, (TStruct) array.get(0, k)));
			}
		}

		// user-plots
		array = (TArray) resumeData.get("userplots");
		if (array != null) {
			for (int k = 0; k < array.size()[1]; k++) {
				this.userplots.add(new PlotDescription(this, (TStruct) array.get(0, k)));
			}
		}
	}

	/**
	 * Set serialized mappings
	 * 
	 * @param targetName
	 *            Name of target
	 * @param newValue
	 *            Serialized object
	 * @return True
	 */
	public boolean setSerializedMapping(String targetName, String newValue) {
		boolean outcome = true;

		int index = -1;
		for (int i = 0; i < this.results.length; i++) {
			if (this.results[i].getName().equals(targetName)) {
				index = i;
				break;
			}
		}
		if (index != -1) {
			this.serializedMappings[index] = newValue;
		}

		return outcome;
	}

	/**
	 * Add a serialized sensitivity analysis object
	 * 
	 * @param newValue
	 *            Serialized object
	 * @return True if operation was successful, false otherwise
	 */
	public boolean addSerializedSensitivityAnalyses(String newValue) {
		boolean outcome = true;

		if (this.serializedSensitivityAnalyses != null) {
			if (this.serializedSensitivityAnalyses.size() == TMappingConstants.MAX_STORED_SENSITIVITY_ANALYSES) {
				this.serializedSensitivityAnalyses.remove(0);
			}

			this.serializedSensitivityAnalyses.add(newValue);
		} else {
			this.serializedSensitivityAnalyses = new ArrayList<String>();
			this.serializedSensitivityAnalyses.add(newValue);
		}

		return outcome;
	}

	/**
	 * Update/replace an existing serialized sensitivity analysis object
	 * 
	 * @param newValue
	 *            Serialized object
	 * @param index
	 *            Index
	 * @return True if operation was successful, false otherwise
	 */
	public boolean updateSerializedSensitivityAnalyses(String newValue, int index) {
		boolean outcome = true;

		if (this.serializedSensitivityAnalyses != null) {
			this.serializedSensitivityAnalyses.set(index, newValue);
		}

		return outcome;
	}

	/**
	 * Set serialized errors-classifier
	 * 
	 * @param serializedErrorsClassifier
	 *            Serialized object
	 */
	public void setSerializedErrorsClassifier(String serializedErrorsClassifier) {
		this.serializedErrorsClassifier = serializedErrorsClassifier;
	}

	/**
	 * Set solver
	 * 
	 * @param solver
	 *            Solver
	 * @throws Exception
	 */
	public void setSolver(Solver solver) throws Exception {
		if ((solver == null) || (this == solver.getProblem())) {
			this.solver = solver;
		} else {
			throw new Exception("Solver for wrong problem");
		}
	}

	/**
	 * List names of user-plots
	 * 
	 * @return List of names
	 */
	public String[] listUserPlots() {
		String[] plots;

		plots = new String[this.userplots.size()];

		for (int k = 0; k < plots.length; k++) {
			plots[k] = this.userplots.get(k).getName();
		}

		return plots;
	}

	/**
	 * Add user-plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void addUserPlot(String name) {
		boolean exists = false;

		for (PlotDescription p : this.userplots) {
			exists |= p.getName().compareTo(name) == 0;
		}

		if (!exists) {
			this.userplots.add(new PlotDescription(this, name));
			if (this.getOptimization() != null) {
				this.getOptimization().fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Add: " + name));
			}
		}
	}

	/**
	 * Remove user-plot
	 * 
	 * @param name
	 *            Name of plot
	 */
	public void rmUserPlot(String name) {
		for (int k = this.userplots.size() - 1; k >= 0; k--) {
			if (this.userplots.get(k).getName().compareTo(name) == 0) {
				this.userplots.remove(k);
				this.getOptimization()
						.fireOptmizerPerformed(new OptimizerEvent(OptimizerEvent.PLOT, "Remove: " + name));
			}
		}
	}

	/**
	 * Get user-plot description
	 * 
	 * @param name
	 *            Name of plot
	 * @return Plot description
	 */
	public PlotDescription getUserPlot(String name) {
		PlotDescription plot = null;
		for (PlotDescription p : this.userplots) {
			if (p.getName().compareTo(name) == 0) {
				plot = p;
			}
		}

		return plot;
	}

	/**
	 * Get settings from problem
	 * 
	 * @return Settings
	 * @throws Exception
	 */
	private TStruct getSettings() throws Exception {
		TStruct settings;
		TArray array;

		settings = new TStruct();

		// variable description
		if (this.variables.length != 0) {
			array = new TArray(1, this.variables.length);
			settings.put("varDesc", array);
			for (int k = 0; k < this.variables.length; k++) {
				array.set(0, k, this.variables[k].getResumeData());
			}
		}

		// result description
		if (this.results.length != 0) {
			array = new TArray(1, this.results.length);
			settings.put("resDesc", array);
			for (int k = 0; k < this.results.length; k++) {
				array.set(0, k, this.results[k].getResumeData());
			}
		}

		// poolsize
		settings.put("poolsize", new TNumeric(this.getPoolSize()));

		// functionTimeout
		settings.put("functionTimeout", new TNumeric(this.getFunctionTimeout()));

		// autosaveInterval
		settings.put("autosaveInterval", new TNumeric(this.getAutosaveInterval()));

		// solver
		if (this.solver != null) {
			settings.put("solver", new TString(SolverList.getSolverListEntry(this.solver.getOptimizerName()).name()));
			settings.put("solverSettings", this.solver.getSettings());
		}

		// init empty solver
		settings.put("initSolver", new TBoolean(this.getInitSolver()));

		// reattempt on error
		settings.put("reattemptError", new TBoolean(this.getReattemptError()));

		// autosave time enabled
		settings.put("autosaveTimeEnabled", new TBoolean(this.getAutosaveTimeEnabled()));

		// autosave time
		settings.put("autosaveTime", new TNumeric(this.getAutosaveTime()));

		// user-plots
		if (!this.userplots.isEmpty()) {
			array = new TArray(1, this.userplots.size());
			settings.put("userplots", array);
			for (int k = 0; k < this.userplots.size(); k++) {
				array.set(0, k, this.userplots.get(k).getResumeData());
			}
		}

		return settings;
	}

	/**
	 * Set settings of problem
	 * 
	 * @param settings
	 *            Settings data
	 * @throws Exception
	 */
	private void setSettings(TStruct settings) throws Exception {
		TNumeric num;
		TBoolean bool;
		TString str;

		// poolsize
		num = (TNumeric) settings.get("poolsize");
		if (num != null) {
			this.setPoolSize((int) num.getReal(0, 0));
		}

		// functionTimeout
		num = (TNumeric) settings.get("functionTimeout");
		if (num != null) {
			this.setFunctionTimeout((int) num.getReal(0, 0));
		}

		// autosaveInterval
		num = (TNumeric) settings.get("autosaveInterval");
		if (num != null) {
			this.setAutosaveInterval((int) num.getReal(0, 0));
		}

		// optimizer settings
		str = (TString) settings.get("solver");
		if (str != null) {
			try {
				this.solver = SolverList.valueOf(str.getString()).getSolver(this);
				this.solver.setSettings((TStruct) settings.get("solverSettings"));
			} catch (Exception ex) {
			}
		}

		// init empty solver
		bool = (TBoolean) settings.get("initSolver");
		if (bool != null) {
			this.setInitSolver(bool.getBoolean());
		}

		// reattempt on error
		bool = (TBoolean) settings.get("reattemptError");
		if (bool != null) {
			this.setReattemptError(bool.getBoolean());
		}

		// autosave time enabled
		bool = (TBoolean) settings.get("autosaveTimeEnabled");
		if (bool != null) {
			this.setAutosaveTimeEnabled(bool.getBoolean());
		}

		// autosave time
		num = (TNumeric) settings.get("autosaveTime");
		if (num != null) {
			this.setAutosaveTime((int) num.getReal(0, 0));
		}
	}

	/**
	 * Export problem settings
	 * 
	 * @param xml
	 *            XML-file to save
	 */
	public void exportSettings(File xml) {
		FileOutputStream out;

		if (!xml.getName().toLowerCase().endsWith(".xml")) {
			xml = new File(xml.getAbsolutePath() + ".xml");
		}

		try {
			out = new FileOutputStream(xml);
			this.getSettings().write(out);
			out.close();
		} catch (Exception ex) {
		}
	}

	/**
	 * Import settings
	 * 
	 * @param settings
	 *            Settings
	 * @return Warning message
	 * @throws Exception
	 */
	String importSettings(TStruct settings) throws Exception {
		int index;
		String ident;
		TData data;
		TArray array, axes;
		TStruct struct, plot;
		VariableDesc varDesc;
		ResultDesc resDesc;
		Parameter[] params;
		PlotDescription plotDesc;
		StringBuffer message;
		Project project;

		project = this.archive.getProject();
		message = new StringBuffer("");

		data = settings.get("varDesc");
		if (data != null) {
			array = (TArray) data;
			for (int k = 0; k < array.size()[1]; k++) {
				struct = (TStruct) array.get(0, k);

				// get name (identifier)
				ident = ((TString) struct.get("name")).getString();
				index = ident.indexOf('.');
				ident = project.getName() + ident.substring(index);

				// check for input parameter
				params = project.getDynamicParams();
				for (Parameter p : params) {
					if (p.getIdentifier().compareTo(ident) == 0) {
						this.addVariable(ident);
					}
				}

				// set values
				varDesc = this.getVariable(ident);
				if (varDesc != null) {
					varDesc.setBounds(((TNumeric) struct.get("lb")).getReal(0, 0),
							((TNumeric) struct.get("steps")).getReal(0, 0),
							((TNumeric) struct.get("ub")).getReal(0, 0));
				} else {
					message.append("\nUnable to add variable: " + ident);
				}
			}
		}

		data = settings.get("resDesc");
		if (data != null) {
			array = (TArray) data;
			for (int k = 0; k < array.size()[1]; k++) {
				struct = (TStruct) array.get(0, k);

				// get name (identifier)
				ident = ((TString) struct.get("name")).getString();
				index = ident.indexOf('.');
				ident = project.getName() + ident.substring(index);

				// check for output parameter
				params = project.getDynamicOutputs();
				for (Parameter p : params) {
					if (p.getIdentifier().compareTo(ident) == 0) {
						this.addResult(ident);
					}
				}

				// set values
				resDesc = this.getResult(ident);
				if (resDesc != null) {
					resDesc.setBounds(((TNumeric) struct.get("lb")).getReal(0, 0),
							((TNumeric) struct.get("ub")).getReal(0, 0));
					resDesc.setObjective(ObjectiveType.valueOf(((TString) struct.get("obj")).getString()));
				} else {
					message.append("\nUnable to add result: " + ident);
				}
			}
		}

		this.setSettings(settings);

		data = settings.get("userplots");
		if (data != null) {
			array = (TArray) data;
			for (int k = 0; k < array.size()[1]; k++) {
				// create plot
				plot = (TStruct) array.get(0, k);
				ident = ((TString) plot.get("name")).getString();
				axes = (TArray) plot.get("axes");
				this.addUserPlot(ident);
				plotDesc = this.getUserPlot(ident);

				// resize plot description
				while (plotDesc.getNumRows() < axes.size()[0]) {
					plotDesc.addRow(0);
				}
				while (plotDesc.getNumColumns() < axes.size()[1]) {
					plotDesc.addColumn(0);
				}

				for (int r = 0; r < axes.size()[0]; r++) {
					for (int c = 0; c < axes.size()[0]; c++) {
						struct = (TStruct) axes.get(r, c);
						// x-axis
						ident = ((TString) struct.get("x")).getString();
						if (!ident.isEmpty()) {
							index = ident.indexOf('.');
							ident = project.getName() + ident.substring(index);
							if ((this.getVariable(ident) == null) && (this.getResult(ident) == null)) {
								ident = "";
							}
						}
						plotDesc.setXAxis(r, c, ident);
						// y-axis
						ident = ((TString) struct.get("y")).getString();
						if (!ident.isEmpty()) {
							index = ident.indexOf('.');
							ident = project.getName() + ident.substring(index);
							if ((this.getVariable(ident) == null) && (this.getResult(ident) == null)) {
								ident = "";
							}
						}
						plotDesc.setYAxis(r, c, ident);
					}
				}
			}
		}

		if (message.length() > 0) {
			return message.toString().substring(1);
		} else {
			return null;
		}
	}

	/**
	 * Reduce archive to number of pareto-front of problem
	 * 
	 * @param archiveName
	 *            Name of new archive
	 * @param numOfParetoFront
	 *            Number of pareto-fronts
	 * @throws Exception
	 */
	public void reduceArchive(String archiveName, int numOfParetoFront) throws Exception {
		Archive arch;
		Memory mem;
		Individual[] pareto;

		arch = this.getArchive().duplicate(archiveName);
		mem = new Memory(this.getArchive());
		mem.individuals.addAll(this.getArchive().getMemory().individuals);

		// reduce memory
		for (int k = 0; k < numOfParetoFront; k++) {
			pareto = mem.getParetoIndividual(this);
			for (Individual ind : pareto) {
				arch.getMemory().addIndividual(ind);
				mem.individuals.remove(ind);
			}
		}

		arch.save();
	}

	@Override
	public String toString() {
		StringBuffer str;

		str = new StringBuffer();
		str.append("Problem: " + this.name + "\n");
		str.append("\tArchive: " + this.archive.getName() + "\n");
		str.append("\tProject: " + this.archive.getMop().getAbsolutePath() + "\n");
		str.append("\tVariables:\n");
		for (int k = 0; k < this.variables.length; k++) {
			str.append("\t\t" + this.variables[k]);
		}
		str.append("\tResults:\n");
		for (int k = 0; k < this.results.length; k++) {
			str.append("\t\t" + this.results[k]);
		}

		return str.toString();
	}

	/**
	 * Get optimization
	 * 
	 * @return Optimization
	 */
	public Optimization getOptimization() {
		if (this.optim == null) {
			try {
				this.optim = new Optimization(this);
			} catch (Exception e) {
				this.optim = null;
				System.err.println(e.getMessage());
				e.printStackTrace();
			}
		}

		return this.optim;
	}
}
