/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import at.lcm.bu21.optim.algorithm.JMSolver;
import at.lcm.bu21.optim.algorithm.Solver;
import at.lcm.bu21.optim.algorithm.SolverList;
import at.lcm.bu21.optim.algorithm.me_nsga2.SsaME_Nsga2;
import at.lcm.bu21.optim.algorithm.nsga2.SsaNsga2;
import at.lcm.bu21.optim.algorithm.spea2.SsaSpea2;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Memory;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.TThreadState;
import jmetal.core.SolutionSet;
import jmetal.metaheuristics.nsgaII.Nsga2Real;
import jmetal.metaheuristics.spea2.Spea2Real;
import jmetal.util.JMException;

public class MappingOptimizationThread extends Thread {

	private MappingManager parentMappingManager;

	private TThreadState optimizationState;

	private int maxEvaluations;

	private int evaluatedInds;

	private int invalidCount;

	private int errorCount;

	private int environmentalSelectionSafeSize;

	private int startingArchiveSize;

	private final AtomicBoolean stopOptimization;

	private Solver solver;

	/** Initial individuals */
	private LinkedList<Individual> startingArchive = new LinkedList<Individual>();

	public MappingOptimizationThread(MappingManager parentMappingManager, AtomicBoolean stopOptimization) {
		this.parentMappingManager = parentMappingManager;
		this.startingArchiveSize = 200;
		this.environmentalSelectionSafeSize = 500;
		this.maxEvaluations = 5000;
		// this.maxEvaluations = 500;
		this.solver = SolverList.getSolverListEntry(parentMappingManager.getProblem().getSolver().getOptimizerName())
				.getSolver(parentMappingManager.getProblem());

		if (this.solver.getOptimizerName().startsWith("Generational Nsga2")) {
			try {
				String name = (new SsaNsga2(parentMappingManager.getProblem())).getOptimizerName();
				this.solver = SolverList.getSolverListEntry(name).getSolver(parentMappingManager.getProblem());
			} catch (JMException e) {
				e.printStackTrace();
			}
		}

		if (this.solver.getOptimizerName().startsWith("Generational Spea2")) {
			try {
				String name = (new SsaSpea2(parentMappingManager.getProblem())).getOptimizerName();
				this.solver = SolverList.getSolverListEntry(name).getSolver(parentMappingManager.getProblem());
			} catch (JMException e) {
				e.printStackTrace();
			}
		}

		if (this.solver.getOptimizerName().startsWith("Generational ME-NSGA-II")) {
			try {
				String name = (new SsaME_Nsga2(parentMappingManager.getProblem())).getOptimizerName();
				this.solver = SolverList.getSolverListEntry(name).getSolver(parentMappingManager.getProblem());
			} catch (JMException e) {
				e.printStackTrace();
			}
		}

		this.stopOptimization = stopOptimization;
		this.optimizationState = TThreadState.RUNNING;
	}

	public List<Individual> extractBestIndividuals(int desiredCount) {
		// extract the most promising individuals
		List<Individual> allIndividuals = new ArrayList<Individual>();
		for (int i = 0; i < this.parentMappingManager.getMappingMemory().getNumIndividuals(); i++) {
			allIndividuals.add(this.parentMappingManager.getMappingMemory().getIndividual(i));
		}

		if (desiredCount >= this.environmentalSelectionSafeSize) {
			while (allIndividuals.size() > desiredCount) {
				allIndividuals = this.reduceSizeViaNonDominatedSorting(allIndividuals, desiredCount, false);
			}
		} else {
			while (allIndividuals.size() > this.environmentalSelectionSafeSize) {
				allIndividuals = this.reduceSizeViaNonDominatedSorting(allIndividuals,
						this.environmentalSelectionSafeSize, false);
			}
			if (this.environmentalSelectionSafeSize >= desiredCount) {
				allIndividuals = this.reduceSizeViaEnvironmentalSelection(allIndividuals, desiredCount, false);
			}
		}
		return allIndividuals;
	}

	private void extractStartingArchive() {
		Individual[] validIndividuals = this.parentMappingManager.getProblem().getArchive().getMemory()
				.getValidIndividual(this.parentMappingManager.getProblem());

		List<Individual> startingArchive = new ArrayList<Individual>();
		startingArchive.addAll(Arrays.asList(validIndividuals));

		// System.out.println("Starting archive size: " +
		// startingArchive.size());
		// int c = 0;
		// for (Individual ind : startingArchive) {
		// for (ResultDesc rDesc : this.parentMappingManager.getProblem()
		// .getResults()) {
		// if (rDesc.isObjective() || rDesc.isConstraint()) {
		// this.parentMappingManager.addLog("Ind " + c + " "
		// + rDesc.getName() + " : "
		// + ind.getResult(rDesc.getName()));
		// }
		// }
		// c++;
		// }

		while (startingArchive.size() > this.environmentalSelectionSafeSize) {
			startingArchive = this.reduceSizeViaNonDominatedSorting(startingArchive, startingArchive.size() / 2, true);
		}

		// System.out.println("Starting archive size: " +
		// startingArchive.size());
		// c = 0;
		// for (Individual ind : startingArchive) {
		// for (ResultDesc rDesc : this.parentMappingManager.getProblem()
		// .getResults()) {
		// if (rDesc.isObjective() || rDesc.isConstraint()) {
		// this.parentMappingManager.addLog("Ind " + c + " "
		// + rDesc.getName() + " : "
		// + ind.getResult(rDesc.getName()));
		// }
		// }
		// c++;
		// }

		startingArchive = this.reduceSizeViaEnvironmentalSelection(startingArchive, this.startingArchiveSize, true);

		for (Individual ind : startingArchive) {
			try {
				this.solver.setResolvedIndividual(ind);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		try {
			Individual[] selectedIndividuals = this.solver.getArchive();
			int i = 0;
			while ((i < selectedIndividuals.length) && (!this.stopOptimization.get())) {

				this.parentMappingManager.evaluateRegressionTargetsOfIndividual(selectedIndividuals[i]);
				this.evaluatedInds++;
				i++;
				this.parentMappingManager
						.updateOptimizationProgress((1.0 * this.evaluatedInds) / (this.maxEvaluations));
				// this.parentMappingManager.addLog("Mapping eval. inds.: "
				// + this.evaluatedInds);
				// this.parentMappingManager.addLog("Mapping memory size: "
				// + this.parentMappingManager.getMappingMemory()
				// .getNumIndividuals());
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public TThreadState getOptimizationState() {
		return this.optimizationState;
	}

	private List<Individual> reduceSizeViaEnvironmentalSelection(List<Individual> originalPopulation,
			int desiredPopulationSize, boolean useRealMemoryReference) {
		List<Individual> reducedPopulation = new ArrayList<Individual>();
		List<Individual> newPopulation = new ArrayList<Individual>();

		try {

			// System.out.println("Original population size: "
			// + originalPopulation.size());
			// int c = 0;
			// for (Individual ind : originalPopulation) {
			// for (ResultDesc rDesc : this.parentMappingManager.getProblem()
			// .getResults()) {
			// if (rDesc.isObjective() || rDesc.isConstraint()) {
			// this.parentMappingManager.addLog("Ind " + c + " "
			// + rDesc.getName() + " : "
			// + ind.getResult(rDesc.getName()));
			// }
			// }
			// c++;
			// }

			SolutionSet originalSolSet = ((JMSolver) this.solver)
					.getSolutionSet(originalPopulation.toArray(new Individual[0]));
			reducedPopulation.addAll(Arrays.asList(((JMSolver) this.solver)
					.getIndividuals(Spea2Real.refineSolutionSet(originalSolSet, desiredPopulationSize))));

			Memory m = this.parentMappingManager.getProblem().getArchive().getMemory();
			if (!useRealMemoryReference) {
				m = this.parentMappingManager.getMappingMemory();
			}

			for (Individual ind : reducedPopulation) {
				newPopulation.add(m.getIndividual(ind));
			}

			// System.out.println("New population size: " +
			// newPopulation.size());
			// c = 0;
			// for (Individual ind : newPopulation) {
			// for (ResultDesc rDesc : this.parentMappingManager.getProblem()
			// .getResults()) {
			// if (rDesc.isObjective() || rDesc.isConstraint()) {
			// this.parentMappingManager.addLog("Ind " + c + " "
			// + rDesc.getName() + " : "
			// + ind.getResult(rDesc.getName()));
			// }
			// }
			// c++;
			// }

			this.parentMappingManager.addLog(
					"Population reduced to " + newPopulation.size() + " individuals via environmental selection!");

		} catch (Exception e) {

			e.printStackTrace();
		}

		return newPopulation;
	}

	private List<Individual> reduceSizeViaNonDominatedSorting(List<Individual> originalPopulation,
			int desiredPopulationSize, boolean useRealMemoryReference) {
		List<Individual> reducedPopulation = new ArrayList<Individual>();
		List<Individual> newPopulation = new ArrayList<Individual>();

		try {
			SolutionSet originalSolSet = ((JMSolver) this.solver)
					.getSolutionSet(originalPopulation.toArray(new Individual[0]));
			int newSize = (originalSolSet.size() / 2) + 1;
			if (newSize < desiredPopulationSize) {
				newSize = desiredPopulationSize;
			}
			reducedPopulation.addAll(
					Arrays.asList(((JMSolver) this.solver).getIndividuals(Nsga2Real.refineSolutionSet(originalSolSet,
							newSize, this.parentMappingManager.getProblem().getNumObjectives()))));

			Memory m = this.parentMappingManager.getProblem().getArchive().getMemory();
			if (!useRealMemoryReference) {
				m = this.parentMappingManager.getMappingMemory();
			}

			for (Individual ind : reducedPopulation) {
				newPopulation.add(m.getIndividual(ind));
			}

			this.parentMappingManager.addLog(
					"Population reduced to " + newPopulation.size() + " individuals via non-dominated sorting!");

		} catch (Exception e) {

			e.printStackTrace();
		}

		return newPopulation;
	}

	@Override
	public void run() {

		// System.out.println("Memory size: "
		// + this.solver.getProblem().getArchive().getMemory()
		// .getNumIndividuals());

		// change optimization state
		this.optimizationState = TThreadState.RUNNING;
		this.parentMappingManager.handleOptimizationStateChange();

		// extract starting population
		this.extractStartingArchive();

		this.invalidCount = 0;
		this.errorCount = 0;

		// run optimization using chosen solver
		while ((this.evaluatedInds < this.maxEvaluations) && (!this.stopOptimization.get())) {
			try {
				Individual ind = this.solver.getNextIndividual();

				if (ind == null) {
					System.out.println("Solver description:" + this.solver.getDescription());
					System.out.println("Solver archive size:" + this.solver.getArchive().length);
				}

				if (!this.parentMappingManager.isIndividualLikelyToBeErroneous(ind)
						|| !this.parentMappingManager.useErrorsClassifier()) {

					this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);
					if (ind.isValid(this.parentMappingManager.getProblem())) {
						this.solver.setResolvedIndividual(ind);
						this.parentMappingManager.getMappingMemory().addIndividual(ind);
					} else {
						this.invalidCount++;
					}
					this.evaluatedInds++;

					this.parentMappingManager
							.updateOptimizationProgress((1.0 * this.evaluatedInds) / (this.maxEvaluations));
				} else {
					this.errorCount++;
				}

				// this.parentMappingManager.addLog("" + this.errorCount);

				if (this.errorCount >= this.maxEvaluations) {
					this.evaluatedInds = this.maxEvaluations;
					this.parentMappingManager
							.updateOptimizationProgress((1.0 * this.evaluatedInds) / (this.maxEvaluations));
				}

				// this.parentMappingManager.addLog("Mapping eval. inds.: "
				// + this.evaluatedInds);
				// this.parentMappingManager.addLog("Mapping memory size: "
				// + this.parentMappingManager.getMappingMemory()
				// .getNumIndividuals());
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		this.parentMappingManager.addLog("Mapping based optimization outcome - invalid count: " + this.invalidCount);
		this.parentMappingManager.addLog("Mapping based optimization outcome - erroneous count: " + this.errorCount);

		// change optimization state
		if (this.parentMappingManager.getMappingMemory().getNumIndividuals() != 0) {
			this.optimizationState = TThreadState.READY;
			this.parentMappingManager.addLog("Mapping based optimization finished!");
		} else {
			this.optimizationState = TThreadState.IDLE;
			this.parentMappingManager.addLog("Mapping based optimization failed!");
		}
		this.parentMappingManager.handleOptimizationStateChange();
	}

	public void setOptimizationState(TThreadState optimizationState) {
		this.optimizationState = optimizationState;
	}
}
