/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.mapping.utils;

import java.util.concurrent.atomic.AtomicBoolean;

import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.core.mapping.MappingManager;
import at.lcm.bu21.optim.core.mapping.utils.dstructures.TMappingConstants.TThreadState;

public class SensitivityAnalysisThread extends Thread {

	private MappingManager parentMappingManager;

	private TThreadState sensitivityAnalysisState;

	private final AtomicBoolean stopSensitivityAnalysis;

	private SensitivityAnalysisHolder resultHolder;

	public TThreadState getSensitivityAnalysisState() {
		return this.sensitivityAnalysisState;
	}

	public SensitivityAnalysisThread(MappingManager parentMappingManager, AtomicBoolean stopSensitivityAnalysis,
			SensitivityAnalysisHolder resultHolder) {
		this.parentMappingManager = parentMappingManager;
		this.stopSensitivityAnalysis = stopSensitivityAnalysis;
		this.sensitivityAnalysisState = TThreadState.RUNNING;
		this.resultHolder = resultHolder;
	}

	@Override
	public void run() {
		try {
			this.parentMappingManager.evaluateRegressionTargetsOfIndividual(this.resultHolder.getBaseIndividual("FE"));
			// this.printIndividual(this.resultHolder.getBaseIndividualTrans());

			int evalInds = 0;
			for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
				for (Individual ind : this.resultHolder.getLocalSensitivityIndividuals().get(vDesc.getName())) {
					this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);

					// this.printIndividual(ind);

					evalInds++;
					this.parentMappingManager.updateSensitivityAnalysisProgress(
							(1.0 * evalInds) / (this.resultHolder.getTotalIndividualsCount()));

					if (this.stopSensitivityAnalysis.get()) {
						break;
					}
				}
			}

			// System.out.println("");
			// System.out.println("");
			// System.out.println("");
			// System.out.println("");
			// System.out.println("");

			for (Individual ind : this.resultHolder.getGlobalSensitivityIndividuals()) {
				this.parentMappingManager.evaluateRegressionTargetsOfIndividual(ind);

				// this.printIndividual(ind);

				evalInds++;
				this.parentMappingManager.updateSensitivityAnalysisProgress(
						(1.0 * evalInds) / (this.resultHolder.getTotalIndividualsCount()));

				if (this.stopSensitivityAnalysis.get()) {
					break;
				}
			}

			this.sensitivityAnalysisState = TThreadState.RUNNING;
		} catch (Exception e) {
			this.parentMappingManager.addError("ERROR in sensitivity analysis thread!", e.getStackTrace());
			this.parentMappingManager.handleSensitivityAnalysisRunErronousEnd();
		}

		if (this.stopSensitivityAnalysis.get()) {
			this.parentMappingManager.handleSensitivityAnalysisRunErronousEnd();
		} else {
			this.parentMappingManager.handleSensitivityAnalysisSucessfulRun();
		}

	}

	public void setSensitivityAnalysisState(TThreadState sensitivityAnalysisState) {
		this.sensitivityAnalysisState = sensitivityAnalysisState;
	}

	@SuppressWarnings("unused")
	private void printIndividual(Individual ind) {
		String v = "";
		String o = "";
		for (VariableDesc vDesc : this.parentMappingManager.getProblem().getVariedVariables()) {
			v += ind.getVariable(vDesc.getName()) + ",";
		}
		for (ResultDesc rDesc : this.parentMappingManager.getProblem().getResults()) {
			if (rDesc.isObjective() || rDesc.isConstraint()) {
				o += ind.getResult(rDesc.getName()) + ",";
			}
		}
		System.out.println("Individual V+O:" + v + " - Objectives: " + o);
	}

}
