/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.core.project;

import java.io.File;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.LinkedList;

import javax.swing.SwingUtilities;

import at.lcm.bu21.general.dtypes.TData;
import at.lcm.bu21.general.dtypes.TNumeric;
import at.lcm.bu21.general.plot.LinkPlot;
import at.lcm.bu21.general.plot.Plot2D;
import at.lcm.bu21.general.plot.Plot2Line;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.docu.DocTexWriter;
import at.lcm.bu21.magopt.structure.ParamField;
import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.magopt.structure.Project;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.core.VariableDesc;
import at.lcm.bu21.optim.gui.project.PanelPlots;
import magopt.utils.MopFile;
import magopt.utils.ProgressBar;

/**
 * Create new project and doc-files from selected optimum
 * 
 * @author Werner Koppelsttter
 */
public class SelectOptimum {

	/** Index of progress-bar */
	private int dispIndex = 0;

	/** Size of plot for save */
	private final static int SAVE_WITDH = 450;

	/** Size of plot for save */
	private final static int SAVE_HEIGHT = 340;

	/**
	 * Create new project with selected individuals form save-point
	 * 
	 * @param mop
	 *            Name of new project
	 * @param prob
	 *            Problem definition
	 * @param list
	 *            List of individuals
	 * @param selected
	 *            Selected individuals
	 * @param suppressed
	 *            Suppressed individuals
	 * @param doc
	 *            Create doc-files
	 * @return Project with selected individuals
	 * @throws Exception
	 */
	public Project createProject(MopFile mop, Problem prob, Individual[] list, int[] selected, int[] suppressed,
			boolean doc) throws Exception {
		Project project;
		Individual ind;
		final ProgressBar bar;
		VariableDesc[] varDesc;

		bar = new ProgressBar("Create project with selected individuals", selected.length + 2);

		// clone base-project
		project = new Project("project", "", mop);
		prob.getArchive().getProject().cloneParam(project, prob.getArchive().getBasedIndex());
		project.refresh(true);
		project.deleteDocFilesOpt(0);
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				bar.setValue(1);
			}
		});

		// duplicate dataset
		varDesc = prob.getVariables();
		for (int k = 0; k < selected.length; k++) {
			project.duplicate(0);
			project.setIndex(k + 1);
			ind = list[selected[k]];

			for (int v = 0; v < varDesc.length; v++) {
				project.getObject(varDesc[v].getName()).setData(new TNumeric(ind.getVariable(v)));
			}
			project.refresh();

			if (doc) {
				// creat doc-files
				this.createDocFiles(project, prob, list, selected[k], suppressed);
			}
			this.dispIndex = k + 2;
			SwingUtilities.invokeLater(new Runnable() {
				@Override
				public void run() {
					bar.setValue(SelectOptimum.this.dispIndex);
				}
			});

		}

		// finish
		project.setIndex(0);
		project.reduce();
		project.save();

		bar.dispose();

		return project;
	}

	/**
	 * Create doc-files for index
	 * 
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @param suppressed
	 *            Indices of suppressed individuals
	 * @throws Exception
	 */
	private void createDocFiles(Project project, Problem prob, Individual[] pareto, int index, int[] suppressed)
			throws Exception {
		File docfiles;

		// create directory
		docfiles = FileUtils.getFile(project.getFilesDirectory().getPath(), "docFiles");
		docfiles.mkdirs();

		// plots
		this.createObjectivePlots(docfiles, project, prob, pareto, index, suppressed);
		this.createVariableHistPlots(docfiles, project, prob, pareto, index, suppressed);
		this.createQualityPlot(docfiles, project, prob);

		// tex-tables
		this.createDocTexOptimization(docfiles, project, prob);
		this.createDocTexFixed(docfiles, project, prob);
		this.createDocTexVaried(docfiles, project, prob, pareto, index);
		this.createDocTexObjective(docfiles, project, prob, pareto, index);
		this.createDocTexConstraint(docfiles, project, prob, pareto, index);

		project.resetDocFiles();
	}

	/**
	 * Create objective-plots
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @param suppressed
	 *            Indices of suppressed individuals
	 * @throws Exception
	 */
	private void createObjectivePlots(File docfiles, Project project, Problem prob, Individual[] pareto, int index,
			int[] suppressed) throws Exception {
		SubPlot sub;
		LinkPlot link;

		sub = new SubPlot(1, 1);
		PanelPlots.initObjectives(sub, prob);
		link = new LinkPlot();
		link.addPlot(sub);

		LinkedList<Individual[]> p = new LinkedList<Individual[]>();
		p.add(pareto);
		PanelPlots.drawObjectives(sub, prob, p, -1, false);

		// All objectives without Selection
		link.reset();
		this.saveObjectivePlots(docfiles, project, prob, sub, "FigOpt", true);

		// All objectives with Selection
		link.select(0, index, true);
		link.reset();
		this.saveObjectivePlots(docfiles, project, prob, sub, "FigOpt_withSelection", true);

		// Reduced objectives without Selection
		link.clearSelect();
		for (int k : suppressed) {
			link.suppress(0, k, true);
		}
		link.reset();
		this.saveObjectivePlots(docfiles, project, prob, sub, "FigOptRed", false);

		// Reduced objectives with Selection
		link.select(0, index, true);
		link.reset();
		this.saveObjectivePlots(docfiles, project, prob, sub, "FigOptRed_withSelection", false);

	}

	/**
	 * Save objective plots
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param sub
	 *            Sub-plot for objectives
	 * @param name
	 *            Name of plot
	 * @param detail
	 *            Save each plot of sub-plot separately
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void saveObjectivePlots(File docfiles, Project project, Problem prob, SubPlot sub, String name,
			boolean detail) throws IOException, NoSuchAlgorithmException {
		File file;
		int[] size;
		Plot2D plot;
		String tmp;
		int o1, o2;
		ResultDesc[] objDesc;

		// save sub-plot
		file = new File(docfiles, name + ".eps");
		file = project.getUniqueDocFileName(file);
		sub.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
		project.registerDocFile(name, file.getPath(), name);

		// save plots of sub-plot
		if (detail) {
			objDesc = prob.getObjectives();
			if (objDesc.length == 1) {
				// save plot with on objective
				plot = (Plot2D) sub.getPlot(0, 0);
				tmp = name + "_1_1_" + objDesc[0].getName().substring(objDesc[0].getName().lastIndexOf(".") + 1) + "_"
						+ objDesc[0].getName().substring(objDesc[0].getName().lastIndexOf(".") + 1);
				file = new File(docfiles, tmp + ".eps");
				file = project.getUniqueDocFileName(file);
				plot.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
				project.registerDocFile(tmp, file.getPath(), tmp);

			} else {
				// save each plot
				o1 = 0;
				o2 = 1;
				size = sub.getSize();
				for (int r = 0; r < size[0]; r++) {
					for (int c = 0; c < size[1]; c++) {
						plot = (Plot2D) sub.getPlot(r, c);
						if (plot != null) {
							tmp = name + "_" + (o1 + 1) + "_" + (o2 + 1) + "_"
									+ objDesc[o1].getName().substring(objDesc[o1].getName().lastIndexOf(".") + 1) + "_"
									+ objDesc[o2].getName().substring(objDesc[o2].getName().lastIndexOf(".") + 1);
							file = new File(docfiles, tmp + ".eps");
							file = project.getUniqueDocFileName(file);
							plot.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
							project.registerDocFile(tmp, file.getPath(), tmp);

							o2++;
							if (o2 >= objDesc.length) {
								o1++;
								o2 = o1 + 1;
							}
						}
					}
				}
			}
		}
	}

	/**
	 * Create variable-plots
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @param suppressed
	 *            Indices of suppressed individuals
	 * @throws Exception
	 */
	private void createVariableHistPlots(File docfiles, Project project, Problem prob, Individual[] pareto, int index,
			int[] suppressed) throws Exception {
		SubPlot sub;
		LinkPlot link;

		sub = new SubPlot(1, 1);
		PanelPlots.initVariablesHist(sub, prob);
		link = new LinkPlot();
		link.addPlot(sub);

		LinkedList<Individual[]> p = new LinkedList<Individual[]>();
		p.add(pareto);
		PanelPlots.drawVariablesHist(sub, prob, p, -1, false);

		// All variables without selection
		link.reset();
		this.saveVariableHistPlots(docfiles, project, prob, sub, "FigOptVar", true);

		// All variables with selection
		link.select(0, index, true);
		link.reset();
		this.saveVariableHistPlots(docfiles, project, prob, sub, "FigOptVar_withSelection", true);

		// Reduced variables without selection
		link.clearSelect();
		for (int k : suppressed) {
			link.suppress(0, k, true);
		}
		link.reset();
		this.saveVariableHistPlots(docfiles, project, prob, sub, "FigOptVarRed", false);

		// Reduced variables with selection
		link.select(0, index, true);
		link.reset();
		this.saveVariableHistPlots(docfiles, project, prob, sub, "FigOptVarRed_withSelection", false);

	}

	/**
	 * Save variable plots
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param sub
	 *            Sub-plot for objectives
	 * @param name
	 *            Name of plot
	 * @param detail
	 *            Save each plot of sub-plot separately
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void saveVariableHistPlots(File docfiles, Project project, Problem prob, SubPlot sub, String name,
			boolean detail) throws IOException, NoSuchAlgorithmException {
		File file;
		int[] size;
		Plot2D plot;
		String tmp;
		int v;
		VariableDesc[] varDesc;

		// save sub-plot
		file = new File(docfiles, name + ".eps");
		file = project.getUniqueDocFileName(file);
		sub.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
		project.registerDocFile(name, file.getPath(), name);

		// save plots of sub-plot
		if (detail) {
			varDesc = prob.getVariedVariables();
			size = sub.getSize();
			v = 0;
			for (int c = 0; c < size[1]; c++) {
				for (int r = 0; r < size[0]; r++) {
					plot = (Plot2D) sub.getPlot(r, c);
					if (plot != null) {
						tmp = name + "_" + (v + 1) + "_"
								+ varDesc[v].getName().substring(varDesc[v].getName().lastIndexOf(".") + 1);
						file = new File(docfiles, tmp + ".eps");
						file = project.getUniqueDocFileName(file);
						plot.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
						project.registerDocFile(tmp, file.getPath(), tmp);

						v++;
					}
				}
			}
		}
	}

	/**
	 * Create quality-plot
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @throws Exception
	 */
	private void createQualityPlot(File docfiles, Project project, Problem prob) throws Exception {
		Plot2Line plot;
		File file;
		String name;

		plot = new Plot2Line();
		PanelPlots.initQuality(plot);
		PanelPlots.drawQuality(plot, prob, true, false);

		plot.reset();
		name = "FigOptQuality";
		file = new File(docfiles, name + ".eps");
		file = project.getUniqueDocFileName(file);
		plot.save(file, SelectOptimum.SAVE_WITDH, SelectOptimum.SAVE_HEIGHT);
		project.registerDocFile(name, file.getPath(), name);
	}

	/**
	 * Create tex-table of optimization
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void createDocTexOptimization(File docfiles, Project project, Problem prob)
			throws IOException, NoSuchAlgorithmException {
		String name;
		String tablestruct;
		File file;
		DocTexWriter content;

		name = "TableOptimization";

		file = new File(docfiles, name + ".tex");
		file = project.getUniqueDocFileName(file);
		content = new DocTexWriter(file);

		tablestruct = "|l|l|";
		content.add("\\ifdefined\\usehtml\\begin{tabular}{" + tablestruct + "}\\else\\begin{longtable}[l]{"
				+ tablestruct + "}\\fi\n");
		content.add("\t\\hline\n");
		content.add("\t Archivename & ");
		content.addString(prob.getArchive().getName());
		content.add(" \\tabularnewline \\hline\n");
		content.add("\t Problemname & ");
		content.addString(prob.getName());
		content.add(" \\tabularnewline \\hline\n");
		content.add("\t No. Savepoints & " + prob.getNumSavePoints() + " \\tabularnewline \\hline\n");
		content.add("\\ifdefined\\usehtml\\end{tabular}\\else\\end{longtable}\\fi\n");

		content.write();

		project.registerDocFile(name, file.getPath(), name);
	}

	/**
	 * Create tex-table with non used variables for optimization
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void createDocTexFixed(File docfiles, Project project, Problem prob)
			throws IOException, NoSuchAlgorithmException {
		String[] name;
		String tablestruct, tableHeading;
		File file;
		DocTexWriter content;
		ParamField[][] params;
		LinkedList<ParamField> list;
		TData data;
		ParamField field;
		int startIndex;
		VariableDesc[] varDesc;

		name = new String[] { "TableFixedParameters", "TableFixedParametersNotConstant" };
		tablestruct = "|m{7cm}|>{\\centering}p{1.8cm}|>{\\centering}p{1.4cm}||>{\\raggedleft}p{1.4cm}|";
		tableHeading = "\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Unit} & \\textbf{Value} \\tabularnewline \\hline \\hline\n";

		params = new ParamField[2][];

		// list parameter from Projects
		list = new LinkedList<ParamField>();
		for (Parameter param : project.getParameters(ParamField.class)) {
			field = (ParamField) param;
			if (field.getUsed() && !field.getReadOnly()) {
				data = field.getData();
				if (data instanceof TNumeric) {
					if (((TNumeric) data).isScalar()) {
						list.add(field);
					}
				}
			}
		}

		// remove variables
		varDesc = prob.getVariables();
		for (int k = list.size() - 1; k >= 0; k--) {
			String ident = list.get(k).getIdentifier();
			for (int m = 0; m < varDesc.length; m++) {
				if (ident.compareTo(varDesc[m].getName()) == 0) {
					list.remove(k);
				}
			}
		}
		params[0] = list.toArray(new ParamField[0]);

		// remove constant
		for (int k = list.size() - 1; k >= 0; k--) {
			if (list.get(k).getConstant()) {
				list.remove(k);
			}
		}
		params[1] = list.toArray(new ParamField[0]);

		// write tex-files
		startIndex = project.getName().length() + 1;
		for (int n = 0; n < name.length; n++) {
			file = new File(docfiles, name[n] + ".tex");
			file = project.getUniqueDocFileName(file);
			content = new DocTexWriter(file);

			content.add("\\ifdefined\\usehtml\\begin{tabular}{" + tablestruct + "}\\else\\begin{longtable}[l]{"
					+ tablestruct + "}\\fi\n");
			content.add(tableHeading);

			// write table content
			for (int k = 0; k < params[n].length; k++) {
				field = params[n][k];
				content.add("\t");
				content.add("\\parbox[t]{7cm}{");
				if (field.getDescription().isEmpty()) {
					content.add("\\vspace*{1.25ex}");
				} else {
					content.addString(field.getDescription()); // Description
					content.add(" \\\\[0.5ex]");
				}
				content.add(" \\tiny ");

				content.addString(field.getIdentifier().substring(startIndex)); // Identifier
				content.add("\\\\[-1ex] } & ");
				if (field.getTexString().isEmpty()) {
					content.addString(field.getName()); // Parameter
				} else {
					content.add("$" + field.getTexString() + "$"); // Parameter
				}
				content.add(" & ");
				content.addString(field.getUnit()); // Unit
				content.add(" & ");
				content.add("$" + String.format("%.5g", ((TNumeric) field.getData()).getReal(0, 0)) + "$ "); // Value
				content.add("\\tabularnewline \\hline\n");
			}
			content.add("\\ifdefined\\usehtml\\end{tabular}\\else\\end{longtable}\\fi\n");

			content.write();

			project.registerDocFile(name[n], file.getPath(), name[n]);
		}
	}

	/**
	 * Create tex-table with used variables for optimization
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void createDocTexVaried(File docfiles, Project project, Problem prob, Individual[] pareto, int index)
			throws IOException, NoSuchAlgorithmException {
		String[] name, tablestruct, tableheading;
		File file;
		DocTexWriter content;
		ParamField field;
		int startIndex;
		VariableDesc[] varDesc;

		name = new String[] { "TableVariedParameters", "TableVariedParameters_withSelection" };
		tablestruct = new String[] {
				"|m{4cm}|>{\\centering}p{1.6cm}|>{\\centering}p{1.2cm}||>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}|",
				"|m{4cm}|>{\\centering}p{1.6cm}|>{\\centering}p{1.2cm}||>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}||>{\\raggedleft}p{1.2cm}|" };
		tableheading = new String[] {
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Unit} & \\textbf{Min} & \\textbf{Step} & \\textbf{Max} \\tabularnewline \\hline \\hline\n",
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Unit} & \\textbf{Min} & \\textbf{Step} & \\textbf{Max}& \\textbf{Value} \\tabularnewline \\hline \\hline\n" };

		startIndex = project.getName().length() + 1;
		varDesc = prob.getVariables();
		for (int n = 0; n < name.length; n++) {
			file = new File(docfiles, name[n] + ".tex");
			file = project.getUniqueDocFileName(file);
			content = new DocTexWriter(file);

			content.add("\\ifdefined\\usehtml\\begin{tabular}{" + tablestruct[n] + "}\\else\\begin{longtable}[l]{"
					+ tablestruct[n] + "}\\fi\n");
			content.add(tableheading[n]);

			for (int k = 0; k < varDesc.length; k++) {
				field = (ParamField) project.getObject(varDesc[k].getName());
				content.add("\t");
				content.add("\\parbox[t]{4cm}{");
				if (field.getDescription().isEmpty()) {
					content.add("\\vspace*{1.25ex}");
				} else {
					content.addString(field.getDescription()); // Description
					content.add(" \\\\[0.5ex]");
				}
				content.add(" \\tiny ");
				content.addString(field.getIdentifier().substring(startIndex)); // Identifier
				content.add("\\\\[-1ex] } & ");
				if (field.getTexString().isEmpty()) {
					content.addString(field.getName());
				} else {
					content.add("$" + field.getTexString() + "$"); // Parameter
				}
				content.add(" & ");
				content.addString(field.getUnit()); // Unit
				content.add(" & ");
				content.add("$" + varDesc[k].getLB() + " $"); // Min
				content.add(" & ");
				content.add("$" + varDesc[k].getStepSize() + " $"); // Stepsize
				content.add(" & ");
				content.add("$" + varDesc[k].getUB() + " $"); // Max
				if (n == 1) {
					content.add("& $" + String.format("%.5g", pareto[index].getVariable(k)) + " $ ");// Value
				}
				content.add("\\tabularnewline \\hline\n");
			}
			content.add("\\ifdefined\\usehtml\\end{tabular}\\else\\end{longtable}\\fi\n");

			content.write();

			project.registerDocFile(name[n], file.getPath(), name[n]);
		}

	}

	/**
	 * Create tex-table with objectives
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void createDocTexObjective(File docfiles, Project project, Problem prob, Individual[] pareto, int index)
			throws IOException, NoSuchAlgorithmException {
		String[] name, tablestruct, tableheading;
		File file;
		DocTexWriter content;
		ParamField field;
		int startIndex;
		ResultDesc[] resDesc;

		name = new String[] { "TableObjectives", "TableObjectives_withSelection" };
		tablestruct = new String[] { "|m{6cm}|>{\\centering}p{1.8cm}|>{\\centering}p{1.0cm}|",
				"|m{6cm}|>{\\centering}p{1.8cm}|>{\\centering}p{1.0cm}|>{\\centering}p{1.4cm}||>{\\raggedleft}p{1.4cm}|" };
		tableheading = new String[] {
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Type} \\tabularnewline \\hline \\hline\n",
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Type} & \\textbf{Unit} & \\textbf{Value} \\tabularnewline \\hline \\hline\n" };

		for (int n = 0; n < name.length; n++) {
			file = new File(docfiles, name[n] + ".tex");
			file = project.getUniqueDocFileName(file);
			content = new DocTexWriter(file);

			content.add("\\ifdefined\\usehtml\\begin{tabular}{" + tablestruct[n] + "}\\else\\begin{longtable}[l]{"
					+ tablestruct[n] + "}\\fi\n");
			content.add(tableheading[n]);

			startIndex = project.getName().length() + 1;
			resDesc = prob.getResults();
			for (int k = 0; k < resDesc.length; k++) {
				if (resDesc[k].isObjective()) {
					field = (ParamField) project.getObject(resDesc[k].getName());
					content.add("\t");
					content.add("\\parbox[t]{7cm}{");
					if (field.getDescription().isEmpty()) {
						content.add("\\vspace*{1.25ex}");
					} else {
						content.addString(field.getDescription()); // Description
						content.add(" \\\\[0.5ex]");
					}
					content.add(" \\tiny ");
					content.addString(field.getIdentifier().substring(startIndex)); // Identifier
					content.add("\\\\[-1ex] } & ");
					if (field.getTexString().isEmpty()) {
						content.addString(field.getName()); // Parameter
					} else {
						content.add("$" + field.getTexString() + "$"); // Parameter
					}
					content.add(" & ");
					content.addString(resDesc[k].getObjective().name());
					if (n == 1) {
						content.add(" & ");
						content.addString(field.getUnit()); // Unit
						content.add(" & ");
						content.add("$" + String.format("%.5g", pareto[index].getResult(k)) + " $ "); // Value
					}
					content.add("\\tabularnewline \\hline\n");
				}
			}
			content.add("\\ifdefined\\usehtml\\end{tabular}\\else\\end{longtable}\\fi\n");

			content.write();

			project.registerDocFile(name[n], file.getPath(), name[n]);
		}
	}

	/**
	 * Create tex-table with contraints
	 * 
	 * @param docfiles
	 *            Directory of doc-files
	 * @param project
	 *            Project with current dataset
	 * @param prob
	 *            Problem definition
	 * @param pareto
	 *            Pareto individuals
	 * @param index
	 *            Index of individuals
	 * @throws IOException
	 * @throws NoSuchAlgorithmException
	 */
	private void createDocTexConstraint(File docfiles, Project project, Problem prob, Individual[] pareto, int index)
			throws IOException, NoSuchAlgorithmException {
		String[] name, tablestruct, tableheading;
		File file;
		DocTexWriter content;
		ParamField field;
		int startIndex;
		ResultDesc[] resDesc;

		name = new String[] { "TableConstraints", "TableConstraints_withSelection" };
		tablestruct = new String[] {
				"|m{5.2cm}|>{\\centering}p{1.8cm}|>{\\centering}p{1.2cm}||>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}|",
				"|m{5.2cm}|>{\\centering}p{1.8cm}|>{\\centering}p{1.2cm}||>{\\raggedleft}p{0.6cm}|>{\\raggedleft}p{0.6cm}||>{\\raggedleft}p{1.4cm}|" };
		tableheading = new String[] {
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Unit} & \\textbf{Min} & \\textbf{Max} \\tabularnewline \\hline \\hline\n",
				"\t\\hline \\textbf{Description}\\tiny /\\moparam{Identifier} & \\textbf{Parameter} & \\textbf{Unit} & \\textbf{Min} & \\textbf{Max} & \\textbf{Value} \\tabularnewline \\hline \\hline\n" };

		for (int n = 0; n < name.length; n++) {
			file = new File(docfiles, name[n] + ".tex");
			file = project.getUniqueDocFileName(file);
			content = new DocTexWriter(file);

			content.add("\\ifdefined\\usehtml\\begin{tabular}{" + tablestruct[n] + "}\\else\\begin{longtable}[l]{"
					+ tablestruct[n] + "}\\fi\n");
			content.add(tableheading[n]);

			startIndex = project.getName().length() + 1;
			resDesc = prob.getResults();
			for (int k = 0; k < resDesc.length; k++) {
				if (resDesc[k].isConstraint()) {
					field = (ParamField) project.getObject(resDesc[k].getName());
					content.add("\t");
					content.add("\\parbox[t]{6.4cm}{");
					if (field.getDescription().isEmpty()) {
						content.add("\\vspace*{1.25ex}");
					} else {
						content.addString(field.getDescription()); // Description
						content.add(" \\\\[0.5ex]");
					}
					content.add(" \\tiny ");
					content.addString(field.getIdentifier().substring(startIndex)); // Identifier
					content.add("\\\\[-1ex] } & ");
					if (field.getTexString().isEmpty()) {
						content.addString(field.getName()); // Parameter
					} else {
						content.add("$" + field.getTexString() + "$"); // Parameter
					}
					content.add(" & ");
					content.addString(field.getUnit()); // Unit
					content.add(" & ");
					content.add("$" + resDesc[k].getLB() + " $"); // Min
					content.add(" & ");
					content.add("$" + resDesc[k].getUB() + " $ "); // Max
					if (n == 1) {
						content.add(" & ");
						content.add("$" + String.format("%.5g", pareto[index].getResult(k)) + " $ "); // Value
					}
					content.add("\\tabularnewline \\hline\n");
				}
			}
			content.add("\\ifdefined\\usehtml\\end{tabular}\\else\\end{longtable}\\fi\n");

			content.write();

			project.registerDocFile(name[n], file.getPath(), name[n]);
		}
	}
}
