/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 249 $
 * $Date: 2020-04-14 20:42:56 +0200 (Tue, 14 Apr 2020) $
 */
package at.lcm.bu21.optim.gui.project;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.LinkedList;

import javax.swing.DefaultCellEditor;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import at.lcm.bu21.general.plot.HOLD;
import at.lcm.bu21.general.plot.LinkPlot;
import at.lcm.bu21.general.plot.SubPlot;
import at.lcm.bu21.optim.core.Archive;
import at.lcm.bu21.optim.core.Individual;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;
import at.lcm.bu21.optim.gui.OptimizerViews;
import jobmanager.server.Job;
import magopt.gui.helpers.JProjectChooser;
import magopt.utils.MopFile;

/**
 * Compare different Problem with the same objectives
 * 
 * @author Werner Koppelsttter
 * 
 */
public class PanelCompare extends JPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** View-object of optimizer */
	private OptimizerViews _views;

	/** Sub-plot for objective-plot */
	private SubPlot _sub;

	/** Table-model of settings */
	private CompareModel _model;

	/** Settings-table of selected problems */
	private JTable _table;

	/** List of selected problems */
	private LinkedList<CompareObject> _compare = new LinkedList<PanelCompare.CompareObject>();

	/**
	 * Connects column indices with column header names.
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum CompareColumn {
		/** Optimization project */
		MOP(0, "Project", String.class),
		/** Optimization archive */
		ARCHIVE(1, "Archive", String.class),
		/** Optimization problem */
		PROBLEM(2, "Problem", String.class),
		/** Valid selection */
		VALID(3, "Valid", Boolean.class),
		/** Plot color */
		COLOR(4, "Color", ColorSelect.class);

		/** Column number */
		private final int column;

		/** Column name */
		private final String name;

		/** Column type */
		private final Class<?> type;

		/**
		 * Create enumeration
		 * 
		 * @param column
		 *            Column number
		 * @param name
		 *            Column name
		 * @param type
		 *            Column type
		 */
		private CompareColumn(int column, String name, Class<?> type) {
			this.column = column;
			this.name = name;
			this.type = type;
		}

		/**
		 * Get column number
		 * 
		 * @return Number
		 */
		protected int getValue() {
			return this.column;
		}

		/**
		 * Get column name
		 * 
		 * @return Name
		 */
		protected String getName() {
			return this.name;
		}

		/**
		 * Get column type
		 * 
		 * @return Type
		 */
		protected Class<?> getType() {
			return this.type;
		}

		/**
		 * Get element of enumeration with to the index
		 * 
		 * @param index
		 *            Index of column
		 * @return Enumeration element
		 */
		static CompareColumn getColumn(int index) {
			CompareColumn comp = null;

			for (CompareColumn c : CompareColumn.values()) {
				if (c.getValue() == index) {
					comp = c;
				}
			}

			return comp;
		}
	}

	/**
	 * Enumeration of possible colors
	 * 
	 * @author Werner Koppelsttter
	 */
	public enum ColorSelect {
		/** blue */
		BLUE("blue", Color.blue),
		/** green */
		GREEN("green", Color.green.darker()),
		/** red */
		RED("red", Color.red),
		/** cyan */
		CYAN("cyan", Color.cyan),
		/** magenta */
		MAGENTA("magenta", Color.magenta),
		/** yellow */
		YELLOW("yellow", Color.yellow.darker()),
		/** black */
		BLACK("black", Color.black);

		/** Name of color */
		private final String name;

		/** Color value */
		private final Color color;

		/**
		 * Create color enumeration
		 * 
		 * @param name
		 *            Name of color
		 * @param color
		 *            Color object
		 */
		private ColorSelect(String name, Color color) {
			this.name = name;
			this.color = color;
		}

		/**
		 * Get name of color
		 * 
		 * @return Name
		 */
		protected String getName() {
			return this.name;
		}

		/**
		 * Get color objext
		 * 
		 * @return Color
		 */
		protected Color getColor() {
			return this.color;
		}

		@Override
		public String toString() {
			return this.name;
		}
	}

	/**
	 * Compare object with selected problem
	 * 
	 * @author Werner Koppelsttter
	 */
	private class CompareObject {

		/** Mop-file */
		private MopFile _mop = null;

		/** Archive */
		private Archive _arch = null;

		/** Problem */
		private Problem _prob = null;

		/** Selected color */
		private ColorSelect _color = null;

		/**
		 * Get selected mop-file
		 * 
		 * @return Mop-file
		 */
		public MopFile getMopFile() {
			MopFile mop;
			Archive arch;

			arch = this.getArchive();
			if (arch != null) {
				mop = arch.getMop();
			} else {
				mop = this._mop;
			}

			return mop;
		}

		/**
		 * Select mop-file
		 * 
		 * @param mop
		 *            Mop-file
		 */
		public void setMopFile(MopFile mop) {
			if ((this._mop == null) || (mop == null)
					|| ((this._mop != mop) && (this._mop.getAbsolutePath().compareTo(mop.getAbsolutePath()) != 0))) {
				// reset archive and problem
				this._arch = null;
				this._prob = null;
			}
			this._mop = mop;
		}

		/**
		 * Get selected archive
		 * 
		 * @return Archive
		 */
		public Archive getArchive() {
			Archive arch;
			Problem prob;

			prob = this.getProblem();
			if (prob != null) {
				arch = prob.getArchive();
			} else {
				arch = this._arch;
			}

			return arch;
		}

		/**
		 * Select archive
		 * 
		 * @param arch
		 *            Archive
		 */
		public void setArchive(Archive arch) {
			if (this._arch != arch) {
				this._arch = arch;
				// reset problem
				this._prob = null;
			}
		}

		/**
		 * Get selected problem
		 * 
		 * @return Problem
		 */
		public Problem getProblem() {
			return this._prob;
		}

		/**
		 * Select problem
		 * 
		 * @param prob
		 *            Problem
		 */
		public void setProblem(Problem prob) {
			this._prob = prob;
		}

		/**
		 * Get selected color
		 * 
		 * @return Color-enumeration
		 */
		public ColorSelect getColor() {
			return this._color;
		}

		/**
		 * Select color
		 * 
		 * @param color
		 *            Color-enumeration
		 */
		public void setColor(ColorSelect color) {
			this._color = color;
		}

		/**
		 * Is selected problem valid. Check if there are equal objectives
		 * 
		 * @param prob
		 *            Problem to check
		 * @return Is valid
		 */
		public boolean isValid(Problem prob) {
			boolean valid = false, val;
			ResultDesc[] res1, res2;
			String name1, name2;

			if (this.getProblem() != null) {
				// check for equal objectives
//				if (prob.getNumSavePoints() > 0) {
				if (prob.getNumObjectives() == this.getProblem().getNumObjectives()) {
					valid = true;
					res1 = prob.getObjectives();
					res2 = this.getProblem().getObjectives();
					for (int k1 = 0; k1 < res1.length; k1++) {
						name1 = res1[k1].getName();
						name1 = name1.substring(name1.indexOf('.'));
						val = false;
						for (int k2 = 0; k2 < res2.length; k2++) {
							name2 = res2[k2].getName();
							name2 = name2.substring(name2.indexOf('.'));
							val |= name1.compareTo(name2) == 0;
						}
						valid &= val;
					}
				}
//				}
			}

			return valid;
		}
	}

	/**
	 * Table model for compare setting table.
	 * 
	 * @author Werner Koppelsttter
	 */
	public class CompareModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public int getColumnCount() {
			return CompareColumn.values().length;
		}

		@Override
		public int getRowCount() {
			return PanelCompare.this._compare.size() + 1;
		}

		@Override
		public Object getValueAt(int row, int column) {
			CompareObject comp;
			Object obj = null;

			if (row == (this.getRowCount() - 1)) {
				comp = new CompareObject();
				comp.setColor(ColorSelect.values()[row % ColorSelect.values().length]);
			} else {
				comp = PanelCompare.this._compare.get(row);
			}

			if (column == CompareColumn.MOP.getValue()) {
				if (comp.getMopFile() != null) {
					obj = comp.getMopFile().getName();
				}
			} else if (column == CompareColumn.ARCHIVE.getValue()) {
				if (comp.getArchive() != null) {
					obj = comp.getArchive().getName();
				}
			} else if (column == CompareColumn.PROBLEM.getValue()) {
				if (comp.getProblem() != null) {
					obj = comp.getProblem().getName();
				}
			} else if (column == CompareColumn.COLOR.getValue()) {
				obj = comp.getColor();
			} else if (column == CompareColumn.VALID.getValue()) {
				obj = comp.isValid(PanelCompare.this._compare.getFirst().getProblem());
			} else {
				obj = comp;
			}

			return obj;
		}

		@Override
		public String getColumnName(int column) {
			for (CompareColumn e : CompareColumn.values()) {
				if (e.getValue() == column) {
					return e.getName();
				}
			}
			return null;
		}

		@Override
		public boolean isCellEditable(int row, int column) {
			boolean edit = false;

			if (row > 0) {
				if (column == CompareColumn.MOP.getValue()) {
					edit = true;
				} else if (column == CompareColumn.ARCHIVE.getValue()) {
					edit = PanelCompare.this._model.getValueAt(row, CompareColumn.MOP.getValue()) != null;
				} else if (column == CompareColumn.PROBLEM.getValue()) {
					edit = PanelCompare.this._model.getValueAt(row, CompareColumn.ARCHIVE.getValue()) != null;
				} else if (column == CompareColumn.COLOR.getValue()) {
					edit = true;
				}
			}

			return edit;
		}

		@Override
		public Class<?> getColumnClass(int column) {
			return CompareColumn.getColumn(column).getType();
		}

		@Override
		public void setValueAt(Object value, int row, int col) {
			CompareObject comp;

			if (row == (this.getRowCount() - 1)) {
				PanelCompare.this._compare.add((CompareObject) this.getValueAt(row, -1));
			}
			comp = PanelCompare.this._compare.get(row);

			if (col == CompareColumn.MOP.getValue()) {
				if (value == null) {
					comp.setMopFile(null);
					this.fireTableDataChanged();
				} else if (((String) value).compareTo("...") == 0) {
					JProjectChooser chooser = new JProjectChooser(JProjectChooser.PROJECT_ONLY, PanelCompare.this._views
							.getOptimization().getProblem().getArchive().getMop().getParentFile());
					int returnVal = chooser.showOpenDialog(PanelCompare.this);
					if (returnVal == JFileChooser.APPROVE_OPTION) {
						comp.setMopFile(new MopFile(chooser.getSelectedFile()));
						this.fireTableDataChanged();
					}
				}
			} else if (col == CompareColumn.ARCHIVE.getValue()) {
				try {
					if (value != null) {
						comp.setArchive(Archive.load(comp.getMopFile(), (String) value));
					} else {
						comp.setArchive(null);
					}
					this.fireTableDataChanged();
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else if (col == CompareColumn.PROBLEM.getValue()) {
				try {
					if (value != null) {
						comp.setProblem(comp.getArchive().readProblem((String) value));
					} else {
						comp.setProblem(null);
					}
					this.fireTableDataChanged();
				} catch (Exception e) {
					e.printStackTrace();
				}
			} else if (col == CompareColumn.COLOR.getValue()) {
				comp.setColor((ColorSelect) value);
				this.fireTableDataChanged();
			}
		}
	}

	/**
	 * Create a compare panel
	 * 
	 * @param views
	 *            Optimizer views object
	 */
	public PanelCompare(OptimizerViews views) {
		JTabbedPane tabbedPane;
		JPanel plots;
		Problem prob;
		JScrollPane scroll;
		CompareObject obj;
		LinkPlot link;

		this._views = views;

		obj = new CompareObject();
		obj.setProblem(this._views.getOptimization().getProblem());
		obj.setColor(ColorSelect.BLUE);
		this._compare.add(obj);

		this.setLayout(new BorderLayout(0, 0));

		tabbedPane = new JTabbedPane(SwingConstants.TOP);
		this.add(tabbedPane);

		scroll = new JScrollPane();
		tabbedPane.addTab("Settings", null, scroll, null);

		this._model = new CompareModel();

		this._table = new JTable(this._model);
		scroll.setViewportView(this._table);

		this._table.getTableHeader().setReorderingAllowed(false);
		this._table.setDefaultEditor(String.class, new DefaultCellEditor(new JComboBox()) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row,
					int column) {
				DefaultComboBoxModel model = ((DefaultComboBoxModel) ((JComboBox) this.getComponent()).getModel());
				model.removeAllElements();
				model.addElement(null);
				if (column == CompareColumn.MOP.getValue()) {
					if (value != null) {
						model.addElement(value);
					}
					model.addElement("...");
				} else if (column == CompareColumn.ARCHIVE.getValue()) {
					MopFile mop = ((CompareObject) PanelCompare.this._model.getValueAt(row, -1)).getMopFile();
					if (mop != null) {
						try {
							for (String name : Archive.listArchives(mop)) {
								model.addElement(name);
							}
						} catch (Exception e) {
						}
					}
				} else if (column == CompareColumn.PROBLEM.getValue()) {
					Archive arch = ((CompareObject) PanelCompare.this._model.getValueAt(row, -1)).getArchive();
					if (arch != null) {
						for (String name : arch.listProblems()) {
							model.addElement(name);
						}
					}
				}
				return super.getTableCellEditorComponent(table, value, isSelected, row, column);
			}

		});
		this._table.setDefaultRenderer(String.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
				if (column == CompareColumn.MOP.getValue()) {
					MopFile mop = ((CompareObject) PanelCompare.this._model.getValueAt(row, -1)).getMopFile();
					if (mop != null) {
						this.setText(mop.getName());
						this.setToolTipText(mop.getPath());
					} else {
						this.setText(null);
						this.setToolTipText(null);
					}
				} else {
					this.setToolTipText(this.getText());
				}

				return this;
			}

		});
		this._table.setDefaultEditor(ColorSelect.class, new DefaultCellEditor(new JComboBox(ColorSelect.values())));
		this._table.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				Job.refreshList();
				try {
					switch (e.getKeyCode()) {
					case KeyEvent.VK_DELETE:
						int row = PanelCompare.this._table.getSelectedRow();
						if (row > 0) {
							PanelCompare.this._compare.remove(row);
							PanelCompare.this._model.fireTableDataChanged();
						}
						break;
					}
				} catch (Exception e1) {
					// e1.printStackTrace();
				}
			}
		});

		// min column width
		this._table.getColumnModel().getColumn(CompareColumn.MOP.getValue()).setMinWidth(50);
		this._table.getColumnModel().getColumn(CompareColumn.ARCHIVE.getValue()).setMinWidth(50);
		this._table.getColumnModel().getColumn(CompareColumn.PROBLEM.getValue()).setMinWidth(50);
		this._table.getColumnModel().getColumn(CompareColumn.VALID.getValue()).setMinWidth(40);
		this._table.getColumnModel().getColumn(CompareColumn.COLOR.getValue()).setMinWidth(50);
		// preferred column width
		this._table.getColumnModel().getColumn(CompareColumn.VALID.getValue()).setMaxWidth(60);
		this._table.getColumnModel().getColumn(CompareColumn.COLOR.getValue()).setMaxWidth(80);

		plots = new JPanel();
		plots.setLayout(new BorderLayout());
		this._sub = new SubPlot(1, 1);
		plots.add(this._sub.getPanel(), BorderLayout.CENTER);
		tabbedPane.addTab("Plots", null, plots, null);

		prob = this._views.getOptimization().getProblem();
		try {
			PanelPlots.initObjectives(this._sub, prob);
			this.refreshPlot();
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		for (int r = 0; r < this._sub.getSize()[0]; r++) {
			for (int c = 0; c < this._sub.getSize()[1]; c++) {
				if (this._sub.getPlot(r, c) != null) {
					this._sub.getPlot(r, c).setHold(HOLD.ALL);
				}
			}
		}
		link = new LinkPlot();
		link.addPlot(this._sub);

		tabbedPane.addChangeListener(new ChangeListener() {
			@Override
			public void stateChanged(ChangeEvent event) {
				if (((JTabbedPane) event.getSource()).getSelectedIndex() > 0) {
					PanelCompare.this.refreshPlot();
				}
			}
		});
	}

	/**
	 * Refresh plots
	 */
	public void refreshPlot() {
		LinkedList<Individual[]> paretos;
		LinkedList<Color> colors;
		Problem baseProb, prob;

		baseProb = PanelCompare.this._compare.getFirst().getProblem();
		paretos = new LinkedList<Individual[]>();
		colors = new LinkedList<Color>();

		// get pareto-fronts
		for (CompareObject comp : PanelCompare.this._compare) {
			if (comp.isValid(baseProb)) {
				prob = comp.getProblem();
				colors.add(comp.getColor().getColor());
				if (prob.getNumSavePoints() > 0) {
					paretos.add(prob.getSavePoint(prob.getNumSavePoints() - 1).getFullPareto());
				} else {
					paretos.add(new Individual[0]);
				}
			}
		}

		// plot paretos
		try {
			// set color-order
			Color[] col = colors.toArray(new Color[0]);
			SubPlot sub = PanelCompare.this._sub;
			for (int r = 0; r < sub.getSize()[0]; r++) {
				for (int c = 0; c < sub.getSize()[1]; c++) {
					if (sub.getPlot(r, c) != null) {
						sub.getPlot(r, c).setColorOrder(col);
					}
				}
			}
			// plot
			PanelPlots.drawObjectives(sub, baseProb, paretos, -1, false);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}
