/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 247 $
 * $Date: 2020-03-19 10:13:52 +0100 (Thu, 19 Mar 2020) $
 */
package at.lcm.bu21.optim.gui.setting;

import java.awt.Color;
import java.awt.Component;

import javax.swing.DefaultCellEditor;
import javax.swing.GroupLayout;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.border.LineBorder;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import at.lcm.bu21.magopt.structure.Parameter;
import at.lcm.bu21.optim.core.ObjectiveType;
import at.lcm.bu21.optim.core.Problem;
import at.lcm.bu21.optim.core.ResultDesc;

/**
 * Panel to change results descriptions
 * 
 * @author Gnther Weidenholzer
 * @author Werner Koppelsttter
 */
public class ChooseResults extends SettingsPanel {

	/** Serialization version number */
	private static final long serialVersionUID = 1L;

	/** Problem where results are defined */
	private Problem problem = null;

	/** Table model for results */
	private ResultTableModel model;

	/** List of available result parameters */
	private Parameter[] ress;

	/** Refresh panel */
	private PanelRefresh refresh = null;

	/** Panel enabled */
	private boolean enable = true;

	/** Table to configure results */
	private JTable tableResults;

	/**
	 * Definition of table-columns
	 * 
	 * @author Werner Koppelsttter
	 */
	private enum ResColumn {
		/** Name */
		NAME(0, "Name", String.class),
		/** Description */
		DESCRIPTION(1, "Description", String.class),
		/** Selected */
		SELECT(2, "Select", Boolean.class),
		/** Min limit */
		OBJECTIVE(3, "IsObjective", ObjectiveType.class),
		/** Min limit */
		MIN(4, "Min constraint", Double.class),
		/** Max limit */
		MAX(5, "Max constraint", Double.class);

		/** Column number */
		private final int column;

		/** Column name */
		private final String name;

		/** Type of column object */
		private final Class<?> type;

		/**
		 * Create enumeration
		 * 
		 * @param column
		 *            Column number
		 * @param name
		 *            Column name
		 * @param tpye
		 *            Column type
		 */
		private ResColumn(int column, String name, Class<?> tpye) {
			this.column = column;
			this.name = name;
			this.type = tpye;
		}

		/**
		 * Get column number
		 * 
		 * @return Number
		 */
		int getValue() {
			return this.column;
		}

		/**
		 * Get column name
		 * 
		 * @return Name
		 */
		String getName() {
			return this.name;
		}

		/**
		 * Get column type
		 * 
		 * @return Type
		 */
		Class<?> getType() {
			return this.type;
		}

		/**
		 * Get element of enumeration with to the index
		 * 
		 * @param index
		 *            Index of column
		 * @return Enumeration element
		 */
		static ResColumn getResColumn(int index) {
			ResColumn res = null;

			for (ResColumn r : ResColumn.values()) {
				if (r.getValue() == index) {
					res = r;
				}
			}

			return res;
		}

	}

	/**
	 * Table model for choosing results
	 * 
	 * @author Werner Koppelsttter
	 */
	private class ResultTableModel extends AbstractTableModel {

		/** Serialization version number */
		private static final long serialVersionUID = 1L;

		@Override
		public String getColumnName(int column) {
			return ResColumn.getResColumn(column).getName();
		}

		@Override
		public Class<?> getColumnClass(int column) {
			return ResColumn.getResColumn(column).getType();
		}

		@Override
		public int getColumnCount() {
			return ResColumn.values().length;
		}

		@Override
		public int getRowCount() {
			return ChooseResults.this.ress.length;
		}

		@Override
		public boolean isCellEditable(int row, int column) {
			boolean edit;

			switch (ResColumn.getResColumn(column)) {
			case SELECT:
				edit = !ChooseResults.this.problem.getArchive().isLocked();
				break;
			case OBJECTIVE:
			case MIN:
			case MAX:
				edit = (ChooseResults.this.getResDesc(ChooseResults.this.ress[row].getIdentifier()) != null);
				break;
			default:
				edit = false;
				break;
			}

			edit &= ChooseResults.this.enable;

			return edit;
		}

		@Override
		public Object getValueAt(int row, int column) {
			Object obj = null;
			Parameter param;
			ResultDesc resDesc = null;

			param = ChooseResults.this.ress[row];
			resDesc = ChooseResults.this.getResDesc(param.getIdentifier());

			switch (ResColumn.getResColumn(column)) {
			case NAME:
				obj = param.getIdentifier();
				break;
			case DESCRIPTION:
				obj = param.getDescription();
				break;
			case SELECT:
				obj = (resDesc != null);
				break;
			case OBJECTIVE:
				if (resDesc != null) {
					obj = resDesc.getObjective();
				}
				break;
			case MIN:
				if ((resDesc != null) && !Double.isNaN(resDesc.getLB())) {
					obj = resDesc.getLB();
				}
				break;
			case MAX:
				if ((resDesc != null) && !Double.isNaN(resDesc.getUB())) {
					obj = resDesc.getUB();
				}
				break;
			default:
				obj = param;
				break;
			}

			return obj;
		}

		@Override
		public void setValueAt(Object aValue, int row, int column) {
			ResultDesc resDesc;
			String name;
			double value;

			name = ChooseResults.this.ress[row].getIdentifier();
			resDesc = ChooseResults.this.getResDesc(name);

			switch (ResColumn.getResColumn(column)) {
			case SELECT:
				if (!ChooseResults.this.problem.getArchive().isLocked()) {
					if ((Boolean) aValue) {
						ChooseResults.this.problem.addResult(name);
					} else {
						ChooseResults.this.problem.rmResult(name);
					}
				}
				break;
			case OBJECTIVE:
				if (resDesc != null) {
					if (resDesc.getObjective() != (ObjectiveType) aValue) {
						ChooseResults.this.problem.clearMapping();
					}
					resDesc.setObjective((ObjectiveType) aValue);
				}
				break;
			case MIN:
				if (resDesc != null) {
					if (aValue == null) {
						value = Double.NaN;
					} else {
						value = (Double) aValue;
					}
					resDesc.setBounds(value, resDesc.getUB());
				}
				break;
			case MAX:
				if (resDesc != null) {
					if (aValue == null) {
						value = Double.NaN;
					} else {
						value = (Double) aValue;
					}
					resDesc.setBounds(resDesc.getLB(), value);
				}
				break;
			default:
				break;
			}

			this.fireTableDataChanged();

			// Select cell for checkbox editor
			if ((ResColumn.getResColumn(column) == ResColumn.SELECT)
					|| (ResColumn.getResColumn(column) == ResColumn.OBJECTIVE)) {
				ChooseResults.this.tableResults.setRowSelectionInterval(row, row);
				ChooseResults.this.tableResults.setColumnSelectionInterval(column, column);
			}

		}
	}

	/**
	 * Validate problem
	 * 
	 * @param prob
	 *            Problem
	 * 
	 * @return Message if not valid
	 */
	public static String validate(Problem prob) {
		String message = null;

		if (prob.getNumObjectives() < 1) {
			message = "Select at least one objective!";
		} else {
			for (ResultDesc res : prob.getResults()) {
				if (!res.isValid()) {
					message = "Please complete settings";
				}
			}
		}

		return message;
	}

	/**
	 * Creates new form choose-results
	 * 
	 * @param problem
	 *            Problem to set results
	 * @param proj
	 *            Project to problem
	 * @param refresh
	 *            Refresh panel
	 */
	public ChooseResults(Problem problem, PanelRefresh refresh) {
		this(problem);
		this.refresh = refresh;
	}

	/**
	 * Create panel. Used by WindowsBuilder
	 */
	private ChooseResults() {
		this.model = new ResultTableModel();

		this.initComponents();
	}

	/**
	 * Creates new form choose-results
	 * 
	 * @param problem
	 *            Problem to set results
	 */
	public ChooseResults(Problem problem) {
		this();
		this.problem = problem;

		if (problem.getArchive().getProject() != null) {
			this.ress = problem.getArchive().getProject().getDynamicOutputs();
		} else {
			this.ress = new Parameter[0];
		}

		this.tableResults.setDefaultRenderer(Double.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			{
				this.setHorizontalAlignment(SwingConstants.RIGHT);
			}

			@Override
			public void setValue(Object value) {
				this.setText((value == null) ? "" : ((Double) value).toString());
			}

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {
				ResultDesc res;

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);

				this.setToolTipText(this.getText());

				if (isSelected) {
					table.getSelectionBackground();
				} else {
					switch (ResColumn.getResColumn(column)) {
					case MIN:
					case MAX:
						res = ChooseResults.this.getResDesc(ChooseResults.this.ress[row].getIdentifier());
						if ((res != null) && !res.isValid()) {
							this.setBackground(Color.RED);
						} else {
							this.setBackground(table.getBackground());
						}
						break;
					default:
						this.setBackground(table.getBackground());
						break;
					}
				}

				return this;
			}

		});
		this.tableResults.setDefaultRenderer(String.class, new DefaultTableCellRenderer() {
			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			@Override
			public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected,
					boolean hasFocus, int row, int column) {

				super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
				this.setToolTipText(this.getText());

				return this;
			}
		});

		JTextField tf = new JTextField();
		tf.setHorizontalAlignment(SwingConstants.RIGHT);
		this.tableResults.setDefaultEditor(Number.class, new DefaultCellEditor(tf) {

			/** Serialization version number */
			private static final long serialVersionUID = 1L;

			/** Value of cell */
			Double value;

			@Override
			public boolean stopCellEditing() {
				String s = (String) super.getCellEditorValue();

				if ((s != null) && !s.isEmpty()) {
					try {
						this.value = Double.parseDouble(s);
					} catch (Exception e) {
						((JComponent) this.getComponent()).setBorder(new LineBorder(Color.red));
						return false;
					}
				} else {
					this.value = null;
				}

				return super.stopCellEditing();
			}

			@Override
			public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row,
					int column) {
				JTextField tf;
				this.value = null;
				((JComponent) this.getComponent()).setBorder(new LineBorder(Color.black));
				tf = (JTextField) super.getTableCellEditorComponent(table, value, isSelected, row, column);
				tf.selectAll();

				return tf;
			}

			@Override
			public Object getCellEditorValue() {
				return this.value;
			}

		});
		this.tableResults.setDefaultEditor(ObjectiveType.class,
				new DefaultCellEditor(new JComboBox<ObjectiveType>(ObjectiveType.values())));

		this.tableResults.getColumnModel().getColumn(0).setMinWidth(250);
		this.tableResults.getColumnModel().getColumn(1).setMinWidth(150);
		this.tableResults.getModel().addTableModelListener(new TableModelListener() {

			@Override
			public void tableChanged(TableModelEvent e) {
				ChooseResults.this.refresh();
			}
		});

		this.refreshEnabled();
	}

	/**
	 * Refresh panel
	 */
	protected void refresh() {
		if (this.refresh != null) {
			this.refresh.refresh();
		}
	}

	@Override
	public String validateContents() {
		return ChooseResults.validate(this.getProblem());
	}

	/**
	 * Are settings enabled
	 * 
	 * @return Setting enable
	 */
	public boolean isSettingEnabled() {
		return this.enable;
	}

	/**
	 * Set settings enabled
	 * 
	 * @param enable
	 *            Settings enable
	 */
	public void setSettingEnabled(boolean enable) {
		this.enable = enable;
		this.refreshEnabled();
	}

	/**
	 * Refresh enable state of components
	 */
	private void refreshEnabled() {
		this.tableResults.setEnabled(this.enable && !this.problem.isLocked());
	}

	/**
	 * Get result-description to the name
	 * 
	 * @param name
	 *            Name of result
	 * @return Result-description
	 */
	private ResultDesc getResDesc(String name) {
		ResultDesc resDesc = null;

		for (ResultDesc res : this.problem.getResults()) {
			if (res.getName().compareTo(name) == 0) {
				resDesc = res;
			}
		}

		return resDesc;
	}

	/**
	 * Get problem of panel
	 * 
	 * @return Problem
	 */
	public Problem getProblem() {
		return this.problem;
	}

	/**
	 * Initialize components
	 */
	private void initComponents() {
		JScrollPane jScrollResults;

		// table
		this.tableResults = new JTable(this.model);
		this.tableResults.setColumnSelectionAllowed(true);
		this.tableResults.getTableHeader().setReorderingAllowed(false);
		jScrollResults = new JScrollPane();
		jScrollResults.setViewportView(this.tableResults);
		this.tableResults.getColumnModel().getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

		// layout
		GroupLayout layout = new GroupLayout(this);
		this.setLayout(layout);
		layout.setHorizontalGroup(layout.createSequentialGroup().addContainerGap()
				.addComponent(jScrollResults, GroupLayout.DEFAULT_SIZE, 724, Short.MAX_VALUE).addContainerGap());
		layout.setVerticalGroup(layout.createSequentialGroup().addContainerGap()
				.addComponent(jScrollResults, GroupLayout.DEFAULT_SIZE, 525, Short.MAX_VALUE).addContainerGap());
	}
}
