# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10062 $
$LastChangedDate:: 2020-11-20 13:22:02 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import ast
import pandas as pd
from symspace.console.script_utils import ScriptUtils
from symspace.generic_gui_qt import GUIFormThread
from symspace.wrapper.symoptutils import SyMOptUtils
from symspace.wrapper.symoptwrapper import SymArchive, SymProblem


def main(problem: SymProblem):
    default_project_file = problem.getArchive().getProject().getFile(get_abs_path=True)
    default_excel_file = os.path.join(os.path.dirname(default_project_file), 'pareto.xlsx')
    var_names = problem.getArchive().getVariables()
    individual_types = ['all', 'valid', 'no_error', 'error', 'pareto']

    form_title = 'Add individuals to buffer'
    form_elements = [
        ('Select Option', 'combobox', ['From Excel File', 'From Optimization Problem', 'From Custom Input']),
        ('\nFrom Excel File', 'label'),
        ('Select Exel file', 'browse_file', default_excel_file, 'Excel Files (*.xlsx *.xls)'),
        ('Map Short Identifier', 'checkbox', True),
        ('\nFrom Optimization Problem', 'label'),
        ('Select project file', 'browse_file', default_project_file, 'SyMSpace Files (*.mop)'),
        ('Individual Type', 'combobox', individual_types),
        ('Problem Selection in the next step', 'label'),
        ('\nFrom Custom Input', 'label'),
    ]
    for var_name in var_names:
        form_elements.append((var_name, 'line', '[]'))
    form = GUIFormThread(form_title, form_elements)

    form.start()
    if form.accept:
        option = form.result['Select Option']
        excel_file = form.result['Select Exel file']
        map_short_ids = form.result['Map Short Identifier']
        project_file_other = form.result['Select project file']
        individual_type = form.result['Individual Type']
        var_data = dict()
        for var_name in var_names:
            var_data[var_name] = ast.literal_eval(form.result[var_name])

        if option == 'From Excel File' and os.path.isfile(excel_file):
            addIndividualsFromExcel(problem, excel_file, map_short_ids)
        elif option == 'From Excel File' and not os.path.isfile(excel_file):
            raise FileNotFoundError('Selected file does not exist: "{}"'.format(excel_file))
        elif option == 'From Optimization Problem' and os.path.isfile(project_file_other):
            addIndividualsFromProblem(problem, project_file_other, individual_type)
        elif option == 'From Optimization Problem' and not os.path.isfile(project_file_other):
            raise FileNotFoundError('Selected file does not exist: "{}"'.format(project_file_other))
        elif option == 'From Custom Input':
            addIndividualsFromCustomInput(problem, var_data)
        else:
            raise NotImplementedError('Invalid Option: {}'.format(option))
    else:
        print('Cancelled by user')


def addIndividualsFromProblem(problem: SymProblem, project_file_other: str, individual_type: str):
    """
    Adds individuals of another optimization problem to the buffer

    Parameters
    ----------
    problem: SymProblem
        Current optimization problem
    project_file_other: str
        Project file to import problem from
    individual_type: str
        Type of individuals to use
    """
    archive_problem_list = ScriptUtils.getProblemsList(project_file_other)

    form_title = 'Select Problem'
    form_elements = [
        ('Selected project file', 'label', project_file_other),
        ('Archive//Problem Name', 'combobox', archive_problem_list),
    ]
    form = GUIFormThread(form_title, form_elements)
    form.start()

    if form.accept:
        archive_problem_other = form.result['Archive//Problem Name']

        archive_name_other, problem_name_other = ScriptUtils.parseArchiveProblemName(archive_problem_other)
        archive_other = SymArchive.load(project_file_other, archive_name_other)
        problem_other = archive_other.readProblem(problem_name_other)
        SyMOptUtils.addExistingIndividuals2Buffer(problem, problem_other, individual_type)

        print('Update finished')
    else:
        print('Cancelled by user')


def addIndividualsFromExcel(problem: SymProblem, excel_file: str, map_ids: bool):
    """
    Adds individuals of excel file to the buffer

    Parameters
    ----------
    problem: SymProblem
        Current optimization problem
    excel_file: str
        Excel file to import from e.g. a Pareto-Front excel export
    map_ids: bool
        Map short parameter identifiers (used in Pareto-Front excel export) to the full parameter identifier
    """
    archive = problem.getArchive()
    df = pd.read_excel(excel_file)
    df = SyMOptUtils.renameDataFrame(archive, df, map_ids)

    SyMOptUtils.addCustomIndividuals2Buffer(problem, df)
    print('Update finished')


def addIndividualsFromCustomInput(problem: SymProblem, variable_data: dict):
    """
    Adds individuals from custom input data to the buffer

    Parameters
    ----------
    problem: SymProblem
        Current optimization problem
    variable_data: dict
        Dictionary with data of the individual variables
    """
    df = pd.DataFrame.from_dict(variable_data)
    SyMOptUtils.addCustomIndividuals2Buffer(problem, df)
    print('Update finished')


if __name__ == '__main__':
    """
    This script adds individuals to the current optimization buffer
    
    1. Option: from an Excel file (e.g. Pareto-Front excel export)
    2. Option: from an existing optimization problem
    3. Option: from custom input (e.g Variables: a=[1, 2, 3], b=[4, 4, 4]

    This script needs to be executed in the SyMSpace-Optimizer Python Console.
    """
    _params = locals()
    _problem = ScriptUtils.getProblem(_params)

    if _problem is not None:
        ScriptUtils.run(main, _problem)
    else:
        ScriptUtils.showNoProblemWarning()
