# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10062 $
$LastChangedDate:: 2020-11-20 13:22:02 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import numpy as np
import matplotlib.pyplot as plt
from symspace.wrapper.symoptwrapper import SymArchive, SymProblem
from symspace.wrapper.symplot import SymSubPlot, SymPlot2Scatter, SymPlot2Line
from symspace.console.script_utils import ScriptUtils
from symspace.generic_gui_qt import GUIFormThread

try:
    '''latest symspace version only'''
    from symspace.plotting.matplot import FigureMPL
except ImportError:
    FigureMPL = None


def plotFromProject(prj_file: str, a_name: str, p_name: str, java_plot: bool):
    archive = SymArchive.load(prj_file, a_name)
    problem = archive.readProblem(p_name)
    plotFromProblem(problem, java_plot)


def plotMPL(d: dict, objectives_names: list, objectives_type: list, plot_optimum: bool):
    """plot with matplotlib"""
    fig = plt.Figure()
    n_rows = len(objectives_names)
    n_cols = 1  # adjust plot columns
    for index, key in enumerate(objectives_names):
        ylabel = '{} ({})'.format(key.split('.')[-1], objectives_type[index])
        ax = fig.add_subplot(n_rows, n_cols, 1 + index)
        x = d[key][0]
        y = d[key][1]

        if objectives_type[index] == 'MIN':
            y_flat = []
            x_flat = []
            for sp_index in range(len(x)):
                y_flat.append(np.min(y[sp_index]))
                x_flat.append(sp_index)
            ax.plot(x_flat, y_flat, 'r', marker='o')
        elif objectives_type[index] == 'MAX':
            y_flat = []
            x_flat = []
            for sp_index in range(len(x)):
                y_flat.append(np.max(y[sp_index]))
                x_flat.append(sp_index)
            ax.plot(x_flat, y_flat, 'r', marker='o')

        if not plot_optimum:
            y_flat = [item for sublist in y for item in sublist]
            x_flat = [item for sublist in x for item in sublist]
            ax.scatter(x_flat, y_flat)

        ax.set_xlabel('Savepoints')
        ax.set_ylabel(ylabel)
        ax.grid(True)

    fig_mpl = FigureMPL(fig)
    fig_mpl.showQT()


def plotJava(d: dict, objectives_names: list, objectives_type: list, plot_optimum: bool):
    """plot with java plot"""
    n_rows = len(objectives_names)
    sym_subplot = SymSubPlot(n_rows, 1)
    for index, key in enumerate(objectives_names):
        ylabel = '{} ({})'.format(key.split('.')[-1], objectives_type[index])
        x = d[key][0]
        y = d[key][1]

        if plot_optimum and objectives_type[index] == 'MIN':
            y_flat = []
            x_flat = []
            for sp_index in range(len(x)):
                y_flat.append(np.min(y[sp_index]))
                x_flat.append(sp_index)
            sym_subplot[index, 0] = SymPlot2Line()
            sym_subplot[index, 0].setXLimit(0, len(x_flat))
        elif plot_optimum and objectives_type[index] == 'MAX':
            y_flat = []
            x_flat = []
            for sp_index in range(len(x)):
                y_flat.append(np.max(y[sp_index]))
                x_flat.append(sp_index)

            sym_subplot[index, 0] = SymPlot2Line()
            sym_subplot[index, 0].setXLimit(0, len(x_flat))
        else:
            y_flat = [item for sublist in y for item in sublist]
            x_flat = [item for sublist in x for item in sublist]
            sym_subplot[index, 0] = SymPlot2Scatter()

        sym_subplot[index, 0].addSignal('', x_flat, y_flat)
        sym_subplot[index, 0].setXLabel('Savepoints')
        sym_subplot[index, 0].setYLabel(ylabel)

    sym_subplot.show()


def plotFromProblem(problem: SymProblem, java_plot: bool):
    local_pareto = False  # set full or local pareto
    plot_optimum = False  # plot only Min or Max of each Savepoint

    objectives_desc = problem.getObjectives()

    objectives_names = [desc.getName() for desc in objectives_desc]
    objectives_type = [str(desc.getObjective()).split('.')[-1] for desc in objectives_desc]

    d = {}
    for key in objectives_names:
        d[key] = [[], []]

    for sp_index in range(problem.getNumSavePoints()):
        # print('SavePoint[{:d}]'.format(sp_index))
        save_point = problem.getSavePoint(sp_index)
        if local_pareto:
            pareto = save_point.getLocalPareto()
        else:
            pareto = save_point.getFullPareto()
        for key in objectives_names:
            d[key][0].append([])
            d[key][1].append([])
        for i_index, individual in enumerate(pareto):
            # print('\tIndividual[{:d}]'.format(i_index))
            for name in objectives_names:
                objective_value = individual.getResult(name)
                # print('\t\t', name, objective_value)
                d[name][0][sp_index].append(sp_index)
                d[name][1][sp_index].append(objective_value)

    if java_plot:
        plotJava(d, objectives_names, objectives_type, plot_optimum)
    else:
        plotMPL(d, objectives_names, objectives_type, plot_optimum)


def main(params: dict):
    project = ScriptUtils.selectProjectFromParams(params)
    if project is not None:
        prj_file = project.getFile(get_abs_path=True)
        archive_problem_list = ScriptUtils.getProblemsList(prj_file)

        form_title = 'Plot Objectives vs. SavePoints'
        form_elements = [
            ('Archive//Problem Name', 'combobox', archive_problem_list),
            ('Plotting Framework', 'combobox', ['Java', 'Matplotlib'])
        ]
        form = GUIFormThread(form_title, form_elements)

        form.start()
        if form.accept:
            archive_problem = form.result['Archive//Problem Name']
            archive_name, problem_name = ScriptUtils.parseArchiveProblemName(archive_problem)
            plot_framework = form.result['Plotting Framework']

            java_plot = True if plot_framework.lower() == 'java' else False

            plotFromProject(prj_file, archive_name, problem_name, java_plot)
            print('Update finished')
        else:
            print('Cancelled by user')


if __name__ == '__main__':
    """
    This script plots Objectives vs. SavePoints

    This script needs to be executed in the SyMSpace Python Console.
    """
    _params = locals()
    ScriptUtils.run(main, _params)

