#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10062 $
$LastChangedDate:: 2020-11-20 13:22:02 +0100#$

This file is part of SyMSpace.

Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import sys
import warnings
import subprocess
from symspace.wrapper.symwrapper import SymProject, SymParameter
from symspace.wrapper.symoptwrapper import SymProblem, SymArchive
from symspace.generic_gui_qt import GUIMessageThread, GUIFormThread


class ScriptUtils:
    """Helper class to run scripts in the symspace console"""

    @staticmethod
    def getProject(params: dict, manual_assigned: bool = False) -> SymProject:
        """
        Get the project variable in the params dictionary

        Parameters
        ----------
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used

        Returns
        -------
        project : SymProject
        """
        if '__project__' in params and not manual_assigned:
            '''run macro'''
            project = params['__project__']
        elif 'CurProject' in params:
            '''manual assignment in symspace center'''
            project = params['CurProject']
        elif 'OptProject' in params:
            '''manual assignment in optimizer'''
            project = params['OptProject']
        else:
            project = None
        return project

    @staticmethod
    def getParameter(params: dict, manual_assigned: bool = False) -> SymParameter:
        """
        Get the parameter variable in the params dictionary

        Parameters
        ----------
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used

        Returns
        -------
        parameter : SymParameter
        """
        if '__parameter__' in params and not manual_assigned:
            '''run macro'''
            parameter = params['__parameter__']
        elif 'CurProject' in params:
            '''manual assignment in symspace center'''
            parameter = params['CurObject']
        else:
            parameter = None
        return parameter

    @staticmethod
    def getProblem(params: dict, manual_assigned: bool = False) -> SymProblem:
        """
        Get the problem variable in the params dictionary

        Parameters
        ----------
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used

        Returns
        -------
        problem : SymProblem
        """
        if '__problem__' in params and not manual_assigned:
            '''run macro'''
            problem = params['__problem__']
        elif 'Problem' in params:
            '''manual assignment in optimizer'''
            problem = params['Problem']
        else:
            problem = None
        return problem

    @staticmethod
    def getController(params: dict, manual_assigned: bool = False) -> 'jep.PyObject':
        """
        Get the controller variable in the params dictionary

        Parameters
        ----------
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used

        Returns
        -------
        controller : jep.PyObject
        """
        if '__controller__' in params and not manual_assigned:
            '''run macro'''
            controller = params['__controller__']
        elif 'Problem' in params:
            '''manual assignment in optimizer'''
            controller = params['Controller']
        else:
            controller = None
        return controller

    @staticmethod
    def setProject(value: SymProject, params: dict, manual_assigned: bool = False):
        """
        Set the project variable in the params dictionary

        Parameters
        ----------
        value :SymProject
            Value to use
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used
        """
        if manual_assigned:
            params['CurProject'] = value  # for symspace center
            params['OptProject'] = value  # for optimizer
        else:
            params['__project__'] = value

    @staticmethod
    def setParameter(value: SymParameter, params: dict, manual_assigned: bool = False):
        """
        Set the parameter variable in the params dictionary

        Parameters
        ----------
        value :SymParameter
            Value to use
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used
        """
        if manual_assigned:
            params['CurObject'] = value
        else:
            params['__parameter__'] = value

    @staticmethod
    def setProblem(value: SymProblem, params: dict, manual_assigned: bool = False):
        """
        Set the problem variable in the params dictionary

        Parameters
        ----------
        value :SymProblem
            Value to use
        params: dict
            Dictionary which usually contains all local variables
        manual_assigned: bool
            Select if the manual assignment or macro assignment names should be used
        """
        if manual_assigned:
            params['Problem'] = value
        else:
            params['__problem__'] = value

    @classmethod
    def run(cls, fnc_handle, *args, **kwargs) -> None:
        """
        Run a function in the symspace console with standardized error handling.

        Examples
        --------

        >>> from symspace.console.script_utils import ScriptUtils
        >>> def main(*args, **kwargs): print('Do sth', args, kwargs)
        >>> ScriptUtils.run(main)
        >>> ScriptUtils.run(main, 1, 2, a='a', b='b')

        """
        try:
            fnc_handle(*args, **kwargs)

        except Exception as ex:
            cls.showError(str(ex))
            raise ex

    @staticmethod
    def showNoProjectWarning():
        """Show the standard warning in console and gui"""
        warnings.warn('Run this script via "Run macro (Ctrl+M)"!')
        msg_elements = {'icon': 'warning',
                        'text': 'No SyMSpace project found!\n\n'
                                'Open a SyMSpace project and run this script via Run macro (Ctrl+M)!',
                        'buttons': 'ok'}
        msg = GUIMessageThread('Warning', msg_elements)
        msg.start()
        print('Script finished with error')

    @staticmethod
    def showNoProblemWarning():
        """Show the standard warning in console and gui"""
        warnings.warn('Run this script via "Run macro (Ctrl+M)"!')
        msg_elements = {'icon': 'warning',
                        'text': 'No SyMSpace optimization problem found!\n\n '
                                'Open a SyMSpace optimization problem and run this script via Run macro (Ctrl+M)!',
                        'buttons': 'ok'}
        msg = GUIMessageThread('Warning', msg_elements)
        msg.start()
        print('Script finished with error')

    @staticmethod
    def showError(text: str):
        """
        Show a specific warning in console and gui

        Parameters
        ----------
        text : str
            error message
        """
        warnings.warn(text)
        msg_elements = {'icon': 'critical', 'text': text, 'buttons': 'ok'}
        msg = GUIMessageThread('Error', msg_elements)
        msg.start()

        print('Script finished with error')

    @classmethod
    def selectProjects(cls, default_path: str = None, n_projects: int = 1) -> list:
        """
        Open a gui to select project files.

        Parameters
        ----------

        n_projects : int
        default_path : str
            Default path of the project file

        Returns
        -------
        project_list : list
            list of projects
        """
        if default_path is None:
            default_path = ''
        form_title = 'SyMSpace file'
        form_elements = []

        for index in range(n_projects):
            form_elements.append(
                ('Select project file {}'.format(index + 1), 'browse_file', default_path, 'SyMSpace Files (*.mop)'))
        form = GUIFormThread(form_title, form_elements)

        form.start()

        project_list = []
        if form.accept:

            for index in range(n_projects):
                file_path = form.result['Select project file {}'.format(index + 1)]
                if file_path.endswith('.mop'):
                    project = SymProject.load(file_path)
                    project_list.append(project)
                else:
                    file_ending = file_path.split('.')[-1]
                    raise Exception('Filetype "{}" is not supported'.format(file_ending))
            # status = 0
        else:
            print('Cancelled by user')
            # status = 1
        return project_list

    @classmethod
    def selectProjectFromParams(cls, params: dict) -> SymProject:
        """
        Helper function to select a project

        Parameters
        ----------
        params: dict
            Dictionary which usually contains all local variables

        Returns
        -------
         selected_project : SymProject
        """
        project = cls.getProject(params)
        problem = cls.getProblem(params)

        if problem is not None:
            '''1. case: if called from an optimizer console with open problem
               Instead of __project__/OptProject the project of the problem will be used!
            '''
            project_path = problem.getArchive().getProject().getFile(get_abs_path=True)
        elif project is not None:
            '''2. case: if called from an symspace console with an open project'''
            project_path = project.getFile(get_abs_path=True)
        else:
            '''3. case: if called from an symspace or optimizer console without an open project/problem'''
            project_path = None
        project_list = cls.selectProjects(project_path)
        if len(tuple(project_list)) == 0:
            selected_project = None
        else:
            selected_project = project_list[0]
        return selected_project

    @staticmethod
    def getProblemsList(prj_file: str) -> list:
        """
        Get a list of all archive-problem names in an optimization project

        Parameters
        ----------
        prj_file : str
            Project file path

        Returns
        -------
        archive_problem_list : list
            List of strings representing the archive and problem name
        """
        archive_problem_list = []
        for a_name in SymArchive.listArchives(prj_file):
            archive = SymArchive.load(prj_file, a_name)
            for p_name in archive.listProblems():
                archive_problem_list.append('{}//{}'.format(a_name, p_name))
        if len(archive_problem_list) == 0:
            raise Exception('The selected project does not contain optimization problems!')
        return archive_problem_list

    @staticmethod
    def getObjectivesList(project: SymProject) -> list:
        """
        Get a list of all objectives in an optimization project

        Parameters
        ----------
        project: SymProject

        Returns
        -------
        objectives: list
            List of objective parameter short identifiers
        """
        objectives = []
        prj_file = project.getFile(get_abs_path=True)

        for a_name in SymArchive.listArchives(prj_file):
            archive = SymArchive.load(prj_file, a_name)

            for p_name in archive.listProblems():
                problem = archive.readProblem(p_name)
                objectives_list = [desc.getName() for desc in problem.getObjectives()]
                objectives.extend(objectives_list)

        objectives = sorted(set(objectives))
        for index, param_key in enumerate(objectives):
            objectives[index] = project.getObject(param_key).getShortIdentifier()
        return objectives

    @staticmethod
    def parseArchiveProblemName(archive_problem: str) -> tuple:
        """
        Parse the archive and problem name from a string

        Parameters
        ----------
        archive_problem: str
            String representing the archive and problem name

        Returns
        -------
        tuple
            archive_name: str
            problem_name: str
        """
        archive_name = archive_problem.split('//')[0]
        problem_name = archive_problem.split('//')[1]
        return archive_name, problem_name

    @staticmethod
    def openExplorer(file: str):
        """Open the function directory in the system file manager"""
        try:
            if sys.platform == "win32":
                '''Windows'''
                subprocess.call("explorer " + file)
            elif sys.platform == "linux" or sys.platform == "linux2":
                '''linux'''
                raise NotImplementedError
            elif sys.platform == "darwin":
                ''' OS X'''
                raise NotImplementedError
        except Exception as e:
            print(e)
            print('Unable to open dir')
