# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10096 $
$LastChangedDate:: 2021-01-21 16:08:58 +0100#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import shutil
import pandas as pd
import bokeh.plotting as bkh
import bokeh.models
import bokeh.transform as bkt

try:
    from symspace.wrapper.symwrapper import SymProject, SymFunction, SymField
except ImportError:
    print('SyMSpace import failed')

class IndividualsPlot:
    image_key = 'img'
    legend_key = 'Individuals'

    @classmethod
    def fromXLS(cls, xls_file: str, variables: tuple):
        df = pd.read_excel(xls_file)
        cwd = os.path.dirname(xls_file)
        return cls(df, variables, cwd)

    def __init__(self, df: pd.DataFrame, variables: tuple, cwd: str):
        self.df = df
        self.variables = variables
        self.cwd = cwd

    def toXLS(self, xls_file:str):
        self.df.to_excel(xls_file, index=False)

    def renameDF(self, d_rename:dict=None, remove_min_max: bool = True):
        if d_rename is None:
            d_rename = {}

        if remove_min_max:
            for key in self.df.keys():
                if key not in d_rename:
                    d_rename[key] = key.replace(' (MAX)', '').replace(' (MIN)', '')

        self.df.rename(columns=d_rename, inplace=True)

    def setProjectIndices(self, sym_file: str, delete_doc_files:bool =True):
        sym_prj = SymProject.load(sym_file)

        if delete_doc_files:
            for index in range(sym_prj.getLength()):
                sym_prj.setIndex(index)
                sym_prj.deleteDocumentationFiles()

        n_diff = self.df.shape[0] - sym_prj.getLength()
        if n_diff > 0:
            for i in range(n_diff):
                print('Adding index {:d}/{:d}'.format(i + 1, n_diff))
                sym_prj.duplicate()
        elif n_diff < 0:
            for i in range(abs(n_diff)):
                print('Removing index {:d}/{:d}'.format(i + 1, abs(n_diff)))
                sym_prj.reduce()

        param_ids = dict()
        for param in sym_prj.getDynamicParams():
            if isinstance(param, SymField) and not param.getReadOnly():

                for key in self.df.keys():
                    if param.getShortIdentifier() == key:
                        param_ids[key] = param.getIdentifier()

        for index, row in self.df.iterrows():
            sym_prj.setIndex(index)
            print('Index', index)
            for short_id, full_id in param_ids.items():
                val = row[short_id]
                param = sym_prj[full_id]
                param.setData(val)
                print('\tSet {} = {}'.format(short_id, val))

        sym_prj.save()

    def addDocFiles(self, sym_file: str, param_id: str):

        image_dir = os.path.join(self.cwd, 'Images')

        if not os.path.isdir(image_dir):
            os.makedirs(image_dir)

        sym_prj = SymProject.load(sym_file)

        if self.df.shape[0] != sym_prj.getLength():
            raise Exception('Number of individuals in the dataset is not equal to the project indices!')

        images = []

        for index in range(sym_prj.getLength()):
            print('Index:', index)
            sym_prj.setIndex(index)

            sym_fnc = sym_prj[param_id]  # type: SymFunction
            if len(sym_fnc.getDocFiles()) == 0:
                sym_fnc.generateDocumentationFiles()
            doc_file = sym_fnc.getDocFiles()[0]['file']

            if isinstance(doc_file, str) and os.path.isfile(doc_file):
                doc_file_copy = os.path.join(image_dir, 'Index_{:03d}.png'.format(index))
                shutil.copy2(doc_file, doc_file_copy)
                images.append(os.path.relpath(doc_file_copy, self.cwd))
            else:
                images.append('not available')

        self.df[self.image_key] = images


    def plotBKH(self, title: str, xy_pairs: tuple,
                        show_vars: bool = False, show_img: bool = False,
                        img_height: int = 200, img_width: int = 200,
                        fnc_legend = None) -> str:


        html_file = os.path.join(self.cwd, '{}.html'.format(title))

        data = dict()
        for key in self.df.keys():
            data[key] = self.df[key].values

        if fnc_legend is not None:
            fnc_legend(self.legend_key, data)

        cds = []
        if self.legend_key not in data:
            cds.append(bkh.ColumnDataSource(data))  # no groups
            color = "blue"
        else:
            groups = sorted(set(data[self.legend_key]), reverse=True)
            color = bkt.factor_cmap(self.legend_key, 'Category10_10', groups)

            dft1 = pd.DataFrame.from_dict(data)
            for group in groups:
                dft2 = dft1.loc[dft1[self.legend_key] == group] # type: pd.DataFrame
                cds.append(bkh.ColumnDataSource(dft2))


        hover = self.__getHoverBKH(img_width, show_vars, show_img)

        tools = ['pan', 'box_zoom', 'wheel_zoom', 'lasso_select', 'tap', 'reset', 'save', hover]

        figs = []
        for x_name, y_name in xy_pairs:
            fig = bkh.figure(tools=tools, x_axis_label=x_name, y_axis_label=y_name)
            for cd in cds:
                fig.circle(x_name, y_name, source=cd, legend=self.legend_key, size=10, color=color)

            fig.min_border_top = img_height // 2
            fig.min_border_bottom = img_height // 2
            fig.min_border_left = img_width // 2
            fig.min_border_right = img_width // 2
            fig.legend.click_policy = 'hide'  # or 'mute'
            fig.legend.location = 'bottom_right'

            figs.append(fig)

        ncols = 3
        grid = bkh.gridplot(figs, ncols=ncols)
        bkh.output_file(html_file)

        bkh.save(grid)
        bkh.show(grid)
        return html_file

    def __getHoverBKH(self, img_width: int, show_vars: bool, show_img: bool):

        if show_img:
            div_img = """
            <div>
                <img
                    src="@img" width="{}" alt="@img"
                    style="float: left; margin: 0px 15px 15px 0px;"
                    border="2"
                ></img>
            </div>
            """.format(img_width)
        else:
            div_img = ''

        div_vars = '<div>'
        if show_vars:
            for var_name in self.variables:
                div_vars += '<span style="float: right font-size: 17px; font-weight: bold;">{0}: @{0}</span><br />'.format(
                    var_name)
        div_vars += '</div>'

        div_loc = """
        <div>
            <span style="font-size: 15px;">Location</span>
            <span style="font-size: 10px; color: #696;">($x, $y)</span><br />
            <span style="font-size: 15px;">Index</span>
            <span style="font-size: 10px; color: #966;">[$index]</span>
        </div>
        """
        hover = bokeh.models.HoverTool(
            tooltips="""
                <div>{}{}{}
                </div>
                """.format(div_loc, div_img, div_vars)
        )
        # hover.attachment = 'vertical'
        # hover.anchor = 'bottom_right'
        return hover


def selectorC2(key: str, data: dict):
    data[key] = []
    for value in data['c2']:
        if value >= 50:
            data[key].append('c2>=50')
        elif 30 <= value < 50:
            data[key].append('c2>=30')
        elif 0 <= value < 30:
            data[key].append('c2>=0')
        else:
            data[key].append('c2>=?')



if __name__ == '__main__':
    _d_rename = {'Ram.lges': 'lges'}
    _variables = ('a1', 'a2', 'c1', 'c2', 'd1', 'd2', 'd3', 'l1', 'l2', 'lges')

    _xy_pairs = (

        ('ram_y_rigid_body_velocity_max', 'ram_kinetic_energy_max'),
        ('striker_kinetic_energy_max', 'ram_kinetic_energy_max'),
        ('rod_kinetic_energy_max', 'ram_kinetic_energy_max'),

    )

    # _wdir = r'C:\Users\kevin\Downloads\Optimzer'
    _wdir = r'C:\Users\kevin\Downloads\Optimizer_v2'


    _xls_file = os.path.join(_wdir, 'export_pareto.xlsx')
    _xls_file_2 = os.path.join(_wdir, 'export_pareto_2.xlsx')
    _sym_file = os.path.join(_wdir, 'Hilti_AP3_static_mesh_v3_Pareto.docFiles.mop')


    # _ip = IndividualsPlot.fromXLS(_xls_file, _variables)
    # _ip.renameDF(_d_rename)
    #
    # # _ip.setProjectIndices(_sym_file)
    # _ip.addDocFiles(_sym_file, 'Preprocessing.GenerateMesh.Ram')
    # _ip.toXLS(_xls_file_2)

    _ip = IndividualsPlot.fromXLS(_xls_file_2, _variables)
    _ip.plotBKH('Valid',_xy_pairs, fnc_legend=selectorC2, show_vars=True, show_img=True)
