# !/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision: 10184 $
$LastChangedDate:: 2021-05-05 14:10:10 +0200#$

This file is part of SyMSpace.

Copyright (c) 2020, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""
import os
import sys
import re
import shutil
import compileall
import importlib.util
# import logging
import importlib.machinery
from importlib.machinery import PathFinder
from symspace.pyutils.version_util import VersionUtil
from symspace.osutils.pathutils import Paths
# from symspace.utils.logger import LoggerClass


class ZipPackage:
    """Helper class to create and load zip-file python distributions"""

    file_extension = 'zip'
    regex_pattern = r"^(?P<pkg_name>\w+)-" \
                    r"(?P<pkg_version>[\d]+\.[\d]+\.[\d]+)-" \
                    r"(?P<pyVersion>[\w\.]+)\." \
                    r"(?P<file_extension>\w+)$"

    # def __init__(self, loglevel=logging.WARNING):
    #     super().__init__(loglevel)

    @classmethod
    def createFromPkgDir(cls, pkg_dir: str, target_dir: str = None, pkg_version: str = None,
                         compile_bytecode: bool = True, dry_run: bool = False) -> str:
        """
        Create a distribution from a package directory

        Parameters
        ----------
        pkg_dir: str
            Package directory
        target_dir: str
            Target directory for the created distribution file
        pkg_version: str
            Package version
        compile_bytecode: bool
            Compile py to pyc
        dry_run: bool
            Don't create a file (but return the file path, where it would be if it was created)

        Examples
        --------
        >>> zip_file = ZipPackage.createFromPkgDir('x:\\my_pkg')

        Returns
        -------
        file: str
            Zip-Distribution file
        """
        version_tag = 'py{}'.format(VersionUtil.getPythonVersion())
        pkg_name = os.path.basename(pkg_dir)
        parent_folder = os.path.dirname(pkg_dir)
        folder_name = os.path.basename(pkg_dir)

        if pkg_version is None:
            pkg_version = '0.0.0'
        if target_dir is None:
            try:
                target_dir = os.path.join(Paths.getSyMSpaceRoot(), 'py_dist')
            except Exception as e:
                print(e)
                target_dir = os.path.join(parent_folder, 'py_dist')

        target_path = os.path.join(target_dir, '{}-{}-{}'.format(pkg_name, pkg_version, version_tag))
        target_file = '{}.{}'.format(target_path, cls.file_extension)

        if dry_run:
            file = target_file
        else:
            if compile_bytecode:
                '''compile py to pyc
                Python will not attempt to modify the .zip archive by adding the corresponding .pyc file,
                meaning that if a ZIP archive doesn’t contain .pyc files, importing may be rather slow.
                '''
                success = compileall.compile_dir(pkg_dir, quiet=True)
                if not success:
                    raise Exception('Bytecode compilation failed for package: {}'.format(pkg_dir))

            '''create zip file from package'''
            file = shutil.make_archive(target_path, format=cls.file_extension, root_dir=parent_folder,
                                       base_dir=folder_name)
            if target_file != file:
                raise Exception('Unexpected error: {} != {}'.format(target_file, file))
            if not os.path.isfile(file):
                raise FileNotFoundError('File does not exist: {}'.format(file))
            if not os.path.getsize(file) > 0:
                raise FileExistsError('File is empty: {}'.format(file))
            print('File created:', file)

        return file

    @classmethod
    def isDistFile(cls, file: str, info=None):
        flag = False

        if file.endswith(cls.file_extension):
            '''extract package name and version'''
            match_obj = re.search(cls.regex_pattern, os.path.basename(file))
            if match_obj is not None:
                if isinstance(info, dict):
                    info['pkg_name'] = match_obj.group(1)
                    info['pkg_version'] = match_obj.group(2)
                    info['version_tag'] = match_obj.group(3)
                flag = True
        return flag

    @classmethod
    def getFilesFromDir(cls, directory: str) -> list:
        """
        Find zip-distribution files in a directory

        Parameters
        ----------
        directory: str
            Directory to search

        Examples
        --------
        >>> zip_files = ZipPackage.getFilesFromDir('x:\\dist')

        Returns
        -------
        include_files: list
            List of found zip-distribution files
        """
        include_files = []
        for file in os.listdir(directory):
            '''look for matching files in the directory'''
            if cls.isDistFile(file):
                file_path = os.path.join(directory, file)
                include_files.append(file_path)

        return include_files

    @classmethod
    def extendImportPath(cls, include_files: list) -> list:
        """
        Try to extend the python import path (=sys.path) with these files

        Parameters
        ----------
        include_files : list
            List of distribution files to use

        Examples
        --------
        >>> imported_files = ZipPackage.extendImportPath(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-1.0.0-py3.6.egg'])

        Returns
        -------
        files: list
            List of files, which have been added to the the python import path (=sys.path)
        """
        latest_pkgs = cls.findLatestPackages(include_files)
        files = cls._getImportableFiles(latest_pkgs)
        sys.path.extend(files)
        return files

    @classmethod
    def findLatestPackages(cls, include_files: list) -> dict:
        """
        Find the latest packages in the list of files

        Parameters
        ----------
        include_files : list
            List of distribution files to use

        Examples
        --------
        >>> latest_pkgs = ZipPackage.findLatestPackages(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-1.0.0-py3.6.egg'])

        Returns
        -------
        pkgs: dict
            Collection of latest-packages
        """
        pkgs = cls._findPackages(include_files)
        latest_pkgs = cls._getLatestPackages(pkgs)
        return latest_pkgs

    @classmethod
    def _findPackages(cls, files: list) -> dict:
        """
        Analyze zip-distribution files for package informations

        Parameters
        ----------
        files: list
            List of files to analyze

        Examples
        --------
        >>> zp = ZipPackage()
        >>> pkgs = zp._findPackages(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-2.0.0-py3.6.zip'])

        Returns
        -------
        pkgs :dict
            Collection of retrieved package information
        """
        pkgs = dict()
        for file_path in set(files):
            info = dict()
            if cls.isDistFile(file_path, info):
                pkg_name = info['pkg_name']
                pkg_version = info['pkg_version']
                version_tag = info['version_tag']
                pkg_py_version = cls._getPkgPyVersion(version_tag)

                spec = PathFinder.find_spec(pkg_name, path=[file_path, ])
                if spec is not None and spec.origin != '<unknown>':
                    if pkg_name not in pkgs:
                        pkgs[pkg_name] = dict(py_version=[], pkg_version=[], file=[])

                    pkgs[pkg_name]['py_version'].append(pkg_py_version)
                    pkgs[pkg_name]['pkg_version'].append(pkg_version)
                    pkgs[pkg_name]['file'].append(file_path)

        return pkgs

    @classmethod
    def _getImportableFiles(cls, latest_pkgs: dict) -> list:
        """
        Get importable distribution files form the collection of latest packages

        Parameters
        ----------
        latest_pkgs : dict
            Collection of latest distribution package information

        Examples
        --------
        >>> zp = ZipPackage()
        >>> latest_pkgs = zp.findLatestPackages(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-2.0.0-py3.6.zip'])
        >>> zip_files = zp._getImportableFiles(latest_pkgs)

        Returns
        -------
        files : list
            List of files, which could be added to the import path (=sys.path)
        """
        # py_version = VersionUtil.getPythonVersion()
        files = []
        for _pkg_name, _pkg in latest_pkgs.items():
            module_spec = importlib.util.find_spec(_pkg_name)

            if _pkg['file'] not in sys.path and module_spec is None:
                # if os.path.isdir(folder):
                #     '''check if the source package folder is in the same location as the zip file'''
                #     self.logger.warning(
                #         'Folder and {}-file of the same package "{}" detected in {}!'
                #             .format(self.file_extension, _pkg_name, self.directory))

                # if len(_pkg['pkg_version']) > 0 and isinstance(_pkg['pkg_version'], list):
                #     self.logger.warning(
                #         'Multiple {}-files of package "{}" detected.'
                #         ' The latest package version "{}" will be loaded!'
                #             .format(self.file_extension, _pkg_name, _pkg['pkg_version']))
                # if _pkg['py_version'] != py_version:
                #     self.logger.warning(
                #         'Package "{}" was created with python version "{}" '
                #         'and might be incompatible with the used version "{}"'
                #             .format(_pkg_name, _pkg['py_version'], py_version))

                files.append(_pkg['file'])
            # else:
            #     self.logger.debug('Already loaded package "{}" from "{}"'.format(_pkg_name, module_spec.origin))

        return files

    @staticmethod
    def _getLatestPackages(pkgs: dict) -> dict:
        """
        Get the lastest zip-distribution package from the pkgs dictionary

        Parameters
        ----------
        pkgs: dict
            Collection of packages

        Examples
        --------
        >>> zp = ZipPackage()
        >>> pkgs = zp._findPackages(['x:\\my_pkg-1.0.0-py3.6.zip', 'x:\\my_pkg-2.0.0-py3.6.zip'])
        >>> latest_pkgs = zp._getLatestPackages(pkgs)

        Returns
        -------
        pkgs: dict
            Collection of latest-packages
        """
        latest_pkgs = dict()

        for _pkg_name, _pkg in pkgs.items():
            for index, _pkg_version in enumerate(_pkg['pkg_version']):
                pkg_info = dict(py_version=_pkg['py_version'][index], pkg_version=_pkg_version,
                                file=_pkg['file'][index])
                if _pkg_name in latest_pkgs:
                    if VersionUtil.parse_version(latest_pkgs[_pkg_name]['pkg_version']) < \
                            VersionUtil.parse_version(_pkg_version):
                        latest_pkgs[_pkg_name] = pkg_info
                else:
                    latest_pkgs[_pkg_name] = pkg_info

        return latest_pkgs

    @staticmethod
    def _getPkgPyVersion(version_tag: str) -> str:
        """
        Get the python version from the version tag of a package

        Parameters
        ----------
        version_tag: str

        Returns
        -------
        pkg_py_version : str
        """
        match_obj = re.search(r'^py(?P<pkg_name>[\d]+\.[\d]+)', version_tag)
        if match_obj is None:
            pkg_py_version = '0.0'
        else:
            pkg_py_version = match_obj.group(1)
        return pkg_py_version
