#!/usr/bin/python
# -*- coding: <UTF-8> -*-

"""
Python Version : 3.6

$LastChangedRevision$
$LastChangedDate::                           $


Copyright (c) 2018, Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
All rights reserved.
"""

"""
This module defines tools used for the SyMSpace tolerance evaluation module ´symtol´.
"""
import time
import symspace.wrapper.symwrapper as sw
import os
import sys
import logging


class StdConsoleFilter(logging.Filter):

    def filter(self, record):
        retval = True
        if (record.levelno == logging.WARNING) | (record.levelno == logging.ERROR) | (record.levelno == logging.CRITICAL):
            retval = False
        return retval


def setStandardLogger(fname):

    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    # add FileHandler for error logging
    # fname should be something like ..._error.log
    fh = logging.FileHandler(fname)
    fh.setLevel(logging.WARNING)
    f_formatter = logging.Formatter('%(asctime)s %(name)s: %(levelname)s - %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    fh.setFormatter(f_formatter)

    # add configuration for console logging (up to warning)
    ch = logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(logging.INFO)
    c_formatter = logging.Formatter('%(asctime)s: %(levelname)s - %(message)s', datefmt='%a, %H:%M:%S')
    ch.setFormatter(c_formatter)
    f = StdConsoleFilter()
    ch.addFilter(f)

    # add configuration for console logging
    ceh = logging.StreamHandler(stream=sys.stderr)
    ceh.setLevel(logging.WARNING)
    ceh.setFormatter(c_formatter)

    logger.addHandler(fh)
    logger.addHandler(ch)
    logger.addHandler(ceh)

    return logger


def tprint(s, pre=""):
    """Prints string s with time index.

    Parameters
    ----------
    s : str
        string to print after time index
    pre : str
        string to print before time index

    Returns
    -------
    str
        formatted string

    """
    ret_val = "{}{:s}\t... {:s}".format(pre, time.strftime("%H:%M:%S"), s)
    print(ret_val)

    return ret_val


def checkSymProject(glob):
    # glob = globals().copy() has to be defined outside!

    print('\n' + '*'*30)
    print("searching for SyMSpace project handle ...\n")

    sp = []
    ret = None

    for g in glob:
        if isinstance(glob[g], sw.SymProject):
            sp.append(g)

    if len(sp) == 0:
        print(
            "No SymProject object found in the workspace of the current Python console!\n--> Assign CurProject to SyMSpace Python Console (ctrl + P)")
    elif len(sp) > 1:
        print("Multiple SymProject objects available in the current Python console:")
        for p in sp:
            print(f"\t{p}")
        print(
            "Only keep one SymProject object and delete the others or define the variable 'sp' explicitely in the runTolAnalysis script, e.g., sp = CurProject.")
    else:  # len(sp)==1
        sp = sp[0]
        print(f"SymProject '{sp}' found.")
        ret = glob[sp]

    return ret


class RunEveryN:

    def __init__(self, N, func):
        self.N = N
        self.run = func
        self.cnt = 0

    def __call__(self, *args, **kwargs):

        self.cnt += 1

        if self.cnt == self.N:
            self.run(*args, **kwargs)
            self.cnt = 0


class ControlFile:

    def __init__(self, simProject: sw.SymProject):

        self.fname = os.path.join(simProject.getDirectory(), simProject.getFile().split('.mop')[0] + '.ctrl')

        """if control file exists it is removed and newly initialized"""
        if os.path.isfile(self.fname):
            os.remove(self.fname)

        fhandle = open(self.fname, 'w')
        fhandle.write('stop=False\n')
        fhandle.write('finish=False\n')
        fhandle.close()

        self.last_modified = os.path.getmtime(self.fname)

        logger = logging.getLogger(self.__class__.__name__)
        logger.info(f"Control file {self.fname} sucessfully created.")
        #print('\n' + '*' * 30)
        #tprint(f"Control file {self.fname} sucessfully created")

    def remove(self):
        if os.path.isfile(self.fname):
            logger = logging.getLogger(self.__class__.__name__+'.remove')
            try:
                os.remove(self.fname)
                logger.info(f"Control file {self.fname} sucessfully removed.")
            except IOError as e:
                errno, strerror = e.args
                logger.error(f"I/O error({errno}): {strerror}")
            except:
                logger.error('Unexpected error when removing control file.')

    def modified(self):
        """
        Returns
        -------
        bool
            True if control file is modified, False otherwise

        """
        # tprint("#### check for control-file modification ####")
        modified = False
        if os.path.getmtime(self.fname) != self.last_modified:
            logger = logging.getLogger(self.__class__.__name__ + '.modified')
            logger.info('Control file seems to be modified.')
            #print('\n' + '*' * 30)
            #tprint('Control file seems to be modified!')
            modified = True

        return modified

    def getControls(self):
        """
        return dictionary with all controls, e.g.: finish and stop
        every control has to be in a seperate line and have the format: key=value (no colons for key)

        Returns
        -------
        dict
            dictionary holding all controls

        """
        self.last_modified = os.path.getmtime(self.fname)
        with open(self.fname, 'r') as fhandle:
            return eval('dict(' + ', '.join(fhandle.read().splitlines()) + ')')
