/*
 * This file is part of SyMSpace.
 * 
 * Copyright(c)2020,Linz Center of Mechatronics GmbH (LCM) http://www.lcm.at/
 * All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify it under 
 * the terms of the GNU General Public License, version 2.0 (GPLv2), as published 
 * by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A 
 * PARTICULAR PURPOSE. See the GNU General Public License, version 2.0, for more 
 * details.
 * The GPLv2 can be found in the licenses folder as gpl-2.0.txt.
 * 
 * $Revision: 10084 $
 * $Date: 2020-12-22 14:17:15 +0100 (Tue, 22 Dec 2020) $
 */
package jobmanager.server;

import java.awt.Desktop;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Iterator;
import java.util.concurrent.ConcurrentSkipListMap;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamReader;

import at.lcm.bu21.general.utils.FileUtils;
import at.lcm.bu21.magopt.structure.event.Event;
import at.lcm.bu21.magopt.structure.event.EventType;
import jobmanager.JobDesc;
import jobmanager.JobStatus;
import jobmanager.JobType;
import jobmanager.JobUniverse;
import jobmanager.logging.Logger;

/**
 * Job representation
 * 
 * @author Werner Koppelsttter
 */
public class Job implements Comparable<Job> {

	/** Working directory of the JobManager */
	private static File workingDirectory;
	/** List of jobs */
	public static ConcurrentSkipListMap<String, Job> joblist = new ConcurrentSkipListMap<String, Job>();
	/** Next job-id */
	private static long nextid = 0;

	/** Log of error-jobs */
	public static Logger errorLog;

	/** job-id */
	private long id;
	/** Project name */
	private String project = "Undefined";
	/** Dataset name */
	private String dataset = "Undefined";
	/** Parameter name */
	private String parameter = "Undefined";

	/** Name of job */
	private String name = "Undefined";
	/** Identifier of job */
	private String identifier;
	/** Docker-image of job */
	private String dockerImage;
	/** Command of job */
	private String command;
	/** Arguments to execute */
	private String[] args;
	/** List of linked files */
	private String[] linkedFile = new String[0];
	/** Requirement of job */
	private String requirement;
	/** Timeout of job */
	private long timeout = Long.MAX_VALUE;
	/** Type of job */
	private JobType type = JobType.UNDEFINED;
	/** Universe of job */
	private JobUniverse universe = JobUniverse.JAVA;
	/** Status of Job */
	private JobStatus status = JobStatus.UNDEFINED;
	/** Adapter for execution queue */
	private JobAdapter adapter = null;
	/** Runtime of job-execution */
	private int runtime = 0;
	/** Host of job-execution */
	private String host = null;
	/** Last change of status */
	private long statusChanged = 0;

	static {
		Job.workingDirectory = JobDesc.getWorkingDir();
		if (!Job.workingDirectory.exists()) {
			Job.workingDirectory.mkdirs();
		}

		try {
			Job.errorLog = new Logger("ErrorJob.txt");
		} catch (IOException e) {
			System.err.println(e.getMessage());
			e.printStackTrace();
		}
	}

	/**
	 * Refresh job list
	 */
	public static void refreshList() {
		new Thread() {
			@Override
			public void run() {
				Job job;
				File[] files;
				Iterator<Job> iter;

				// add jobs
				files = Job.workingDirectory.listFiles();
				for (File f : files) {
					if (f.isDirectory()) {
						job = Job.joblist.get(f.getName());
						if (job == null) {
							new Job(f.getName());
						}
					}
				}

				// remove Jobs
				iter = Job.joblist.values().iterator();
				while (iter.hasNext()) {
					job = iter.next();
					if ((job.getAdapter() == null) && !job.getJobDirectory().exists()) {
						Job.joblist.remove(job.getIdentifier());
					}
				}

				JobManager.refreshList();
			}
		}.start();
	}

	/**
	 * Create Job for adapter
	 * 
	 * @param type    Type of job
	 * @param adapter Job adapter
	 */
	public Job(JobType type, JobAdapter adapter) {
		int k = 0;

		do {
			this.identifier = "Undefined_" + k++ + "_";
		} while (Job.joblist.containsKey(this.identifier));

		this.id = Job.nextid++;
		this.type = type;
		this.setStatus(JobStatus.UNDEFINED);
		this.adapter = adapter;
		Job.joblist.put(this.getIdentifier(), this);
	}

	/**
	 * Add a existing job
	 * 
	 * @param identifier
	 */
	Job(String identifier) {
		int index;

		this.identifier = identifier;
		index = identifier.lastIndexOf('(');
		if (index >= 0) {
			this.name = identifier.substring(0, index);
		} else {
			this.name = identifier;
		}
		this.id = Job.nextid++;
		this.setStatus(JobStatus.UNDEFINED);
		Job.joblist.put(this.getIdentifier(), this);
	}

	/**
	 * Create a new job
	 * 
	 * @param project   Project name
	 * @param dataset   Dataset name
	 * @param parameter Parameter name
	 * @param name      Job name
	 * @param universe  Job universe
	 * @param args      command
	 */
	public Job(String project, String dataset, String parameter, String name, JobUniverse universe, String[] args) {
		this(name + "_" + Long.toString(System.currentTimeMillis(), 36) + "_" + Long.toString(System.nanoTime(), 36)
				+ "_");
		this.project = project;
		this.dataset = dataset;
		this.parameter = parameter;
		this.name = name;
		this.universe = universe;
		this.args = args;
		this.setStatus(JobStatus.CREATED);

		new File(Job.workingDirectory, this.identifier).mkdirs();
	}

	/**
	 * Update object
	 */
	public void update() {

		// FIXME refreshList (Open view???)

		// remove undefined jobs (1 hour)
		if (this.getStatus() == JobStatus.UNDEFINED) {
			if ((System.currentTimeMillis() - this.statusChanged) > (60 * 60 * 1000)) {
				this.remove();
			}
		}

		// remove old jobs (3 days)
		if (this.getStatus() == JobStatus.COMPLETED) {
			if ((System.currentTimeMillis() - this.statusChanged) > (3 * 24 * 60 * 60 * 1000)) {
				this.remove();
			}
		}

		if (this.getStatus() == JobStatus.REMOVED) {
			if (this.adapter == null) {
				if (this.getJobDirectory().exists()) {
					FileUtils.delete(this.getJobDirectory());
				} else {
					Job.joblist.remove(this.getIdentifier());
				}
			}
		}

		if (this.getStatus() == JobStatus.RESTART) {
			if (this.adapter == null) {
				this.setStatus(JobStatus.SUBMIT);
			}
		}

		if (this.getStatus() == JobStatus.SUBMIT) {
			if (this.adapter == null) {
				this.submit();
			}
		}

		if (this.getStatus() == JobStatus.COMPLETE) {
			if (this.adapter == null) {
				if (new File(this.getJobDirectory(), "Error.xml").exists()) {
					Event err;
					try {
						InputStream in = new FileInputStream(new File(this.getJobDirectory(), "Error.xml"));
						XMLInputFactory xmlFactory = XMLInputFactory.newInstance();
						XMLStreamReader xmlReader = xmlFactory.createXMLStreamReader(in, "utf-8");
						err = new Event(EventType.ERROR, "", "", null);

						// read file
						while (xmlReader.getEventType() != XMLStreamConstants.START_ELEMENT) {
							xmlReader.next();
						}
						err.xml2obj(xmlReader);
						if (xmlReader.getEventType() != XMLStreamConstants.END_DOCUMENT) {
							err = null;
						}
						xmlReader.close();
					} catch (Exception e) {
						System.err.println(e.getMessage());
						e.printStackTrace();
						err = null;
					}

					if (err != null) {
						this.addErrorLog("Error in job: " + this.getHost() + "\t" + err.getDescription());
					} else {
						this.addErrorLog("Error in job: " + this.getHost());
					}

					this.setStatus(JobStatus.ERROR);
				} else {
					this.setStatus(JobStatus.COMPLETED);
				}
			}
		}
	}

	/**
	 * Get local directory of job
	 * 
	 * @return Local directory
	 */
	public File getJobDirectory() {
		return new File(Job.workingDirectory, this.identifier);
	}

	/**
	 * Linked file
	 * 
	 * @param file File to link
	 */
	public void linkFile(File file) {
		if ((file != null) && file.exists()) {
			this.linkedFile = Arrays.copyOf(this.linkedFile, this.linkedFile.length + 1);
			this.linkedFile[this.linkedFile.length - 1] = file.getAbsolutePath();
		}
	}

	/**
	 * Get list of files to transfer
	 * 
	 * @param sep Separator
	 * @return List of files
	 */
	public String getTransferFiles(String sep) {
		StringBuffer list = new StringBuffer();

		// add linked files
		for (String f : this.linkedFile) {
			list.append(f);
			list.append(sep);
		}

		// remove last separator
		if (list.length() > 0) {
			list.delete(list.length() - sep.length(), list.length());
		}

		return list.toString();
	}

	public String[] getLinkedFiles() {
		String[] files;

		files = new String[this.linkedFile.length];
		for (int k = 0; k < this.linkedFile.length; k++) {
			files[k] = this.linkedFile[k];
		}
		return files;
	}

	/**
	 * Get list of library files
	 * 
	 * @param sep  Separator
	 * @param path Add with path
	 * @return List of library
	 */
	public String getLibraryFiles(String sep, boolean path) {
		StringBuffer list = new StringBuffer();

		// add linked files
		for (String f : this.linkedFile) {
			if (f.toLowerCase().endsWith(".jar")) {
				if (list.length() != 0) {
					list.append(sep);
				}

				if (path) {
					list.append(f);
				} else {
					list.append(new File(f).getName());
				}
			}
		}

		return list.toString();
	}

	/**
	 * Get Project name
	 * 
	 * @return Name of project
	 */
	public String getProject() {
		return this.project;
	}

	/**
	 * Get dataset name
	 * 
	 * @return Name of dataset
	 */
	public String getDataset() {
		return this.dataset;
	}

	/**
	 * Get parameter name
	 * 
	 * @return Name of Parameter
	 */
	public String getParameter() {
		return this.parameter;
	}

	/**
	 * Get job name
	 * 
	 * @return Job name
	 */
	public String getName() {
		return this.name;
	}

	/**
	 * Get identifier of job
	 * 
	 * @return Identifier
	 */
	public String getIdentifier() {
		return this.identifier;
	}

	/**
	 * Get arguments of job
	 * 
	 * @return Array with arguments
	 */
	public String[] getArguments() {
		return this.args;
	}

	/**
	 * Set Arguments of job
	 * 
	 * @param args Array of arguments
	 */
	public void setArguments(String[] args) {
		this.args = args;
	}

	/**
	 * Get job requirements
	 * 
	 * @return Requirements
	 */
	public String getRequirement() {
		return this.requirement;
	}

	/**
	 * Set job requirements
	 * 
	 * @param requirement Requirements
	 */
	public void setRequirement(String requirement) {
		this.requirement = requirement;
	}

	/**
	 * Get job timeout in seconds
	 * 
	 * @return Timeout in seconds
	 */
	public long getTimeout() {
		return this.timeout;
	}

	/**
	 * Set job timeout in seconds
	 * 
	 * @param timeout Timeout in seconds
	 */
	public void setTimeout(long timeout) {
		this.timeout = timeout;
	}

	/**
	 * Get status of job
	 * 
	 * @return Status
	 */
	public JobStatus getStatus() {
		return this.status;
	}

	/**
	 * Set/Update status of job
	 * 
	 * @param status New status
	 */
	public void setStatus(JobStatus status) {
		this.status = status;
		this.statusChanged = System.currentTimeMillis();
	}

	/**
	 * Get type of job
	 * 
	 * @return Type of job
	 */
	public JobType getType() {
		return this.type;
	}

	/**
	 * Set type of job
	 * 
	 * @param type Type of job
	 */
	public void setType(JobType type) {
		this.type = type;
	}

	/**
	 * Get universe of job
	 * 
	 * @return Universe of job
	 */
	public JobUniverse getUniverse() {
		return this.universe;
	}

	/**
	 * Set universe of job
	 * 
	 * @param universe Universe of job
	 */
	public void setUniverse(JobUniverse universe) {
		this.universe = universe;
	}

	/**
	 * Get docker-image of job
	 * 
	 * @return Docker-image name
	 */
	public String getDockerImage() {
		return this.dockerImage;
	}

	/**
	 * Set docker-image of job
	 * 
	 * @param dockerImage Universe of job
	 */
	public void setDockerImage(String dockerImage) {
		this.dockerImage = dockerImage;
	}

	/**
	 * Get command of job
	 * 
	 * @return Command
	 */
	public String getCommand() {
		return this.command;
	}

	/**
	 * Set command of job
	 * 
	 * @param cmd Command of job
	 */
	public void setCommand(String cmd) {
		this.command = cmd;
	}

	/**
	 * Get a String representation of the object
	 * 
	 * @return String representation
	 */
	@Override
	public String toString() {
		String str;

		str = new File(this.project).getName() + "  " + this.dataset + "  " + this.parameter + "  " + this.name + "  "
				+ this.getStatusString();
		if (this.adapter != null) {
			str += "  " + this.adapter;
		}
		return str;
	}

	/**
	 * Get staus-string
	 * 
	 * @return Status-string
	 */
	public String getStatusString() {
		String status;

		status = this.getStatus().toString();
		switch (this.getType()) {
		case LOCAL:
			status += " (L)";
			break;
		case CLUSTER:
			status += " (C)";
			break;
		default:
			status += " (?)";
		}

		return status;
	}

	/**
	 * Submit a job to an execution-queue
	 */
	public void submit() {
		File[] files;
		boolean contain;
		Client client;

		if (this.adapter == null) {
			if (this.getStatus() == JobStatus.SUBMIT) {
				// remove directories and unlinked files
				files = this.getJobDirectory().listFiles();
				for (File f : files) {
					if (f.isDirectory()) {
						FileUtils.delete(f);
					} else {
						contain = false;
						for (String l : this.linkedFile) {
							contain |= l.equals(f.getAbsolutePath());
						}
						if (!contain) {
							f.delete();
						}
					}
				}
				// Reset runtime/host
				this.setRuntime(0);
				this.setHost(null);

				client = JobManager.getClient(this.type);
				if (client == null) {
					this.type = JobType.LOCAL;
					client = JobManager.getClient(this.type);
				}
				if (client != null) {
					this.adapter = client.addJob(this);
				}
			}
		}
	}

	/**
	 * Mark job to remove from JobManager
	 */
	public void remove() {
		this.setStatus(JobStatus.REMOVED);
	}

	/**
	 * Mark job to remove from JobManager
	 */
	public void stop() {
		this.setStatus(JobStatus.STOP);
	}

	/**
	 * Restart job
	 */
	public void restart() {
		this.setStatus(JobStatus.RESTART);
	}

	/**
	 * Remove job-adapter for execution-queue
	 */
	public void clearAdapter() {
		this.adapter = null;
	}

	/**
	 * Get Adapter of job
	 * 
	 * @return Job-adapter
	 */
	public JobAdapter getAdapter() {
		return this.adapter;
	}

	/**
	 * Open working directory of job
	 */
	public void openDirectory() {
		if (Desktop.isDesktopSupported()) {
			try {
				Desktop.getDesktop().open(this.getJobDirectory());
			} catch (IOException e) {
			}
		}
	}

	@Override
	public int compareTo(Job other) {
		return (int) (this.id - other.id);
	}

	/**
	 * Format time in seconds to string
	 * 
	 * @param time Time to format (seconds)
	 * 
	 * @return String of time
	 */
	protected String formatTime(int time) {
		String str = null;

		if (time != 0) {
			str = String.format("%02d", time % 60);
			time /= 60;

			str = String.format("%02d:%s", time % 60, str);
			time /= 60;

			str = String.format("%02d:%s", time % 24, str);
			time /= 24;

			str = String.format("%02d+%s", time % 100, str);
		}

		return str;
	}

	/**
	 * Get runtime of job-execution
	 * 
	 * @return Runtime (sec)
	 */
	public String getRuntime() {
		return this.formatTime(this.runtime);
	}

	/**
	 * Set Runtime of job-execution
	 * 
	 * @param runtime Runtime (sec)
	 */
	public void setRuntime(int runtime) {
		this.runtime = runtime;
	}

	/**
	 * Get host of job-execution
	 * 
	 * @return Host
	 */
	public String getHost() {
		return this.host;
	}

	/**
	 * Set host of job-execution
	 * 
	 * @param host Host
	 */
	public void setHost(String host) {
		this.host = host;
	}

	/**
	 * Add error-log message
	 * 
	 * @param message
	 */
	public void addErrorLog(String message) {
		if (Job.errorLog != null) {
			try {
				Job.errorLog.write(message);
			} catch (IOException e) {
				System.err.println(e.getMessage());
				e.printStackTrace();
			}
		}
	}
}
